# SCRIPT FOR SENDING PUSHOVER NOTIFICATIONS WITH A GIF FILE ATTACHMENT
#
# Version: 3
# Created by jaydeel at ipcamtalk.com
# Copyright: (2024) jaydeel
#
# Changelog: 
# V3 - 2025-03-16    - BREAKING CHANGE: must replace old user-settings-file
#                    - ADDED: script now supports temporary DAT files created in 'C:\ProgramData\Blue Iris\temp' (v5.9.9.29+);
#                      temporary DAT files are created when: 1) using AI, and 2) setting 'Save AI analysis DAT files' is not enabled
#                    - ADDED: argument -IgnoreDatAge to force use of temporary DAT files regardless of age (useful for testing)
#                    - FIXED: bugs when using -AlertRec, -DbRecNo, and -UI3Rec arguments
#                    - FIXED: bug in logic that retries getting the alert's JSON clipstats object
#                    - IMPROVED: argument -PushService and user-setting $push_service now support simultaneously using both 
#                      notification services 'pushover' and 'telegram'
#                    - CHANGED: Pushover API now accepts attachment sizes up to 5.0MB (previously 2.5 MB)
#                    - CHANGED: added a short wait, if needed, for Blue Iris to finish writing an alert's DAT file
#                    - CHANGED: improved validation method for checking if BVR-extracted JPG files are ready for FFmpeg
#                    - CHANGED: script now shows BVR extraction delay timing information in outputs
#                    - CHANGED: default notification title now includes the script version
#                    - CHANGED: all instances of .ToString() now use "F" ('fixed-point') format
#                    - CHANGED: function BIJsonGetAlertList() & BIJsonAlertListParse() now support alert duration property 'msec'
# V2.3 - 2024-12-10  - BREAKING CHANGE: must replace old user-settings-file
#                    - FIXED: various bugs encountered w/ older versions of Blue Iris (tested back to 5.7.3)
#                      bugs associated with checking for already loaded Win32 assemblies (multiple instances)
#                      bug where assembly 'System.Drawing' was not available in the powershell.exe environment
#                      bug in function SendTelegram(); returned item not always a string
#                      bug in MakeGIF() while appending a BI-style date to a cloned clipstats object's Unix timestamp
#                    - FIXED: bug where old variable $DbRec should have been new variable $DbRecNo (originally changed in V2)
#                    - ADDED: new column 'PSVersion' to script's CSV log file (for future debugging info)
#                    - ADDED: new user setting to support older Blue Iris software versions: $require_new_json_session_keys_flg 
#                    - CHANGED: function BiJsonClipstats() now ensures that database record id values are prefixed with '@'
#                    - CHANGED: all JSON-based functions now utilize global variable $script:JsonCallForceFlg
#                      Note: Blue Iris automatically logs out of inactive JSON sessions
#                    - IMPROVED: various optimizations / efficiency gains:
#                      script now avoids unnecessary attempts to reload assembly 'System.Drawing'
#                      eliminated unnecessary -match comparisons and replaced w/ faster non-regex operators
#                      better logic for loading of assemblies in functions Format-Color(), PauseOutput()
#                      disabled progress bar in function BIJsonGetAlertList() when using -Latest argument
#                      stopped logging out of the used JSON session in BIJsonClipstats() & BIJsonGetAlertlist()
#                      argument -DevMode now includes detailed timing info, including: 1) timing info for JSON cmds, and 
#                      2) JSON alertlist response for argument -Latest
#                      added code to test 2 additional methods for verifying successful extraction of JPGs from BVR files; 
#                      settled on the original method used in V2.2
#                      misc updates to functions: MakeGIF(), SendPushover(), SendTelgram(), Format-Color(), PauseOutput()
# V2.2 - 2024-11-16  - BREAKING CHANGE: must replace old user-settings-file
#                    - FIXED: bug where ffmpeg was not honoring user-settings-file variable $gif_height
#                    - FIXED: bug where argument -AlertMemo was not triggering automatic right-trimming of GIF
#                    - FIXED: bug where function PostCustomBiLogEntry() did not trim messages to BI limit (128 characters)
#                    - FIXED: bug where function PostCustomBiLogEntry() failed to work when using BI user credentials
#                    - ADDED: new argument -Examples to provide the user with easy-to-try example commands
#                    - ADDED: new argument -Devices to support custom device(s) for notifications (Pushover only); comma-delimited
#                    - ADDED: new column 'ScriptHost' to script's CSV log file
#                    - ADDED: function MakeGIF() now outputs a script performance hashtable; potentially useful for debugging bottlenecks
#                    - ADDED: user setting '$jpg_images_from_substream_flg' to enable extracting JPGs from BVR default-stream
#                    - ADDED: user setting '$jpg_images_match_gif_height_flg' to enable extracting full-sized JPGs from BVR files
#                    - ADDED: user setting '$append_performance_tag_details_flg' to hide default performance details notification tag
#                    - ADDED: warning if GIF archive pruning operation time is becoming excessive
#                    - CHANGED: removed user-setting $reg_bi_alerts_folder_path; script now retrieves the setting from the Windows registry
#                    - CHANGED: removed user-setting $bi_webserver; script now retrieves the setting from the Windows registry
#                    - CHANGED: removed user-setting $dat_image_interval_msec; script now retrieves the setting from the Windows registry
#                    - CHANGED: functions BIJsonGetAlertList() & BIJsonAlertListParse() now support argument -LatestFlg
#                    - IMPROVED: script can now use BI credentials to create GIFs using JPG files extracted from BVR files
#                    - IMPROVED: logic & timing for verifying successful extraction of JPGs from BVR files
#                    - IMPROVED: script now validates user-setting $bi_new_folder_paths against folders in the BI settings Storage tab
# V2.1 - 2024-09-21  - BREAKING CHANGE: must replace old user-settings file
#                    - ADDED: script now supports Telegram Bot notifications (alternative to Pushover); 
#                      added 3 new user settings to support this feature; 
#                      added 1 new column (TelegramMs') to the script's CSV log file;
#                      added push method ('pushover' or 'telegram') to BI logfile entry
#                    - ADDED: added argument -PushService ('Pushover' or 'Telegram') to support on-demand 
#                      service selection (case-insensitive)
#                    - ADDED: argument -Sound to support on-demand sounds (Pushover notifications only)
#                    - ADDED: script can now make GIFs via BI Clipboard clips (or Alerts created via Clipboard clips)
#                    - ADDED: user setting '$append_performance_tag_flg' to hide default performance notification tag 'GifMakeTime: N msec'
#                    - ADDED: user setting '$experienced_user_flg' to hide output intended for first-time users
#                    - FIXED: script now handles DAT files with JPG images stored oldest to newest; effective BI 5.9.4.11+
#                    - FIXED: bug where script did not try BVR extraction when DAT files contained fewer than required JPG images
#                    - FIXED: bug preventing script from automatically using the BVR file when the DAT file not exist
#                      (thanks snz at ipcamtalk)
#                    - FIXED: bug where script ignored argument -NoArchive
#                    - FIXED: bug in BIJsonGetAlertList(); previously failed to exclude cloned camera entries
#                    - FIXED: bug in BIJsonAlertListParse(); previously failed to exclude cloned camera entries
#                    - CHANGED: changed user setting '$append_DAT_tag_flg' to 'append_gif_method_tag_flg';
#                      used to hide default notification tags, like ' (viaDAT:n)' or ' (viaClip)' or ' (viaLive)'
#                    - CHANGED: reformatted notification tags for DAT-generated GIFs from '(DATn)' to '(viaDAT:n)'
#                    - CHANGED: made the following changes to the BI logfile entry to shorten the entry (127 characters max):
#                      replaced script basename (typically BI_gif_tool') to 'gif_tool'; saved 3 chars;
#                      changed duration ' sec' units '[320 H, 2 FPS, 6.0 sec, 724.5 KB]' to '[320 H, 2 FPS, 6.0 s, 724.5 KB]'; saved 2 chars;
#                      removed redundant ' ms' units '[0 ms, 8559 ms, 246 ms, 9021 ms]' to '[0, 8559, 246, 9021 ms]'; saved 9 chars;
#                      removed substring 'method=' (leaving just 'viaDat', 'viaClip', 'viaLive'); saved 7 characters
#                      removed entry 'rec='; saved at least 4 chars;
#                      removed substring 'source='; saved 7 chars;
#                      total characters saved = 3+2+9+7+4+7-8 for added push method substring = 24
#                    - IMPROVED: made function IsBiVersion() more efficient (gets BI version from registry instead of JSON call)
#                    - IMPROVED: -Live argument now supports the use of non-default folder naming patterns in 
#                      Camera settings > Record > Video
#                    - IMPROVED: updated functions: FindBytes() (V3), IsBiVersion() (V2)
# V2 - 2024-03-18    - BREAKING CHANGE: must replace old user-settings file 
#                    - BREAKING CHANGE: changed name of argument -DebRec to -DbRecNo; to clarify that script is 
#                      expecting a record ID number, not a string like "@xxxx"
#                    - BREAKING CHANGE: deprecated redundant argument -AlertDB (use -AlertRec instead)
#                    - ADDED: script now supports multiple Blue Iris 'New' folders
#                    - ADDED: argument -Latest (use only w/ -Camera); makes GIF from the most recent alert
#                    - ADDED: argument -DatImageMsec to improve DAT-derived GIF duration estimates;
#                      should be the same as the value specified for setting 'Analyze one image each x msec'
#                      (BI > Camera settings >  Alert > AI Configuration)
#                    - ADDED: argument -DevMode now exports each camera's JSON alertlist (to 'app_folder\_devmode')
#                    - ADDED: added descriptive/explanatory '_README.txt' files to all programmatically 
#                      created folders
#                    - ADDED: script's CSV log file automatically renamed for version changes
#                    - IMPROVED: changed method for making -Live GIFs (now using live stream instead of BVR)
#                    - IMPROVED: supports the use of non-default folder naming patterns in Camera settings > Record > Video
#                      e.g., %Y-%m-%d\&CAM\&CAM.%Y%m%d_%H%M%S
#                    - IMPROVED: args -TrimBegPct & -TrimEndPct now ALSO work on DAT-derived GIFs; this
#                      works by skipping the the extraction of an appropriate number of JPGs from the DAT file
#                    - IMPROVED: argument -TrimBegPct may be negative to include pre-trigger images
#                      (BVR-derived GIFs only)
#                    - IMPROVED: added 2 columns ('PushoverMs' & 'TotalMs') to the script's performance log
#                    - IMPROVED: updated -Help output to clarify usage options
#                    - IMPROVED: enhanced validation of the script's arguments
#  	                 - IMPROVED: script now provides feedback if commandline includes unsupported arguments 
#                    - CHANGED: deprecated argument -NukeDatFile; WHY? per Blue Iris Support a 
#                      DB/Regenerate action is always REQUIRED after deleting a DAT file. Obviously, 
#                      this must be avoided because it momentarily stops all recording DURING AN EVENT. 
#                      Instead, consider batch deleting no-longer-needed DAT files via my 'BI_DAT_tool.ps1' script
#                    - CHANGED: deprecated argument -NoArchiveDat; WHY? has no use after deprecating -NukeDatFile
#                    - CHANGED: deprecated user-file setting $dat_archive_lifetime_hours; WHY? has no use after deprecating -NukeDatFile
#                    - FIXED: null source filespec for BVR when using argument -Live
#                    - FIXED: untrapped error if Blue Iris service is not running
#                    - FIXED: bug in function SendPushover()
#                    - IMPROVED: updated functions BIJsonCall(), BIJsonGetAlertList(), $FindBytes()
#                    - IMPROVED: function BIJsonAlertListParse() is now more tolerant of database issues
# V1.9  - 2023-09-07 - ADDED: a repair is now attempted for situations where the JPG start- & end-of-image 
#                      byte markers appear to be unbalanced
#                    - IMPROVED: JPG validation function 'IsJpegFileValid()' now also tests for 
#                      valid end-of-image byte marker
#                    - CHANGED: the usage guide (shown via argument -Help) is no longer displayed by default
#                    - CHANGED: argument -NukeDatFile has been disabled until Blue Iris support provides 
#                      a safe method to delete DAT files
#                    - FIXED: SendPushover() call; lacked grave accent (`) after -Camera param line; 
#                      caused malformed value
#                    - FIXED: DAT-extracted JPG files that previously lacked the req'd JPEG 
#                      Interchange Format end-of-image byte marker ('FF D9')
# V1.8  - 2023-07-12 - IMPROVED: handling of warning if user appears to be trying to use 
#                      parameter -ImagesCnt with DAT files
#                    - FIXED: argument -Priority; mis-typed as [String] instead of [Int]; 
#                      PS7 forgiving; PS5 not
# V1.7  - 2023-07-09 - ADDED: error trapping for non-existent database record IDs
#                    - IMPROVED: shortened message in BI logfile, and 'gif_source' entry in script's CSV log (removed path)
#                    - CHANGED: logged GIF duration values units from msec to seconds 
#                      (REQUIRES RENAMING OLD CSV FILE)
#                    - FIXED: BI logfile message; missing source file
# V1.6  - 2023-06-30 - ADDED: warning if user appears to be trying to use parameter -ImagesCnt 
#                      with DAT files
#                    - IMPROVED: error trapping & feedback for DAT files with insufficient images
#                    - IMPROVED: error trapping when trying to get $gif_jpg_WxH from one of the JPG files
#                    - IMPROVED: ffmpeg issues (if any) are now displayed in script output to the console
#                    - FIXED: handling of function ExtractDatImages() when its returned object is empty
# V1.5  - 2023-06-28 - IMPROVED: performance log (*.csv file) now reporting GIF's JPG WxH instead of  Alert's WxH
#                    - IMPROVED: function IsJpegFileValid() is now releasing resources ($img.Dispose()) 
#                      to reduce risk of memory leak
#                    - FIXED: function CopyFileAsSequence()
# V1.4  - 2023-06-27 - ADDED: error trap when ffmpeg fails to successfully create a GIF file
#                    - IMPROVED: GIF size reduction algorithm now meets the Pushover API limit 
#                      (GIF can no longer have <1 image)
#                    - IMPROVED: Blue Iris logfile logging levels now reflect the script's output if it 
#                      contains 'ERR/' or 'WARNING:'
# V1.3  - 2023-06-27 - ADDED: argument -NoArchiveDat
#                    - FIXED: issue that prevented the logging of the source file & alert record to 
#                      the CSV performance log file
# V1.2  - 2023-06-26 - ADDED: sleep status bar SleepWithStatusBar()
#                    - FIXED: problem retrieving actively recording BVR filename (PowerShell v5)
# V1.1  - 2023-06-26 - CHANGED: modified default value for user setting '$gif_trim_end_cust_obj'
# V1    - 2023-06-26 - initial release
#
# REQUIRES:
#    CHOCOLATEY - Download and install the Chocolatey package manager: https://chocolatey.org/install 
#    FFMPEG     - Install the latest ffmpeg build using [choco install ffmpeg] ... https://www.gyan.dev/ffmpeg/builds/
#               - Home: https://www.ffmpeg.org/download.html          
#
# USAGE:      {this script name}.ps1 Msg [Priority AttachFile Quality Width, TargetKB Quiet]
# ARGUMENTS:  -AlertRec      = (String) alert record, typically via Blue Iris macro &ATTACH_DB
#             -AlertDB       = (String) [DEPRECATED] alert record, typ. via Blue Iris macro &ATTACH_DB
#             -AlertMemo     = (String) alert AI memo, typ. via Blue Iris macro &MEMO
#             -Live          = (Switch) create GIF from specified camera's stream
#             -Camera        = (String) camera shortname (required with -Live argument)
#             -UI3Rec        = (String) hyphenated numeric values following UI3 URL argument '&rec=', 
#                                       like 'ui3.htm?rec=208953322-1181690'
#             -DbRecNo       = (String) Blue Iris database record number for clip, like '208953322' (not prefixed with '@')
#             -Seconds       = (Integer) GIF duration; w/ -Live argument only; default is 10 seconds
#             -DatImageMsec  = (Integer) camera-specific msec between DAT images (applies to DAT-derived GIFs only)
#             -BvrDelayMsec  = (Integer) camera-specific msec to wait before extracting images (applies to BVR-derived GIFs only)
#             -Latest        = (Switch) uses DAT file from the most recent alert
#             -IgnoreDat     = (Switch) ignores 'DAT' file (if it exists); 
#                                       by default, a DAT file is always used if it contains N or more JPGs
#                                       N is specified in the user-settings file '$gif_DAT_reqd_images_cnt'; the default is 4
#             -NukeDatFile   = (Switch) [DEPRECATED] removes DAT file from Blue Iris storage
#             -NoArchiveDat  = (Switch) [DEPRECATED] disables automatic archiving of DAT files (if used to create GIF)
#             -ImagesCnt     = (Integer) count of JPGs to use in the GIF animation
#             -Height        = (Integer) GIF'S height dimension; units pixels
#             -FPS           = (Integer) GIF frames per second
#             -TrimBegPct    = (Integer) percentage of alert clip to trim from GIF animation (left side)
#             -TrimEndPct    = (Integer) percentage of alert clip to trim from GIF animation (right side)
#             -PushService   = (String[]) 'Pushover' or 'Telegram'; overrides user settings file '$push_service'
#             -Title         = (String) custom title for notification
#             -Msg           = (String) notification message - may contain HTML
#             -Priority      = (Integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
#             -Sound         = (String) custom sound for notification (Pushover only)
#             -Devices       = (String) custom device(s) for notification (Pushover only); comma-delimited
#             -NoPush        = (Switch) disables sending push service notification (enabled by default)
#                                       NOTE: also disables management of GIF size to under 5.0 MB
#             -NoBiLog       = (Switch) disables logging to the blue Iris logfile (enabled by default)
#             -NoArchive     = (Switch) disables automatic archiving of GIF files (enabled by default)
#             -Quiet         = (Switch) disables ALL background messaging, and all Console messaging 
#                                       EXCEPT warnings and return object
#             -IgnoreDatAge  = (Switch) ignores age of temporary 'DAT' files (if exists); 
#             -Demo          = (Switch) disables sending notifications, and posting to the Blue logfile
#             -Help          = (Switch) enable display of usage guide
#             -Examples      = (Switch) enable display of examples guide
#             -RtnHashtable  = (Switch) returns Powershell hashtable instead of default JSON string (background only)
# RETURNS:    (hashtable) @{}
#
# EXAMPLES:
#    To send Pushover with GIF from Blue Iris 'On alert...' action
#       Parameters: "C:\ps_scripts\BI_gof_tool.ps1" -AlertRec `&ALERT_DB -Msg """<b>&CAM</b> &MEMO <a href='&WAN/alerts/&ALERT_DB?fulljpeg'>Hi-Res</a> <a href='&WAN/ui3.htm?rec=&ALERT_DB'>UI3</a>""" -Priority 2 -UseDatFile -ImagesCnt 10 -FPS 2 -Height 320 -TrimBegPct 0 -TrimEndPct 50
#

# MUST BE THE FIRST LINE IN A SCRIPT
param(
	[String]$AlertRec,
	#[String]$AlertDB,
	[String]$AlertMemo,
	[Switch]$Live,
	[String]$Camera,
	[Int]$Seconds,    # Default = 10
	[Int]$DatImageMsec,
	[Int]$BvrDelayMsec,
	[String]$UI3Rec,
	[String]$DbRecNo,
	[Switch]$Latest,
	[Switch]$IgnoreDat,
	[Switch]$NukeDatFile,  # Deprecated, forced to $false (argument kept for backward compatibility)
	[Int]$ImagesCnt,  # Default = 10
	[Int]$Height,     # Default = 320
	[Int]$FPS,        # Default = 2
	[Int]$TrimBegPct,
	[Int]$TrimEndPct,
	[String[]]$PushService,
	[String]$Title,
	[String]$Msg,
	[Int]$Priority,
	[String]$Sound,
	[String]$Devices,
	[Switch]$NoPush,
	[Switch]$NoBiLog,
	[Switch]$NoArchive,
	[Switch]$Quiet,
	[Switch]$IgnoreDatAge,
	[Switch]$Demo,
	[Switch]$Help,
	[Switch]$Examples,
	[Switch]$RtnHashtable,
	[Switch]$DevMode,
	[Switch]$Sim1,  # [OPTIONAL] remove ALL 'memo':'value' pairs from raw JSON 'alertlist' response
	[Switch]$Sim2,  # [OPTIONAL] remove 'memo':'value' pairs for 3 random alerts from JSON 'alertlist' response (per iteration)
	[Switch]$Sim3,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ all 'memo'='NA' params
	[Switch]$Sim4,  # [OPTIONAL] simulate BIJsonGetAlertList() response w/ 3 random 'memo'='NA' params (per iteration)
	[String]$Sim5   # [OPTIONAL] simulate BIJsonGetAlertList() response w/ zero date values; arg value is CamName
)

# NOTE: PowerShell recognizes parameters by the shortest unique match, meaning -title, -t, or -ti are the same parameter. 
# If you have 2 parameters like -title and -tag then the shortest unique match would be -ti and -ta respectively, 
# and using -t would throw an ambiguous parameter error.

# For debugging
# Write-Host ("PSBoundParameters = " + ($PSBoundParameters | ConvertTo-Json -Depth 100))
# Write-Host "AlertRec = $AlertRec"
# Write-Host "AlertMemo = $AlertMemo"
# Write-Host "Camera = $Camera"
# Write-Host "Live = $Live"
# Write-Host "DbRecNo = $DbRecNo"

# Write-Host "UI3Rec = $UI3Rec"
# Write-Host "Seconds = $Seconds"
# Write-Host "DatImageMsec = $DatImageMsec"
# Write-Host "Msg = $Msg"
# Write-Host "Title = $Title"
# Write-Host "Priority = $Priority"
# Write-Host "Sound = $Sound"
# Write-Host "IgnoreDat = $IgnoreDat"
# Write-Host "NukeDatFile = $NukeDatFile"
# if(!$PSBoundParameters.ContainsKey('ImagesCnt')) {
	# Write-Host "argument -ImagesCnt not used"
# }
# Write-Host "ImagesCnt = $ImagesCnt"
# Write-Host "Height = $Height"
# Write-Host "TrimBegPct = $TrimBegPct"
# Write-Host "TrimEndPct = $TrimEndPct"
# Write-Host "FPS = $FPS"
# Write-Host "Quiet = $Quiet"
# Write-Host "IgnoreDatAge = $IgnoreDatAge"
# Write-Host "Demo = $Demo"

#Set-StrictMode  -Version 3  # https://tinyurl.com/mrx3w9v4
#Start-Transcript  # default destination is C:\Users\{user}l\Documents

$ScriptVersion = "V3"

# GATHER/VALIDATE SCRIPT & ENVIRONMENT INFO
if (1 -eq 1) {  # Use to collapse this code section while editing

	#$ScriptName = $MyInvocation.MyCommand.Name
	$ScriptName = Split-Path -leaf $PSCommandpath
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')
	$ScriptBaseName = $ScriptName.Replace('.ps1', '')
	$ScriptPath = $PSCommandPath.Replace("\$ScriptName", '')

	# For debugging
	#Write-Host "ScriptName = $ScriptName"
	#Write-Host "ScriptPath = $ScriptPath"
	#Write-Host "ScriptBaseName = $ScriptBaseName"
	#Write-Host "ScriptPath = $ScriptPath"

	# INITIALIZE VARS
	$PsVersion = $PSVersionTable.PSVersion.Major
	$PsVersionFull = $PSVersionTable.PSVersion.ToString() + " (" + $PSVersionTable.PSEdition + ")"
	# NOTE: Scripts running from Windows Task Scheduler have a parent process named 'svchost'.
	#       https://stackoverflow.com/a/65894924
	$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
	# ... create lookup table (used by script's CSV log)
	$HostLabels = @{}
	$HostLabels['svchost'] = 'TaskScheduler'
	$HostLabels['explorer'] = 'Powershell or Powershell-ISE'
	$HostLabels['WindowsTerminal'] = 'WindowsTerminal'

	# For testing
	# Write-Host "ParentProcess = $ParentProcess"
	# Write-Host "PsVersion = $PsVersion"
	# Write-Host "PsVersionFull = $PsVersionFull"

	# ... Get friendly host name
	if ($HostLabels[$ParentProcess]) {
		$script_host_str = $HostLabels[$ParentProcess]
	} else {
		$script_host_str = $ParentProcess
	}
	# NOTE: Scripts called by another script have a 'PSCommandPath'.
	#       Method 'PSCommandPath' returns null if script is executed via the terminal
	#       https://stackoverflow.com/a/23370618/13499984
	$CallingScript = $MyInvocation.PSCommandPath

	# DETERMINE IF RUNNING IN BACKGROUND (IN CLI MODE, OR FROM TASK SCHEDULER, BLUE IRIS, etc.)
	$FromConsole = $False  # Default
	if ($null -eq $CallingScript) {
		$FromConsole = $True
	} elseif ($CallingScript -match $ScriptName) {  # Detect recursive calls
		$FromConsole = $True
	}

	# Get the current username and domain (hostname)
	# NOTE: sometimes needed for msg.exe (e.g. from PowerShell ISE)
	# ... tip: from the console, you can type 'hostname' or 'whoami'
	$my_servername = (whoami).Split('\')[0]
	$server_username = (whoami).Split('\')[1]

	# For debugging
	#msg.exe $server_username /TIME:10 "entered script '$ScriptName'"
	#msg.exe $server_username /TIME:10 "PsVersion ($ScriptName) = $PsVersion"
	#msg.exe $server_username /TIME:10 "ParentProcess ($ScriptName) = $ParentProcess"
	#msg.exe $server_username /TIME:10 "FromConsole ($ScriptName) = $FromConsole"
	#msg.exe $server_username /TIME:5 "From script... $ScriptName`n-----`nPowerShell version is... $PsVersion`nParent process is... '$ParentProcess'`nCalling script name... $CallingScript"

	# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
	if ($bg_msgs_on -and !$FromConsole) {
		$ExecPolicy = Get-ExecutionPolicy
		$exec_policy_str = "PSExecutionPolicy = $ExecPolicy"
		if ($env:PSExecutionPolicyPreference) {
			$ExecPolicyPref = $env:PSExecutionPolicyPreference
			$exec_policy_str = "Session ExecutionPolicy = $ExecPolicyPref"
		}
		$msg_sec = 10
		$bg_msg = "Script '$ScriptName' is running in the background" + 
		"`n$exec_policy_str" + 
		"`nPsVersion = $PsVersion" + 
		"`nParentProcess = $ParentProcess" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"

		msg.exe $server_username /TIME:$msg_sec $bg_msg
	}

	# FEEDBACK UNEXPECTED COMMANDLINE ARGUMENTS
	#Write-Host ("params = " + ($PSBoundParameters | ConvertTo-Json))
	$cmdline_args_passed = [String[]]$PSBoundParameters.Keys
	#Write-Host "Extra Args w/ values = $Args"
	$cmdline_args_bad = ($Args | where{$_ -like "-*"}) | ForEach-Object{$_ -Replace '-', ''}
	if ($cmdline_args_bad) {
		$cmdline_args_bad = [String[]]$cmdline_args_bad
	} else {
		$cmdline_args_bad = @()
	}

	# For debugging
	#Write-Host "Cmdline Args Passed = $cmdline_args_passed"
	#Write-Host "Cmdline Args Bad = $cmdline_args_bad"

	if ($cmdline_args_bad.Count -gt 0) {
		if ($FromConsole) {
			Write-Host ("HEADS-UP: commandline contains unsupported argument(s): [ " + ($cmdline_args_bad -Join ', ') + " ]") -ForegroundColor Green
			return
		}
	}
}

# For testing - Simulate background execution
# $FromConsole = $false

# IMPORT USER SETTINGS...
# Note: The 'user settings' script is expected to be in same folder as this script
$user_settings_filename = $ScriptName.Replace(".ps1", "_user_settings.ps1")
try {
	. $user_settings_filename
	
	# For debugging
	#Write-Host "Loaded user settings from '$user_settings_filename'"
}
catch {
	return "ERR/failed to load user settings file '$user_settings_filename'`nTip: try executing it directly to confirm it exists and reveal syntax errors"
}

# ===============================================

# FUNCTIONS
if (1 -eq 1) {  # Use to collapse this code section while editing

	function PopupBackgroundMsg() {
	# PURPOSE:  Show background message via msg.exe when script is not running from the console
	#           e.g., from a Blue Iris action set, or Windows Task Scheduler
	# WHY USE?  Mainly for debugging script while running in background.
	#           Parameter -InBackground simplifies using msg.exe w/o wrapping every call in a conditional
	# ARGS:     see param lines below
	# RETURNS:  nothing
	#
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V2 (2024/11/13) added debugging code; added test for null -Message and -UserName

		param(
			[String]$Message,   # [REQUIRED]
			[Int]$Seconds,      # [REQUIRED]
			[String]$UserName,  # [REQUIRED]
			[Boolean]$InBackground = $true
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		# For debugging
		# Write-Host "Entered function: $fn_name() ..."	
		# Write-Host "$fn_name() Message = $Message"	
		# Write-Host "$fn_name() UserName = $UserName"	
		# Write-Host "$fn_name() InBackground = $InBackground"
		
		if ($Message -and $UserName) {
			if ($InBackground) {
				msg.exe $UserName /TIME:$Seconds $Message
			}
		}
	}

	function BIJsonCall() {
	# PURPOSE:   To get a Blue Iris JSON interface response from the Blue Iris web server.
	#            Intended to be called by other PowerShell scripts.
	# ARGUMENTS: -JsonCmd  = (hashtable) like '@{cmd=xxx},@{cmd=xxx,camera=name},...'
	#            -ForceFlg = [OPTIONAL] (boolean) forces new login (gets a new authenticated session key)
	#            -RetryNo  = [OPTIONAL] (int) if >0, this call is a recursive retry attempt
	# RETURNS:   (object) PowerShell PSCustomObject
	#            .or. [String] "ERR/*"
	# GLOBALS:   $script:FromConsole, $script:Quiet 
	#            $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
	#            $script:Quiet       = (boolean) $true suppresses verbose output to the console
	# REFS:      1) Blue Iris Help PDF: chapter 'Administration', section 'JSON Interface'
	#            2) https://www.houselogix.com/docs/blue-iris/BlueIris/json.htm  (old!)
	#            3) https://ipcamtalk.com/threads/json-interface.21491/
	#            4) https://ipcamtalk.com/threads/json-help-please.40381/#post-372770
	# NOTES:
	# 1) This function saves the BI JSON Interface session key to a file for reuse.
	#    WHY? To avoid unnecessary new connections while an existing connection is still active
	#    (opening a new session is 10-100X slower than using an open session w/ saved session key).
	#    The file is stored in folder 'c:\temp\$ScriptBaseName\bi_session_key_{host}.txt'
	# 2) This function will recursively call itself if a JSON cmd's previous
	#     $json_response.result value is 'fail' and the reason is anything but 'Access denied'.
	#    The number of recursive attempts is limited to $max_retries.
	#    The script will wait a short time (100 msec) before each retry.
	# 3) In the event that the JSON call fails with reason 'Access denied' (Admin account required),
	#    the script will exit with an ERR
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V4 (8/29/2024) replaced *Flg args w/ global vars
	# V5 (11/18/2024) removed "(DB record not found)" from null result $err_msg
	# V6 (11/21/2024) added timers for comparing session key get methods (from file, or renewed); 
	#                 improved DevMode output to provide the user w/ more informative 'under-the-hood' feedback
	#                 changed 'x -eq $null' instances to '$null -eq x';  likewise for '-ne $null instances
	#                 changed type of param 'JsonCmd' from 'hashtable' to 'ordered'
	# V7 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]
	# V8 (2/3/2025) named unnamed timers
	
		param (
			[System.Collections.Specialized.OrderedDictionary]$JsonCmd,
			[Boolean]$ForceFlg,
			[Int]$RetryNo
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name() ..."	
		# Write-Host ("$fn_name() JsonCmd = " + ($JsonCmd | ConvertTo-Json))
		# Write-Host "$fn_name() ForceFlg = $ForceFlg"
		# Write-Host "$fn_name() RetryNo = $RetryNo"

		if ($script:FromConsole -and $script:DevMode) {
			Write-Host ""
			Write-Host ("-" * 40) -ForegroundColor Cyan
			Write-Host ("(dev mode) Entered function $fn_name()") -ForegroundColor Cyan
		}

		# HANDLE COMMANDLINE ARGS
		if ($null -eq $JsonCmd) {
			$err_msg = "ERR/expecting an argument: JsonCmd (PowerShell ordered hashtable)"
			return $err_msg
		}
		
		# For debugging
		# Write-Host ("requested JsonCmd type = " + ($JsonCmd.GetType()))
		# Write-Host ("modified JsonCmd type = " + ($json_cmd.GetType()))

		# System vars
		$ScriptName = Split-Path -leaf $PSCommandpath
		$ScriptBaseName = $ScriptName.Replace('.ps1', '')
		$ParentProcess = (Get-Process -Id (Get-CimInstance Win32_Process -Filter "ProcessID = $pid").ParentProcessId).Name
		$PsVersion = $PSVersionTable.PSVersion.Major
		# Get the current server domain (hostname) and username
		# ... used for background messaging via msg.exe
		# ... tip: from the console, you can type 'hostname' or 'whoami'
		$server_name = (whoami).Split('\')[0]
		$server_username = (whoami).Split('\')[1]

		# Function for encrypting the JSON cmd $response_hash
		function encryptMD5($str) {
			# PURPOSE: Get MD5 hash for a string.
			# ARGS:    $str = (String) string to be encrypted
			# RETURNS: (String) MD5 encrypted string... like, '0753a0e20c9027ccab70f4291a21aa08'
			# REF:     https://infosecscout.com/get-md5-hash-in-powershell/
			# NOTES:   (from REF)
			#	Unfortunately, PowerShell provides no direct function to generate a hash from a string.
			#	However, PowerShell offers a cmdlet to generate MD5 hash for a file (Get-FileHash)
			# 	This cmdlet can also be used to get the MD5 hash for a string, by opening a stream and hashing it.
			# USEFUL RESOURCE: https://www.md5hashgenerator.com/
			#   for example, try 'MD5string' ... should return '0753a0e20c9027ccab70f4291a21aa08'

			$stringAsStream = [System.IO.MemoryStream]::new()
			$writer = [System.IO.StreamWriter]::new($stringAsStream)
			$writer.write($str)
			$writer.Flush()
			$stringAsStream.Position = 0
			$hash_obj = Get-FileHash -InputStream $stringAsStream -Algorithm MD5
			
			# For debugging
			# $str = $hash_obj | Out-String
			# Write-Host "hash_obj = $str"

			return $hash_obj.Hash
		}
		
		# HANDLE COMMANDLINE ARGS
		# MAX RETRIES FOR JSON CMD
		# ... This function will recursively call itself if a JSON cmd's previous
		#     $json_response.result value is 'fail' and the reason is anything but 'Access denied'
		$max_retries = 3
		# ... Wait time before attempting a 'Invalid session' recursive call 
		$retry_delay_ms = 100
		
		if (!$RetryNo) {
			$retry_cntr = 0
		} else {
			$retry_cntr = $RetryNo
			
			if ($script:FromConsole -and $script:DevMode) {
				Write-Host ("(dev mode) This is retry #$retry_cntr for this JsonCmd") -ForegroundColor Cyan
			}
		}
		
		# For testing
		# $retry_cntr = $max_retries + 1

		# Abort if we've exceeded the retries 'Invalid session' break variable
		if ($retry_cntr -gt $max_retries) {
			$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' failed $max_retries retries; unable to open a valid session"
			return $err_msg
		}

		# HANDLE FILE USED TO SAVE THE SESSION KEY
		# ... NOTE: if a path is not specified, the default Windows location is 'C:\Users\<username>"
		$path = "C:\temp\$ScriptBaseName"
		if (!(Test-Path -LiteralPath $path)) {
			New-Item -ItemType Directory -Force -Path $path
		}
		# ... NOTE: we use a different filename when running the script via different processes;
		#     this avoids collisions when simultaneously running scripts via:
		#     'scvhost' (Task Scheduler), 'WindowsTerminal', 'explorer' (Powershell ISE)
		$base_name = "$path\bi_session_key"
		$session_filespec = $base_name + "_" + $ParentProcess + ".txt"
		$found_saved_session_file_flg = (Test-Path -LiteralPath $session_filespec)

		# HANDLE COMMANDLINE ARGS
		# Override reuse of saved session_key
		if ($ForceFlg) {
			$found_saved_session_file_flg = $false
		}

		# ========================
		# PERFORM API CALLS

		# Make the URL
		$uri = "http://" + $script:bi_webserver + "/json";

		# For debugging
		# Write-Host "$fn_name() uri = $uri"

		# Flag 'logout' JSON cmd requests
		# ... NOTE: All JSON 'logout' cmds should be performed by the calling script!
		if ($JsonCmd.cmd -eq 'logout') {  # REMINDER: Powershell string comparisons are case-insensitive by default
			$logging_out_flg = $true
		} else {
			$logging_out_flg = $false
		}
		
		# For debugging
		if ($logging_out_flg) {
			# Write-Host "$fn_name() detected JSON cmd to log out of BI API session"
		}

		# SPECIAL HANDLING FOR 'geofence' JSON COMMAND REQUESTS
		$dev_uuid = $null  # Default
		$json_cmd_keys = $JsonCmd.Keys
		if ($json_cmd_keys -contains 'uuid') {
			if ($null -ne $JsonCmd.uuid) {
				$dev_uuid = $JsonCmd.uuid
			}
		}

		# For debugging
		# Write-Host "$fn_name() dev_uuid = $dev_uuid"

		# EXECUTE API CALL

		# Start performance timer
		# Note: typically takes 10-20+ msec if NOT using a saved session key (creating it anew on every call)
		#                       1+ msec if using a saved session key
		#                       5-10 msec if refreshing an expired saved session key
		$timer_json_auth_connection = [Diagnostics.Stopwatch]::StartNew()
		
		if (!$found_saved_session_file_flg) {

			# Start performance timer
			# Note: typically takes 5-20 msec
			$timer_get_auth_session_key_1 = [Diagnostics.Stopwatch]::StartNew()
			
			# ================================================================
			# 1.1 PRELIMINARY LOGIN
			
			# Create, then encode the JSON 'login' command
			$parameters = @{
				cmd = 'login'
			}
			$json_params = $parameters | ConvertTo-Json

			# For debugging
			# Write-host "$fn_name() (1.1) login #1 json_params... `n$json_params"
			# msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 json_params... `n$json_params"

			# Start performance timer
			# $timer_login_prelimary = [Diagnostics.Stopwatch]::StartNew()

			# EXECUTE THE JSON REQUEST
			# Returns a PSCustomObject
			$json_response_obj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

			# Stop timer
			# $timer_login_prelimary.Stop()
			# $elapsed_msec = [Math]::Round($timer_login_prelimary.elapsed.totalmilliseconds,1)
			# $elapsed_msec_str = $elapsed_msec.ToString("F3")
			# Write-Host ("$fn_name() json request timer (1.1) = $elapsed_msec_str msec")

			# For debugging
			# Write-Host ("$fn_name() (1.1) jsonResponse... `n" + ($json_response_obj  | ConvertTo-Json))
			# msg.exe $server_username /TIME:15 "$fn_name() (1.1) login #1 jsonResponse... `n$json_response_obj"
			
			# Extract the session key
			$session = $json_response_obj.session
			
			# ================================================================
			# 1.2 AUTHENTICATE THE CONNECTION

			# Create MD5 hash-encoded credentials string
			$response_str = $bi_user_name + ":" + $session + ":" + $bi_user_pw
			$response_hash = encryptMD5 $response_str

			# Create, then encode a modified JSON 'login' command
			if ($dev_uuid) {
				$parameters = @{
					cmd      = 'login'
					session  = $session
					response = $response_hash
					uuid     = $dev_uuid
				}
			} else {
				$parameters = @{
					cmd      = 'login'
					session  = $session
					response = $response_hash
				}
			}
			$json_params = $parameters | ConvertTo-Json

			# For debugging
			# Write-Host "$fn_name() (1.2A) login #2 (authorizing json login) json_params... `n$json_params"
			# msg.exe $server_username /TIME:15 "$fn_name() (1.2A) authorizing json login ... `n$json_params"
			
			# EXECUTE THE JSON REQUEST
			# Returns a PSCustomObject
			$json_response_obj = $json_params | Invoke-RestMethod -Uri $uri -Method Post

			# For debugging
			# Write-Host ("$fn_name() (1.2A) json_response_obj.result = " + $json_response_obj.result)
			# Write-Host ("$fn_name() (1.2A) jsonResponse... `n" + ($json_response_obj  | ConvertTo-Json))
			# msg.exe $server_username /TIME:15 "$fn_name() (1.2A) jsonResponse... `n$json_response_obj"
			
			# Handle failure
			if ($json_response_obj.result -eq 'fail') {
				$err_msg = "ERR/unable to authenticate Blue Iris JSON interface"
				return $err_msg
			}
			
			# Extract the authenticated session key
			$auth_session_key = $json_response_obj.session

			# Stop timer
			$timer_get_auth_session_key_1.Stop()
			
			if ($script:FromConsole -and $script:DevMode) {
				$elapsed_msec = [Math]::Round($timer_get_auth_session_key_1.elapsed.totalmilliseconds,1)
				$elapsed_msec_str = $elapsed_msec.ToString("F3")
				Write-Host ("(dev mode) time to renew session_key = $elapsed_msec_str msec") -ForegroundColor Cyan
			}

			# For debugging
			# Write-Host "$fn_name() (1.2A) login #2 auth_session_key = '$auth_session_key'"
			#msg.exe $server_username /TIME:5 "$fn_name() auth_session_key... $auth_session_key"
		
			# Save the authenticated session key to disk file
			# NOTE: THE SESSION ID IS ALWAYS SAVED, WHETHER USED OR NOT
			$auth_session_key | Set-Content "$session_filespec" -NoNewline

		} else {

			# Start performance timer
			# Note: typically takes 0.5-0.6 msec
			$timer_get_auth_session_key_2 = [Diagnostics.Stopwatch]::StartNew()
		
			# Retrieve the saved session key from a disk file
			# Note: This typically takes only ~0.5 msec
			# KEY POINT: 
			#   Force typing the output of 'Get-Content' to [String] was essential when running script in
			#   in Powershell ISE. Otherwise the output was similar to an object.
			#   *** But in actuality it is more complicated ***
			#   Per https://stackoverflow.com/a/55881978/13499984
			#      When 'Get-Content' is assigned to a variable, its output is an array of .NET [String] objects, 
			#      each representing a line from the input file. However, Get-Content also decorates these string 
			#      objects with additional properties that provide helpful metadata about where the strings came from.

			$saved_session_key = [String](Get-Content -Path "$session_filespec")

			$auth_session_key = $saved_session_key

			# For debugging
			#if (!$logging_out_flg) {
				# Write-Host ("$fn_name() (1.2B) saved_session_key type = " + $saved_session_key.GetType())
				# Write-Host "$fn_name() (1.2B) saved_session_key... $saved_session_key"
				# msg.exe * /TIME:5 "$fn_name() (1.2B) saved_session_key... $saved_session_key"
			#}

			# Stop timer
			$timer_get_auth_session_key_2.Stop()
			
			if ($script:FromConsole -and $script:DevMode) {
				$elapsed_msec = [Math]::Round($timer_get_auth_session_key_2.elapsed.totalmilliseconds,1)
				$elapsed_msec_str = $elapsed_msec.ToString("F3")
				Write-Host ("(dev mode) time to get `"session`" key from file = $elapsed_msec_str msec") -ForegroundColor Cyan
			}
			
		}

		# Stop timer
		$timer_json_auth_connection.Stop()
		
		if ($script:FromConsole -and $script:DevMode) {
			$elapsed_msec = [Math]::Round($timer_json_auth_connection.elapsed.totalmilliseconds,1)
			$elapsed_msec_str = $elapsed_msec.ToString("F3")
			Write-Host ("(dev mode) time to attempt connection authentication = $elapsed_msec_str msec") -ForegroundColor Cyan
		}

		# ================================================================
		# 1.3 EXECUTE THE REQUESTED JSON COMMAND
			
		# In the event that this JSON call fails with reason 'Invalid session',
		# the script will automatically attempt up to $max_retries

		if ($null -eq $dev_uuid) {
		
			# Modify, then encode the requested JSON login command
			$JsonCmd.session = $auth_session_key
			$json_params = $JsonCmd | ConvertTo-Json

			if ($script:FromConsole -and $script:DevMode) {
				Write-Host ("(dev mode) JsonCmd = $json_params") -Foregroundcolor Cyan
			}

			# EXECUTE THE JSON REQUEST

			# 1. Start performance timer
			$timer_json_cmd = [Diagnostics.Stopwatch]::StartNew()

			# 2. Do it!
			$params = @{
				Uri         = $uri
				#Headers     = @{"Content-Type"  = "application/json"}
				Method      = 'POST'
				Body        = $json_params
				ContentType = 'application/json'
			}
			# Returns a PSCustomObject
			$json_response_obj = Invoke-RestMethod @params
			# Alternate syntaxes (pick one)
			#$json_response_obj = Invoke-RestMethod -Uri $uri -Method Post -Body ($JsonCmd | ConvertTo-Json) #>
			#$json_response_obj = ($JsonCmd | ConvertTo-Json) | Invoke-RestMethod -Uri $uri -Method Post

			# 3. Stop timer
			# ... NOTE: JSON cmd execution time is displayed later - after outputting the cmd's result status
			$timer_json_cmd.Stop()

			# For debugging
			#if ($JsonCmd.cmd -match 'update') {
				# Write-Host ("$fn_name() (1.3) requested JsonCmd '" + $JsonCmd.cmd + "'")
				# Write-Host ("$fn_name() (1.3) json_response_obj type = " + $json_response_obj.GetType())
				# Write-Host ("$fn_name() (1.3) json_response_obj cnt = " + $json_response_obj.Count)
				# Write-Host ("$fn_name() (1.3) json_response_obj.result = " + $json_response_obj.result)
				# Write-Host ("$fn_name() (1.3) JsonCmd '" + $JsonCmd.cmd + "' response = `n" + ($json_response_obj | ConvertTo-Json))
				# Write-Host ("$fn_name() (1.3) json_response_obj = `n" + ($json_response_obj | ConvertTo-Json))
			#}
			
			# Test for issues
			if (1 -eq 1) {  # Use to collapse this code section while editing
				
				$retry_flg = $false

				if ($null -eq $json_response_obj) {
					$retry_flg = $true 
					
					if ($script:FromConsole -and $script:DevMode) {
						Write-Host ("(dev mode) JsonCmd response is null") -ForegroundColor Cyan
					}
				}

				if ($json_response_obj.Count -eq 0) {
					$retry_flg = $true 
					
					if ($script:FromConsole -and $script:DevMode) {
						Write-Host ("(dev mode) JsonCmd response is empty") -ForegroundColor Cyan
					}				
				}

				if ($json_response_obj -is [String]) {
					$retry_flg = $true 
					
					if ($script:FromConsole -and $script:DevMode) {
						if ($json_response_obj.StartsWith('ERR/')) {
							Write-Host ("(dev mode) JsonCmd ERR; response: '$json_response_obj'") -ForegroundColor Cyan
						} else {
							Write-Host ("(dev mode) unexpected JsonCmd response: '$json_response_obj'") -ForegroundColor Cyan
						}
					}
				}				

				if ($json_response_obj.result -eq 'fail') {
					
					if ($script:FromConsole -and $script:DevMode) {
						$pfx = "(dev mode) JsonCmd response is `"fail`""
					}

					# For debugging
					# Write-Host ("$fn_name() failed json_response_obj = `n" + ($json_response_obj | ConvertTo-Json))
					
					# Try to determine reason
					if ($json_response_obj.data) {

						$data_obj = $json_response_obj.data

						# For debugging
						# Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data = " + ($data_obj | ConvertTo-Json))
						
						if ($data_obj.reason) {
							$reason = $data_obj.reason
						
							# For debugging
							# Write-Host ("$fn_name() JsonCmd '" + $JsonCmd.cmd + "' response.data.reason = $reason")
							
							# For testing
							# $reason = 'Invalid session'
							# $reason = 'Access denied'

							if ($script:FromConsole -and $script:DevMode) {
								Write-Host ($pfx + "; reason: '$reason'") -ForegroundColor Cyan
							}

							if ($reason -match 'Invalid session') {
								$retry_flg = $true
							} elseif ($reason -match 'Access denied') {
								$err_msg ="ERR/JsonCmd '" + $JsonCmd.cmd + "' requires an Administrator account"
								return $err_msg							
							} else {  # Unknown failed reason (just in case)
								$err_msg = "ERR/JsonCmd '" + $JsonCmd.cmd + "' failed with reason '$reason'"
								return $err_msg
							}
							
						} else {  # No reason identified (is null)
							$retry_flg = $true
						}
					} else {  # No reason available
						$retry_flg = $true
					
						if ($script:FromConsole -and $script:DevMode) {
							Write-Host ($pfx + "; no reason found") -ForegroundColor Cyan
							Write-Host ($json_response_obj | ConvertTo-Json) -ForegroundColor Cyan
						}
					}
				} else {
					if ($script:FromConsole -and $script:DevMode) {
						Write-Host ("(dev mode) JsonCmd response is `"" + $json_response_obj.result + "`"") -ForegroundColor Cyan
					}
				}

			}

			if ($script:FromConsole -and $script:DevMode) {
				$elapsed_msec = [Math]::Round($timer_json_cmd.elapsed.totalmilliseconds,1)
				$elapsed_msec_str = $elapsed_msec.ToString("F3")
				Write-Host ("(dev mode) time to execute json cmd `"" + $JsonCmd.cmd + "`" = $elapsed_msec_str msec") -ForegroundColor Cyan
			}
			
			# For debugging
			# Write-Host "retry_flg = $retry_flg"

			if ($retry_flg) {
							
				# Update retry counter
				$retry_cntr++
		
				if ($script:FromConsole -and $script:DevMode) {
					Write-Host ("(dev mode) RETRYING JsonCmd; attempt #$retry_cntr ($max_retries max)") -ForegroundColor Cyan
				}

				# RECURSIVE CALL							
				# Returns a powershell PSCustomObject (or string 'ERR/*')
				$json_response_obj = `
					BIJsonCall `
					-JsonCmd $JsonCmd `
					-ForceFlg $true `
					-RetryNo $retry_cntr
						
			}
		}

		# For debugging
		if ($json_response_obj) {
			# Write-Host ("$fn_name() (1.3) jsonResponse type = " + $json_response_obj.GetType())
			# Write-Host ("$fn_name() (1.3) jsonResponse... `n" + ($json_response_obj  | ConvertTo-Json))
			# msg.exe $server_username /TIME:5 "$fn_name() (1.3) jsonResponse... $json_response_obj"
		} 

		# For  testing
		# $json_response_obj = $null
		# $json_response_obj = ''

		return $json_response_obj

	}

	function RegGetValue() {
	# PURPOSE:  Get current value of a registry entry
	# ARGS:     see param lines below
	# RETURNS:  registry value, user-provided default value, or error message
	#
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V2 (2024/07/01)

		param(
			[String]$KeyPath,  # like, 'HKLM:SOFTWARE\Perspective Software\Blue Iris\clips'
			[String]$KeyName,  # like, 'compact'
			$DefaultValue      # value to return if $KeyName does not exist 
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		# For debugging
		#Write-Host "Entered $fn_name() ..."
		#Write-Host "$fn_name() -KeyPath = '$KeyPath'"
		#Write-Host "$fn_name() -KeyName = '$KeyName'"
		#Write-Host "$fn_name() -DefaultValue = '$DefaultValue'"
		
		try {
			$reg_obj = Get-ItemProperty -Path $KeyPath -Name $KeyName -ErrorAction Stop
		
			# For debugging
			#Write-Host ("reg_obj = " + ($reg_obj | ConvertTo-Json -Depth 100))
			
			if ($reg_obj) {
				if (!($null -eq $reg_obj.$KeyName)) {
					return $reg_obj.$KeyName
				}
			}
		}
		catch {
			# continue
		}	

		if ($DefaultValue) {
			$rtn_value = $DefaultValue
		} else {
			$err_msg = "ERR/unable to get registry key '$KeyPath\$KeyName'"
			$rtn_value = $err_msg
		}
		
		return $rtn_value
	}

	function GetBiWebserverIpPort() {
	# PURPOSE:   Get the Blue Iris web server's LAN IP and Port
	# ARGUMENTS: None           
	# RETURNS:   string, like "192.168.1.3:81" 
	#            or $null if failure
	# REQUIRES:  function RegGetValue()
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (9/30/2024)
	# V2 (11/15/2024) Changed failure return value to $null
	# V3 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# Write-Host "Entered $fn_name()..."

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_val = $null  # Assume failure
		
		# 1) GET LAN IP VALUE
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\server'
		$key_name = 'lanip'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$server_lan_ip = $reg_get_value
				}
			} else {
				$server_lan_ip = $reg_get_value
			}
		}
		
		# 2) GET WEB SERVER PORT VALUE
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris\server'
		$key_name = 'port'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$server_port = $reg_get_value
					# Build the IP & Port address
					$rtn_val = $server_lan_ip + ":" + $server_port
				}
			} else {
				$server_port = $reg_get_value
				# Build the IP & Port address
				$rtn_val = $server_lan_ip + ":" + $server_port
			}
		}

		$server_port = $reg_get_value

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"

		return $rtn_val
	}

	function GetBiVersion() {
	# PURPOSE:  Get Blue Iris application's current software version
	# ARGS:     None
	# RETURNS:  string, like "5.1.2.9"
	#           or $null if failure
	# REQUIRES: RegGetValue(), Convert-HexToVersion()
	# GLOBALS:  Uses global vars: $FromConsole
	#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1
	# V2 (11/15/2024) Changed failure return value to $null
	# V3 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]
	
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name()..."
		# Write-Host "$fn_name() FromConsole = $script:FromConsole"

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_val = $null  # Assume failure

		# GET CURRENT VERSION
		
		# Get current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$key_path = 'HKLM:\SOFTWARE\Perspective Software\Blue Iris'
		$key_name = 'version'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name

		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"

		# Abort if GET failed
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					# Get the current version string
					$bi_version_hex = $reg_get_value.ToString("X")  # like hex 5090704
					$rtn_val = Convert-HexToVersion -Hex $bi_version_hex  # like '5.9.7.4'
				}
			} else {
				# Get the current version string
				$bi_version_hex = $reg_get_value.ToString("X")  # like hex 5090704
				$rtn_val = Convert-HexToVersion -Hex $bi_version_hex  # like '5.9.7.4'
			}
		}

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"

		return $rtn_val
	}

	function IsBiVersion() {
	# PURPOSE:   To confirm if the CURRENT Blue Iris version exceeds a specified UPDATE version
	# ARGUMENTS: -CompareToVersion = [REQUIRED] (String) compare version, '5.7.2.3'
	#            -CurrentVersion   = [OPTIONAL] (String) current version
	#            Note: supports version strings 'X.000.000.000' to 'X.999.999.999'
	#            Note: Using arg -CurrentVersion improves performance if function is called many times via a loop
	# RETURNS:   boolean; $true current BI version exceeds specified version
	# REQUIRES:  function GetBiVersion()
	#
	# By jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# VERSION:   2
		
		param (
			[String]$CompareToVersion,
			[String]$CurrentVersion
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# Write-Host "Entered $fn_name()..."
		# Write-Host "$fn_name() CompareToVersion = $CompareToVersion"
		# Write-Host "$fn_name() CurrentVersion = $CurrentVersion"
		# Write-Host "$fn_name() FromConsole = $script:FromConsole"

		# Start timer
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_flg = $false  # Assume failure

		if (!$CurrentVersion) {
			
			# GET CURRENT VERSION
			$bi_version = GetBiVersion

			# Abort if GET failed
			if ($null -eq $bi_version) {
				if ($script:FromConsole) {
					Write-Host "$fn_name() ERR/REG request failed"
				}
				return $false
			}
			
		} else {
			$bi_version = $CurrentVersion
		}
		
		# Make comparator integer arrays
		[int[]]$my_ver_ary = $CompareToVersion.Split('.')
		[int[]]$bi_ver_ary = $bi_version.Split('.')

		# Make 'prime.subprime' version comparators, like 5.8 vs 5.7
		$my_ver_prime_dot = $my_ver_ary[0] + $my_ver_ary[1]/1000
		$bi_ver_prime_dot = $bi_ver_ary[0] + $bi_ver_ary[1]/1000

		# For debugging
		# Write-Host "bi_version_hex (registry) = $bi_version_hex"
		# Write-Host "bi_version = $bi_version"
		# Write-Host ("my_ver_ary = " + ($my_ver_ary | ConvertTo-Json))
		# Write-Host ("bi_version_ary = " + ($bi_ver_ary | ConvertTo-Json))
		# Write-Host "my_ver_prime_dot = $my_ver_prime_dot"
		# Write-Host "bi_ver_prime_dot = $bi_ver_prime_dot"
		
		if ($my_ver_prime_dot -eq $bi_ver_prime_dot) {

			# Make sub-subprime version comparators, 5.001 for 5.8.5.1, 999.020 for 5.8.999.20
					
			$mult1 = [Math]::Pow(10, 0)
			$mult2 = [Math]::Pow(10, -3)

			if ($my_ver_ary[2] -eq 0 -and $my_ver_ary[3] -eq 0) {
				$my_ver_sub = 0.000
			} else {
				$my_ver_sub = $my_ver_ary[2]*$mult1 + $my_ver_ary[3]*$mult2
			}

			if ($bi_ver_ary[2] -eq 0 -and $bi_ver_ary[3] -eq 0) {
				$bi_ver_sub = 0.000
			} else {
				$bi_ver_sub = $bi_ver_ary[2]*$mult1 + $bi_ver_ary[3]*$mult2
			}

			# For debugging
			# Write-Host ("my_ver_sub = " + ("{0:n3}" -f $my_ver_sub))
			# Write-Host ("bi_ver_sub = " + ("{0:n3}" -f $bi_ver_sub))

			$rtn_flg = ($bi_ver_sub -ge $my_ver_sub) 


		} else {
			$rtn_flg = ($bi_ver_prime_dot -ge $my_ver_prime_dot)
		}

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer ('$CompareToVersion') = $elapsed_msec_str msec")

		return $rtn_flg
	}

	function GetBiTempFolder() {
	# PURPOSE:  Get the 'temp' folder path from the Windows registry
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Options\temppath
	# ARGS:     none
	# RETURNS:  string, like "C:\ProgramData\Blue Iris\temp"
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (2/17/2025)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."

		# For debugging
		# $timer = [Diagnostics.Stopwatch]::StartNew()

		$rtn_val = $null  # Assume failure

		# GET 'NEW' FOLDER PATH
		# ... KEY POINT: User may create additional folder via Camera settings > Record tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\options"
		$key_name = "temppath"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rtn_val = $reg_get_value
				}
			} else {
				$rtn_val = $reg_get_value
			}
		}

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetBiNewFolder() {
	# PURPOSE:  Get the 'New' folder path from the Windows registry
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\0
	# ARGS:     none
	# RETURNS:  string, like "D:\BlueIris\New"
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (11/15/2024)
	# V2 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."

		# For debugging
		# $timer = [Diagnostics.Stopwatch]::StartNew()

		$rtn_val = $null  # Assume failure

		# GET 'NEW' FOLDER PATH
		# ... KEY POINT: User may create additional folder via Camera settings > Record tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\0"
		$key_name = "path"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rtn_val = $reg_get_value
				}
			} else {
				$rtn_val = $reg_get_value
			}
		}

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetBiAlertsFolder() {
	# PURPOSE:  Get the 'Alerts' folder path from the Windows registry
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\2
	# ARGS:     none
	# RETURNS:  string, like "D:\BlueIris\Alerts"
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (11/15/2024)
	# V2 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."

		# For debugging
		# $timer = [Diagnostics.Stopwatch]::StartNew()

		$rtn_val = $null  # Assume failure

		# GET ALERT FOLDER PATH
		# ... From the Camera settings > Record tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\2"
		$key_name = "path"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rtn_val = $reg_get_value
				}
			} else {
				$rtn_val = $reg_get_value
			}
		}

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetBiStorageFolders() {
	# PURPOSE:  Get the 'Alerts' folder path from the Windows registry
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\2
	# ARGS:     none
	# RETURNS:  ordered hashtable, like 
	#           @{ 'New'='D:\BlueIris\New','Storage'='D:\BlueIris\Stored','Alerts'='D:\BlueIris\New','Aux 1'='D:\BlueIris\Aux1'...)
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# NOTES:    This function takes 100-150 msec to execute; if you need only a single folder, then it would be
	#           more efficient to use instead functions like GetBiAlertsFolder() or GetBiNewFolder()
	#           The reason for this inefficiency is that the function calls RegGetValue() 32X
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (11/16/2024)
	# V2 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."

		# For debugging
		# $timer = [Diagnostics.Stopwatch]::StartNew()
		
		$rtn_val = $null  # Assume failure

		$reg_bi_folders_ht = [ordered]@{}
		$err_msgs = @()
		for ($i = 0; $i -le 15; $i++) {
			
			$success_flg = $true  # assume success
			
			# GET BLUE IRIS FOLDER NAME
			# ... From the Camera settings > Record tab
			$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\$i"
			$key_name = "name"

			# Get regkey's current value
			# ... returns value, $null, or string 'ERR/*'
			$reg_get_value = `
				RegGetValue `
				-KeyPath $key_path `
				-KeyName $key_name
				
			# For debugging
			# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
			# Write-Host "$fn_name() reg_get_value = $reg_get_value"
			
			if ($null -eq $reg_get_value) {
				$err_msgs += "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
				$success_flg = $false
			} else {
				if ($reg_get_value -is [String]) {
					if ($reg_get_value.StartsWith('ERR/')) {
						$err_msgs += $reg_get_value
						$success_flg = $false
					} else {
						$key = $reg_get_value
					}
				} else {
					$key = $reg_get_value
				}
			}
			
			if ($success_flg) {
				
				# GET BLUE IRIS FOLDER PATH
				# ... From the Camera settings > Record tab
				$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\clips\folders\$i"
				$key_name = "path"

				# Get regkey's current value
				# ... returns value, $null, or string 'ERR/*'
				$reg_get_value = `
					RegGetValue `
					-KeyPath $key_path `
					-KeyName $key_name
				
				if ($null -eq $reg_get_value) {
					$reg_bi_folders_ht.Add($key, '')
				} else {
					if ($reg_get_value -is [String]) {
						if ($reg_get_value.StartsWith('ERR/')) {
							$reg_bi_folders_ht.Add($key, '')
						} else {
							$reg_bi_folders_ht.Add($key, $reg_get_value)
						}
					} else {
						$reg_bi_folders_ht.Add($key, $reg_get_value)
					}
				}
			}
			
		}
		
		# For debugging
		# Write-Host ("reg_bi_folders_ht keys = " + $reg_bi_folders_ht.Keys)
		
		if ($reg_bi_folders_ht.Keys.Count -eq 16 -and $err_msgs.Count -eq 0) {
			$rtn_val = $reg_bi_folders_ht
		} else {
			if ($script:FromConsole) {
				Write-Warning "unable to retrieve all Blue Iris storage folders"
				Write-Host ($err_msgs | ConvertTo-Json) -ForegroundColor Yellow
			}
		}

		# Stop timer
		# $timer.Stop()
		# $elapsed_msec = [Math]::Round($timer.elapsed.totalmilliseconds,1)
		# $elapsed_msec_str = $elapsed_msec.ToString("F2")
		# Write-Host ("$fn_name() timer = $elapsed_msec_str msec")

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetPreTrigRecMsec() {
	# PURPOSE:  Get Pre-trigger record time registry entry for specified camera and profile
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\FrontDoor\Clips\1\rectime
	# ARGS:     see param lines below
	# RETURNS:  integer, like 2000
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (4/23/2024)
	# V2 (11/15/2024) Changed failure return value to $null
	# V3 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		param(
			[String]$CamLongName,   # Blue Iris camera longname
			[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() CamLongName = '$CamLongName'"
		# Write-Host "$fn_name() CamProfileNo = $CamProfileNo"

		if (!$CamLongName) {
			$err_msg = "ERR/$fn_name() requires camera shortname and longname"
			return $err_msg
		}
		
		$rtn_val = $null  # Assume failure

		# GET CAMERA's 'Pre-trigger playback time'
		# ... From the Camera settings > Record tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Clips\" + $CamProfileNo
		$key_name = "rectime"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rec_time = $reg_get_value  # tenths of a second
					$rtn_val = 1000 * ($rec_time / 10 )
				}
			} else {
				$rec_time = $reg_get_value  # tenths of a second
				$rtn_val = 1000 * ($rec_time / 10 )
			}
		}

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetPreTrigPlaybackMsec() {
	# PURPOSE:  Get Pre-trigger playback time registry entry for specified camera and profile
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\FrontDoor\Clips\1\playtime
	# ARGS:     see param lines below
	# RETURNS:  integer, like 2000
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (4/23/2024)
	# V2 (11/15/2024) Changed failure return value to $null
	# V3 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		param(
			[String]$CamLongName,   # Blue Iris camera longname
			[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() CamLongName = '$CamLongName'"
		# Write-Host "$fn_name() CamProfileNo = $CamProfileNo"

		if (!$CamLongName) {
			$err_msg = "ERR/$fn_name() requires camera longname"
			return $err_msg
		}
		
		$rtn_val = $null  # Assume failure

		# GET CAMERA's 'Pre-trigger playback time'
		# ... From the Camera settings > Record tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Clips\" + $CamProfileNo
		$key_name = "playtime"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$playback_time = $reg_get_value  # tenths of a second
					$rtn_val = 1000 * ($playback_time / 10 )
				}
			} else {
				$playback_time = $reg_get_value  # tenths of a second
				$rtn_val = 1000 * ($playback_time / 10 )
			}
		}

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetConfirmAlertWithAI() {
	# PURPOSE:  Get 'confirm alert with AI' entry for specified camera and profile
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\CamX\Alerts\ProfX\smart
	# ARGS:     see param lines below
	# RETURNS:  integer, like 2
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2025) jaydeel
	# 
	# V1 (2/18/2025)

		param(
			[String]$CamLongName,   # Blue Iris camera longname
			[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() CamLongName = '$CamLongName'"
		# Write-Host "$fn_name() CamProfileNo = $CamProfileNo"

		if (!$CamLongName) {
			$err_msg = "ERR/$fn_name() requires camera longname"
			return $err_msg
		}
		
		$rtn_val = $null  # Assume failure

		# GET CAMERA's 'Pre-trigger playback time'
		# ... From the Camera settings > Alerts tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Alerts\" + $CamProfileNo
		$key_name = "smart"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rtn_val = ($reg_get_value -eq 2)
				}
			} else {
				$rtn_val = ($reg_get_value -eq 2)
			}
		}

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetAiAnalyzeOneEachMsec() {
	# PURPOSE:  Get AI configuration "Analyze one image each" entry for specified camera and profile
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\FrontDoor\Alerts\1\smartextraint4
	#           Expected values are: (0=50ms, 1=100ms, 2=200ms, 3=250ms, 4=500ms, 5=750ms, 6=1s, 7=2s, 8=3s, 9=4s, 10=5s)
	#           KEY POINT: Applies to Blue Iris v5.8.0+
	# ARGS:     see param lines below
	# RETURNS:  integer, like 500
	#           or $null if failure
	# REQUIRES: RegGetValue(), IsBiVersion()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	# 
	# V1 (11/15/2024)
	# V2 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		param(
			[String]$CamLongName,   # Blue Iris camera longname
			[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() -CamLongName = '$CamLongName'"
		# Write-Host "$fn_name() -CamProfileNo = $CamProfileNo"

		if (!$CamLongName) {
			$err_msg = "ERR/$fn_name() requires camera shortname and longname"
			return $err_msg
		}
		
		if (!(IsBiVersion "5.8.0")) {
			$err_msg = "ERR/$fn_name() requires Blue Iris version 5.8.0+"
			return $err_msg
		}

		$rtn_val = $null  # Assume failure
		
		# (0=50ms, 1=100ms, 2=200ms, 3=250ms, 4=333ms, 5=500ms, 6=750ms, 7=1s, 8=2s, 9=3s, 10=4s, 11=5s, 12='Prog')
		$lookup_obj = @{
			'0' = 50
			'1' = 100
			'2' = 200
			'3' = 250
			'4' = 333
			'5' = 500
			'6' = 750
			'7' = 1000
			'8' = 2000
			'9' = 3000
			'10' = 4000
			'11' = 5000
			'12' = 'Prog'
		}

		# GET CAMERA's AI configuration "Analyze one image each" entry
		# ... From the Camera settings > Alert tab > AI configuration button dialog
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Alerts\" + $CamProfileNo
		$key_name = "smartextraint4"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rtn_val = $lookup_obj[$reg_get_value.ToString()]
				}
			} else {
				$rtn_val = $lookup_obj[$reg_get_value.ToString()]
			}
		}

		# For debugging
		#Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function GetAiSaveDatFile() {
	# PURPOSE:  Get 'Save AI analysis .DAT files' entry for specified camera and profile
	#           like, HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\CamX\Alerts\ProfX\smartdetails
	# ARGS:     see param lines below
	# RETURNS:  [integer] 0 or 1
	#           or $null if failure
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2025) jaydeel
	# 
	# V1 (2/18/2025)

		param(
			[String]$CamLongName,   # Blue Iris camera longname
			[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() CamLongName = '$CamLongName'"
		# Write-Host "$fn_name() CamProfileNo = $CamProfileNo"

		if (!$CamLongName) {
			$err_msg = "ERR/$fn_name() requires camera longname"
			return $err_msg
		}
		
		$rtn_val = $null  # Assume failure

		# GET CAMERA's 'Save AI analysis .DAT files'
		# ... From the Camera settings > Alerts tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Alerts\" + $CamProfileNo
		$key_name = "smartdetails"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		# Write-Host ("$fn_name() reg_get_value type = " + $reg_get_value.GetType())
		# Write-Host "$fn_name() reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				} else {
					$rtn_val = $reg_get_value
				}
			} else {
				$rtn_val = $reg_get_value
			}
		}

		# For debugging
		# Write-Host "$fn_name() rtn_val = $rtn_val"
		
		return $rtn_val
		
	}

	function IsCamRecContinuous() {
	# PURPOSE:  Get video recording mode registry entry for for specified camera and profile
	#           like, HKLM\SOFTWARE\Perspective Software\Blue Iris\Cameras\DriveWay-1\Clips\5\continuous
	# ARGS:     see param lines below
	# RETURNS:  boolean, or error message string like "ERR/..."
	# REQUIRES: RegGetValue()
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1 (9/7/2024)
	# V2 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]

		param(
			[String]$CamLongName,   # Blue Iris camera longname
			[String]$CamProfileNo   # Blue Iris camera profile (1-7); caller should validate
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		#Write-Host "Entered $fn_name() ..."
		#Write-Host "$fn_name() -CamLongName = '$CamLongName'"
		#Write-Host "$fn_name() -CamProfileNo = $CamProfileNo"

		if (!$CamLongName) {
			$err_msg = "ERR/$fn_name() requires camera shortname and longname"
			return $err_msg
		}

		# GET CAMERA's 'Pre-trigger playback time'
		# ... From the Camera settings > Record tab
		$key_path = "HKLM:\SOFTWARE\Perspective Software\Blue Iris\Cameras\" + $CamLongName + "\Clips\" + $CamProfileNo
		$key_name = "continuous"

		# Get regkey's  current setting's value
		# ... returns value, $null, or string 'ERR/*'
		$reg_get_value = `
			RegGetValue `
			-KeyPath $key_path `
			-KeyName $key_name
			
		# For debugging
		#Write-Host "reg_get_value = $reg_get_value"
		
		if ($null -eq $reg_get_value) {
			$err_msg = "ERR/$fn_name() unable to GET registry key value '$key_path\$key_name"
			return $err_msg
		} else {
			if ($reg_get_value -is [String]) {
				if ($reg_get_value.StartsWith('ERR/')) {
					return $reg_get_value
				}
			}
		}

		# Expecting: 0=When triggered, 1=Continuous, 2=Off, 3=Periodic, each, 4=Triggered+periodic: each, 5=Continuous+Alerts, 6=When alerted, 7=Continuous+Triggered)
		$val = [Int]$reg_get_value
		$rtn_flg = (@(2,5,7) -contains $val)

		# For debugging
		#Write-Host "rtn_val = $rtn_val"
		
		return $rtn_flg
		
	}

	function BIJsonClipstats() {
	# PURPOSE:  To get blue iris json 'clipstats' cmd properties
	# PARAMS:   -Clips = [REQUIRED] (String) comma-delimited QUOTED DB locators, like "@926894649","@926871636"
	#                    NOTE: extensions accepted, but not required, e.g., "@926894649.bvr",
	#           -Props = [OPTIONAL] (String) comma-delimited JSON properties, like 'name,push,count'
	# RETURNS:  1. (ordered hashtable) if multiple clips and/or properties requested
	#           2. (String) if a single clip and a single property requested
	#           3  'NA' if the clip is specified and the property does not exists
	# GLOBALS:  $script:JsonCallForceFlg, $script:FromConsole, $script:Quiet, $script:DevMode, $script:DevModePath 
	#
	# EXAMPLE RETURN OBJECT:
	# {
	# 	"result": "success",
	# 	"data": {
	#		"@152866668.bvr": {
	#			"camera": "DW1",
	#			"path": "@152720932.bvr",
	#			"file": "DW1.20230530_060000.1965063.17-0.jpg",
	#			"msec": 3599166,
	#			"alertmsec": 27426,
	#			"res": "2688x1520"
	#			"dummy": "NA"
	#		},
	#		<next clip>,
	#       "@123456789.bvr": "NA"
	#   }
	# }
	# ... OR ...
	# {
	#	"result": "fail",
	#	"data": {}"
	# {
	#
	# USAGE EXAMPLE
	#		$response_obj = BIJsonClipstats -Clips -Props 'camera,path,file,msec,alertmsec,res'
	#
	#       From the response, set these variables:
	#		$clipstats_result = $response_obj[0] 
	#		$clipstats_data = $response_obj[1]
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V3 (9/4/2024) replaced *Flg args w/ global vars; now avoids error if argument -Props contains duplicate items
	# V4 (11/17/2024) added code to make sure all -Clips argument values are prefixed with '@'
	# V5 (11/20/2024) added support for global var $script:JsonCallForceFlg
	# V6 (11/21/2024) disabled code for logging out of active JSON session
	# V7 (12/8/2024) changed coding [ $x.GetType() -eq "string"] to [ $x -is [String] ]

		# MUST BE THE FIRST LINE IN A SCRIPT
		param(
			[String]$Clips,
			[String]$Props
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"	
		# Write-Host "$fn_name() Clips = '$Clips'"
		# Write-Host "$fn_name() Props = '$Props'"
		
		# HANDLE REQUIRED COMMANDLINE ARGS
		if (!$Clips) {
			$err_msg = "ERR/$fn_name() expecting an argument (-Clips)"
			return $err_msg
		}		
		$my_clips_ary = $Clips.split(',')						
		# Write-Host ("$fn_name() my_clips_ary BEFORE = " + ($my_clips_ary | ConvertTo-Json))
		# Make sure each database record begins with '@'
		$i = 0
		foreach ($clip in $my_clips_ary) {
			if (!$clip.StartsWith('@')) {
				$my_clips_ary[$i] = "@" + $clip
			}
			$i++
		}
		# Write-Host ("$fn_name() my_clips_ary AFTER = " + ($my_clips_ary | ConvertTo-Json))
		
		# HANDLE OPTIONAL COMMANDLINE ARGS
		$my_props_ary = @()
		if ($Props) {
			# Ensure that requested properties are unique
			$my_props_ary = $Props.split(',') | Select-Object -Unique
		}

		# For debugging
		#Write-Host ("$fn_name() my_clips_ary = " + $my_clips_ary)
		#Write-Host ("$fn_name() my_clips_ary cnt = " + $my_clips_ary.length)
		#Write-Host ("$fn_name() my_props_ary = " + $my_props_ary)

		# INITIALIZE
		$rtn_obj = [ordered]@{
			result = 'fail'
			data = @{}
		}
		$rtn_obj_data = [ordered]@{}
		$x = ""

		if ($my_clips_ary.length -gt 0) {
			$i = 1;
			foreach ($clip in $my_clips_ary) {
				
				if ($script:FromConsole -and !$script:Quiet) {
					if ($i -eq 1) {
						#Write-Host "$fn_name() processing clip '$clip'"
					} else {
						#Write-Host "`n$fn_name() processing clip '$clip'"
					}
				}

				# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
				# ... REQUIRES COMPANION SCRIPT 'BI_json_call.ps1'
				#     WHICH RETURNS A POWERSHELL PSCustomObject
				$jsonCmd = [ordered]@{
					cmd = 'clipstats'
					path = $clip
				}
				
				# Returns a powershell PSCustomObject (or string 'ERR/*')
				$json_response_obj = `
					BIJsonCall `
					-JsonCmd $jsonCmd `
					-ForceFlg $script:JsonCallForceFlg

				# Trap various errors
				if (1 -eq 1) {  # Use to collapse this code section while editing

					# For debugging
					# Write-Host ("$fn_name() json_response_obj type = " + $json_response_obj.GetType())
					# Write-Host ("$fn_name() json_response_obj = " + ($json_response_obj | ConvertTo-Json))
					# Write-Host ("$fn_name() json_response_obj.result = " + ($json_response_obj.result))

					# If GET failed, process next clip
					<# if ($json_response_obj -is [String]) {
						if ($json_response_obj.StartsWith('ERR/')) {
							if ($script:FromConsole) {
								Write-Host $json_response_obj -ForegroundColor Red
							}
							$rtn_obj_data[$clip] = 'NA'
							break
						}
					} #>
					if ($json_response_obj -is [String] -and $json_response_obj -match 'ERR/') {
						if ($script:FromConsole) {
							Write-Host $json_response_obj -ForegroundColor Red
						}
						$rtn_obj_data[$clip] = 'NA'
						break
					}
					if ($json_response_obj.result -eq "fail") {

						# For debugging
						#Write-Host ("$fn_name() json_response_obj.result type = " + (($json_response_obj.result).GetType()))
						#Write-Host ("$fn_name() son_response_obj.result = " + ($json_response_obj.result | ConvertTo-Json))

						if ($script:FromConsole) {
							Write-Host ("$fn_name() json_response_obj.result = fail") -ForegroundColor Red
							Write-Host ($json_response_obj | ConvertTo-Json) -ForegroundColor Red
						}
						$rtn_obj_data[$clip] = 'NA'
						break
					}
					
				}

				# Extract data from the JSON command response
				if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

					$data_obj = $json_response_obj.data

					# For debugging
					#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
					#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))

					# $data_obj is a PSCustomObject like
					# @{
					#	camera=DW1;
					#	path=@746553655.bvr;
					#	file=DW1.20230428_220000.1572042.17-0.jpg; --OR-- 2023-04-28\\DW2\\DW1.20230428_220000.1572042.17-0.jpg
					#	offset=1572009;
					#	date=1682744400;
					#	color=16711808;
					#	flags=134217985;
					#	zones=0;
					#	memo=;
					#	plate=;
					#	res=2688x1520;
					#	msec=2814017;
					#	alertmsec=24520:
					#	camprofile=7;
					#	camschedule=1;
					#	camsmart=2;
					#	camzones=1;
					#	campreset=30;
					#	filesize=46m55s (419.5M);
					#	filetype=bvr H264 New
					# }

					# Convert PSCustomObject into a hashtable
					# REF: https://tinyurl.com/ddatr635
					$hashtable = @{}
					foreach ($property in $data_obj.psobject.properties.name) {
						$hashtable[$property] = $data_obj.$property
					}
					$data_obj_keys = $hashtable.Keys

					# For debugging
					# Write-Host ("$fn_name() hashtable = " + ($hashtable | ConvertTo-Json))
					# Write-Host "$fn_name() data_obj_keys = $data_obj_keys"

					$new_obj = [ordered]@{}		
					if ($my_props_ary.length -gt 0) {
						
						# LIST REQUESTED PROPERTIES
						foreach ($clipProp in $my_props_ary) {
							if ($data_obj.$clipProp -or $data_obj.$clipProp -eq 0) {
								$new_obj.Add($clipProp, $data_obj.$clipProp)
								$x = $data_obj.$clipProp

								# For debugging
								#Write-Host ($clipProp + " = "  + $data_obj.$clipProp)

							} else {
								$new_obj.Add($clipProp, "NA")
								$x = "NA"

								# For debugging
								#Write-Host ($clipProp + " = NA")
							}
						}
						$rtn_obj_data[$clip] = $new_obj
						
					} else {
						
						# LIST ALL PROPERTIES
						foreach ($clip_prop in $data_obj_keys) {
							if ($data_obj.$clip_prop -or $data_obj.$clip_prop -eq 0) {
								if ($script:FromConsole -and !$script:Quiet) {
									#Write-Host ($clip_prop + " = "  + $data_obj.$clip_prop)
								}							
								$new_obj.Add($clip_prop, $data_obj.$clip_prop)
							} else {
								if ($script:FromConsole -and !$script:Quiet) {
									#Write-Host ($clip_prop + " = NA")
								}							
								$new_obj.Add($clip_prop, "NA")
							}
						}
						$rtn_obj_data[$clip] = $new_obj
					}
					$i++
				}
			}

			if ($rtn_obj_data.Count -gt 0) {
				$rtn_obj = [ordered]@{
					result = 'success'
					data = $rtn_obj_data
				}
			}

			# LOG OUT OF API SESSION
			<# $jsonCmd = [ordered]@{cmd = 'logout'}

			# Returns a powershell PSCustomObject (or string 'ERR/*')
			$json_response_obj = `
				BIJsonCall `
				-JsonCmd $jsonCmd `
				-ForceFlg $script:JsonCallForceFlg
				
			if ($script:FromConsole) {
				if ($json_response_obj.result -eq 'fail') {
					Write-Warning "BI API: log out fail; connection not terminated"
				} else {
					if (!$script:Quiet) {
						# Write-Host "BI API: log out successful; connection terminated"	
					}
				}
			} #>
			
		}

		# Handle single clip, single property request
		if ($my_clips_ary.length -eq 1 -and $my_props_ary.length -eq 1) {
			$rtn_obj = $x
		}
									
		<# if ($script:FromConsole -and !$script:Quiet) {
			Write-Host "Clipstats:"
			Write-Host "----------"
			Write-Host ($rtn_obj | ConvertTo-Json)
		} #>

		return $rtn_obj
	}

	function BIJsonAlertListParse() {
	# PURPOSE:  To get and parse the Blue Iris JSON interface cmd 'alertlist' response for a SINGLE CAMERA
	#           To retrieve all or, if requested, a subset of 'alertlist' properties for each Alert
	#           NOTE: returns non-existent requested 'alertlist' properties as 'alert_prop':'NA'
	# PARAMS:   -CamName         = [REQUIRED] camera shortname
	#           -Pattern         = process Alerts with descriptor substring; must be used w/ -CamName
	#           -AlertProps      = (array) like 'memo,offset', comma delimited
	#             			        options: camera, clip, color, date, file, filesize, flags memo, 
	#                                     newalerts, newalerttime, offset, path, res, zones
	#           -MemoTag         = (String) process only those protected Alerts with descriptor substring
	#           -DatabaseView    = (String) default is 'alerts'
	#                              options: all, new, stored, alerts, aux1, aux2, aux3, aux4, aux5, aux6, aux7,
	#                                       flagged, export, archive, people, vehicles, confirmed, or canceled                             
	#                              Note: When view='flagged', clip items are included too, but they don't have msec metadata.
	#           -AlertPath        = (String) process single Alert with alertlist [path] like '@123456789'
	#           -ExcludeClonesFlg = (boolean) exclude clones of cam names in -Cameras
	#           -LatestFlg        = (boolean) $true returns only the most recent alert for each -Cameras
	#           -SilentFlg        = (boolean) true suppresses all output to the console, including warnings
	# RETURNS:  (object) PowerShell hashtable, @('cam':camName,'alert_prop':value}
	# REQUIRES: Functions: BIJsonCall(), IsBiVersion(), GetBiVersion()
	# GLOBALS:  $script:BiVersion, $script:JsonCallForceFlg, $script:FromConsole, $script:Quiet, 
	#           $script:DevMode, $script:DevModePath 
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V5 (8/29/2024) improved BI version checking; replaced *Flg args w/ global vars
	# V6 (9/9/2024) fixed bug where alertlist response unexpectedly contains cloned camera entries
	# V7 (11/11/2024) fixed bug where alertlist response unexpectedly contains cloned camera entries
	# V8 (11/11/2024) added -LatestFlg support
	# V9 (11/20/2024) added support for global var $script:JsonCallForceFlg
	# V10 (12/8/2024) changed coding [ $x.GetType() -eq "string"] to [ $x -is [String] ]
	#                 changed [ $x = [array]$y ] typing to [ $x = @($y) ]
	# V11 (3/13/2025) added alert duration (property 'msec') to output


		param(
			[String]$CamName,           # [REQUIRED] camera shortname
			[String]$Pattern,           # must be used with -CamName
			[String[]]$AlertProps,      # list of BI JSON 'alertlist' cmd properties (parameters)
			[String]$MemoTag,           # substring in Alert's memo
			[String]$DatabaseView,      # default is 'alerts'
			[String]$AlertPath,         # like '@123456789' (JSON alertlist [path] value)
			[Boolean]$ExcludeClonesFlg, # $true excludes $CamName's cloned camera entries 
			[Boolean]$LatestFlg,        # $true returns only the most recent alert for each -Cameras
			[Boolean]$SilentFlg         # suppresses all output to the console, including warnings
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() CamName = $CamName"
		# Write-Host "$fn_name() Pattern = $Pattern"
		# Write-Host ("$fn_name() AlertProps type = " + $AlertProps.GetType())
		# Write-Host ("$fn_name() AlertProps = " + ($AlertProps | ConvertTo-Json))
		# Write-Host "$fn_name() MemoTag = $MemoTag"
		# Write-Host "$fn_name() DatabaseView = $DatabaseView"
		# Write-Host "$fn_name() AlertPath = $AlertPath"
		# Write-Host "$fn_name() ExcludeClonesFlg = $ExcludeClonesFlg"
		# Write-Host "$fn_name() LatestFlg = $LatestFlg"
		# Write-Host "$fn_name() SilentFlg = $SilentFlg"
		# Write-Host ("$fn_name() global FromConsole = " + $script:FromConsole)
		# Write-Host ("$fn_name() global Quiet = " + $script:Quiet)"
		# Write-Host ("$fn_name() global DevMode = " + $script:DevMode)
		# Write-Host ("$fn_name() global DevModePath = " + $script:DevModePath)

		#Set-StrictMode  -Version 3  # https://tinyurl.com/mrx3w9v4
		# PARSE ARGS
		if ($Pattern -and !$CamName) {
			$err_msg = "ERR/argument -Pattern must be used with -CamName [$fn_name]"
			return $err_msg
		}
		if (!$CamName) {
			$err_msg = "ERR/expecting at least 1 argument (-CamName) [$fn_name]"
			return $err_msg
		}

		# If searching only a single AlertMemo, we can avoid GETTING all alerts 
		# via the JSON 'alertlist' cmd parameter 'search'
		$search_param_val = $null
		if ($MemoTag) {
			$search_param_val = $MemoTag
		}
		#Write-Host "$fn_name() search_param_val = $search_param_val"
		
		if (!$DatabaseView) {
			$DatabaseView = 'all'
		}
		if (!$AlertPath) {
			$AlertPath = "*"
		}

		# Initialize master collector for all alerts in database 
		$al_props_obj = @{}
		$al_props_obj.Add( 'cameras', @() )
		$al_props_obj.Add( 'paths', @() )
		$al_props_obj.Add( 'clips', @() )
		$al_props_obj.Add( 'files', @() )
		$al_props_obj.Add( 'memos', @() )
		$al_props_obj.Add( 'msecs', @() )
		$al_props_obj.Add( 'offsets', @() )
		$al_props_obj.Add( 'flagss', @() )  # double 's' intentional!
		$al_props_obj.Add( 'ress', @() )    # double 's' intentional!
		$al_props_obj.Add( 'zoness', @() )  # double 's' intentional!
		$al_props_obj.Add( 'dates', @() )
		$al_props_obj.Add( 'colors', @() )
		$al_props_obj.Add( 'filesizes', @() )
		
		# For debugging
		#Write-Host ("$fn_name() al_props_obj type = " + $al_props_obj.GetType())
		#Write-Host ("$fn_name() al_props = " + ($al_props_obj | ConvertTo-Json))

		# ---------------------------------------------------------------------------------------
		# KEY POINT: Prior to version 5.7.9.3 BI's 'alertlist' json cmd returned only 1000 items; 
		#            As of version 5.7.9.3 BI's 'alertlist' json cmd returns 2048 items; 
		#            so we must re-send the json cmd until we have retrieved all alerts.
		#            REF: https://ipcamtalk.com/threads/blue-iris-ui3.23528/post-788744
		
		$alertlist_2048_flg = `
			IsBiVersion `
			-CompareToVersion '5.7.9.3' `
			-CurrentVersion $script:BiVersion

		if ($alertlist_2048_flg) {
			$json_alertlist_max = 2048;
			#$json_cliplist_max = 4096;
		} else {
			$json_alertlist_max = 1000;  # Default (tested against a BI 5.6.3.7 server by bp2008@ipcamtalk)
			#$json_cliplist_max = 1200;   # Default (tested against a BI 5.6.3.7 server by bp2008@ipcamtalk)
		}
		# ---------------------------------------------------------------------------------------

		# Initialize WHILE LOOP variables
		$paths_ary = @('x')	# primary break loop variable... starts with a dummy value
		$oldest_date = 0  # oldest date found 
		$newest_date = 0  # newest date found 
		$skipped_next_json_call_flg = $false
		
		# PREP FOR SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
		# WHY? Per Blue Iris support...
		#     Prior to version 5.8.2.4, the alertlist 'memo' field was included only if it was not blank.
		#     Starting with version 5.8.3.4 it is now always included (REF: 12/18/2023 email from Ken at blueirissupport.com)
		$alertlist_memo_always_flg = `
			IsBiVersion `
			-CompareToVersion '5.8.2.4' `
			-CurrentVersion $script:BiVersion
		
		# Initialize WHILE-LOOP break variables
		$prev_loop_alerts_cnt = 0
		if ($LatestFlg) {
			$max_tries = 1
		} else {
			$max_tries = 10
		}
		$iter = 1;

		# WHILE LOOP - Get JSON 'alertlist' data in batches until the cmd returns no new data
		while ($paths_ary.Count -gt 0) {
			
			if ($script:DevMode -and $script:FromConsole) {
				if ($iter -eq 1) {
					Write-Host ""
					#Write-Host "(dev mode) ----- $CamName" -ForegroundColor Cyan
				} else {
					#Write-Host "(dev mode) -----" -ForegroundColor Cyan
				}
				Write-Host "(dev mode) ----- $CamName" -ForegroundColor Cyan
				Write-Host "(dev mode) JSON alertlist iteration #$iter" -ForegroundColor Cyan
			}

			# Initialize collector for this JSON 'alertlist' call
			$tmp_obj = @{}
			$tmp_obj.Add( 'cameras', @() )
			$tmp_obj.Add( 'paths', @() )
			$tmp_obj.Add( 'clips', @() )
			$tmp_obj.Add( 'files', @() )
			$tmp_obj.Add( 'memos', @() )
			$tmp_obj.Add( 'msecs', @() )
			$tmp_obj.Add( 'offsets', @() )
			$tmp_obj.Add( 'flagss', @() )  # double 's' intentional!
			$tmp_obj.Add( 'ress', @() )    # double 's' intentional!
			$tmp_obj.Add( 'zoness', @() )  # double 's' intentional!
			$tmp_obj.Add( 'dates', @() )
			$tmp_obj.Add( 'colors', @() )
			$tmp_obj.Add( 'filesizes', @() )
			#$tmp_obj.Add( 'AAAA', @() )
			
			# For debugging
			#Write-Host ("$fn_name() tmp_obj type = " + $tmp_obj.GetType())
			#Write-Host ("$fn_name() tmp_obj = " + ($tmp_obj | ConvertTo-Json))
			#Write-Host ("$fn_name() tmp_obj.paths type  = " + ($tmp_obj.paths).GetType())

			# GET JSON DATA FROM THE BLUE IRIS WEB SERVER

			# First call params - Default
			$jsonCmd = [ordered]@{
				cmd    = 'alertlist'
				camera = $CamName
				search = $search_param_val
				view   = $DatabaseView
			}
			
			# Subsequent call params
			if ($iter -gt 1) {
				
				# For debugging
				#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
				#Write-Host "$fn_name() oldest_date = $oldest_date"
				
				if ($oldest_date -gt 0) {  #Just in case
					# Params for the next set of 'alertlist' data
					$jsonCmd = [ordered]@{
						cmd     = 'alertlist'
						camera  = $CamName
						search  = $search_param_val
						view    = $DatabaseView
						enddate = $oldest_date   # oldest time found (so far); per Ken @ Blue Iris Support 
					}
				}
			}
			
			# For debugging
			# Write-Host ("$fn_name() jsonCmd = " + ($jsonCmd | ConvertTo-Json))

			# FIRST JSON CALL
			# Returns a powershell PSCustomObject (or string 'ERR/*')
			$json_response_obj = `
				BIJsonCall `
				-JsonCmd $jsonCmd `
				-ForceFlg $script:JsonCallForceFlg

			# Trap various errors
			if (1 -eq 1) {  # Use to collapse this code section while editing

				# For debugging
				# Write-Host ("$fn_name() jsonCmd response = " + ($json_response_obj | ConvertTo-Json))
				if ($iter -gt 1) {
					# Write-Host ("$fn_name() jsonCmd response type = " + $json_response_obj.GetType())
					# Write-Host ("$fn_name() jsonCmd response = " + ($json_response_obj | ConvertTo-Json))
				}
				
				if ($json_response_obj -is [String] -and $json_response_obj -match 'ERR/') {

					# For debugging
					# Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
					# msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj = $json_response_obj"

					break
				}
				if ($json_response_obj.result -eq 'fail') {

					# For debugging
					# Write-Host "$fn_name() BIJsonCall() json_response_obj.result = 'fail'`n"
					# msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj.result = 'fail'"

					break
				}
			}

			if ($json_response_obj.result -eq 'success') {

				# Test for $json_response_obj.data object
				if (Get-Member -InputObject $json_response_obj -name "data" -Membertype Properties) {

					$data_obj = $json_response_obj.data

					if ($data_obj) {
						
						# For debugging
						# Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
						# Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))
						# Write-Host ("$fn_name() data_obj cnt = " + ($data_obj.Count))
						# Write-Host ("$fn_name() data_obj.camera = " + $data_obj.camera + "'")
						# Write-Host ("$fn_name() data_obj.path type = " + ($data_obj.path).GetType())
						# Write-Host ("$fn_name() data_obj.memo type = " + ($data_obj.memo).GetType())

						# KEY POINT: $paths_ary is the WHILE LOOP's primary break variable
						try {
							# Get array of all 'path' values in the $data_obj
							$paths_ary = [String[]]$data_obj.path  					
						}
						catch {
							$err_msg = "ERR/JSON 'alertlist' response for '$CamName' contains no 'path' entries (database corrupt?)"
							return $err_msg
						}
						
						# Handle cases where alertlist response ALSO contains entries from $CamName's clones
						# ... the goal is to remove those entries from $data_obj
						if ($ExcludeClonesFlg) {
							# Get array of all UNIQUE 'camera' values in the $data_obj
							$obj_cams_ary = $data_obj.camera | Select-Object -Unique
							# ... alertlist contains cloned cam names if the this UNIQUE array contains more that one element 
							if ($obj_cams_ary.length -gt 1) {
								# Remove cloned camera entries from object
								$data_obj = $data_obj | Where-Object { $_.camera -eq $CamName }
								# Refresh $paths_ary
								$paths_ary = [String[]]$data_obj.path  					
							}
						}
						
						# Get the alertlist dates array
						# KEY POINT: using try:catch as a redundant quality test
						try {
							# Get array of all 'date' values in the $data_obj
							$dates_ary = [int[]]$data_obj.date  					
						}
						catch {
							$err_msg = "ERR/JSON 'alertlist' response for '$CamName' contains no 'date' entries (database corrupt?)"
							return $err_msg
						}					
		
						# PERFORM SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
						# WHY? Per Blue Iris support...
						#     Prior to version 5.8.2.4, the alertlist 'memo' field was included only if it was not blank.
						#     Starting with version 5.8.3.4 it is now always included (REF: 12/18/2023 email from Ken at blueirissupport.com)
						$memo_fix_cntr = 0
						if ($alertlist_memo_always_flg) {
							$memos_ary = [String[]]$data_obj.memo
							#Write-Host ("$fn_name() memos_ary** = " + ($memos_ary | ConvertTo-Json))
						} else {
							# For debugging
							#Write-Host "$fn_name() Special handling of Alert memos" 
							
							try {
								# KEY POINT: This code line fails if ANY Alert in the JSON 'alertlist' response lacks a 'memo'
								$memos_ary = [String[]]$data_obj.memo  
							} 
							catch {
								# REPAIR THE 'BAD' ALERTS
								
								# For debugging
								Write-Warning "$fn_name() Found alerts w/ empty memos; preparing to change values to 'NA'..." 
								
								$i = 0
								foreach ($item in $data_obj) {
									try {
										$x = $item.memo
									}
									Catch {
										$data_obj[$i] | Add-Member -MemberType NoteProperty -Name "memo" -Value "NA"
										$memo_fix_cntr++
									}
									$i++
								}
								$memos_ary = [String[]]$data_obj.memo
								
								if ($data_obj['memos'].Count -ne $tmp_obj['paths'].Count) {
									$err_msg = "ERR/unable to fix missing 'memo':'value' pairs in the JSON 'alertlist' response"
									return $err_msg
								}
								
								if ($script:FromConsole -and !$SilentFlg) {
									Write-Warning "repaired $memo_fix_cntr Alerts in the JSON 'alertlist' response that lacked a 'memo' key; now 'memo':'NA'"
								}
							}
						}

						# For debugging
						if ($memo_fix_cntr -gt 0) {
							Write-Host ("$fn_name() memos_ary cnt AFTER repair = " + $memos_ary.Count)
							Write-Host ("$fn_name() memos_ary AFTER repair = " + ($memos_ary | ConvertTo-Json))
						}

						# For debugging
						#Write-Host ("$fn_name() memos_ary = " + ($memos_ary | ConvertTo-Json))

						# Initialize vars
						$this_iter_alerts_cnt = $paths_ary.Count
						$this_iter_dates_cnt = $dates_ary.Count

						# Break loop if this iteration is a single result with same oldest date
						# THIS IN AN EDGE CASE where item #2048 has 'date' X, and item #2049 also has 'date' X
						if ($this_iter_alerts_cnt -eq 1) {
							if ($dates_ary[$this_iter_dates_cnt-1] -eq $oldest_date) {
								if ($script:DevMode -and $script:FromConsole) {
									Write-Host ("(dev mode) `"alertlist`" count this iteration = " + $this_iter_alerts_cnt) -ForegroundColor Cyan
									# Show $dates_ary endpoints
									Write-Host ("(dev mode) fist date = dates_ary[0] = " + $dates_ary[0]) -ForegroundColor Cyan 
									Write-Host ("(dev mode) last date = dates_ary[" + ($this_iter_dates_cnt-1) + "] = " + $dates_ary[$this_iter_dates_cnt-1]) -ForegroundColor Cyan
									}
									Write-Host ""
								break
							}
						}

						# Break loop if item's 'path' and 'date' parameter counts do not match
						# SOMETHING'S WRONG
						if ($this_iter_alerts_cnt -ne $this_iter_alerts_cnt) {
							if ($script:DevMode -and $script:FromConsole) {
								Write-Host "(dev mode) terminating 'alertlist' call; response is unexpected ('path' & 'date' counts unequal)" -ForegroundColor Cyan
							}
							break
						}

						# Break loop if JSON 'alertlist' call returns the same number of alerts as the 
						# previous result (<2048 items)
						if ($this_iter_alerts_cnt -lt $json_alertlist_max -and ($this_iter_alerts_cnt -eq $prev_loop_alerts_cnt)) {
							if ($al_props_obj['paths'] -contains $paths_ary[0]) {
								if ($script:DevMode -and $script:FromConsole) {
									Write-Host "(dev mode) ignoring JSON alertlist response w/ same alert count as previous response" -ForegroundColor Cyan
								}
								break
							}
						}
						
						$prev_loop_alerts_cnt = $this_iter_alerts_cnt
						
						if ($script:DevMode -and $script:FromConsole) {
							Write-Host "(dev mode) `"alertlist`" count this iteration = $this_iter_alerts_cnt" -ForegroundColor Cyan
							# Show $dates_ary endpoints
							Write-Host ("(dev mode) fist date = dates_ary[0] = " + $dates_ary[0]) -ForegroundColor Cyan 
							Write-Host ("(dev mode) last date = dates_ary[" + ($this_iter_dates_cnt-1) + "] = " + $dates_ary[$this_iter_dates_cnt-1]) -ForegroundColor Cyan
							Write-Host ""
						}

						# For iterations > 1, remove the first alert in the response
						# THIS IN AN EDGE CASE where item #2048 has 'date' X, and item #2049 also has 'date' X
						if ($iter -gt 1) {
							
							if ($this_iter_alerts_cnt -ge 1 -and $this_iter_alerts_cnt -le $json_alertlist_max) {
								
								# REMOVE THE FIRST OBJECT ELEMENT 
								$data_obj = $data_obj[1 .. ($data_obj.Count - 1)]

								# SPECIAL HANDLING FOR THE 'MEMO' PARAMETER
								$memos_ary = $memos_ary[1 .. ($memos_ary.Count - 1)]
								
								# KEY POINT: 
								# Note that we do not update var $paths_ary at this point
								# Why? Because it is the while loop's break variable and we still have work to do
								
								# For debugging
								#Write-Host ("$fn_name() data_obj cnt AFTER = " + ($data_obj.Count))
							}
						}

						# For debugging
						#if ($iter -eq 2) {
							#$data_obj_keys += $($data_obj[0] | Get-Member -MemberType *Property).Name
							#Write-Host ("$fn_name() data_obj type = " + $data_obj.GetType())
							#Write-Host ("$fn_name() data_obj keys = " + ($data_obj_keys | ConvertTo-Json))
							#Write-Host ("$fn_name() data_obj = " + ($data_obj | ConvertTo-Json))
							#Write-Host ("$fn_name() data_obj.date type = " + ($data_obj.date).GetType())
						#}

						# Get all dates for this set of Alerts
						# ... This SHOULD BE an array of UTC timestamps, ordered largest to smallest
						# ... NOTE: this array is also used for QUALITY testing
						[Int[]]$dates_ary = $data_obj.date
						$dates_cnt = $dates_ary.Count

						# HANDLE SIMULATIONS
						if ($script:Sim5) {  #simulate response w/ zero-date (timestamp) values; arg value is CamName
							if ($Cameras -contains $script:Sim5) {
								# Set first date to zero
								#$dates_ary[0] = 0
								#$data_obj[0].date = 0

								# Set last date to zero
								#$dates_ary[$dates_cnt-1] = 0
								#$data_obj[$dates_cnt-1].date = 0
							
								# Set 2nd date to zero
								$dates_ary[1] = 0
								$data_obj[1].date = 0

								# Set 2nd to last date to zero
								$dates_ary[$dates_cnt-2] = 0						
								$data_obj[$dates_cnt-2].date = 0
							
								# For debugging
								Write-Host ("SIM5 fist date = dates_ary[0] = " + $dates_ary[0])
								Write-Host ("SIM5 last date = dates_ary[" + ($dates_cnt-1) + "] = " + $dates_ary[$dates_cnt-1])
							}
						}
						
						# Get the oldest/newest dates for this set of Alerts 
						# ... The oldest date is needed for the next JSON 'alertlist' call, if necessary
						$newest_date = $dates_ary[0]
						$oldest_date = $dates_ary[$dates_cnt-1]
						$max_date = ($dates_ary | Measure-Object -Maximum).Maximum
						$min_date = ($dates_ary | Measure-Object -Minimum).Minimum
						$dates_sum = ($dates_ary | Measure-Object -Sum).Sum
						
						# For debugging
						#Write-Host ("$fn_name() dates_ary type = " + $dates_ary.GetType())
						#Write-Host "$fn_name() dates_cnt = $dates_cnt"
						#Write-Host "$fn_name() dates_ary = $dates_ary"
						#Write-Host ("$fn_name() newest_date type = " + $newest_date.GetType())
						#Write-Host "$fn_name() newest_date = $newest_date"
						#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
						#Write-Host "$fn_name() oldest_date = $oldest_date"
						#Write-Host "$fn_name()dates_sum = $dates_sum"
						#Write-Host ("$fn_name() data_obj.date = " + ($data_obj.date | ConvertTo-Json))

						# Check for more QUALITY issues...
						# ... 1. Check if all dates are zero (FATAL)
						if ($dates_sum -eq 0 ) {
							$skipped_next_json_call_flg = $true
							if ($script:FromConsole -and !$SilentFlg) {
								Write-Warning "JSON 'alertlist' response for '$CamName' contains only zero 'date' entries (database corrupt?); find w/ -DevMode"
							}
							break
						}
						# ... 2. Check if the first array element is the max date (IT SHOULD BE)
						$first_date_is_not_max = ($newest_date -lt $max_date)
						# ... 3. Check that the first and last array elements are NOT zero (THEY SHOULD BE)
						$first_date_is_zero = ($newest_date -eq 0)
						$last_date_is_zero = ($oldest_date -eq 0)
						$zero_dates_cntr = 0  # Initialize counter
						# ... 4. Check if any OTHER alerts have dates with zero values
						#        Make hashtable like @("date1":count, "date2":count ...}
						#        if zero-dates are found, $ht will have and entry like {..., "0": 23, ...}
						# ... REF: https://stackoverflow.com/a/12640991
						$dates_ary | Group | % { $ht = @{} } { $ht[$_.Name] = $_.Count } { $ht } | Out-Null
						
						# For debugging
						#Write-Host ("ht = " + ($ht | ConvertTo-Json))
						#Write-Host ("ht['0'] = " + $ht['0'])
						
						if ($ht['0']) {
							$zero_dates_cntr += $ht['0']
							$dates_between_zero_cntr = $ht['0']
							
							# For debugging
							#Write-Host ("ht['0'] type = " + ($ht['0']).GetType())
							
						} else {
							$dates_between_zero_cntr = 0
						}
						if ($first_date_is_zero -and $dates_between_zero_cntr -gt 0) {
							$dates_between_zero_cntr--
						}
						if ($last_date_is_zero -and $dates_between_zero_cntr -gt 0) {
							$dates_between_zero_cntr--
						}
						if ($script:FromConsole -and !$SilentFlg) {
							if ($first_date_is_zero) {
								if ($last_date_is_zero) {
									Write-Warning "JSON 'alertlist' response #$iter for '$CamName' begins and ends with an Alert with zero-date values"
								} else {
									Write-Warning "JSON 'alertlist' response #$iter for '$CamName' begins with an Alert with a zero-date value"
								}
							} else {
								if ($first_date_is_not_max) {
									Write-Warning "JSON 'alertlist' response #$iter for '$CamName' does not have the newest date in the first entry (as expected)"
								}
								if ($last_date_is_zero) {
									Write-Warning "JSON 'alertlist' response #$iter for '$CamName' ends with an Alert with a zero-date value"
								}
							}
							if ($dates_between_zero_cntr -gt 0) {
								Write-Warning "JSON 'alertlist' response #$iter for '$CamName' has $zero_dates_cntr zero-date values; find w/ -DevMode"
							}
						}
						
						# For debugging
						#Write-Host ("$fn_name() newest_date type = " + $newest_date.GetType())
						#Write-Host "$fn_name() newest_date = $newest_date"
						#Write-Host ("$fn_name() newest_date for iter #" + $iter + " ($this_iter_alerts_cnt alerts) = " + $newest_date)
						#Write-Host ("$fn_name() oldest_date type = " + $oldest_date.GetType())
						#Write-Host "$fn_name() oldest_date = $oldest_date"

						# Load collectors
						$tmp_obj['cameras']   = $data_obj.camera  # an array (adds a column of values to the hashtable object)
						$tmp_obj['paths']     = $data_obj.path
						$tmp_obj['clips']     = $data_obj.clip
						$tmp_obj['files']     = $data_obj.file
						#$tmp_obj['memos']     = $data_obj.memo   # We handle '$data_obj.memo' differently 
						$tmp_obj['memos']     = $memos_ary
						$tmp_obj['msecs']     = $data_obj.msec
						$tmp_obj['offsets']   = $data_obj.offset
						$tmp_obj['flagss']    = $data_obj.flags
						$tmp_obj['ress']      = $data_obj.res
						$tmp_obj['zoness']    = $data_obj.zones
						$tmp_obj['dates']     = $data_obj.date	# Note: $date_ary[] is reverse-ordered.
						$tmp_obj['colors']    = $data_obj.color
						$tmp_obj['filesizes'] = $data_obj.filesize

						# For debugging
						#Write-Host ("$fn_name() tmp_obj['memos'] = " + ($tmp_obj['memos'] | ConvertTo-Json))
		
						# HANDLE SIMULATIONS
						if ($script:Sim1) {  # remove ALL 'memo':'value' pairs from raw JSON 'alertlist' response
							foreach ($item in $data_obj) {
								$item.PSObject.Properties.Remove('memo')
							}
							if ($script:FromConsole) {
								Write-Host "(test mode) Simulation #1: removed ALL 'memo':'value' pairs from JSON 'alertlist' response" -Foregroundcolor Green
								#Write-Host ("(test mode) Simulation #1: data_obj AFTER = " + ($data_obj | ConvertTo-Json)) -Foregroundcolor Green
							}
						}
						if ($script:Sim2) {  # remove 'memo':'value' pairs for 3 random alerts from JSON 'alertlist' response
							try {
								$tmp_obj['memos'] = @($data_obj.memo)
								$total_cnt = $tmp_obj['paths'].Count
								$chgd_memos_cnt = [Math]::min(3,$alerts_cnt-1)
								$chgd_idxs_ary = @()
								$chgd_alerts_obj = @()
								$i = 0
								while ($i -lt $chgd_memos_cnt) {
									$idx = Get-Random -Maximum $total_cnt
									# Only protected alerts
									#Write-Host ("flags = " + ($data_obj[$idx]).flags)
									$flags_ary = `
										MakeBitArray `
										-DecVal ($data_obj[$idx]).flags `
										-NoBits 32
									# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
									if ($flags_ary[2] -eq 1) {
										if (!($chgd_idxs_ary -contains $idx)) {
											$data_obj[$idx].PSObject.Properties.Remove('memo')
											$chgd_alerts_obj += $data_obj[$idx]
											$chgd_idxs_ary += $idx
											$i++
										}
									}
								}
								if ($script:FromConsole) {
									Write-Host "(test mode) Simulation #2: removed JSON alertlist 'memo' entry from $chgd_memos_cnt random protected Alerts" -Foregroundcolor Green
									Write-Host ("(test mode) Simulation #2: changed alertlist entries = " + ($chgd_alerts_obj | ConvertTo-Json)) -Foregroundcolor Green
								}
							} catch {
								if ($script:FromConsole) {
									Write-Host "(test mode) Simulation #2: unable to run simulation; no Alerts have a 'memo':'value' pair"
								}
							}
						}
						if ($script:Sim3) {  # simulate response w/ all 'memo'='NA' params
							$tmp_obj['memos'] = 1..$tmp_obj['paths'].Count | % { 'NA' }
							if ($script:FromConsole) {
								Write-Host "(test mode) Simulation #3: set 'memo'='NA' for all Alerts in $fn_name() response" -Foregroundcolor Green
								Write-Host ("(test mode) Simulation #3: tmp_obj['memos'] modified = " + ($tmp_obj['memos'] | ConvertTo-Json)) -Foregroundcolor Green
							}
						}
						if ($script:Sim4) {  # simulate response w/ 3 random 'memo'='NA' params
							$memos_cnt = $tmp_obj['memos'].Count
							if ($memos_cnt -gt 0) {
								#Write-Host "memos_cnt = $memos_cnt"
								$memos_cnt = $memos_cnt
								$chgd_memos_cnt = [Math]::min(3,$memos_cnt-1)
								$chgd_idxs_ary = @()
								$chgd_alerts_ary = @()
								$i = 0
								while ($i -lt $chgd_memos_cnt) {
									$idx = Get-Random -Maximum $memos_cnt
									#Write-Host "idx = $idx"
									# Only protected alerts
									#Write-Host ("flags = " + ($data_obj[$idx]).flags)
									$flags_ary = `
										MakeBitArray `
										-DecVal ($data_obj[$idx]).flags `
										-NoBits 32

									# Flags: 2^0=flagged, 2^1=protected, 2^6=archive, 2^9=export, 2^28 confirmed
									if ($flags_ary[2] -eq 1) {
										if (!($chgd_idxs_ary -contains $idx)) {
											if ($tmp_obj['memos'][$idx] -ne 'NA') {
												$tmp_obj['memos'][$idx] = 'NA'
												$chgd_alerts_ary += (($data_obj[$idx]).file).Replace('.jpg','')
												$chgd_idxs_ary += $idx
												$i++
											}
										}
									}
								}
								if ($script:FromConsole) {
									Write-Host "(test mode) Simulation #4: set 'memo'='NA' for $chgd_memos_cnt random Alerts in $fn_name() response" -Foregroundcolor Green
									#Write-Host ("(test mode) Simulation #4: tmp_obj['memos'] modified = " + ($tmp_obj['memos'] | ConvertTo-Json)) -Foregroundcolor Green
									Write-Host ("(test mode) Simulation #4: modified AlertIDs = " + ($chgd_alerts_ary | ConvertTo-Json)) -Foregroundcolor Green
								}
							}
						}

						# For debugging
						
						#Write-Host ("$fn_name() tmp_obj['paths'] cnt = " + $tmp_obj['paths'].Count)
						#Write-Host ("$fn_name() tmp_obj['paths'] = " + ($tmp_obj['paths'] | ConvertTo-Json))
						#Write-Host ("$fn_name() tmp_obj['flagss'] cnt = " + $tmp_obj['flagss'].Count)
						#Write-Host ("$fn_name() tmp_obj['flagss'] = " + ($tmp_obj['flagss'] | ConvertTo-Json))
						#Write-Host ("$fn_name() tmp_obj['memos'] dnt = " + $tmp_obj['memos'].Count)
						#Write-Host ("$fn_name() tmp_obj['memos'] = " + ($tmp_obj['memos'] | ConvertTo-Json))

						# Merge results with the master accumulator
						$al_props_obj['cameras']   = $al_props_obj['cameras']    + $tmp_obj['cameras']
						$al_props_obj['paths']     = $al_props_obj['paths']      + $tmp_obj['paths']
						$al_props_obj['clips']     = $al_props_obj['clips']      + $tmp_obj['clips']
						$al_props_obj['files']     = $al_props_obj['files']      + $tmp_obj['files']
						$al_props_obj['memos']     = $al_props_obj['memos']      + $tmp_obj['memos']
						$al_props_obj['msecs']     = $al_props_obj['msecs']      + $tmp_obj['msecs']	
						$al_props_obj['offsets']   = $al_props_obj['offsets']    + $tmp_obj['offsets']	
						$al_props_obj['flagss']    = $al_props_obj['flagss']     + $tmp_obj['flagss']	
						$al_props_obj['ress']      = $al_props_obj['ress']       + $tmp_obj['ress']	
						$al_props_obj['zoness']    = $al_props_obj['zoness']     + $tmp_obj['zoness']	
						$al_props_obj['dates']     = $al_props_obj['dates']      + $tmp_obj['dates']	
						$al_props_obj['colors']    = $al_props_obj['colors']     + $tmp_obj['colors']	
						$al_props_obj['filesizes'] = $al_props_obj['filesizes']  + $tmp_obj['filesizes']	
						
						# Stop requesting data if the last date is zero
						if ($oldest_date -eq 0) {
							$skipped_next_json_call_flg = $true
							if ($script:FromConsole -and !$SilentFlg) {
								Write-Warning "skipping next JSON 'alertlist' call because this call's oldest date is zero"
							}
							break
						}
					
					} else {
						break
					}
				} else {
					break
				}
			} else {
				break
			}

			# Avoid infinite loop
			if ($iter -gt $max_tries) {
				break
			}
			
			$iter++
		}
		
		# For debugging
		#Write-Host "Broke out of While loop"

		# TEST FOR MORE QUALITY ISSUES (indicators of a possibly corrupt database)
		# ... Test if 'clip' values contain '@-1.jpg' entries (WARNING)
		[String[]]$clips_ary = $al_props_obj['clips']
		if ($clips_ary -contains '@-1.jpg') {
			$clips_cnt = $clips_ary.Count
			$bad_clip_cntr = 0
			foreach($item in $clips_ary) {
				if ($item -eq '@-1.jpg') {
					$bad_clip_cntr++
				}
			}
			if ($script:FromConsole -and !$SilentFlg) {
				Write-Warning "JSON 'alertlist' response for '$CamName' has $bad_clip_cntr/$clips_cnt 'clip'='@-1.jpg' entries; find w/ -DevMode"
			}
		}

		# For debugging
		# Write-Host ("$fn_name() al_props_obj type = " + ($al_props_obj.GetType()))
		# Write-Host ("$fn_name() al_props_obj keys = " + $al_props_obj.Keys)
		# Write-Host ("$fn_name() al_props_obj = " + ($al_props_obj | ConvertTo-Json))
		# Write-Host ("$fn_name() al_props_obj['cameras'] = " + ($al_props_obj['cameras'] | ConvertTo-Json))
		# Write-Host ("$fn_name() al_props_obj['memos'] = " + ($al_props_obj['memos'] | ConvertTo-Json))
		# Write-Host ("$fn_name() al_props_obj['flagss'] = " + ($al_props_obj['flagss'] | ConvertTo-Json))
		# Write-Host ("$fn_name() al_props_obj.paths = " + $al_props_obj.paths)
		
		# DEV MODE - Save each camera's 'alertlist' data
		if ($script:DevMode -and $script:FromConsole) {
			if (!$script:Sim1 -and !$script:Sim2 -and !$script:Sim3 -and !$script:Sim4) {
				
				$al_objects_ary = @()  # Initialize accumulator; an array of hashtables
				$keys = @($al_props_obj.Keys)
				$all_alerts_cnt = $al_props_obj['paths'].Count
				$alertlist_filespec = "$DevModePath\alertlist.$CamName.json"
				
				# For debugging
				#Write-Host ("$fn_name() keys = " + ($keys | ConvertTo-Json))
				#Write-Host "$fn_name() all_alerts_cnt = $all_alerts_cnt"
				#Write-Host "$fn_name() alertlist_filespec = '$alertlist_filespec'"
				
				if ($all_alerts_cnt -gt 0) {

					# Possible keys in $al_props_obj
					$keys1 = @('cameras','memos','paths','clips','files','filesizes','dates','offsets')
					$keys2 = @('flagss','zoness')

					for ($i=0; $i -lt $all_alerts_cnt; $i++) {
						$ht = @{}  # Initialize hashtable
						# Modify $al_props_obj key for use in hashtable 
						
						#Write-Host ""
						
						foreach ($key in $keys) {
							if ($keys1 -contains $key) {
								$key_new = $key.TrimEnd('s')
							} else {  # $keys_2 -contains $key
								$key_new = $key.Replace('ss','s')
							}
							
							# For debugging
							#Write-Host "key = $key"
							#Write-Host ("al_props_obj[$key] type = " + ($al_props_obj[$key]).GetType())
							#Write-Host ("al_props_obj[$key] = " + ($al_props_obj[$key] | ConvertTo-Json))
							#Write-Host ("al_props_obj[$key][0] = " + $al_props_obj[$key][0])
							#Write-Host ("al_props_obj[$key][" + ($all_alerts_cnt-1) + "] = " + $al_props_obj[$key][$all_alerts_cnt-1])
							#Write-Host ("al_props_obj[$key][$i] = " + $al_props_obj[$key][$i])
							
							$ht.Add($key_new, $al_props_obj[$key][$i])	
							#exit
							
						}
						$al_objects_ary += $ht
					}

					# For debugging
					#Write-Host ("al_list_ary = " + ($al_list_ary | ConvertTo-Json))

					$alertlist_saved_flg = $false
					if (!$Cameras -and !$Pattern) {
						$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"
						$alertlist_saved_flg = $true
					} else {
						if ($Cameras) {
							$match_str = ($Cameras -Join '|')
							if ($match_str -match $CamName) {
								$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"	
								$alertlist_saved_flg = $true
							}
						}				
						if ($Pattern) {
							if ($Pattern -match $CamName) {
								$al_objects_ary | ConvertTo-Json | Out-File -FilePath "$alertlist_filespec"	
								$alertlist_saved_flg = $true
							}
						}				
					}
					if ($alertlist_saved_flg) {		
						Write-Host "(dev mode) `"alertlist`" response saved to disk ($all_alerts_cnt items) --> '$alertlist_filespec'" -ForegroundColor Cyan
							#Write-Host ""
					}
					
					# Top open this *.json file in Microsoft Excel...
					# 1. Data > Get Data > From File > From JSON
					# 2. Navigate to *json file
					# 3. Click 'Import' 
					# 4. Click 'To Table'
					# 5. Click 'OK'
					# 6. Click icon by 'Column1'
					# 7. Click 'OK'
					# 8. Click 'Close & Load'
					
				}  else {
					Write-Host "(dev mode) HEADS-UP: no alerts; unable to save alertlist to disk ($alertlist_filespec)" -ForegroundColor Cyan
				}
			} else {
				Write-Host "(dev mode) HEADS-UP: cam 'alertlist' *NOT* saved to disk when using arguments -SimX" -ForegroundColor Cyan
			}
		}

		# Get all alerts files, paths
		$all_alerts_paths = $al_props_obj.paths
		$all_alerts_files = $al_props_obj.files
		if ($LatestFlg) {
			$all_alerts_cnt = 1
		} else {
			$all_alerts_cnt = $all_alerts_paths.Count
		}
		
		$all_alerts_zones = $al_props_obj.zoness

		# For debugging
		# Write-Host ("$fn_name() AlertPath = " + $AlertPath)
		# Write-Host ("$fn_name() all_alerts_paths cnt = " + $all_alerts_paths.Count)
		# Write-Host ("$fn_name() all_alerts_paths = " + ($all_alerts_paths | ConvertTo-Json))
		# Write-Host ("$fn_name() all_alerts_files = " + ($all_alerts_files | ConvertTo-Json))
		# Write-Host ("$fn_name() all_alerts_zones type = " + $all_alerts_zones.GetType())
		# Write-Host ("$fn_name() all_alerts_zones cnt = " + $all_alerts_zones.Count)
		# Write-Host ("$fn_name() all_alerts_zones = " + ($all_alerts_zones | ConvertTo-Json))
		# Write-Host "$fn_name() Found $all_alerts_cnt total items`n"

		# Initialize collectors
		$rtn_obj = [ordered]@{}
		$new_obj = [ordered]@{}
		if ($AlertProps.Count -gt 0) {
			foreach ($alert_prop in $AlertProps) {
				$new_obj.Add($alert_prop, @())
			}
		} else {
			foreach ($al_props_key in $al_props_obj_keys) {
				$new_obj.Add($al_props_key, @())
			}
		}
		
		# For debugging
		#Write-Host ("$fn_name() new_obj = " + ($new_obj | ConvertTo-Json))

		# WALK THROUGH ALERTS LIST
		for ($i=0; $i -lt $all_alerts_cnt; $i++) { 
		
			# For debugging
			# Write-Host "`nalert idx = $i..."
				
			if ($AlertPath -ne "*") {
				if (!($all_alerts_paths[$i] -match $AlertPath)) {
					Continue
				} else {
					# For debugging
					#Write-Host "Found match for AlertPath ($AlertPath)"
				}
			}
			
			if ($Pattern.Length -gt 0) {
				$test1 = $all_alerts_paths[$i] -match $Pattern
				$test2 = $all_alerts_files[$i] -match $Pattern
				if (!$test1 -and !$test2) {
					Continue
				} else {
					# For debugging
					if ($test1) {
						Write-Host ("Pattern '$Pattern' is a match for '" + $all_alerts_paths[$i] + "'")
					} else {
						Write-Host ("Pattern '$Pattern' is a match for '" + $all_alerts_files[$i] + "'")
					}
				}
			}

			if ($AlertProps.Count -gt 0) {
				
				# GET REQUESTED PROPERTIES

				foreach ($alert_prop in $AlertProps) {

					$key = $alert_prop + "s"  # $al_props_obj's keys are plural
					
					$arr = @($al_props_obj.$key)
					
					# For debugging
					#if ($key -eq 'memos') {
						#Write-Host ("$fn_name() arr[$key] = " + ($arr | ConvertTo-Json))
					#}
					
					if (!($null -eq $arr)) {  # KEY POINT: $arr can be a single element array @(0)
						
						$val = $arr[$i]
						
						# For debugging
						#if ($arr[$i] -eq '') {
							#Write-Host ($key + " = '" + $arr[$i] + "'")
						#}

						# Append value to collector
						if ($arr[$i] -or $arr[$i] -eq 0) {
							$new_obj.$alert_prop += $arr[$i]				
						} else {
							$new_obj.$alert_prop += ''
						}
						
						# For debugging
						#if ($arr[$i] -eq '') {
							#Write-Host ("new_obj.$alert_prop = " + ($new_obj.$alert_prop | ConvertTo-Json))
						#}

						# For debugging
						#Write-Host ("arr = " + $arr)
						#Write-Host ("item #" + ($i+1) + " $alert_prop = " + $arr[$i])

					} else {
						if ($i -eq 0) {
							if ($script:FromConsole) {
								Write-Host ("WARNING/nonexistent AlertProp '$alert_prop', expecting: [" + ($AlertProps -Join ',') + "]")
							}
						}
					}
				}
				
			} else {
				
				# GET ALL PROPERTIES
				
				# Get all property names
				if ($i -lt 2) {
					
					# NOTE: $al_props_obj's keys are plural
					$al_props_obj_keys = $al_props_obj.Keys

					# For debugging
					#Write-Host ("al_props_obj_keys = " + $al_props_obj_keys)
					
				}
				
				foreach ($al_props_key in $al_props_obj_keys) {

					# For debugging
					#Write-host ("new_obj.$al_props_key type = " + ($new_obj.$al_props_key).GetType())

					# Get all values for the property (an array)
					$vals = $al_props_obj.$al_props_key

					# Remove the terminal 's' from $al_props_key
					$new_obj_key = $al_props_key.Substring(0,$al_props_key.Length-1) 
					
					if ($vals) {
						$new_obj.$new_obj_key += $vals[$i]					
					} else {
						#$new_obj.$new_obj_key += $null
						$new_obj.$new_obj_key += ''
					}
					
					#For debugging
					#Write-Host ("item #" + ($i+1) + " $new_obj_key = " + $vals[$i])
				}
			}
				
			# For debugging
			#Write-host ("new_obj = " + ($new_obj | ConvertTo-Json))
		}
		
		$rtn_obj[$CamName] = $new_obj

		# For debugging
		#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
		#Write-Host "rtn_obj keys = "$rtn_obj.Keys
		#Write-Host ("rtn_obj keys = " + ($rtn_obj | ConvertTo-Json))
		#Write-Host ("rtn_obj = " + ($rtn_obj | ConvertTo-Json))s
		#Write-Host ("rtn_obj.FE = " + ($rtn_obj.FE | ConvertTo-Json))
		#Write-Host ("rtn_obj.DW1.BPS = '" + $rtn_obj.DW1.BPS + "'")
		#Write-Host ("rtn_obj.GT.path = '" + $rtn_obj.GT.path + "'")
		#Write-Host ("rtn_obj.GT.memo = '" + $rtn_obj.GT.memo + "'")
		#Write-Host ("rtn_obj.GT.flags = '" + $rtn_obj.GT.flags + "'")

		return $rtn_obj
	}

	function BIJsonGetAlertList() {
	# PURPOSE:  To get properties from the Blue Iris JSON interface cmd 'alertlist'
	# USAGE:    {this_script_name}.ps1 prop cam(s) prop(s)
	# PARAMS:   -Cameras        = (String) filter results for Alerts with specified camera shortnames(comma-delimited),  
	#                                      like 'DW1','DW2,'FD'
	#           -Patterns       = (String) filter results for Alerts with specified descriptor substrings
	#           -DatabaseView   = (String) default is 'alerts'
	#                             options: all, new, stored, alerts, aux1, aux2, aux3, aux4, aux5, aux6, aux7,
	#                                      flagged, export, archive, people, vehicles, confirmed, or cancelled (yes, double 'll')
	#                             Note: When view='flagged', clip items are included too, but they don't have msec metadata.
	#           -MemoTag        = (String) filter results for Alerts with specified memo substrings
	#           -AlertProps     = (String) return only these JSON 'alertlist' cmd params (comma-delimited )
	#                                      like 'enabled,profile,motion'
	#           -LatestFlg      = (boolean) $true returns only the most recent alerts for each -Cameras
	# RETURNS:  (hashtable) @{ Cam1: @{ prop1:[values], ... , propN:[values] }, Cam2: @{ prop1:[values], ... , propN:[values] }, ... }
	# REQUIRES: Function: BIJsonAlertListParse()
	# GLOBALS:  $script:JsonCallForceFlg, $script:FromConsole, $script:Quiet, $script:DevModeFlg, $script:DevModePath
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V10 (6/8/2024) added -SilentFlg support
	# V11 (11/11/2024) added -LatestFlg support
	# V12 (11/18/2024) removed arguments -FromConsoleFlg, -QuietFlg, -SilentFlg, -DevModeFlg, -DevModePath - now using globals
	# V13 (11/20/2024) added support for global var $script:JsonCallForceFlg
	# V14 (11/21/2024) disabled code for logging out of active JSON session
	# V15 (11/24/2024) disabled progress bar when using argument -LatestFlg
	# V16 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]
	# V17 (3/13/2025) added alert duration (property 'msec') to output

		param(
			[String[]]$Cameras,
			[String[]]$Patterns,
			[String]$DatabaseView,  # default is 'alerts'
			[String]$MemoTag,
			[String[]]$AlertProps,
			[Boolean]$LatestFlg,
			[Boolean]$SilentFlg
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function $fn_name() ..."
		# Write-Host "$fn_name() Cameras = $Cameras"
		# Write-Host "$fn_name() Patterns = $Patterns"
		# Write-Host "$fn_name() DatabaseView = $DatabaseView"
		# Write-Host ("$fn_name() AlertProps = " + ($AlertProps | ConvertTo-Json))
		# Write-Host "$fn_name() MemoTag = $MemoTag"
		# Write-Host "$fn_name() SilentFlg = $SilentFlg"
		# Write-Host "$fn_name() Sim1 = $script:Sim1"
		# Write-Host "$fn_name() Sim3 = $script:Sim3"

		# HANDLE ARGS
		if ($Cameras) {
			$my_cams_ary = $Cameras -Split ','
		} else {
			$my_cams_ary = @()
		}
		$my_cams_cnt = $my_cams_ary.Count

		if ($Patterns) {
			$my_patterns_ary = $Patterns -Split ','
		} else {
			$my_patterns_ary = @('')
		}
		$my_patterns_cnt = $my_patterns_ary.Count
		
		if ($AlertProps.Count -eq 0) {
			$my_props_ary = @('camera', 'path', 'clip', 'file', 'memo', 'offset', 'msec', 'triggeroffset', 'flags', 'res', 'zones', 'date', 'color', 'filesize')
		} else {
			$my_props_ary = $AlertProps -Split ','
		}
		$my_props_cnt = $my_props_ary.Count

		# For debugging
		# Write-Host ("$fn_name() my_cams_ary type = " + $my_cams_ary.GetType())
		# Write-Host ("$fn_name() my_cams_ary = " + $my_cams_ary)
		# Write-Host ("$fn_name() my_cams_ary cnt = " + $my_cams_ary.Count)
		# Write-Host "my_patterns_ary = $my_patterns_ary"
		# Write-Host "my_flags_ary = $my_flags_ary"
		# Write-Host ("$fn_name() my_props_ary cnt = " + $my_props_ary.Count)
		# Write-Host ("$fn_name() my_props_ary = " + ($my_props_ary | ConvertTo-Json))

		# INITIALIZE
		$rtn_obj = @{}
		$progress_bar_cntr = 0
		#$cntr = 1
		$patterns_found_ary = @()  # Used to avoid looking for already found patterns
		$i = 0
		foreach ($my_cam in $my_cams_ary) {

			# For debugging
			#Write-Host ("$fn_name() " + $my_cam + ":")
			
			# Show progress bar
			if ($script:FromConsole) {
				if (!$LatestFlg) {
					$progress_bar_cntr++
					$WriteProgressParams = @{
						Activity         = "Parsing alertlist data..."
						Status           = "Progress:  cam '$my_cam' ($progress_bar_cntr of $my_cams_cnt requested cams)"
						PercentComplete  = (($progress_bar_cntr / $my_cams_cnt) * 100)
					}
					# Note: If loop is too fast, you might need a delay to get the progress bar to update
					#Start-Sleep -Milliseconds 150				
					Write-Progress @WriteProgressParams
				}
			}
			
			$pattern_cntr = 0
			foreach ($pattern in $my_patterns_ary) {
				
				# For debugging
				#Write-Host "`n----- processing pattern '$pattern' -----"
				
				# Skip pattern search if already found
				if ($patterns_found_ary -contains $pattern) {
					continue
				}
				
				if ($pattern_cntr -eq 0) {
					$al_props_obj = $null
				}
				
				# Avoid showing the same same warnings over and over
				if ($pattern_cntr -gt 0) {
					$silent_flg = $true
				} else {
					$silent_flg = $SilentFlg
				}

				# GET JSON DATA FROM THE BLUE IRIS WEB SERVER
				<# $response_obj is an ordered hashtable, like
				 @{
				 	DW1: @{
						"file": [
				           filename1,
				           ...,
				           filenameN,
				       ],
				       "flags": [
				           int1,
				           ...,
				           int2,
				       ],
				       "memo": [
				           string1,
				           ...,
				           stringN,
				       ],
					}
				 }
				#>
				$response_obj, $al_props_obj = `
					BIJsonAlertListParse `
					-CamName $my_cam `
					-Pattern $pattern `
					-AlertProps $my_props_ary `
					-MemoTag $MemoTag `
					-DatabaseView $DatabaseView `
					-AlertPropsObj $al_props_obj `
					-LatestFlg $LatestFlg `
					-SilentFlg $silent_flg

				# For debugging
				# Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
				# Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))
				
				if ($response_obj -is [String]) {
					if ($response_obj.StartsWith('ERR/')) {
						return $response_obj
					} else {
						$err_msg = "ERR/unexpected response from AlertListParse(); $response_obj"
						return $err_msg
					}
				}

				# INITIALIZE
				$x = ''

				# EXTRACT PROPERTY VALUE(S)
				# ... Note: null/blank $response_obj.ram observed on occasion
				if ($response_obj) {	

					$response_obj_keys = $response_obj.Keys

					# For debugging
					# Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
					# Write-Host ("$fn_name() response_obj cnt = " + $response_obj.Count)
					# Write-Host ("$fn_name() response_obj keys = " + $response_obj_keys)
					# Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))

					if ($response_obj_keys -contains $my_cam) {
						
						$cam = $my_cam
						$item_obj = $response_obj[$cam]
						$item_obj_keys = $item_obj.Keys

						# $item_obj (and $new_obj below) are System.Collections.Specialized.OrderedDictionary 
						# that should look like...
						<# 
						  "camera": ["DW1"],
						  "path": ["@224803430.bvr"],
						  "clip": ["@224669140.bvr"],
						  "file": ["DW1.20240128_020000.3152644.5-0.jpg"],
						  "memo": ["LD/C;vehicle:93%"],
						  "offset": ["3157680"],
						  "flags": [402980864],
						  "res": ["2688x1520"],
						  "zones": [65],
						  "date": [1705777428],
						  "color": [16711808],
						  "filesize": ["22 sec (523K)"]
						}
						#>					

						# For debugging
						# Write-Host ("$fn_name() item_obj type = " + $item_obj.GetType())
						# Write-Host ("$fn_name() item_obj = " + ($item_obj | ConvertTo-Json))
						# Write-Host ("$fn_name() item_obj_keys = " + ($item_obj_keys | ConvertTo-Json))
						# Write-Host ("$fn_name() item_obj = " + ($item_obj | ConvertTo-Json))
						# Write-Host ("$fn_name() item_obj[0] type = " +  ($item_obj[$item_obj_keys[0]]).GetType())
						# Write-Host ("$fn_name() item_obj[0] = " +  $item_obj[$item_obj_keys[0]])
						
						# Test for empty result
						if (!$item_obj[$item_obj_keys[0]]) {  # looking for $item_obj['camera'] = @()
							continue
						} else {
							if ($my_patterns_ary.Count -gt 1) {
								$patterns_found_ary += $pattern
							}
						}

						$new_obj = [ordered]@{}	
						
						if ($my_props_ary.Length -gt 0) {
							$props_ary = $my_props_ary   # RETURN REQUESTED PROPERTIES
						} else {
							$props_ary = $item_obj_keys  # RETURN ALL PROPERTIES			
						}

						# RETURN PROPERTIES

						foreach ($cam_prop in $props_ary) {
							
							$item_obj_ary = $item_obj.$cam_prop  # expect $item_obj.path = @(value)

							# For debugging
							# Write-Host "$fn_name() Processing property '$cam_prop' ..."
							# Write-Host ("$fn_name() item_obj.$cam_prop type = " + ($item_obj.$cam_prop).GetType())
							# Write-Host ("$fn_name() item_obj.$cam_prop ary = " + ($item_obj_ary | ConvertTo-Json))
							# Write-Host ("$fn_name() cam_prop = $cam_prop, val = '" + $item_obj_ary[0] + "'")
							
							# KEY POINT: $null must put on the left side of operator!!
							# REF: https://stackoverflow.com/a/49411823
							if ($null -eq $item_obj_ary) {
								$new_obj.Add($cam_prop, @('NA'))
								$x = 'NA'  # For single camera, single property requests								
							} else {
								if ($null -eq $item_obj_ary[0]) {
									$new_obj.Add($cam_prop, @('NA'))
									$x = 'NA'  # for single camera, single property requests
								} else {
									$new_obj.Add($cam_prop, $item_obj_ary)
									$x = $item_obj_ary[0]  # for single camera, single property requests
								}
							}
							
							# For debugging
							#Write-Host ("$fn_name() " + $cam_prop + " = '"  + $x + "'")	
						}
						
						# For debugging
						# Write-Host ("$fn_name() new_obj = " + $new_obj)
						# Write-Host ("$fn_name() new_obj = " + ($new_obj | ConvertTo-Json))

						if ($my_patterns_cnt -lt 1) {
							$rtn_obj[$cam] = $new_obj
						} else {
							# Handling multiple -Patterns is a bit more complex
							foreach ($item_obj_key in $item_obj_keys) {
								if ($pattern_cntr -eq 0) {
									# First time...
									if (!$rtn_obj[$cam]) {
										# Initialize the $rtn_obj's $cam keys 
										$rtn_obj.Add($cam, @{})
									}
									# Add & initialize the $rtn_obj[$cam]'s key's prop-value pairs
									$rtn_obj[$cam].Add($item_obj_key, @{})
									$rtn_obj[$cam][$item_obj_key] = @()
								}
								$rtn_obj[$cam][$item_obj_key] += $new_obj[$item_obj_key]
							}
						}

					} else {
						if ($script:FromConsole) {
							$err_msg = "ERR/nonexistent camera"
							return $err_msg
						}
						
						$rtn_obj[$cam] = 'NA'
					}
					
				} else {
					if ($script:FromConsole) {
						Write-Warning "JSON query returned null/blank`n"
					}
				}

				# Handle single camera, single property request
				if ($rtn_obj -ne 'NA') {
					if ($my_cams_ary.Length -eq 1 -and $my_props_ary.Length -eq 1) {
						$rtn_obj = $x
					}
				}
				
				$pattern_cntr++
			}

			$i++

			# For debugging
			# Write-Host ("$fn_name() patterns_found_ary = " + ($patterns_found_ary | ConvertTo-Json))
		}

		# LOG OUT OF API SESSION
		<# $jsonCmd = [ordered]@{cmd = 'logout'}

		# Returns a powershell PSCustomObject (or string 'ERR/*')
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd `
			-ForceFlg $script:JsonCallForceFlg
			
		if ($script:FromConsole) {
			if ($json_response_obj.result -eq 'fail') {
				Write-Warning "BI API: log out fail; connection not terminated"
			} else {
				if (!$script:Quiet) {
					# Write-Host "BI API: log out successful; connection terminated"	
				}
			}
		} #>
		
		# For debugging
		# Write-Host ("$fn_name() rtn_obj = " + ($rtn_obj | ConvertTo-Json))

		return $rtn_obj
	}

	function BIJsonCamListParse() {
	# PURPOSE:   Get Blue Iris web server JSON 'camlist' command response (for all or subset of BI cameras)
	#            Intended to be called by other PowerShell scripts.
	# ARGUMENTS: -CamProps = [OPTIONAL] (list) like 'bps,bps2' ; comma delimited
	# RETURNS:  (object) PowerShell ordered hashtable, @('cam':camName,'camProp':value}
	#		                                   @{camName:{[camProp1:value1],[camProp2:value2],...},...}"
	# REQUIRES:  function BIJsonCall()
	# GLOBALS:   $script:JsonCallForceFlg, $script:FromConsole, $script:Quiet
	#
	# Description of JSON cmd 'camlist' from the Blue Iris Help PDF:
	# 	"returns a list of cameras on the system ordered by group. Cameras not belonging to any
	#	group are shown beneath the "all cameras" group. Disabled cameras are placed at the end of
	#	the list. An array of objects is returned in data (note the [] surrounding a JSON array), each
	#	describing a camera or a camera group."
	#
	# Created by jaydeel at ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1
	# V2 (11/18/2024) removed arguments -QuietFlg, -FromConsoleFlg - now using globals
	# V3 (11/20/2024) added support for global variable $script:JsonCallForceFlg
	# V4 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]
	# V5 (2/6/2025) added dev mode timing output

		# MUST BE THE FIRST LINE IN A SCRIPT
		param(
			[String]$CamProps  # [OPTIONAL] list
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() CamProps = $CamProps"

		# PARSE COMMANDLINE ARGS
		if ($CamProps) {
			$cam_props_ary = $CamProps -Split ','
			#msg.exe * /TIME:5 "camProps... $cam_props_ary"
		} else {
			$cam_props_ary = @()
		}
		#Write-Host "cam_props_ary = $cam_props_ary"

		# PERFORM API CALLS
		$jsonCmd = [ordered]@{
			cmd = 'camlist'
		}
		
		# Returns a powershell PSCustomObject (or string 'ERR/*')
		$json_response_obj = `
			BIJsonCall `
			-JsonCmd $jsonCmd `
			-ForceFlg $script:JsonCallForceFlg

		# Trap various errors
		if (1 -eq 1) {  # Use to collapse this code section while editing

			# For debugging
			# Write-Host ("$fn_name() BIJsonCall() json_response_obj type = " + $json_response_obj.GetType())
			# Write-Host ("$fn_name() BIJsonCall() json_response_obj cnt = " + $json_response_obj.Count)
			if ($null -ne $json_response_obj.result) {
				# Write-Host ("$fn_name() BIJsonCall() json_response_obj.result = " + $json_response_obj.result)
			}
			if ($null -ne $json_response_obj.data) {
				# Write-Host ("$fn_name() BIJsonCall() json_response_obj.data cnt = " + $json_response_obj.data.Count)
			}
			#Write-Host ("$fn_name() BIJsonCall() json_response_obj = " + ($json_response_obj | ConvertTo-Json))
			# msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response... $json_response_obj"
			
			if ($json_response_obj -is [String] -and $json_response_obj.StartsWith('ERR/')) {

				# For debugging
				#Write-Host "$fn_name() BIJsonCall() json_response_obj = $json_response_obj`n"
				#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj = $json_response_obj"

				return $json_response_obj
			}
			if ($json_response_obj.result -eq 'fail') {

				$x  = $json_response_obj.result

				# For debugging
				#Write-Host "$fn_name() BIJsonCall() json_response_obj.result = $x`n"
				#msg.exe * /TIME:10 "$fn_name() BIJsonCall() json_response_obj.result = $x"

				$err_msg = "ERR/JSON request failed"
				return $err_msg
			}
		}
		
		# Start performance timer
		$timer_bi_json_camlist_parse = [Diagnostics.Stopwatch]::StartNew()
		
		# Extract data from the JSON command response
		$data_obj = $json_response_obj.data

		# For error testing
		#$data_obj = $null

		# Extract property value
		# ... Note: null/blank $data_obj.ram observed on occasion
		if ($data_obj) {
			
			$data_cnt = $data_obj.Count
			
			# For debugging
			#Write-Host ("data_obj cnt = " + $data_cnt)
			#Write-Host ("data_obj[0] = `n" + $data_obj[0])
			#Write-Host ("data_obj[" + ($data_cnt - 1) +"] = `n" + $data_obj[$data_cnt-1])

			$rtn_obj = [ordered]@{}

			for ($i=0; $i -lt $data_obj.Length; $i++) { 
			
				$item_obj = $data_obj[$i]
				$option_disp = $item_obj.optionDisplay;
				$option_val  = $item_obj.optionValue;
				
				# For debugging
				#Write-Host "$i : option_val = $option_val, option_disp = $option_disp"
				#Write-Host ("    option_val starts w/ @ = " + $option_val.StartsWith("@"))
				#Write-Host ("    option_disp starts w/ + = " + $option_disp.StartsWith("+"))
				
				if ($option_val -ne "Index" -and !$option_disp.StartsWith("+") -and !$option_val.StartsWith("@")) {
					$new_obj = [ordered]@{}
					
					if ($cam_props_ary.Length -gt 0) {
						# GET REQUESTED PROPERTIES
						
						foreach ($cam_prop in $cam_props_ary) {
							#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
							if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
								$x = $item_obj.$cam_prop
								$new_obj.Add($cam_prop, $x)
								#$new_obj.$cam_prop = $x
							} else {
								$new_obj.Add($cam_prop, $null)
								#$new_obj.$cam_prop = $null
							}
						}
					} else {
						# GET ALL PROPERTIES
						
						# Get all property names in the JSON response (a PowerShell PSCustomObject)
						# ... NOTE: Extracting 'keys' from the PSCustomObject was bit tricky
						#           REF: https://stackoverflow.com/a/18780001/13499984
						$item_obj_keys = @()
						foreach ($item in $item_obj) {
							$item_obj_keys += $($item | Get-Member -MemberType *Property).Name
						}

						# For debugging
						#Write-Host ("item_obj_keys keys = " + $item_obj_keys)
						
						foreach ($cam_prop in $item_obj_keys) {
							#Write-Host ("cam_prop = $cam_prop, val = " + $item_obj.$cam_prop)
							if ($item_obj.$cam_prop -or $item_obj.$cam_prop -eq 0) {
								$x = $item_obj.$cam_prop
								$new_obj.Add($cam_prop, $x)
							} else {
								$new_obj.Add($cam_prop, $null)
							}
						}
					}
					$rtn_obj[$option_val] = $new_obj
				}
			}

			# Stop timer
			$timer_bi_json_camlist_parse.Stop()

			if ($script:FromConsole -and $script:DevMode) {
				$elapsed_msec = [Math]::Round($timer_bi_json_camlist_parse.elapsed.totalmilliseconds,1)
				$bi_json_camlist_parse_time_ms = $elapsed_msec.ToString("F3")
				Write-Host ("(dev mode) time to parse json camlist object = $bi_json_camlist_parse_time_ms msec") -ForegroundColor Cyan
			}

			# @@@ DEV - Retain for testing
			#Write-Host ("rtn_obj cnt = " + $rtn_obj.count)
			#Write-Host "rtn_obj[DW1] = "$rtn_obj.DW1
			#Write-Host "rtn_obj[DW1]['BPS'] = "$rtn_obj['DW1']['BPS']
			#Write-Host "rtn_obj keys = "$rtn_obj.Keys
			
		} else {
			#Write-Host "sorry - JSON query returned null/blank`n"
		}

		return $rtn_obj
	}

	function Convert-HexToVersion() {
	# PURPOSE: To convert a hexadecimal number like 5090704 or 5091014 to a string like 5.9.7.4 or 5.9.10.14
	# Source: ChatGPT (modified by jaydeel)

		param (
			[Parameter(Mandatory=$true)]
			[String]$Hex
		)
	 
		# For debugging
		# Write-Host "Entered $fn_name()..."
		# Write-Host "$fn_name() Hex = '$Hex'"
		
		# Ensure the hex number is a string (if given as a number)
		$HexString = $Hex.ToString()

		# Ensure the hex string has an even number of characters by prepending a '0' if needed
		if ($HexString.Length % 2 -ne 0) {
			$HexString = "0" + $HexString
		} 

		# Split the hex string into pairs of characters
		# ... inserts a period after every two characters in the string
		$pairs = for ($i = 0; $i -lt $hex.Length; $i += 2) {
			$HexString.Substring($i, 2)
		}
		
		# For debugging
		# Write-Host ("$fn_name() pairs = " + ($pairs | ConvertTo-Json))

		# Convert each pair to its decimal equivalent and join them with '.'
		return ($pairs | ForEach-Object { [convert]::ToInt32($_, 16) }) -join '.'
	}

	function FindBytes() {
	# PURPOSE:  Find all starting indices in $Bytes where $Search was found
	# ARGS:     see param lines below
	# RETURNS:  (array) integers, or empty array @()
	# Attributes:   https://stackoverflow.com/a/62511302/13499984
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# Version 4 (5/22/2024) Added arg -Reverse; 
	# Version 3 (5/5/2024) added comments to clarify usage

		param (
			[Byte[]]$Bytes,    # [REQUIRED]
			[Byte[]]$Search,   # [REQUIRED]
			[Int]$StartIndex,  # start searching array $Bytes at index X
			[Int]$StopIndex,   # stop searching array $Bytes at index Y
			[Switch]$Reverse,  # search backwards
			[Switch]$All       # returns all matches, otherwise returns FIRST match only
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "entered function $fn_name()..."
		# Write-Host ("$fn_name() Bytes = " + ($Bytes -Join ' '))
		# Write-Host ("$fn_name() Search = " + ($Search -Join ' '))
		# Write-Host "$fn_name() All = $All"
		# Write-Host "$fn_name() StartIndex = $StartIndex"
		# Write-Host "$fn_name() Reverse = $Reverse"
		
		[Int[]]$rtn_ary = @()
		
		if ($Reverse -and ($StopIndex -gt $StartIndex)) {
			if ($script:FromConsole) {
				Write-Warning "$fn_name() param -Reverse requires -StopIndex <= -StartIndex"
			}
			return $rtn_ary
		}
			
		if (!$Reverse) {
			
			if ($StopIndex) {
				$start_index = $StartIndex
				$stop_index  = $StopIndex
			} else {
				$start_index = $StartIndex
				$stop_index  = $Bytes.Length
			}

			# For debugging
			# Write-Host "$fn_name() start_index = $start_index"
			# Write-Host "$fn_name() stop_index = $stop_index"
		
			#for ($idx = $StartIndex; $idx -le $Bytes.Length - $Search.Length ; $idx++) {
			for ($idx = $start_index; $idx -le $stop_index - $Search.Length ; $idx++) {
				for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
				if ($i -ge $Search.Length) { 
					$rtn_ary += $idx
					if (!$All) { 
						return $rtn_ary
					}
				}
			}
			
		} else {
			
			if ($StopIndex) {
				$start_index = $StartIndex
				$stop_index  = $StopIndex
			} else {
				$start_index = $StartIndex
				$stop_index  = 0
			}

			# For debugging
			# Write-Host "$fn_name() start_index = $start_index"
			# Write-Host "$fn_name() stop_index = $stop_index"
			
			#for ($idx = $StartIndex - $Search.Length ; $idx -ge 0 ; $idx--) {
			for ($idx = $start_index - $Search.Length ; $idx -ge $stop_index ; $idx--) {
				for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
				if ($i -ge $Search.Length) { 
					$rtn_ary += $idx
					if (!$All) {
						return $rtn_ary
					}
				}
			}
			
		}
		
		# KEY POINT: By default, Powershell functions will return $null if attempting to return an empty array.
		# e.g., return @()
		# To avoid this, the options are:
		# 1) return ,$rtn_ary  ... where the comma is the Powershell 'array construction operator'
		# 2) write-output $arr -NoEnumerate 
		# REF: https://stackoverflow.com/a/18477004

		return ,$rtn_ary
	}

	<# function FindBytes_OLD() {
	# PURPOSE:  Find all starting indices in $Bytes where $Search was found
	# ARGS:     see param lines below
	# RETURNS:  (array) integers, or empty array @()
	# Attributes:   https://stackoverflow.com/a/62511302/13499984
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# Version 4 (5/22/2024) Added arg -Reverse; 
	# Version 3 (5/5/2024) added comments to clarify usage

		param (
			[Byte[]]$Bytes,    # [REQUIRED]
			[Byte[]]$Search,   # [REQUIRED]
			[Int]$StartIndex,  # start searching array $Bytes at index N (rare use case)
			[Switch]$Reverse,  # search backwards
			[Switch]$All       # returns all matches, otherwise returns FIRST match only
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "entered function $fn_name()..."
		# Write-Host ("$fn_name() Bytes = " + ($Bytes -Join ' '))
		# Write-Host ("$fn_name() Search = " + ($Search -Join ' '))
		# Write-Host "$fn_name() All = $All"
		# Write-Host "$fn_name() StartIndex = $StartIndex"
		# Write-Host "$fn_name() Reverse = $Reverse"
		
		[Int[]]$rtn_ary = @()
		
		if (!$Reverse) {
			
			for ($idx = $StartIndex; $idx -le $Bytes.Length - $Search.Length ; $idx++) {
				for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
				if ($i -ge $Search.Length) { 
					$rtn_ary += $idx
					if (!$All) { 
						return $rtn_ary
					}
				}
			}
			
		} else {
			
			for ($idx = $StartIndex - $Search.Length ; $idx -ge 0 ; $idx--) {
				for ($i = 0; $i -lt $Search.Length -and $Bytes[$idx + $i] -eq $Search[$i]; $i++) {}
				if ($i -ge $Search.Length) { 
					$rtn_ary += $idx
					if (!$All) {
						return $rtn_ary
					}
				}
			}
			
		}
		
		# KEY POINT: By default, Powershell functions will return $null if attempting to return an empty array.
		# e.g., return @()
		# To avoid this, the options are:
		# 1) return ,$rtn_ary  ... where the comma is the Powershell 'array construction operator'
		# 2) write-output $arr -NoEnumerate 
		# REF: https://stackoverflow.com/a/18477004

		return ,$rtn_ary
	} #>

	function RepairDatFile() {
	# PURPOSE: Try to repair a DAT file with unexpected structure [unbalanced JPG beg/end markers]
	# ARGS:    see param()
	# NOTE:    Repair is attempted only when the count of 'end markers' <> the count of 'beg markers'
	# RETURNS  [array] @([Boolean]rtn_flg, [array]SOI_mrks, [array]EOI_mrks)
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1
	# V2 (12/8/2024) changed '[array]$x = $y' typing to '$x = @($y)'

		param (
			[Byte[]]$Bytes,
			[Array]$BegOfImageMarks,
			[Array]$EndOfImageMarks
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		#Write-Host "Entered $fn_name()..."
		#Write-Host ("$fn_name() BegOfImageMarks type = " + $BegOfImageMarks.GetType())
		#Write-Host ("$fn_name() BegOfImageMarks cnt = " + $BegOfImageMarks.Count)
		#Write-Host "$fn_name() BegOfImageMarks = $BegOfImageMarks"
		#Write-Host ("$fn_name() EndOfImageMarks type = " + $EndOfImageMarks.GetType())
		#Write-Host ("$fn_name() EndOfImageMarks cnt = " + $EndOfImageMarks.Count)
		#Write-Host "$fn_name() EndOfImageMarks = $EndOfImageMarks"
		
		# Initialize vars
		$rtn_flg = $true  # Assume success
		$SOI_mrks_cnt = $BegOfImageMarks.Count
		$EOI_mrks_cnt = $EndOfImageMarks.Count		

		# Must convert arrays to ArrayLists to use method 'RemoveAt()' (later)
		$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$BegOfImageMarks
		$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EndOfImageMarks

		# For debugging
		#Write-Host ("$fn_name() SOI_mrks_lst type = " + $SOI_mrks_lst.GetType())
		#Write-Host ("$fn_name() SOI_mrks_lst cnt = " + $SOI_mrks_lst.Count)
		#Write-Host "$fn_name() SOI_mrks_lst = $SOI_mrks_lst"
		#Write-Host ("$fn_name() EOI_mrks_lst type = " + $EOI_mrks_lst.GetType())
		#Write-Host ("$fn_name() EOI_mrks_lst cnt = " + $EOI_mrks_lst.Count)
		
		if ($SOI_mrks_cnt -gt $EOI_mrks_cnt) {

			for ($i=0; $i -lt $EOI_mrks_cnt; $i++) {
				$indices_delta = [Int]$EndOfImageMarks[$i] - [Int]$BegOfImageMarks[$i]
				
				# For debugging
				#Write-Host ("- end index = [$i] " + $EndOfImageMarks[$i])
				#Write-Host "- indices_delta = $indices_delta"

				if ($indices_delta -lt 0) {
					
					$SOI_mrks_lst_new = $SOI_mrks_lst
					$SOI_mrks_lst_new.RemoveAt($i)

					# For debugging
					#Write-Host "starting recursive call..."
					#Write-Host "- SOI_mrks_lst_new = $SOI_mrks_lst_new"

					# Recursive call
					$arr1 = @($SOI_mrks_lst_new)
					$arr2 = @($EOI_mrks_lst)
					$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
						RepairDatFile `
						-Bytes $bytes `
						-BegOfImageMarks $arr1 `
						-EndOfImageMarks $arr2

					# For debugging
					#Write-Host "recursive result_flg = $result_flg"

					if ($result_flg) {
						$rtn_flg = $true
						$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
						$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
						break
					} else {
						$rtn_flg = $false
					}				
				}
			}					
		}
		
		if ($EOI_mrks_cnt -ge $SOI_mrks_cnt) {

			for ($i=0; $i -lt $SOI_mrks_cnt; $i++) {
				$indices_delta = [Int]$EndOfImageMarks[$i] - [Int]$BegOfImageMarks[$i]
				
				# For debugging
				#Write-Host ("- beg index = [$i] " + $BegOfImageMarks[$i])
				#Write-Host "- indices_delta = $indices_delta"
				
				if ($indices_delta -lt 0) {
					
					$EOI_mrks_lst_new = $EOI_mrks_lst
					$EOI_mrks_lst_new.RemoveAt($i)

					# For debugging
					#Write-Host "starting recursive call..."
					#Write-Host "- EOI_mrks_lst_new = $EOI_mrks_lst_new"

					# Recursive call
					$arr1 = @($SOI_mrks_lst)
					$arr2 = @($EOI_mrks_lst_new)
					$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
						RepairDatFile `
						-Bytes $bytes `
						-BegOfImageMarks $arr1 `
						-EndOfImageMarks $arr2
						
					# For debugging
					#Write-Host "recursive result_flg = $result_flg"

					if ($result_flg) {
						$rtn_flg = $true
						$SOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$SOI_mrk_idxs_ary
						$EOI_mrks_lst = [System.Collections.Generic.List[System.Object]]$EOI_mrk_idxs_ary
						break
					} else {
						$rtn_flg = $false
					}				
				}
			}
		}

		return @($rtn_flg, [array]$SOI_mrks_lst, [array]$EOI_mrks_lst)

	}

	function IsJpegFileValid() {
	# PURPOSE: To confirm that a JPG file is a valid JPG image
	# NOTES:
	#    Using 2 methods:
	#    1. Test image type
	#       REFS:  http://learningpcs.blogspot.com/2011/07/powershell-validate-jpeg-files.html
	#              https://www.powershellgallery.com/packages/PScribo/0.9.1/Content/Src%5CPrivate%5CGet-ImageMimeType.ps1
	#    2. Per the JPEG Interchange Format... any JPEG file or stream that is written according to the 
	#       JPEG Interchange Format means that they must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) 
	#       and end with the EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
	#       These two markers will not be found anywhere else in a JPEG file/stream.
	#       REFS: https://stackoverflow.com/a/46809226
	#             https://www.codeproject.com/Questions/421855/Verifying-a-file-is-JPEG
	# DISCUSSION:
	#    Originally I used only Method #1 in this function. However, I learned later that Method #1 would still return valid
	#    if the image bytes did not terminate with the EOI (End-of-Image) marker (0xFF, 0xD9). These JPG files still opened 
	#    just fine in every windows viewer I tried (Photos, Paint, Paint3D, PhotoScape X, ACDSee...), HOWEVER, if they were 
	#    opened in a web browser (tested Chrome, Edge) the images had a white band across the bottom. 
	#    Additionally, web apps like https://app.roboflow.com/ would not import the JPGs.
	#    So clearly, Method 1 is not sufficient. For redundancy, I've opted to use both methods. 
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V2 (11/24/2024) added testing for loaded assembly

		param(
			[String]$FilePath
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging.
		#msg.exe $script:server_username /TIME:15 "entered function $fn_name()..."
		#Write-Host "entered function $fn_name()..."
		#Write-Host "$fn_name() FilePath = '$FilePath'"
		#msg.exe $script:server_username /TIME:30 "$fn_name() FilePath = '$FilePath'"
		
		# Initialize flags
		$method1_flg = $false  # Validate image file Type
		$method2_flg = $false  # Validate bytes per the 'JPEG Interchange Format' (see Notes above)
		
		# ===========================
		# METHOD 1 - Validate image file Type

		# LOAD REQUIRED ASSEMBLIES
		# WARNING: The following assembly IS ESSENTIAL when executing this function from Task Scheduler (or Blue Iris)
		#          (on the other hand, PS7 and Powershell ISE works just fine without it)
		#          REF: https://tinyurl.com/2p929nhz
		# NOTE: While repeatedly executing 'Add-Type -AssemblyName System.Drawing' is not harmful in terms of 
		# functionality (Powershell automatically checks if the assembly is already loaded), checking whether 
		# the assembly is already loaded improves efficiency and avoids redundancy. (per ChatGPT)
		if (-not [System.Drawing.Image]) {
			Add-Type -AssemblyName System.Drawing
		}
		
		try {
			$img = [System.Drawing.Image]::FromFile("$FilePath");
			$method1_flg = $img.RawFormat.Equals([System.Drawing.Imaging.ImageFormat]::Jpeg)
			$img.Dispose()  # Release resources

		} catch [OutOfMemoryException] {
			$method1_flg = $false
		}

		# ===========================
		# METHOD 2 - Validate bytes per the 'JPEG Interchange Format' (see Notes above)
		
		# Import the file to a byte[] array
		$ImageBytes = [System.IO.File]::ReadAllBytes($FilePath)
		
		# Capture the start and end byte pairs.
		#$bytes_rng1 = $ImageBytes[0],$ImageBytes[1]
		#$bytes_rng2 = $ImageBytes[$ImageBytes.Count-2],$ImageBytes[$ImageBytes.Count-1]
		$bytes_rng1 = $ImageBytes[0..1]
		$bytes_rng2 = $ImageBytes[($ImageBytes.Count-2)..($ImageBytes.Count-1)]

		# Define the required JPG start/end markers
		$markers1 = @(255,216) # hexArray @("FF", "D8")
		$markers2 = @(255,217) # hexArray @("FF", "D9")
		<# $hexArray = "FF", "D8"
		$markers1 = @()
		foreach ($hex in $hexArray) {
			$markers1 += [Int32]"0x$hex"
		}
		$hexArray = "FF", "D9"
		$markers2 = @()
		foreach ($hex in $hexArray) {
			$markers2 += [Int32]"0x$hex"
		} #>

		# Validate the start and end byte pairs.
		# ... Note: Method Compare-Object returns null if the arrays are the identical
		$x1 = Compare-Object $bytes_rng1 $markers1
		$x2 = Compare-Object $bytes_rng2 $markers2
		if ($null -eq $x1 -and $null -eq $x2) {
			$method2_flg = $true
		} else {
			$method2_flg = $false
		}
		
		# For testing - Simulate Method 2 (Validate bytes per the 'JPEG Interchange Format')
		#$method2_flg = $true

		# For debugging
		#Write-Host "markers1 = $markers1, markers2 = $markers2"
		#Write-Host "bytes_rng1 = $bytes_rng1, bytes_rng2 = $bytes_rng2"
		#Write-Host "method1_flg = $method1_flg, method2_flg = $method2_flg"
		#msg.exe $script:server_username /TIME:15 "$fn_name() method1_flg = $method1_flg, method2_flg = $method2_flg"
		
		if ($method1_flg -and $method2_flg) {
			return $true
		} else {
			return $false
		}
	}

	function ExtractDatImages() {
	# PURPOSE:  Extract all JPG files from a Blue Iris *.DAT file, and save them to disk
	#           with filenames like 'img01.jpg', 'img02.jpg', ...
	# ARGS:     see param lines below
	# RETURNS:  1. (hashtable) @{ source=filespec ; datjpgs=N ; data=@(jpgfilelist) }
	#           2. or (String) if issue/fail (NOTE: string must begin with 'ISSUE:', or 'FAIL:')
	# GLOBALS:  Uses global vars: $BiVersion, $FromConsole, $Quiet
	#           $script:BiVersion   = (String) Blue Iris software current version
	#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
	#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
	#
	# hastable return object is like...
	# {
	#   "source": "D:\\BlueIris\\Alerts\\DW1.20240306_080000.1950096.5-0.dat",
	#   "jpgstotal": 13,
	#   "jpgsextracted": 
	#      [
	#         "D:\\BI_gif_tool\\DW1\\img05.jpg",
	#         "D:\\BI_gif_tool\\DW1\\img04.jpg",
	#         "D:\\BI_gif_tool\\DW1\\img03.jpg",
	#         "D:\\BI_gif_tool\\DW1\\img02.jpg",
	#         "D:\\BI_gif_tool\\DW1\\img01.jpg"
	#      ]
	# }
	#
	# ATTRIBUTIONS:
	#    Per Ken @blueirissupport... "The DAT file is custom serialized format. 
	#    To extract JPGs: search the byte content for blocks that start with 'FF D8 FF E0' and go through 'FF D9'. 
	#    To extract the JSON string: after the bitmap there are 2 QWORDs (8 bytes each), then the STRING for the JSON 
	#    (comprised of a DWORD length then the characters)."
	#
	#    Per the JPEG Interchange Format... any JPEG file or stream is written according to the JPEG Interchange 
	#    Format must start with a SOI (Start-of-Image) marker (the two bytes 0xFF, 0xD8) and end with the 
	#    EOI (End-of-Image) marker (two bytes, 0xFF, 0xD9). 
	#    These two markers will not be found anywhere else in a JPEG file/stream.
	#    REF: https://stackoverflow.com/a/46809226
	#
	# KEY POINT: 
	#    JPG images in a BI *.DAT file are stored in reverse chronological order
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# Version   3 (8/29/2024) faster getting JPG image beg/end markers; replaced *Flg' args w/ global vars
		
		param(
			[String]$DatFile,            # DAT file, like 'GT.20230610_070000.41437.17-0.jpg' (via JSON 'clipstats'), or 'alert_BY.dat'
			[String]$DatPath,            # folder path for Blue Iris 'DAT' files (full path) (via user-setting $reg_bi_alerts_folder_path)
			[Boolean]$LatestFlg,         # use DAT file from the most recent alert (that has a DAT file)
			[String]$DestPath,           # folder for the extracted JPG files
			[Int]$DatReqdImages,         # number of images that must be found before extracting any images	
			[Int]$ImagesCnt,             # number of images to be extracted from the BVR file, if the DAT file determined to be 'bad'	
			[Boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
			[Int]$ImagesTrimBegPct,      # percent of images to skip at front
			[Int]$ImagesTrimEndPct       # percent of images to skip at end
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		$PsVersion = $PSVersionTable.PSVersion.Major

		# For debugging
		# Write-Host "Entered function: $fn_name()"	
		# Write-Host "$fn_name() DatFile = '$DatFile'"
		# Write-Host "$fn_name() DatPath = '$DatPath'"
		# Write-Host "$fn_name() LatestFlg = $LatestFlg"
		# Write-Host "$fn_name() DestPath = '$DestPath'"
		# Write-Host ("$fn_name() DatReqdImages type = " + $DatReqdImages.GetType())
		# Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
		# Write-Host "$fn_name() ImagesCntUsedFlg = $ImagesCntUsedFlg"
		# Write-Host "$fn_name() ImagesTrimBegPct = $ImagesTrimBegPct"
		# Write-Host "$fn_name() ImagesTrimEndPct = $ImagesTrimEndPct"
		# Write-Host ("$fn_name() BiVersion = " + $script:BiVersion)
		# Write-Host ("$fn_name() Quiet = " + $script:Quiet)
		# Write-Host ("$fn_name() FromConsole = " + $script:FromConsole)
		# Write-Host "$fn_name() PsVersion = $PsVersion"	
		
		# HANDLE ARGS
		$DestPath = $DestPath.TrimEnd('\')
		
		# Initialize
		$dat_file_found_flg = $false  # Default
		$rtn_obj = [ordered]@{}
		
		# Extract camname from DAT filename
		if ($DatFile.StartsWith('alert_')) {
			# DAT file in the BI Temp folder (typ. C:\ProgramData\Blue Iris\temp)
			$cam = $DatFile.Replace('alert_', '').Replace('.dat', '')
		} else {
			# DAT file in the BI Alerts folder
			$str = Split-Path $DatFile -leaf
			$arr = $str.Split(".")
			$cam = $arr[0]
		}
		
		# For debugging
		# Write-Host "$fn_name() str = '$str'"
		# Write-Host ("$fn_name() arr = " + ($arr | ConvertTo-Json))
		# Write-Host "$fn_name() cam = '$cam'"
		
		# TEST FOR DESTINATION FOLDER FOR JPGs
		$jpg_path =  $DestPath + "\" + $cam
		if (!(Test-Path -LiteralPath $jpg_path)) {
			New-Item -ItemType Directory -Force -Path $jpg_path
		}
		
		# For debugging
		# Write-Host "$fn_name() jpg_path = '$jpg_path'"

		# DELETE EXISTING *.JPG FILES ASYNCHRONOUSLY
		Get-Childitem -Path ($jpg_path + "\img*.jpg") | Remove-Item -force
		
		# CONFIRM DAT FILE EXISTENCE
		# Note: extension replacement should not be needed, but performed just in case
		#       the JSON clipstats file was passed directly
		$dat_file = "$DatPath\$DatFile".Replace('.jpg','.dat')
		if (Test-Path -LiteralPath $dat_file) {
			$dat_file_found_flg = $true
		}
		
		# For debugging
		# Write-Host "$fn_name() dat_file_found_flg = $dat_file_found_flg"

		# PROCEED IF 'DAT' FILE FOUND
		if ($dat_file_found_flg) {
			
			if ($script:FromConsole -and !$script:Quiet) {
				Write-Host "`nAlert's AI 'DAT' File Processing Summary:"
				Write-Host "-----------------------------------------"
				Write-Host "Found DAT file: '$dat_file'"
			}
		
			# LOAD THE 'DAT' FILE
			if ($PsVersion -ge 7) {
				#$bytes = [System.IO.File]::ReadAllBytes("$dat_file")
				# Ref: https://stackoverflow.com/a/63782608/13499984
				$bytes = Get-Content -LiteralPath $dat_file -AsByteStream -Raw
			} else {
				$bytes = Get-Content -LiteralPath $dat_file -Encoding Byte -Raw
			}
			$bytes_cnt = $bytes.Count

			# For debugging
			#Write-Host ("$fn_name() bytes len = " + $bytes.Count)
			#Write-Host $bytes
			#$msg = $bytes.Count
			#msg.exe $script:server_username /TIME:60 "$fn_name() bytes len = $msg"

			# //////////////////////////////////////////////////////////////////////////////////
			
			# FIND INDICES FOR ALL IMAGE DELIMITERS

			# Start timer
			# $timer_dat_img_delim_indicies = [Diagnostics.Stopwatch]::StartNew()
		
			# The first 6 bytes of the file looks like @('05', '81', version)
			$search_ary = $bytes[0..5]

			# For debugging
			# Write-Host ("search_ary = " + ($search_ary -Join ' '))

			$img_delim_mrk_idxs = `
				FindBytes `
				-Bytes $bytes `
				-Search $search_ary `
				-All

			$img_delim_mrk_idxs_cnt = $img_delim_mrk_idxs.Count

			# For debugging
			# Write-Host "$fn_name() img_delim_mrk_idxs_cnt = $img_delim_mrk_idxs_cnt"
			# Write-Host "$fn_name() img_delim_mrk_idxs = $img_delim_mrk_idxs"

			# Stop timer
			# $timer_dat_img_delim_indicies.Stop()
			# $elapsed_msec = [Math]::Round($timer_dat_img_delim_indicies.elapsed.totalmilliseconds,1)
			# $elapsed_msec_str = $elapsed_msec.ToString("F2")
			# Write-Host ("timer_dat_img_delim_indicies = $elapsed_msec_str msec")

			# Determine the padding level to use for hashtable keys and safe JPG & JSON filenames
			# ... left padding is necessary in hashtable keys to ensure that the output table sorts the images in natural order
			# ... left padding is helpful in filenames to ensure that files appear in order when viewed in the Windows File Explorer
			# $pad_level = ([Math]::log10($img_delim_mrk_idxs_cnt) - ([Math]::log10($img_delim_mrk_idxs_cnt) % 1)) + 1

			# Start timer
			# $timer_dat_jpg_indicies = [Diagnostics.Stopwatch]::StartNew()

			# Build jpg byte start/end markers
			# ... JPG begin marker
			#$search_ary1 = @(255,216,255,224) # hexArray @("FF", "D8", "FF", "E0")
			$hexArray = "FF", "D8", "FF", "E0"
			$search_ary1 = @()
			foreach ($hex in $hexArray) {
				$search_ary1 += [Int32]"0x$hex"
			}
			# ... JPG end marker
			#$search_ary2 = @(255,217) # hexArray @("FF", "D9")
			$hexArray = "FF", "D9"
			$search_ary2 = @()
			foreach ($hex in $hexArray) {
				$search_ary2 += [Int32]"0x$hex"
			}

			$found_ary1 = @()
			$found_ary2 = @()

			for ($i = 0; $i -lt $img_delim_mrk_idxs_cnt ; $i++) {
				
				$img_no = $img_delim_mrk_idxs_cnt - $i 
				# $img_no_key = $img_no.ToString().PadLeft($pad_level)
				
				$img_SOI_search_beg_idx = $img_delim_mrk_idxs[$i] + 6
									
				# GET THE INDEX OF THE **NEXT** SOI MARKER
				$img_SOI_idxs = `
					FindBytes `
					-Bytes $bytes `
					-Search $search_ary1 `
					-StartIndex $img_SOI_search_beg_idx
				
				$found_ary1 += $img_SOI_idxs[0]

				# GET THE INDEX OF THE **PREVIOUS** EOI MARKER
				# ... KEY-POINT: at the last delim markers index we need to search backwards twice
				#     1st, from the delim marker
				#     2nd, from the end of array $bytes
				if ($i -gt 0) {
					if ($i -le $img_delim_mrk_idxs_cnt - 1) {
						$img_EOI_idxs = `
							FindBytes `
							-Bytes $bytes `
							-Search $search_ary2 `
							-StartIndex $img_SOI_search_beg_idx `
							-Reverse
							
						$found_ary2 += $img_EOI_idxs[0]

						if ($i -eq $img_delim_mrk_idxs_cnt - 1) {
							$end_idx = $bytes.Length - 1
							$img_EOI_idxs = `
								FindBytes `
								-Bytes $bytes `
								-Search $search_ary2 `
								-StartIndex $end_idx `
								-Reverse
								
							$found_ary2 += $img_EOI_idxs[0]
						}
					}
				}
			}

			$found_ary1_cnt = $found_ary1.Count
			$found_ary2_cnt = $found_ary2.Count

			# For debugging
			# Write-Host "$fn_name() jpg data starting byte pattern = $search_ary1"
			# Write-Host "$fn_name() jpg data ending byte byte pattern = $search_ary2"
			# Write-Host "$fn_name() jpg data start indices = $found_ary1"
			# Write-Host "$fn_name() jpg data end indices = $found_ary2"
			# msg.exe $script:server_username /TIME:60 "$fn_name()... `nfound_ary1_cnt = $found_ary1_cnt `nfound_ary2_cnt = $found_ary2_cnt"
			# Write-Host "$fn_name() found_ary1_cnt = $found_ary1_cnt"

			# Stop timer
			# $timer_dat_jpg_indicies.Stop()
			# $elapsed_msec = [Math]::Round($timer_dat_jpg_indicies.elapsed.totalmilliseconds,1)
			# $elapsed_msec_str = $elapsed_msec.ToString("F2")
			# Write-Host ("timer_dat_jpg_indicies (SOI & EOI) = $elapsed_msec_str msec")

			# //////////////////////////////////////////////////////////////////////////////////

			# Validate expected DAT images count
			if ($found_ary1_cnt -lt $DatReqdImages) {
				<# if ($script:FromConsole) {
					$bad_DAT_msg = "DAT file has fewer images than required ($found_ary1_cnt/$DatReqdImages)... "
					$bad_DAT_msg += "`n   Please navigate to Blue Iris 'Camera settings' > 'Trigger' tab > 'Artificial Intelligence' dialog"
					$bad_DAT_msg += "`n   - Confirm that setting 'To cancel' has a value like DoNotCancel"
					$bad_DAT_msg += "`n   - Confirm that setting '+ real-time images' has a value = $DatReqdImages (or " + ($DatReqdImages-1) + " if using motion-leading image)"
					$bad_DAT_msg += "`nTrying to extract images from BVR instead..."
					Write-Host $bad_DAT_msg -ForegroundColor red
				} #>
				$rtn_obj = "ISSUE: insufficient JPG images count in DAT file"
				return $rtn_obj
			}
			# Validate expected DAT structure
			# ... expect $found_ary1 like '25 113202'
			# ...        $found_ary2 like '113054 227880'
			$bad_DAT_file_flg = $false
			$i = 0
			foreach ($jpg_idx in $found_ary1) {
				if ($i -gt 0) {
					if ([Int]$jpg_idx -ge [Int]$found_ary2[$i - 1]) {
						$bad_DAT_file_flg = $true
						$bad_DAT_msg = "DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
						break
					}
				}
			}
			
			# For debugging
			# Write-Host "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"
			# msg.exe $script:server_username /TIME:60 "$fn_name() bad_DAT_file_flg = $bad_DAT_file_flg"

			$extract_success_flg = $false  # Assume failure

			if (!$bad_DAT_file_flg) {

				# DAT file has at least one JPG file
				if ($found_ary1_cnt -gt 0) {

					# DAT file structure is NOT as expected
					if ($found_ary1_cnt -ne $found_ary2_cnt) {
					
						# Try to repair DAT file marker arrays before giving up
						$result_flg, $SOI_mrk_idxs_ary, $EOI_mrk_idxs_ary = `
							RepairDatFile `
							-Bytes $bytes `
							-BegOfImageMarks $found_ary1 `
							-EndOfImageMarks $found_ary2
							
						# For debugging
						#Write-Host "result_flg = $result_flg"

						if ($result_flg) {
							$found_ary1 = $SOI_mrk_idxs_ary
							$found_ary2 = $EOI_mrk_idxs_ary
							$found_ary1_cnt = $SOI_mrk_idxs_ary.Count
							$found_ary2_cnt = $EOI_mrk_idxs_ary.Count

							if ($script:FromConsole) {
								if ($found_ary1_cnt -gt $found_ary2_cnt) {
									Write-Host "REPAIRED 'beg marker' indices: now = $jpg_SOI_mrk_idxs"
								} else {
									Write-Host "REPAIRED 'end marker' indices; now = $jpg_EOI_mrk_idxs"
								}
							}
						}
					}

					# DAT file structure is as expected
					if ($found_ary1_cnt -eq $found_ary2_cnt) {

						if ($found_ary1_cnt -ge $DatReqdImages) {
				
							if ($ImagesCntUsedFlg) {
								if ($script:FromConsole) {
									Write-Warning "Argument '-ImagesCnt' ignored; it applies only to GIFs created from BVR files"
								}
							}

							$jpgs_cnt = $found_ary1_cnt
							$jpg_idx_beg = [Math]::Floor($found_ary1_cnt*$ImagesTrimBegPct/100)
							$jpg_idx_end = $found_ary1_cnt - [Math]::Floor($found_ary1_cnt*$ImagesTrimEndPct/100)
							$jpgs_cnt_adj = $jpg_idx_end - $jpg_idx_beg
							
							# For debugging
							# Write-Host "$fn_name() jpgs_cnt = $jpgs_cnt"
							# Write-Host "$fn_name() jpgs_cnt_adj = $jpgs_cnt_adj"
							# Write-Host "$fn_name() jpg_idx_end = $jpg_idx_end"
							# Write-Host "$fn_name() jpg_idx_beg = $jpg_idx_beg"

							$rtn_obj['source'] = $dat_file
							$rtn_obj['jpgstotal'] = $jpgs_cnt
							$jpgs = @()
							$jpgs_cntr = 0
							#$bi_dat_viewer_time_ms = 0
							
							# EXTRACT THE JPG IMAGES
							
							# Initialize timer collectors
							$jpg_files_names_msec = 0
							$jpg_files_save_msec = 0
							$jpg_files_validate_msec = 0
							
							# KEY POINT: Before 5.9.4.11. JPGs in DAT files were stored newest-to-oldest
							$dat_images_oldest_first_flg = `
								IsBiVersion `
								-CompareToVersion '5.9.4.11' `
								-CurrentVersion $script:BiVersion
							
							# Start timer
							# $timer_jpgs_extract = [Diagnostics.Stopwatch]::StartNew()

							$i = 0
							foreach ($item in $found_ary1) {
								
								# For debugging
								# Write-Host ("$fn_name() item type = " + $item.GetType())
								# Write-Host "$fn_name() item = $item"

								if ($dat_images_oldest_first_flg) {
									$img_no = $found_ary1_cnt - $i
								} else {
									$img_no = $i + 1
								}
								
								# Extract image if in target range
								if ($img_no -ge $jpg_idx_beg -and $img_no -le $jpg_idx_end) {
								
									$bounding_idx1 = $item
									$bounding_idx2 = $found_ary2[$i] + 1  # Include the EOI marker!
									$img_bytes = $bytes[$bounding_idx1..$bounding_idx2]
								
									if ($img_bytes.Count -gt 0) {	

										# Start timer
										$timer_jpg_filename = [Diagnostics.Stopwatch]::StartNew()

										# Build JPG's filespec
										if ($dat_images_oldest_first_flg) {
											$jpg_filename_idx = $jpgs_cntr + 1
										} else {
											$jpg_filename_idx = $jpgs_cnt_adj - $jpgs_cntr 
										}								
										$jpg_filename_sfx = ([String]$jpg_filename_idx).PadLeft(2,'0')
										$jpg_filespec = "$jpg_path\img$jpg_filename_sfx.jpg"

										# Stop timer
										# $timer_jpg_filename.Stop()
										# $elapsed_msec = [Math]::Round($timer_jpg_filename.elapsed.totalmilliseconds,1)
										# $jpg_files_names_msec += $elapsed_msec
										# $elapsed_msec_str = $elapsed_msec.ToString("F2")
										# Write-Host ($img_no.toString() + " timer_jpg_filename = $elapsed_msec_str msec")

										# Start timer
										# $timer_jpg_save = [Diagnostics.Stopwatch]::StartNew()
									
										# Save the JPG file to disk					
										if ($PsVersion -ge 6) {
											# REF: https://stackoverflow.com/a/63782608/13499984
											Set-Content "$jpg_filespec" -Value $img_bytes -AsByteStream 
										} else {
											Set-Content "$jpg_filespec" -Value $img_bytes -Encoding Byte
										}

										# Stop timer
										# $timer_jpg_save.Stop()
										# $elapsed_msec = [Math]::Round($timer_jpg_save.elapsed.totalmilliseconds,1)
										# $jpg_files_save_msec += $elapsed_msec
										# $elapsed_msec_str = $elapsed_msec.ToString("F2")
										# Write-Host ($img_no.toString() + " timer_jpg_save = $elapsed_msec_str msec")

										# Start timer
										# $timer_jpg_validate = [Diagnostics.Stopwatch]::StartNew()

										# Confirm that the extracted data is a valid JPG file
										$jpg_valid_flg = `
											IsJpegFileValid `
											-FilePath $jpg_filespec

										# For debugging
										#msg.exe $script:server_username /TIME:60 "$jpg_filespec is valid = $jpg_valid_flg"

										# Stop timer
										# $timer_jpg_validate.Stop()
										# $elapsed_msec = [Math]::Round($timer_jpg_validate.elapsed.totalmilliseconds,1)
										# $jpg_files_validate_msec += $elapsed_msec
										# $elapsed_msec_str = $elapsed_msec.ToString("F2")
										# Write-Host ($img_no.toString() + " jpg_files_validate_msec = $elapsed_msec_str msec")
										
										if (!$jpg_valid_flg) {
											if ($script:FromConsole) {
												Write-Warning "DAT file export aborted because '$jpg_filespec' is NOT a valid JPG"
											}
											break
										} else {
											$jpgs_cntr++
											$jpgs += $jpg_filespec
										
											# For debugging
											#Write-Host "$fn_name() jpgs = $jpgs"
											#Write-Host "'$jpg_filespec' is a valid JPG"
										}
									}
								}

								$i++
							}

							# Stop timer
							# $timer_jpgs_extract.Stop()
							# $elapsed_msec = [Math]::Round($timer_jpgs_extract.elapsed.totalmilliseconds,1)
							# $elapsed_msec_str = $elapsed_msec.ToString("F2")
							# Write-Host ("timer_jpg_files_names = $jpg_files_names_msec msec")
							# Write-Host ("timer_jpg_files_save = $jpg_files_save_msec msec")
							# Write-Host ("timer_jpg_files_validate = $jpg_files_validate_msec msec")
							# Write-Host ("timer_jpgs_extract (overall) = $elapsed_msec_str msec")

							#Write-Host "bi_dat_viewer_time_ms = $bi_dat_viewer_time_ms" 

							$rtn_obj['jpgsextracted'] = $jpgs
							
							if ($jpgs.Count -gt 0) {
								$extract_success_flg = $true
							}

							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host ("Extracted $jpgs_cntr of $jpgs_cnt total JPG files to folder '$jpg_path'")
							}
							
						} else {
							$rtn_obj = "FAIL: DAT file contains only $found_ary1_cnt/$DatReqdImages required JPGs"							
						}			

					} else {
						$rtn_obj = "FAIL: DAT file has unexpected structure [unbalanced JPG beg/end markers] ($dat_file)"
					}
				} else {
					$rtn_obj = "FAIL: No JPGs found in DAT file '$dat_file'"			
				}		
			} else {
				$rtn_obj = "FAIL: " + $bad_DAT_msg
			}
		} else {
			$rtn_obj = "FAIL: Unable to find DAT file '$dat_file'"
		}
		
		# For debugging
		#msg.exe $script:server_username /TIME:60 "extract_success_flg = $extract_success_flg"
			
		return $rtn_obj
	}

	function IsFileFresh {
	# PURPOSE:   To test if a file was recently created/modified
	# ARGUMENTS: -FilePath     = [REQUIRED] full filespec for the file to be tested
	#            -CutOffAgeSec = [REQUIRED] file's age in seconds must be less than this number
	#
	# Suggested by ChatGPT (2/17/2025)
	# Modified by jaydeel @ ipcamtalk.com
	#
	# V1 (2/18/2025) 

		param (
			[string]$FilePath,
			[int]$CutOffAgeSec
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name()..."
		
		if (!$FilePath -or !$CutOffAgeSec) {
			return $false
		}

		if (Test-Path $FilePath) {
			
			$fileInfo = Get-Item $FilePath
			$threshold = (Get-Date).AddSeconds(-$CutOffAgeSec)
			
			$rtn_flg = ($fileInfo.LastWriteTime -ge $threshold -or $fileInfo.CreationTime -ge $threshold)

			# For debugging
			<# $ts_str = ($fileInfo.LastWriteTime).ToString("yyyy/MM/dd_HH:mm:ss")
			$file_age = (Get-Date) - $fileInfo.LastWriteTime
			$file_age_in_sec = [math]::Round($file_age.TotalSeconds)
			if (!$rtn_flg) {
				Write-Host ("'$FilePath' last modified $file_age_in_sec sec ago ($ts_str)") -ForegroundColor Red
			} else {
				Write-Host ("'$FilePath' last modified $file_age_in_sec sec ago ($ts_str)")
			} #>
			
			return $rtn_flg
		}

		return $false  # Return false if the file does not exist
	}

	function CopyFileAsSequence() {
	# PURPOSE:  Copy a source file without over writing; append a sequence number '(N)' 
	#           to the end of its basename until it is unique in the destination folder 
	# ARGS:     see param lines below
	# RETURNS:  (String) new_filespec, or 'fail'
	# ATTRIBUTIONS:  https://stackoverflow.com/a/53595350/13499984
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# Version 1

		param(
			[String]$SourceFilespec,  # [REQUIRED]
			[String]$TargetFilespec,  # [REQUIRED]
			[String]$DestPath         # [REQUIRED]
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "$fn_name() SourceFilespec = $SourceFilespec"
		# Write-Host "$fn_name() TargetFilespec = $TargetFilespec"
		# Write-Host "$fn_name() DestPath = $DestPath"
		
		$rtn_str = 'fail'  # Assume failure

		# Just in case
		$DestPath = $DestPath.TrimEnd('\')

		if ((Test-Path -LiteralPath $SourceFilespec)) {

			if ((Test-Path -LiteralPath $DestPath)) {

				$fileInfo = Get-Item -Path $TargetFilespec
				$base_name = $fileInfo.BaseName
				$extension = $fileInfo.Extension

				# Get an array of all filenames (name only) of the files with a similar name already present in the destination folder
				$allFiles = @(Get-ChildItem $DestPath -File -Filter "$base_name*$extension" | Select-Object -ExpandProperty Name)

				# Construct the new filename
				$new_name = $base_name + $extension  # Initialize
				$count = 1
				while ($allFiles -contains $new_name) {
					# add a sequence number in brackets to the end of the basename until it is unique in the destination folder
					$new_name = "{0}({1}){2}" -f $base_name, $count++, $extension
				}

				# Construct the new filespec
				$new_target_fs = Join-Path -Path $DestPath -ChildPath $new_name

				try {
					Copy-Item -Path $SourceFilespec -Destination $new_target_fs
					$rtn_str = $new_target_fs
				} catch {
					$rtn_str = "ERR/failed to create file '$new_target_fs'"
				}
			} else {
				$rtn_str = "ERR/destination folder does not exist ('$DestPath')"
			}
		} else {
			$rtn_str = "ERR/source file does not exist ('$FileName')"
		}

		return $rtn_str

	}

	function PruneOldFiles() {
	# PURPOSE:  Delete old files (> N hours) from a filepath like 'D:\myfolder\*.gif'
	# ARGS:     see param lines below
	# RETURNS:  (array) @(found_files_cnt, pruned_files_cnt, oldest_file_hours, oldest_file_path)
	#           or (String) 'ERR/' message
	# ATTRIBUTIONS:  https://stackoverflow.com/a/46086926
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# Version 1
		
		param(
			[String]$Pattern,  # [REQUIRED] full filespec pattern for targeted files; like 'D:\myfolder\*.jpg'
			[Int]$Hours        # [REQUIRED] maximum age for retained files; use 24 for 1 day, 168 for 1 week... 
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		if (!$Pattern -or !$Hours) {
			return "ERR/$fn_name() requires 2 arguments (-Pattern -Hours)"
		}

		# For debugging
		#Write-Host "Entered function: $fn_name()"
		#Write-Host "$fn_name() Pattern = $Pattern"
		#Write-Host "$fn_name() Hours = $Hours"
		
		# GET FILES MATCHING PATTERN
		$cur_ary = @(get-childitem -Path $Pattern | Select Name)
		if ($cur_ary) {
			$cur_ary_cnt = $cur_ary.Count
		} else {
			$cur_ary_cnt = 0
		}

		# For debugging
		#Write-Host "$fn_name() cur_ary cnt = $cur_ary_cnt"
		#Write-Host "$fn_name() cur_ary = $cur_ary"

		# GET FILES MATCHING AGE REQUIREMENT
		$prune_ary = @()
		if ($cur_ary_cnt -gt 0 -and $Hours) {
			
			$limit = (Get-Date).AddHours(-$Hours)
			
			# For debugging
			#Write-Host "$fn_name() limit = $limit"

			$prune_obj = Get-ChildItem -Path $Pattern | 
				Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
				Select Name

			if ($prune_obj) {
				$prune_ary = $prune_obj.Name
			}
		}
		
		# Get count for return array
		$prune_ary_cnt = $prune_ary.Count
		
		# For debugging
		#Write-Host "$fn_name() prune_ary cnt = $prune_ary"
		#Write-Host ("$fn_name() prune_ary = " + ($prune_ary | ConvertTo-Json))

		# REMOVE FILES MATCHING AGE REQUIREMENT ASYNCHRONOUSLY
		if ($prune_ary_cnt -gt 0) {
			if ($Hours) {
				$null = Get-ChildItem -Path $Pattern | 
					Where-Object { !$_.PSIsContainer -and $_.CreationTime -lt $limit } | 
					Remove-Item -Force
			}
		}

		# GET OLDEST REMAINING FILE
		$oldest_file = dir $Pattern | sort LastWriteTime | select -First 1
		$oldest_file_age_days = ((Get-Date) - $oldest_file.LastWriteTime).Days
		$oldest_file_age_hours = 24 * $oldest_file_age_days
		
		# For debugging
		#Write-Host ("$fn_name() oldest_file = " + $oldest_file.FullName)
		#Write-Host ("$fn_name() oldest_file_age_days = " + $oldest_file_age_days)
		#Write-Host ("$fn_name() oldest_file_age_hours = " + $oldest_file_age_hours)
		
		return  @($cur_ary_cnt, $prune_ary.Count, $oldest_file_age_hours, $oldest_file.Name)
	}

	function Show-HashtableAsTable {
		param (
			[hashtable]$HashTable
		)

		$tableOutput = $HashTable.GetEnumerator() | 
					   Select-Object Key, Value | 
					   Format-Table | 
					   Out-String -Stream
	 
		Write-Output $tableOutput
	}

	function MakeGIF() {
	# PURPOSE:  Make GIF image file to be used as attachment for notifications (Pushover, Telegram)
	# ARGS:     see param lines below
	# RETURNS:  [hashtable] @{
	#			'method'    = $gif_method; 
	#			'file'      = $output_fs; 
	#			'delay'     = $delay_ms; 
	#			'cnt'       = $ImagesCnt; 
	#			'datjpgs'   = $dat_jpgs_cnt; 
	#			'height'    = $Height; 
	#			'fps'       = $FPS; 
	#			'kb'        = $gif_kb ; 
	#			'msec'      = $gif_dur_ms; 
	#			'trimbeg'   = $gif_trim_bef_pct; 
	#			'trimend'   = $gif_trim_end_pct; 
	#			'trimcust'  = $gif_trim_end_cust_pct; 
	#			'almsec'    = $al_dur_ms; 
	#			'alfresh'   = $al_fresh; 
	#			'res'       = $gif_jpg_WxH ; 
	#			'cam'       = $al_cam; 
	#			'memo'      = $al_memo; 
	#			'source'    = $source_fs; 
	#			'archive'   = $archive_gif_fs
	#			'dbrec'     = $clip_db_locator
	#			'timing_ht' = $makegif_timing_ht
	#			}
	#           or (String) if error
	# NOTES:    1. Be sure to check if 'gif_size' = 0, as this indicates the GIF file was not actually created
	#           2. Pushover API attachment maximum size is 2,621,440 bytes (2621 KB)
	# GLOBALS:  Used global vars: $FromConsole, $Quiet, $DevMode 
	#           $script:FromConsole = (boolean) $true notifies function that calling command was executed from the console
	#           $script:Quiet       = (boolean) $true suppresses verbose output to the console
	#           $script:DevMode     = (boolean) $true
	# REQUIRES: IsBiVersion(), BIJsonClipstats(), IsCamRecContinuous(), GetPreTrigRecMsec(), GetPreTrigPlaybackMsec(),
	#           GetAiAnalyzeOneEachMsec(), IsFileFresh(), ExtractDatImages(), CopyFileAsSequence()
	# ATTRIBUTIONS:
	#    Thanks to bp2008@ipcamtalk for providing clues to Blue Iris HTTP command to extract JPGs from a BVR
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V2 (8/29/2024) replaced *Flg args w/ global vars
	# V3 (11/10/2024) added timing hashtable (individual codes sections; for performance & bottle neck debugging)
	# V4 (11/24/2024) added testing for loaded assembly
	# V5 (12/8/2024) changed coding [ $x.GetType() -match "string"] to [ $x -is [String] ]
	#                changed [ $x = [array]$y ] typing to [ $x = @($y) ]
	# V6 (2/6/2025) modified timers to minimize unaccounted execution time ('Other')
	# V7 (2/19/2025) add support for temp DAT files in the Blue Iris temp folder

		
		param(
			[String]$RootPath,           # [REQUIRED] root folder used by this function (full path)
			[Array]$BvrPaths,            # [REQUIRED] full folder paths for Blue Iris' 'BVR' files (typ. /New/ folders)
			[String]$DatPath,            # [REQUIRED] full folder path for Blue Iris' 'DAT' files (typ. /Alerts/ folder)
			[String]$TempPath,           # [REQUIRED] alt full folder path for Blue Iris' 'DAT' files (C:ProgramData/BlueIris/temp/ folder)
			[String]$AlertRec,           # Alert record, like '@123456789'
			[Boolean]$LiveFlg,           # create GIF from specified camera's stream
			[String]$Camera,             # camera shortname (required with -Live argument)
			[Int]$Seconds,               # GIF duration (-1 uses alert's duration via JSON cmd 'clipstats'
			[Boolean]$UseDatFileFlg,     # true extracts JPGs from AI DAT if it exists AND meets req'ts (exists?, sufficient images?)
			[Boolean]$LatestFlg,         # use DAT file from the most recent alert (that has a DAT file)
			[Int]$DatReqdImages,         # number of images that must be found before extracting any images	
			[Boolean]$IgnoreDatAgeFlg,   # true processes a temporary DAT file regardless of whether it is 'fresh'
			[Int]$DatIsFreshMult,        # required age for use of temporary DAT files in 'C:\ProgramData\Blue Iris\temp' (v5.9.9.29+)
			[Int]$DatImageIntervalMsec,  # number of milliseconds between images in the DAT file
			[Int]$BvrDelayMsec,          # delay fetching JPGs from live BVR for this many milliseconds
			[Boolean]$NoPushFlg,         # true prevents limiting GIF size to below 5.0MB
			[Int]$ImagesCnt,             # number of JPG frames to extract from BVR
			[Boolean]$ImagesCntUsedFlg,  # true indicates that the param -ImagesCnt was used on the command line
			[Int]$Height,                # height of image extracted from BVR in pixels
			[Int]$FPS,                   # GIF file frames per second
			[Int]$TrimBegPct,            # trim beginning of alert clip by this percentage
			[Int]$TrimEndPct,            # trim end of alert clip by this percentage
			[Int]$TrimEndAutoPct,        # trim end of alert clip by this custom AI-determined percentage
			[Int]$ClipDurAdjPct          # percentage to reduces alert's duration (after applying -TrimBegPct and -TrimEndPct)
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"	
		# Write-Host "$fn_name() RootPath = '$RootPath'"
		# Write-Host ("$fn_name() BvrPaths = " + ($BvrPaths | ConvertTo-Json))
		# Write-Host "$fn_name() DatPath = '$DatPath'"
		# Write-Host "$fn_name() TempPath = '$TempPath'"
		# Write-Host "$fn_name() AlertRec = '$AlertRec'"
		# Write-Host "$fn_name() LiveFlg = '$LiveFlg'"
		# Write-Host "$fn_name() Camera = '$Camera'"
		# Write-Host "$fn_name() Seconds = $Seconds"
		# Write-Host ("$fn_name() Seconds = " + $Seconds.GetType())
		# Write-Host "$fn_name() UseDatFileFlg = $UseDatFileFlg"
		# msg.exe $script:server_username /TIME:10 "UseDatFileFlg = $UseDatFileFlg"		
		# Write-Host "$fn_name() LatestFlg = $LatestFlg"
		# Write-Host "$fn_name() DatReqdImages = $DatReqdImages"
		# Write-Host "$fn_name() DatIsFreshMult = $DatIsFreshMult"
		# Write-Host "$fn_name() DatImageIntervalMsec = $DatImageIntervalMsec"
		# Write-Host "$fn_name() BvrDelayMsec = $BvrDelayMsec"
		# Write-Host "$fn_name() NoPushFlg = $NoPushFlg"
		# Write-Host "$fn_name() ImagesCnt = $ImagesCnt"
		# Write-Host ("$fn_name() ImagesCntUsedFlg type = " + $ImagesCntUsedFlg.GetType())
		# Write-Host "$fn_name() ImagesCntUsedFlg = $ImagesCntUsedFlg"
		# Write-Host "$fn_name() Height = $Height"
		# Write-Host "$fn_name() FPS = $FPS"
		# Write-Host ("$fn_name() TrimBegPct type = " + $TrimBegPct.GetType())
		# Write-Host "$fn_name() TrimBegPct = $TrimBegPct"
		# Write-Host "$fn_name() TrimEndPct = $TrimEndPct"
		# Write-Host "$fn_name() TrimEndAutoPct = $TrimEndAutoPct"
		# Write-Host "$fn_name() ClipDurAdjPct = $ClipDurAdjPct"

		# Write-Host ("$fn_name() Quiet = " + $script:Quiet)
		# Write-Host ("$fn_name() FromConsole = " + $script:FromConsole"
		# Write-Host ("$fn_name() DevMode = " + $script:DevMode"
		# Write-Host ("$fn_name() reg_cam_rec_is_continuous = " + $script:reg_cam_rec_is_continuous)
		# Write-Host ("$fn_name() reg_cam_pretrig_rectime_msec = " + $script:reg_cam_pretrig_rectime_msec)

		# For debugging
		#$bg_msg =  "Entered function: $fn_name()"
		#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on

		# LOAD REQUIRED ASSEMBLIES
		# WARNING: The following assembly IS ESSENTIAL when executing this function from Task Scheduler (or Blue Iris)
		#          (on the other hand, PS7 and Powershell ISE (PS5.1) works just fine without it)
		#          REF: https://tinyurl.com/2p929nhz
		# NOTE: While repeatedly executing 'Add-Type -AssemblyName System.Drawing' is not harmful in terms of 
		# functionality (Powershell automatically checks if the assembly is already loaded), checking whether 
		# the assembly is already loaded improves efficiency and avoids redundancy. (per ChatGPT)
		if (-not [System.Drawing.Image]) {
			Add-Type -AssemblyName System.Drawing
		}
		
		# Initialize vars
		$alert_rec = $AlertRec
		$db_rec = $DbRecNo
		$clip_db_locator = $null
		$clipstats_success_flg = $false
		$trim_end_pct = 0
		# KEY POINT: may be changed programmatically if a recent DAT file cannot be found
		$use_dat_file_flg = $UseDatFileFlg  # Default per user's request
		
		# Initialize collector
		$makegif_timing_ht = @{}
		
		# PARSE OPTIONS
		# ... LIVE GIF
		if ($LiveFlg) {

			# Override default or requested settings
			$use_dat_file_flg = $false

			# Start performance timer - Get BVR file
			$timer_get_bvr = [Diagnostics.Stopwatch]::StartNew()
		
			# Get the MOST RECENT BVR FILE for the specified camera
			# NOTE: Even though we use the live video stream to grab JPGs for a 'Live' GIF
			#       we still use the active BVR filename to name the GIF file
			$bvr_timestamp = Get-Date -Format "yyyyMMdd_"
			$pattern = "$Camera.$bvr_timestamp*bvr"  # like, 'DW2.20231103'
			$bvr_file_found = $false
			# ... Find the New folder path
			foreach ($path in $BvrPaths) {
				#$bvr_file_obj = ((Get-ChildItem "$path\*" -Recurse -Include $pattern) | Sort -Descending) | Select -First 1
				# This version ALSO works w/ non-default file naming patterns in the Blue Iris Settings > Record tab
				$bvr_file_obj = ((Get-ChildItem "$path\*" -Recurse -Include $pattern) | Sort LastWriteTime) | Select -Last 1
				if ($bvr_file_obj) {
					$bvr_file_found = $true
					break
				}
			}
			
			# For debugging
			#Write-Host "$fn_name() bvr path = '$reg_bi_new_folder_path\*'"
			#Write-Host "$fn_name() pattern = '$pattern'"
			#Write-Host ("$fn_name() bvr_file_obj type = " + $bvr_file_obj.GetType())
			#Write-Host ("$fn_name() bvr_file_obj = " + ($bvr_file_obj | ConvertTo-Json))
			#Write-Host ("$fn_name() bvr_file_obj.FullName = " + $bvr_file_obj.FullName)
			#Write-Host ("$fn_name() bvr_file_obj.Name = " + $bvr_file_obj.Name)
			#Write-Host "$fn_name() bvr_file = '$bvr_file'"

			if ($bvr_file_found) {
				
				if ($bvr_file_obj.FullName) {
					
					# MAKE ELEMENTS FOR THE ARCHIVED GIF FILE NAME				
					$bvr_filename = $bvr_file_obj.Name  # like, 'DW1.20240302_130000.bvr'
					$arr = ($bvr_filename).Split('.')
					# like, ["DW1", "20240302_130000", "bvr"]
					$bvr_path = $arr[0] + "." + $arr[1] + ".bvr"

					# CALCULATE THE NOW OFFSET TIME - based the most recent BVR file's creation time
					$bvr_filespec = $bvr_file_obj.FullName
					$file_dt = (Get-Item -LiteralPath $bvr_filespec).CreationTime
					$diff = (Get-Date).Subtract($file_dt)
					$bvr_offset_ms = [Math]::Floor($diff.TotalMilliSeconds)

					# Make the filename for the archived GIF file				
					# ... like, 'DW1.20240302_130000.1464437.gif'
					$archive_gif_path = $arr[0] + "." + $arr[1] + "." + $bvr_offset_ms + ".gif"
					
					# For debugging
					#Write-Host "$fn_name() bvr_filespec = '$bvr_filespec'"
					#Write-Host "$fn_name() bvr_filename = '$bvr_filename'"
					#Write-Host "$fn_name() bvr_path = '$bvr_path'"
					#Write-Host ("$fn_name() bvr_filename arr = " + ($arr | ConvertTo-Json))
					#Write-Host "$fn_name() bvr_offset_ms = '$bvr_offset_ms'"
					#Write-Host "$fn_name() archive_gif_path = '$archive_gif_path'"
					
					<# if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "Active bvr_filespec:    $bvr_filespec"
						Write-Host "Active bvr_file offset: $bvr_offset_ms msec"
					} #>

					# For debugging
					#$bg_msg = "$fn_name() Active bvr_file:`n$bvr_filespec`nbvr_file offset: $bvr_offset_ms ms"
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on

				} else {
					$err_msg = "ERR/unable to get active BVR filename: '$bvr_filespec'"
					return $err_msg
				}

			} else {
				$err_msg = "ERR/unable to find existing BVR file"
				return $err_msg
			}
			
			# Stop timer
			$timer_get_bvr.Stop()
			$elapsed_msec = [Math]::Round($timer_get_bvr.elapsed.totalmilliseconds,1)
			$get_bvr_time_ms = $elapsed_msec.ToString("F0")
			# Write-Host ("$fn_name() json_alertlist_time_ms = $json_alertlist_time_ms msec")
			
			# Add timing to collector
			$makegif_timing_ht.Add('get_bvr_time_ms', $get_bvr_time_ms)

		}

		# ... ALL OTHER OPTIONS
		#     i.e., if ($LatestFlg -or $AlertRec -or $DbRecNo -or $UI3Rec) {  
		if (!$LiveFlg) {
			
			# HANDLE SPECIAL CASE - LATEST ALERT
			# Get the alert's db locator
			if ($LatestFlg) {

				# Start performance timer
				$timer_get_json_alertlist = [Diagnostics.Stopwatch]::StartNew()
				
				# GET BLUE IRIS DB ALERTLIST
				<#response_obj is like...
				{
				  "DW1": {
					"path": ["@194831594.bvr"],
					"memo": ["vehicle:73%"],
					"date": [1732226250],
					"file": ["DW1.20241121_130000.3445154.17-0.jpg"]
				  }
				}
				#>
				$response_obj = `  # returns a hashtable
					BIJsonGetAlertList `
					-Cameras $Camera `
					-LatestFlg $LatestFlg `
					-AlertProps 'path','file','memo','date','msec' `
					-ExcludeClones
				
				# For debugging
				# Write-Host ("$fn_name() response_obj type = " + $response_obj.GetType())
				# Write-Host ("$fn_name() response_obj cnt = " + $response_obj.Count)
				# Write-Host ("$fn_name() response_obj = " + ($response_obj | ConvertTo-Json))

				# For testing - Simulate a bad JSON alertlist response
				#$response_obj = "ERR/test"
				
				if ($response_obj.Count -eq 0) {
					$timer_get_json_alertlist.Stop()
					$err_msg = "ERR/unexpected response from BIJsonGetAlertList(); returned object is empty"
					return $err_msg
				}					
				if ($response_obj -is [String]) {
					$timer_get_json_alertlist.Stop()
					if ($response_obj.StartsWith('ERR/')) {
						return $response_obj	
					} else {
						$err_msg = "ERR/unexpected response from BIJsonGetAlertList()"
						return $err_msg
					}
				}
				$alertlist_ht = $response_obj
				
				# For debugging
				# Write-Host ("$alertlist_ht = " + ($alertlist_ht | ConvertTo-Json))

				# LOAD THE LATEST ALERT'S PROPERTIES ARRAYS
				$al_paths_ary = @($alertlist_ht.$Camera.path)  # Force array
				$al_files_ary = @($alertlist_ht.$Camera.file)
				$al_memos_ary = @($alertlist_ht.$Camera.memo)
				$al_dates_ary = @($alertlist_ht.$Camera.date)
				$al_durs_ary  = @($alertlist_ht.$Camera.msec)
				
				# MAKE THE LATEST ALERT'S DB LOCATOR
				$clip_db_locator = $al_paths_ary[0].Replace('.bvr', '')
				$clip_db_timestamp = $al_dates_ary[0]  # used only for -DevMode output
			
				# Stop timer
				$timer_get_json_alertlist.Stop()
				$elapsed_msec = [Math]::Round($timer_get_json_alertlist.elapsed.totalmilliseconds,1)
				$json_alertlist_time_ms = $elapsed_msec.ToString("F0")
				# Write-Host ("$fn_name() json_alertlist_time_ms = $json_alertlist_time_ms msec")
				
				# Add timing to collector
				$makegif_timing_ht.Add('json_alertlist_time_ms', $json_alertlist_time_ms)
				
				# Show latest alert's info
				if ($FromConsole -and $DevMode) {
					Write-Host ""
					Write-Host ("-" * 67) -ForegroundColor Cyan 
					Write-Host ("(dev mode) BIJsonGetAlertList() response:") -ForegroundColor Cyan
					Write-Host ("Latest alert = `n" + ($response_obj | ConvertTo-Json)) -ForegroundColor Cyan
					Write-Host ("Latest Alert's db_locator = '$clip_db_locator'") -ForegroundColor Cyan
					$formatteddate = Convert-FromUnixDate $clip_db_timestamp "M/d/yyyy h:mm:ss tt"
					Write-Host ("Latest Alert's db_timestamp = " + $clip_db_timestamp + " <-- " + $formatteddate) -ForegroundColor Cyan
					Write-Host ("-" * 67) -ForegroundColor Cyan 
				}

			}

			# Start performance timer
			$timer_get_json_clipstats = [Diagnostics.Stopwatch]::StartNew()

			# GET THE REQUESTED ALERT'S CLIPSTATS DETAILS
			# ... try up to 3x before giving up, wait 200 msec before each subsequent try
			$clipstats_iter_max = 3
			$clipstats_wait_ms = 200
			for ($clipstats_iter=1; $clipstats_iter -le $clipstats_iter_max; $clipstats_iter++) {
				
				# For debugging
				# Write-Host "$fn_name() clipstats object iteration #$clipstats_iter"
				# msg.exe $script:server_username /TIME:5 "trying to get alert's clipstats, iteration #$clipstats_iter"
			
				# Make the alert's db record
				if ($LatestFlg) {
					# Note: $clip_db_locator already defined for the 'Latest' Alert
				} else {
					if ($alert_rec.Length -gt 0) {
						$clip_db_locator = $alert_rec
					}

					# For debugging
					# Write-Host "$fn_name() clip_db_locator = $clip_db_locator"
					# msg.exe $script:server_username /TIME:20 "clip_db_locator = $clip_db_locator"
					
					if ($null -eq $clip_db_locator) {
						$err_msg = "ERR/DB record is null"
						return $err_msg							
					}
				}
				
				# Ensure db rec is prefixed w/ '@'
				if (!$clip_db_locator.StartsWith('@')) {
					$clip_db_locator += "@" + $clip
				}

				# For testing
				# $clip_db_locator += "2"	# test a BAD locator
				# Write-Host "$fn_name() clip_db_locator MOD = $clip_db_locator"

				# GET THE CLIPSTATS OBJECT						
				# Typically takes 200-300 msec
				
				# Do it!
				<# $clip_stats_obj should be like ... 
				[
					{
						"result": "success"
					},
						"data": {
							"@152866668.bvr": {
								"camera": "DW1",
								"path": "@152720932.bvr",
								"file": "DW1.20230530_060000.1965063.17-0.jpg",
							"date": 123456789
								"offset": 2016632,
							"triggeroffset": 2018632,
								"alertmsec": 27426,
								"camprofile:" 7,
								"res": "2688x1520"
								"memo": "LD/C;person:91%"
							}
						}
					}
				]
				 OR for a non-existent db_locator, the response looks like
				[
					{
						"result": "success"
					},
						"data": {
							"@176867793": "NA"
						}
					}
				[
				#>
				$clip_stats_obj = `  # returns an ordered hashtable
					BIJsonClipstats `
					-Clips $clip_db_locator `
					-Props 'camera,path,file,date,offset,triggeroffset,alertmsec,camprofile,res,memo'
				
				# For debugging
				# Write-Host ("$fn_name() clip_stats_obj type = " + $clip_stats_obj.GetType())
				# Write-Host ("$fn_name() clip_stats_obj cnt = " + $clip_stats_obj.Count)
				# Write-Host ("$fn_name() clip_stats_obj = " + ($clip_stats_obj | ConvertTo-Json))

				# Trap various errors
				$err_msg = ''  # any error will retry getting the $clip_stats_obj
				if ($clip_stats_obj -is [String]) {
					$timer_get_json_clipstats.Stop()
					if (!$clip_stats_obj.StartsWith("ERR/")) {
						$err_msg = "ERR/" + $clip_stats_obj
					} else {
						$err_msg = $clip_stats_obj
					}
					# return $err_msg
				} 
				if ($clip_stats_obj.Count -eq 0) {
					$timer_get_json_clipstats.Stop()
					$err_msg = "ERR/unexpected response from BIJsonClipstats(); returned object is empty"
					# return $err_msg
				}
				if (!($clip_stats_obj['result'] -eq "success")) {
					$timer_get_json_clipstats.Stop()
					$err_msg = "ERR/unable to get properties for DB record ($clip_db_locator) via BIJsonClipstats()"
					# return $err_msg
				}
				if ($clip_stats_obj['data'][$clip_db_locator] -is [String]) {
					if ($clip_stats_obj['data'][$clip_db_locator] -eq 'NA') {
						$timer_get_json_clipstats.Stop()
						$err_msg = "ERR/BIJsonClipstats() returned no properties for DB record ($clip_db_locator)"
						# return $err_msg
					}
				} else {
					if ($clip_stats_obj['data'][$clip_db_locator].Count -eq 0) {
						$timer_get_json_clipstats.Stop()
						$err_msg = "ERR/BIJsonClipstats() returned no properties for DB record ($clip_db_locator)"
						# return $err_msg
					}
				}
				
				# If error, try again or give up?
				if ($err_msg.Length -gt 0) {
					if ($clipstats_iter -lt $clipstats_iter_max) {
						Start-Sleep -Milliseconds $clipstats_wait_ms
						Continue
					} else {
						return $err_msg
					}
				}
				
				# OK to proceed
				$clipstats_success_flg = $true

				# Extract the clipstats data						
				$clipstats_ht = $clip_stats_obj['data']  # an ordered hashtable
				
				# Stop timer
				$elapsed_msec = [Math]::Round($timer_get_json_clipstats.elapsed.totalmilliseconds,1)
				$json_clipstats_time_ms = $elapsed_msec.ToString("F0")
				# Write-Host ("$fn_name() json_clipstats_time_ms = $json_clipstats_time_ms msec")
				
				# Add timing to collector
				$makegif_timing_ht.Add('json_clipstats_time_ms', $json_clipstats_time_ms)
				
				# For debugging
				# Write-Host ("$fn_name() clipstats_ht type = " + $clipstats_ht.GetType())
				# Write-Host ("$fn_name() clipstats_ht = " + ($clipstats_ht | ConvertTo-Json))
						
				# OUTPUT CUSTOMIZED CLIPSTATS OBJECT
				# ... we've added a human-readable date in format that matches the Blue Iris clips list
				if ($script:FromConsole -and !$script:Quiet) {

					# Deep clone the ORDERED hashtable
					# ... necessary to avoid changing the original object
					$json = $clipstats_ht | ConvertTo-Json -Depth 100
					$temp_cloned_obj = $json | ConvertFrom-Json
					$cloned_obj = [ordered]@{}
					foreach ($key in $temp_cloned_obj.PSObject.Properties.Name) {
						$cloned_obj[$key] = $temp_cloned_obj.$key
					}
					# For debugging
					# Write-Host ("$fn_name() cloned_obj = " + ($cloned_obj | ConvertTo-Json))

					$cloned_obj_keys = @($cloned_obj.Keys)
					$co_clip_db_locator = $cloned_obj_keys[0]  # 'co' == 'cloned object'; should match var $clip_db_locator
					
					# For debugging
					# Write-Host "$fn_name() cloned_obj_keys = $cloned_obj_keys"
					# Write-Host "$fn_name() co_clip_db_locator = $co_clip_db_locator"
					# Write-Host ("$fn_name() co_clip_db_locator type = " + $co_clip_db_locator.GetType())
					
					$timestamp = [Int64]($cloned_obj.$co_clip_db_locator.date)
					
					# For debugging
					# Write-Host ("$fn_name() cloned_obj.$co_clip_db_locator = " + ($cloned_obj.$co_clip_db_locator | ConvertTo-Json))
					# Write-Host ("$fn_name() cloned_obj.$co_clip_db_locator.date = " + $cloned_obj.$co_clip_db_locator.date)
					# Write-Host ("$fn_name() cloned_obj.$co_clip_db_locator timestamp = " + $timestamp.GetType())
					# Write-Host ("$fn_name() cloned_obj.$co_clip_db_locator timestamp = " + $timestamp)

					# Append the formated date to the cloned object's 'date' key
					if ($timestamp) {  # Just in case; avoids warning 'You cannot call a method on a null-valued expression'
						$formatteddate = Convert-FromUnixDate $timestamp "M/d/yyyy h:mm:ss tt"
						$cloned_obj.$co_clip_db_locator.date = $timestamp.ToString() + " == " + $formatteddate
					}
					
					Write-Host "Alert's JSON 'clipstats' response:"
					Write-Host "----------------------------------"
					Write-Host ($cloned_obj | ConvertTo-Json)
				}
				
				break
			}

			# LOAD THE ALERT'S PROPERTIES ARRAYS
			if ($LatestFlg) {
				# This is already done for the 'Latest' Alert
			} else {				
				$camname = $clipstats_ht[$clip_db_locator]['camera']
				$temp_ht = [ordered]@{}
				$temp_ht.Add($camname,@{})
				$temp_ht[$camname].Add('path',$clipstats_ht[$clip_db_locator]['path'])
				$temp_ht[$camname].Add('file',$clipstats_ht[$clip_db_locator]['file'])
				$temp_ht[$camname].Add('memo',$clipstats_ht[$clip_db_locator]['memo'])
				$temp_ht[$camname].Add('date',$clipstats_ht[$clip_db_locator]['date'])
				$temp_ht[$camname].Add('alertmsec',$clipstats_ht[$clip_db_locator]['alertmsec'])
			
				# For debugging
				# Write-Host ("$fn_name() temp_ht keys = " + $temp_ht.Keys)
				# Write-Host ("$fn_name() temp_ht = " + ($temp_ht | ConvertTo-Json))

				$al_paths_ary = @($temp_ht.$camname.path)  # Force array
				$al_files_ary = @($temp_ht.$camname.file)
				$al_memos_ary = @($temp_ht.$camname.memo)
				$al_dates_ary = @($temp_ht.$camname.date)
				$al_durs_ary  = @($temp_ht.$camname.alertmsec)
			}

			# GET ALERT'S CAMERA SETTINGS VIA THE REGISTRY
			if (1 -eq 1) {  # Use to collapse this code section while editing
				
				# 1. 'Video recording method'
				# 2. 'Pre-trigger record time'
				# 3. 'Pre-trigger playback time'
				# 4. 'Confirm alert witt AI'
				# 5. 'Save AI analysis .DAT files'
				# 6. 'AI analyze one each msec'

				# Start performance timer - get camera settings
				$timer_get_cam_settings = [Diagnostics.Stopwatch]::StartNew()
			
				# GET CAMERA'S SHORT NAME
				$camname = $clipstats_ht["$clip_db_locator"]['camera']

				# GET CAMERA'S PROFILE
				# Needed for registry calls below
				#$cam_profile_current = $script:cam_profiles_current_ht.$camname
				$cam_profile_current = $clipstats_ht["$clip_db_locator"]['camprofile']

				# GET CAMERA'S LONG NAME
				# Needed for registry calls below
				$cam_longname = $script:cam_lognnames_ht.$camname

				# For debugging
				# Write-Host "camname = '$camname'"
				# Write-Host "cam_profile_current = $cam_profile_current"
				# Write-Host "cam_longname = '$cam_longname'"

				# 1. VIDEO RECORDING METHOD
				$reg_cam_rec_is_continuous = $false  # Default
				$response = `
					IsCamRecContinuous `
					-CamLongName $cam_longname `
					-CamProfileNo $cam_profile_current			

				if ($response -is [Boolean]) { 
					$reg_cam_rec_is_continuous = $response
				}

				# For debugging
				# Write-Host ("response type = " + $response.GetType())
				# Write-Host "response = $response"
				# Write-Host "reg_cam_rec_is_continuous = $reg_cam_rec_is_continuous"

				# 2. PRE-TRIGGER RECORD TIME
				$reg_cam_pretrig_rectime_msec = 0  # Default
				$response = `  # Expecting integer
					GetPreTrigRecMsec `
					-CamLongName $cam_longname `
					-CamProfileNo $cam_profile_current			

				if ($response -is [Int]) { 
					$reg_cam_pretrig_rectime_msec = $response
				}

				# For debugging
				# Write-Host ("response type = " + $response.GetType())
				# Write-Host "response = $response"
				# Write-Host "reg_cam_pretrig_rectime_msec = $reg_cam_pretrig_rectime_msec"

				# 3. PRE-TRIGGER PLAYBACK TIME
				$reg_cam_pretrig_playback_msec = 0  # Default
				$response = `
					GetPreTrigPlaybackMsec `
					-CamLongName $cam_longname `
					-CamProfileNo $cam_profile_current			

				if ($response -is [Int]) { 
					$reg_cam_pretrig_playback_msec = $response
				}

				# For debugging
				# Write-Host ("response type = " + $response.GetType())
				# Write-Host "response = $response"
				# Write-Host "reg_cam_pretrig_playback_msec = $reg_cam_pretrig_playback_msec"

				# 4. CONFIRM ALERT WITH AI
				$cam_ai_enabled = $false  # Default
				$response = `  # Expecting boolean
					GetConfirmAlertWithAI `
					-CamLongName $cam_longname `
					-CamProfileNo $cam_profile_current			

				if ($response -is [Boolean]) { 
					$cam_ai_enabled = $response
				}

				# For debugging
				# Write-Host ("response type = " + $response.GetType())
				# Write-Host "response = $response"
				# Write-Host "cam_ai_enabled = $cam_ai_enabled"

				# 5. SAVE AI ANALYSIS .DAT FILES
				# 6. AI ANALYZE ONE EACH MSEC
				$cam_ai_save_dat_file = $false  # Default
				$reg_cam_ai_analyze_each_msec = 0  # Default
				if ($cam_ai_enabled) {
					
					# 5. SAVE AI ANALYSIS .DAT FILES
					# $cam_ai_save_dat_file = $false  # Default
					$response = `
						GetAiSaveDatFile `
						-CamLongName $cam_longname `
						-CamProfileNo $cam_profile_current			

					if ($response -is [Int]) { 
						$cam_ai_save_dat_file = ($response -eq 1)
					}

					# For debugging
					# Write-Host ("response type = " + $response.GetType())
					# Write-Host "response = $response"
					# Write-Host "cam_ai_save_dat_file = $cam_ai_save_dat_file"

					# 6. AI ANALYZE ONE EACH MSEC
					# $reg_cam_ai_analyze_each_msec = 0  # Default
					$response = `
						GetAiAnalyzeOneEachMsec `
						-CamLongName $cam_longname `
						-CamProfileNo $cam_profile_current			

					if ($response -is [Int]) { 
						$reg_cam_ai_analyze_each_msec = $response
					}

					# For debugging
					# Write-Host ("response type = " + $response.GetType())
					# Write-Host "response = $response"
					# Write-Host "reg_cam_ai_analyze_each_msec = $reg_cam_ai_analyze_each_msec"
					if ($PSBoundParameters.ContainsKey('DatImageMsec')) {
						if ($dat_image_interval_msec -ne $reg_cam_ai_analyze_each_msec) {
							if ($FromConsole) {
								Write-Warning "argument -DatImageMsec value ($dat_image_interval_msec) does not match Alerts > AI configuration' value ($reg_cam_ai_analyze_each_msec)"
								Write-Host ("Using argument's value regardless; check if GIF's time span (last frame - first frame) makes sense") -ForegroundColor Yellow
							}
						}			
					} else {
						if ($reg_cam_ai_analyze_each_msec -ne 0) {
							$dat_image_interval_msec = $reg_cam_ai_analyze_each_msec
						} else {
							$dat_image_interval_msec = 500
						}
					}
				}					
		
				# Stop timer
				$timer_get_cam_settings.Stop()
				$elapsed_msec = [Math]::Round($timer_get_cam_settings.elapsed.totalmilliseconds,1)
				$get_cam_settings_time_ms = $elapsed_msec.ToString("F1")
				# Write-Host ("$fn_name() get_cam_settings_time_ms = $get_cam_settings_time_ms msec")
				
				# Add to collector
				$makegif_timing_ht.Add('get_cam_settings_time_ms', $get_cam_settings_time_ms)
			
				# For debugging
				# Write-Host "$fn_name() camname = '$camname'"
				# Write-Host "$fn_name() cam_profile_current = $cam_profile_current"
				# Write-Host "$fn_name() cam_longname = '$cam_longname'"
				# Write-Host "$fn_name() reg_cam_rec_is_continuous = $reg_cam_rec_is_continuous"
				# Write-Host "$fn_name() reg_cam_pretrig_rectime_msec = $reg_cam_pretrig_rectime_msec"
				# Write-Host "$fn_name() reg_cam_pretrig_playback_msec = $reg_cam_pretrig_playback_msec"
				# Write-Host "$fn_name() cam_ai_enabled = $cam_ai_enabled"
				# Write-Host "$fn_name() cam_ai_save_dat_file = $cam_ai_save_dat_file"
				# Write-Host "$fn_name() reg_cam_ai_analyze_each_msec = $reg_cam_ai_analyze_each_msec"
				
			}
			
			# Make Alert's Properties
			$db_rec        = $al_paths_ary[0].Replace('.bvr','')
			$db_file       = $al_files_ary[0]
			$db_ts_sec     = $al_dates_ary[0]
			$al_msec       = $al_durs_ary[0]
			$al_memo       = $al_memos_ary[0]
			$dat_file      = $db_file.Replace('.jpg','.dat')
			# $dat_filespec  = $DatPath + "\" + $dat_file  # Default
			$now_ts_sec    = [DateTimeOffset]::UtcNow.ToUnixTimeSeconds()
			$alert_age_sec = [Int64]$now_ts_sec - [Int64]$db_ts_sec
				
			# For debugging
			# Write-Host "$fn_name() al_msec = $al_msec"
			# Write-Host ("$fn_name() al_msec type = " + $al_msec.GetType())
			# Write-Host "$fn_name() alert_age_sec = $alert_age_sec"
			
			# Validate DAT file's existence
			if ($UseDatFileFlg) {
					
				# Start timer - DAT file is ready
				$timer_dat_is_ready = [System.Diagnostics.Stopwatch]::StartNew()

				# Start performance timer - BI version check time
				$timer_bi_version_chk = [Diagnostics.Stopwatch]::StartNew()

				$is_v59929 = `  # saved to variable to avoid repeating inefficient JSON calls
					IsBiVersion `
					-CompareToVersion '5.9.9.29' `
					-CurrentVersion $script:BiVersion

				# Stop timer
				$timer_bi_version_chk.Stop()
				$elapsed_msec = [Math]::Round($timer_bi_version_chk.elapsed.totalmilliseconds,1)
				$bi_version_chk_time_ms1 = $elapsed_msec.ToString("F1")

				$makegif_timing_ht.Add('bi_version_chk_time_ms1', $bi_version_chk_time_ms1)
				
				# For debugging
				# Write-Host "$fn_name() is_v59929 = $is_v59929"
				# Write-Host "bi_version_chk_time_ms1 = $bi_version_chk_time_ms1 msec"

				# Initialize flag for using temp folder DAT file instead Alerts folder DAT file
				$use_temp_dat_flg = $false

				# Be willing to wait a bit for Blue Iris to finish writing the DAT file
				$timeout = 2  # Maximum wait time in seconds

				while ($timer_dat_is_ready.Elapsed.TotalSeconds -lt $timeout) {
			
					if ($is_v59929 -and $cam_ai_enabled -and !$cam_ai_save_dat_file) {

						# TEST IF BI IS CREATING TEMPORARY DAT FILES (IN THE BI 'TEMP' FOLDER)
						# ... As of 5.9.9.29, BI now supports DAT files created in C:\ProgramData\Blue Iris\temp;
						# ... These DAT files are created when 1) using AI, and 2) setting 'Save AI analysis DAT files' is not enabled
						#     The filename is like 'alert_{camname}.dat' already
						
						$dat_filespec = $TempPath + "\" + "alert_" + $Camera + ".dat"
						if (Test-Path -LiteralPath $dat_filespec) {

							if ($script:FromConsole -and $script:DevMode) {
								$fileInfo = Get-Item $dat_filespec
								$ts_str = ($fileInfo.LastWriteTime).ToString("yyyy/MM/dd_HH:mm:ss")
								$file_age = (Get-Date) - $fileInfo.LastWriteTime
								$file_age_in_sec = [math]::Round($file_age.TotalSeconds)
								Write-Host ("'$dat_filespec' last modified $file_age_in_sec sec ago ($ts_str)") -ForegroundColor Cyan
							}
							
							if (!$IgnoreDatAge) {
								
								#$cutoff_age_sec = $DatIsFreshMult
								$cutoff_age_sec = [Math]::Ceiling($al_msec/1000) * $DatIsFreshMult

								# For testing
								# $cutoff_age_sec = 30000

								# For debugging
								Write-Host "$fn_name() cutoff_age_sec = $cutoff_age_sec"
								
								# Make sure the temp DAT is 'fresh"
								if (IsFileFresh -FilePath $dat_filespec -CutOffAgeSec $cutoff_age_sec) {
									$use_temp_dat_flg = $true
									$use_dat_file_flg = $true
									break
								} else {
									$err_msg = "ERR/temporary DAT file is not 'fresh' (older than $cutoff_age_sec seconds) '$dat_filespec'"
									return $err_msg
								}
								
							} else {
								
								$use_temp_dat_flg = $true
								$use_dat_file_flg = $true
								break
							}
						}						
					
					} else {
						
						# TEST IF THE DAT FILE IS READY (IN THE BI 'ALERTS' FOLDER)
						
						$dat_filespec = $DatPath + "\" + $dat_file  # Default
						if (Test-Path -LiteralPath $dat_filespec) {
							$use_dat_file_flg = $true
							break
						}
						
					}
					Start-Sleep -Milliseconds 100  # Short sleep interval to reduce CPU usage
				}
				

				# Stop timer
				$timer_dat_is_ready.Stop()
				$elapsed_msec = [Math]::Round($timer_dat_is_ready.elapsed.totalmilliseconds,1)
				$dat_is_ready_time_ms = $elapsed_msec.ToString("F1")
				# Write-Host ("$fn_name() dat_is_ready_time_ms = $dat_is_ready_time_ms msec")
				
				# For debugging
				# Write-Host "$fn_name() use_temp_dat_flg = $use_temp_dat_flg"
				# Write-Host "$fn_name() use_dat_file_flg = $use_dat_file_flg"
				
				# Add to collector
				$makegif_timing_ht.Add('dat_is_ready_time_ms', $dat_is_ready_time_ms)

			}

			# Conditionally Validate Arguments
			if ($use_dat_file_flg -eq $true -and $BvrDelayMsec) {
				if ($script:FromConsole) {
					Write-Warning "argument -BvrDelayMsec is ignored when extracting GIF images from a DAT file"
				}
			}

			# For debugging
			# Write-Host ("$fn_name() al_paths_ary = " + ($al_paths_ary | ConvertTo-Json))
			# Write-Host ("$fn_name() al_files_ary = " + ($al_files_ary | ConvertTo-Json))
			# Write-Host "$fn_name() db_rec = $db_rec"
			# Write-Host "$fn_name() db_file = $db_file"
			# Write-Host "$fn_name() db_memo = $db_memo"
			# Write-Host "$fn_name() dat_file (target) = $dat_file"
			# Write-Host "$fn_name() dat_filespec (target) = $dat_filespec"
			# Write-Host ("$fn_name() al_memos_ary = " + ($al_memos_ary | ConvertTo-Json))
			# Write-Host "$fn_name() al_memo = $al_memo"
			# Write-Host ("$fn_name() unix timestamp type = " + $db_ts_sec.GetType())
			# Write-Host "$fn_name() unix timestamp (sec) = $db_ts_sec"
			# Write-Host ("$fn_name() now unix timestamp type = " + $now_ts_sec.GetType())
			# Write-Host "$fn_name() now unix timestamp (sec) = $now_ts_sec"
			# Write-Host "$fn_name() unix timestamp delta (sec) = $alert_age_sec"
			
			# HANDLE DB RECORD NUMBER ARGUMENT
			if ($DbRecNo.Length -gt 0) {
				
				# Also supports UI3 style 'rec='
				# Example: http://192.168.1.3:81/ui3.htm?rec=213168393-3165057&cam=GT
				# UI3Rec = '213168393-3165057'
				
				$hypen_idx = $DbRecNo.IndexOf('-')
				if ($hypen_idx) {
					
					$arr = $DbRecNo.Split('-')
					$db_rec = "@" + ($arr[0]).ToString()
					
					# Currently not used
					if ($arr[1]) {
						$dbrec_offset = [Int]$arr[1]
					} else {
						$dbrec_offset = $null
					}
					
					# For debugging
					#Write-Host "Database record ID = $alert_rec"
					#Write-Host "Database record offset = $dbrec_offset"
					
				} else {
					$db_rec = "@" + $DbRecNo.ToString()
					$dbrec_offset = $null
				}

				# For debugging
				# Write-Host "$fn_name() db_rec = $db_rec"
				# Write-Host "$fn_name() dbrec_offset = $dbrec_offset"
			}
		}
				
		# ... Trim source clip
		# ... KEY POINT: a custom TrimEnd takes precedence
		if ($TrimEndAutoPct -gt 0) {
			$trim_end_pct = $TrimEndAutoPct
		} else {
			$trim_end_pct = $TrimEndPct
		}

		# For debugging
		# Write-Host "$fn_name() alert_rec = $alert_rec"
		# msg.exe $script:server_username /TIME:10 "alert_rec = '$alert_rec'`ndb_rec = '$db_rec'"		
		# Write-Host ("$fn_name() db_rec len = " + $db_rec.Length)
		
		# Initialize
		$gif_method = "viaClip"  # Default; it's the fallback option if a requested DAT file not found
		$delay_ms = 0
		$dat_jpgs_cnt = 0
		$gif_images_cnt = $ImagesCnt
		$gif_dur_adj_pct = 0
		$alert_is_fresh = $null  # Default is undetermined
			
		# TRY TO CREATE A GIF

		if ($clipstats_success_flg -or $LiveFlg -or $LatestFlg) {
		#if ($LiveFlg -or $LatestFlg) {

			# PREP TO MAKE GIF FROM LIVE STREAM
			if ($LiveFlg) {

				# For debugging
				#Write-Host "`n$fn_name() Preparing to make LIVE GIF from camera stream..."	
				#msg.exe $script:server_username /TIME:20 "Preparing to make LIVE GIF from camera stream..."	

				$al_cam    = $Camera
				$al_path   = $bvr_path
				$al_file   = $bvr_filename
				$al_offset = $bvr_offset_ms
				$al_dur_ms = $Seconds * 1000
				$al_res    = "WxH"
				$delay_ms  = 0

				# For debugging
				#Write-Host "$fn_name() al_cam = $al_cam"
				#Write-Host "$fn_name() al_path = '$al_path'"
				#Write-Host "$fn_name() al_file = '$al_file'"

				# Adjust source clip offset
				$gif_offset = $al_offset + ($TrimBegPct/100)*$al_dur_ms
				
				# Set GIF duration (round up to nearest whole second)
				$gif_dur_ms = [Math]::Floor($al_dur_ms)
	
				# Calculate time delta between each extracted BVR image (round up to nearest whole second)
				$step_ms = [Math]::ceiling($gif_dur_ms / ($gif_images_cnt - 1))

				# For debugging
				# Write-Host "$fn_name() al_cam = $al_cam"
				# Write-Host "$fn_name() al_path = $al_path"
				# Write-Host "$fn_name() al_offset = $al_offset"
				# Write-Host "$fn_name() gif_offset = $gif_offset"
				# Write-Host "$fn_name() al_dur_ms = $al_dur_ms"
				# Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"
				# Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
				# Write-Host "$fn_name() step_ms = $step_ms"
				
			}
			
			# PREP TO MAKE GIF FROM EITHER BVR FILE -OR- DAT FILE
			if (!$LiveFlg) {  # if($clipstats_success_flg -or $LatestFlg)
				
				# For debugging
				# Write-Host "`n$fn_name() Preparing to make GIF from GIF or BVR file..."	
				# msg.exe $script:server_username /TIME:20 "Preparing to make GIF from GIF or BVR file..."	

				# Because BIJsonClipstats() accepts multiple AlertRec's, we need to use only the first item
				$clipstats_ht_obj = $clipstats_ht[0]
				$clipstats_ht_obj_keys = $clipstats_ht_obj.Keys
			
				# For debugging
				# Write-Host ("$fn_name() clipstats_ht_obj type = " + $clipstats_ht_obj.GetType())
				# Write-Host ("$fn_name() clipstats_ht_obj = " + ($clipstats_ht_obj | ConvertTo-Json))
				# Write-Host "clipstats_ht_obj_keys = $clipstats_ht_obj_keys"

				$pre_trig_playback_msec = $reg_cam_pretrig_playback_msec  # Default
				$pre_trig_playback_calc_flg = $false

				# Start performance timer - BI version check times
				$timer_bi_version_chk = [Diagnostics.Stopwatch]::StartNew()

				# JSON 'clipstats' cmd key 'triggeroffset' was added to Blue Iris 5.9.2.1
				$is_v5921 = `  # saved to variable to avoid repeating inefficient JSON calls
					IsBiVersion `
					-CompareToVersion '5.9.2.1' `
					-CurrentVersion $script:BiVersion
				
				# The need for argument BvrDelayMsec (and var $bvr_delay_msec) was eliminated in Blue Iris 5.9.9.22.
				# ... the /time/ web server interface now supports MJPEG video for easy browser playback 
				#     and relative time specification (i.e., 1 minute ago). 
				$is_v59922 = `  # saved to variable to avoid repeating inefficient JSON calls
					IsBiVersion `
					-CompareToVersion '5.9.9.22' `
					-CurrentVersion $script:BiVersion

				# Stop timer
				$timer_bi_version_chk.Stop()
				$elapsed_msec = [Math]::Round($timer_bi_version_chk.elapsed.totalmilliseconds,1)
				$bi_version_chk_time_ms2 = $elapsed_msec.ToString("F1")

				$makegif_timing_ht.Add('bi_version_chk_time_ms2', $bi_version_chk_time_ms2)
				
				# For debugging
				# Write-Host "$fn_name() is_v5921 = $is_v5921"
				# Write-Host "$fn_name() is_v59922 = $is_v59922"
				# Write-Host "bi_version_chk_time_ms2 = $bi_version_chk_time_ms2 msec"

				if ($is_v5921) {
					if ($clipstats_ht_obj_keys -contains 'triggeroffset') {  # just in case
						if ($clipstats_ht_obj['triggeroffset'].ToString() -ne 'NA') {
							$delta = [Int]$clipstats_ht_obj.triggeroffset - [Int]$clipstats_ht_obj.offset

							# For debugging
							# Write-Host "delta = $delta"
							
							if ($delta -gt 0) {
								$pre_trig_playback_msec = $delta
								$pre_trig_playback_calc_flg = $true
							}
						}
					}
				}

				if ($script:FromConsole -and !$script:Quiet) {
					if ($is_v5921 -and $pre_trig_playback_calc_flg) {
						Write-Host ("NOTE: `"alertmsec`" includes the pre-trigger playback time (triggeroffset - offset) = " + $pre_trig_playback_msec + " msec")
					} else {
						if ($AlertRec -or $DbRecNo -or $UI3Rec) {
							Write-Host ("NOTE: `"alertmsec`" typically includes the pre-trigger playback time, possibly = " + $pre_trig_playback_msec + " msec [*]")
							Write-Host ("[*] the current registry value for '" + $camname + "' in profile '" + $cam_profile_current + "'")
						} else {
							Write-Host ("NOTE: `"alertmsec`" typically includes the pre-trigger playback time, possibly = " + $pre_trig_playback_msec + " msec [*]")
							Write-Host ("[*] the current registry value for '" + $Camera + "' in profile '" + $script:cam_profile_current + "'")
						}
					}
				}
				
				# GET ALERT'S DETAILS
				if (!($clipstats_ht_obj -is [String] -and $clipstats_ht_obj -eq 'NA')) {						

					# Initialize
					$alert_is_fresh_forced = $false

					# Start performance timer - get alert's details
					# ... the execution time of this code section is deemed inconsequential; typically < 10 ms
					$timer_get_alert_details = [Diagnostics.Stopwatch]::StartNew()
				
					# $clipstats_ht_obj is via BIJsonClipstats -Clips $AlertRec -Props 'camera,path,file,msec,alertmsec,res'
					# NOTE: $al_file can be like:
					#       1) DW2.20231027_120000.49763.17-0.jpg 
					#          ... this the typical filespec
					#       2) 2023-10-27\DW2\DW2.20231027_120000.49763.17-0.jpg
					#          ... non-default file naming patterns in the Blue Iris Settings > Record tab
					$al_cam     = $clipstats_ht_obj.camera
					$al_path    = $clipstats_ht_obj.path
					$al_file    = $clipstats_ht_obj.file  
					$al_res     = $clipstats_ht_obj.res  # WxH
					$al_offset  = [Int]$clipstats_ht_obj.offset
					$al_dur_ms  = [Int]$clipstats_ht_obj.alertmsec
					$al_memo    = $clipstats_ht_obj.memo
					
					# Make DAT filename
					$dat_file = $al_file.Replace('.jpg', '.dat')

					# Change alert duration, if requested
					$al_dur_ms_orig = $al_dur_ms 
					if ($Seconds -gt 0) {
						$al_dur_ms = $Seconds * 1000
					}

					# For debugging
					#Write-Host "$fn_name() al_dur_ms_orig = $al_dur_ms_orig"
					#Write-Host "$fn_name() al_dur_ms_cust = $al_dur_ms"

					# Determine if alert is 'fresh' (just happened)
					# ... currently using the alert's original duration (in seconds) as the basis for this determination
					#$alert_is_fresh = ($alert_age_sec -le $al_dur_ms/1000)
					$alert_is_fresh = ($alert_age_sec -le $al_dur_ms_orig/1000)
					
					# 2025-02-03 - FRESH ALERT SUPPORT CURRENTLY UNSUPPORTED
					# ... blueirissupport is still not providing access the recording video buffer
					#$alert_is_fresh = $true
					# $alert_is_fresh = $false
					# $alert_is_fresh_forced = $true

					# For testing
					
					if ($FromConsole -and $DevMode) {
						if ($alert_is_fresh_forced) {
							Write-Host ("(dev mode) alert_is_fresh = $alert_is_fresh **FORCED**") -ForegroundColor Cyan
						} else {
							Write-Host ("(dev mode) alert_is_fresh = $alert_is_fresh") -ForegroundColor Cyan
						}
						Write-Host ("(dev mode) Note: alert is fresh when: its age (" + $alert_age_sec + " sec) < its db duration (" + ($al_dur_ms_orig/1000) + " sec)") -ForegroundColor Cyan
					}
					
					# TEST IF BVR EXISTS IN A BLUE IRIS 'NEW' FOLDER
					$bvr_file_found = $false
					if ($al_file) {
						
						$al_file_ary = $al_file.Split('.')  
						# Expecting array like @('DW2', '20231027_120000', '49763', '17-0', 'jpg')
						# ...         or like  @('2023-10-27\DW2', '20231027_120000', '49763', '17-0', 'jpg')
						$bvr_file = $al_file_ary[0] + "." + $al_file_ary[1] + ".bvr"  
						# $bvr_file should be like 'DW2.20231027_120000.bvr'
						# ...              or like '2023-10-27\DW2.20231027_120000.bvr'
						foreach ($path in $BvrPaths) {
							$bvr_filespec = $path + "\" + $bvr_file
							if (Test-Path -Path "$bvr_filespec") {
								$bvr_file_found = $true
								break
							}
						}

					} else {
						$timer_get_alert_details.Stop()
						$err_msg = "ERR/non-existent database record id"
						return $err_msg									
					}
						
					# For debugging
					# Write-Host "$fn_name() bvr_file_found = $bvr_file_found"
					
					# ALTERNATE CASE ... If the BVR is not located in a New folder, check if 
					# it is a Clipboard clip -OR- a Clipboard clip saved as Alert
					# Like "Clipboard\DW2.20240330_175026-175045.378.bvr"
					# Note: substring '170526' is a time, like 17:05:26
					if (!$bvr_file_found) {
						
						$al_file_ary1 = $al_file.Split('\')  # NOTE: the Split method ALWAYS returns an array
						# $al_file_ary1 is like @('Clipboard', 'DW2.20240330_175026-175045.378.bvr')
						#                    or @('DW2.20240330_175026-175045.378.bvr')
						$al_file_ary2 = ($al_file_ary1[$al_file_ary1.Count - 1]).Split('.')  
						# $al_file_ary2 is like @('DW2', '20240330_175026-175045', '378', 'bvr')
						$al_file_ary3 = ($al_file_ary2[1]).Split('_')  
						# $al_file_ary3 is like @('20240330', '175026-175045')
						$al_file_ary4 = ($al_file_ary3[1]).Split('-')  
						# $al_file_ary4 is like @('175026, '175045')
						$hour_str = $al_file_ary3[1].Substring(0,2)
						# REF: https://devblogs.microsoft.com/scripting/powertip-use-powershell-to-create-repeating-letters/
						$zeros_str = "0"*($al_file_ary4[0].Length - 2)  
						$bvr_file = $al_file_ary2[0] + "." + $al_file_ary3[0] + "_" + $hour_str + $zeros_str + ".bvr"  
						$bvr_file_found = $false
						$bvr_filespec = $path + "\" + $bvr_file
						if (Test-Path -Path "$bvr_filespec") {
							$bvr_file_found = $true
							# Override default or requested settings
							$use_dat_file_flg = $false
						}

						# For debugging
						# Write-Host ("$fn_name() al_file_ary1 = " + ($al_file_ary1 | ConvertTo-Json))
						# Write-Host ("$fn_name() al_file_ary2 = " + ($al_file_ary2 | ConvertTo-Json))
						# Write-Host ("$fn_name() al_file_ary3 = " + ($al_file_ary3 | ConvertTo-Json))
						# Write-Host ("$fn_name() al_file_ary4 = " + ($al_file_ary4 | ConvertTo-Json))
						# Write-Host "$fn_name() bvr_filespec = $bvr_filespec"
						
					}
					
					if ($bvr_file_found) {
						
						# For debugging
						# Write-Host "$fn_name() al_file = $al_file"
						# Write-Host "$fn_name() bvr_filespec = $bvr_filespec"
						# Write-Host "$fn_name() bvr_file_path = $bvr_filespec_path"
						# Write-Host "$fn_name() reg_bi_new_folder_path = $reg_bi_new_folder_path" 
						
						# SPECIAL CASE...
						# Trap rare instances when 'clipstats' yields a zero value for 'alertmsec'
						if ($DbRecNo) {
							if ($al_dur_ms -eq 0) {
								$timer_get_alert_details.Stop()
								$err_msg = "ERR/argument '-AlertRec', '-DbRecNo', or '-UI3Rec' does not point to an EXISTING Alert record (clipstats 'alertmsec' = 0)"
								return $err_msg			
							}
						}

						# MAKE ADJUSTMENTS
						
						# 1. Adjust source clip offset
						$gif_offset = $al_offset + ($TrimBegPct/100)*$al_dur_ms

						# 2. Adjust GIF's duration, rounding up to nearest whole second
						# ... NOTE: $trim_end_pct already accounts for $TrimEndAutoPct adjustment, which applies only 
						#     if user-setting $gif_trim_end_cust_obj is configured
						if ($TrimBegPct -ge 0) {
							$gif_dur_ms = [Math]::Floor( $al_dur_ms - ($TrimBegPct/100)*$al_dur_ms - ($trim_end_pct/100)*$al_dur_ms )
						} else {
							$gif_dur_ms = [Math]::Floor( $al_dur_ms - ($TrimBegPct/100)*$al_dur_ms - (($trim_end_pct - $TrimBegPct)/100)*$al_dur_ms )
						}
						
						# Show -DevMode details
						if ($script:DevMode -and $script:FromConsole) {

							Write-Host ""
							
							Write-Host ("(dev mode) alert's age (sec) = " + $alert_age_sec + " ( = NOW - alert's timestamp = " + $now_ts_sec + " - " + $db_ts_sec + ")") -ForegroundColor Cyan
							
							Write-Host ("(dev mode) gif offset (msec) BEFORE = " + $al_offset + ", AFTER = " + $gif_offset + ", ADJ = " + ($gif_offset - $al_offset) + " (" + 100*(($gif_offset - $al_offset)/$al_dur_ms) + "% of BEFORE duration)") -ForegroundColor Cyan
							
							Write-Host ("(dev mode) gif duration (msec) BEFORE = " + $al_dur_ms + ", AFTER = " + $gif_dur_ms + ", ADJ = " + ($al_dur_ms - $gif_dur_ms) + " (" + 100*(($al_dur_ms - $gif_dur_ms)/$al_dur_ms) + "% of BEFORE duration)") -ForegroundColor Cyan
						}
						
						# Display adjustments (if any)
						if ($gif_dur_ms -lt $al_dur_ms) {
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "`nGIF Duration Adjustment:"
								Write-Host "------------------------"
								Write-Host "TrimBegPct/TrimEndPct = $TrimBegPct/" -NoNewLine
								if ($TrimEndAutoPct -gt 0) {
									Write-Host ("$trim_end_pct (auto-determined from argument -AlertMemo)") -ForegroundColor Green
								} else {
									Write-Host $trim_end_pct
								}
								Write-Host "Alert's offset adj = $gif_offset msec"
								Write-Host "Alert's duration = $al_dur_ms msec"
								Write-Host "GIF's duration   = $gif_dur_ms msec"
							}
						}
			
						# Calculate time delta between each extracted BVR image (round up to nearest whole second)
						$step_ms = [Math]::ceiling($gif_dur_ms / ($gif_images_cnt - 1))

						# For debugging
						#Write-Host "$fn_name() al_cam = $al_cam"
						#Write-Host "$fn_name() al_path = $al_path"
						#Write-Host "$fn_name() al_offset = $al_offset"
						#Write-Host "$fn_name() gif_offset = $gif_offset"
						#Write-Host "$fn_name() al_dur_ms = $al_dur_ms"
						#Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"
						#Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
						#Write-Host "$fn_name() step_ms = $step_ms"
						
					} else {
						$timer_get_alert_details.Stop()
						$err_msg = "ERR/unable to find existing BVR file"
						return $err_msg
					}

					# Stop timer
					$timer_get_alert_details.Stop()
					$elapsed_msec = [Math]::Round($timer_get_alert_details.elapsed.totalmilliseconds,1)
					$get_alert_details_time_ms = $elapsed_msec.ToString("F1")
					#Write-Host ("$fn_name() get_alert_details_time_ms = $get_alert_details_time_ms msec")
				
					# Add to collector
					$makegif_timing_ht.Add('get_alert_details_time_ms', $get_alert_details_time_ms)
					
				} else {
					$timer_get_alert_details.Stop()
					$err_msg = "ERR/non-existent alert record '$alert_rec'"
					return $err_msg
				}
							
			}				
			
			# MAKE SUBFOLDER FOR THE GIF FILE AND ITS JPGS
			# Note: $RootPath is user-defined ('*_user_settings.ps1' file)
			$path = $RootPath + "\" + $al_cam  # like 'D:\pushover_gifs\DW1' 
			if (!(Test-Path -LiteralPath $path)) {
				New-Item -ItemType Directory -Force -Path $path
			}
		
			# DELETE ANY EXISTING *.JPG AND *.GIF FILES ASYNCHRONOUSLY

			# Start performance timer - Delete previous GIF and JPG files
			$timer_jpg_delete = [Diagnostics.Stopwatch]::StartNew()

			Get-Childitem -Path ($path + "\img*.jpg") | Remove-Item -force
			Get-Childitem -Path ($path + "\*.gif") | Remove-Item -force

			# Stop timer
			$timer_jpg_delete.Stop()
			$elapsed_msec = [Math]::Round($timer_jpg_delete.elapsed.totalmilliseconds,1)
			$jpg_delete_time_ms = $elapsed_msec.ToString("F1")

			# Add timing to collector
			$makegif_timing_ht.Add('jpg_delete_time_ms', $jpg_delete_time_ms)
			
			# For debugging
			# Write-Host "jpg_delete_time_ms = $jpg_delete_time_ms msec"

			# TRY GETTING THE GIF'S JPGs - 2 WAYS
			# ====================================================================
			# 1. GET JPGs FROM ALERT'S DAT FILE
			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			$dat_img_extract_time_ms = 'NA'
			if ($use_dat_file_flg -eq $true) {

				# For debugging
				# $bg_msg = "$fn_name()`nGetting JPGs from Blue Iris DAT file..."
				# Write-Host ("`n" + $bg_msg)
				# PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on

				if ($script:FromConsole -and !$script:Quiet) {
					#Write-Host ""
					#Write-Host "Attempting to extract JPGs from AI Analysis 'DAT' file..."
				}
				
				# Negative left trims are not ignored for DAT file extractions
				if ($TrimBegPct -lt 0) {
					$TrimBegPct = 0
					
					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host ""
						Write-Warning ("Negative -TrimBegPct argument is ignored for DAT file extractions; changed to zero")
					}
				}
				
				# EXTRACT THE JPGS					
					
				# Start performance timer
				$timer_dat_jpg_extract = [Diagnostics.Stopwatch]::StartNew()
				
				<# $extract_obj is like...
				 {
				   "source": "D:\\BlueIris\\Alerts\\DW1.20240306_080000.1950096.5-0.dat",
				   "jpgstotal": 13,
				   "jpgsextracted": 
					  [
						 "D:\\BI_gif_tool\\DW1\\img05.jpg",
						 "D:\\BI_gif_tool\\DW1\\img04.jpg",
						 "D:\\BI_gif_tool\\DW1\\img03.jpg",
						 "D:\\BI_gif_tool\\DW1\\img02.jpg",
						 "D:\\BI_gif_tool\\DW1\\img01.jpg"
					  ]
				 }
				 -OR-
				 [string] error or issue
				#>

				if ($use_temp_dat_flg) {

					$dat_file = "alert_" + $Camera + ".dat"
					
					$extract_obj = `  # returns hashtable, or string if error or issue
						ExtractDatImages `
						-DatFile $dat_file `
						-DatPath $TempPath `
						-DestPath $RootPath `
						-DatReqdImages $DatReqdImages `
						-LatestFlg $LatestFlg `
						-ImagesCnt $ImagesCnt `
						-ImagesCntUsedFlg $ImagesCntUsedFlg `
						-ImagesTrimBegPct $TrimBegPct `
						-ImagesTrimEndPct $trim_end_pct
						
				} else {

					$extract_obj = `  # returns hashtable, or string if error or issue
						ExtractDatImages `
						-DatFile $dat_file `
						-DatPath $reg_bi_alerts_folder_path `
						-DestPath $RootPath `
						-DatReqdImages $DatReqdImages `
						-LatestFlg $LatestFlg `
						-ImagesCnt $ImagesCnt `
						-ImagesCntUsedFlg $ImagesCntUsedFlg `
						-ImagesTrimBegPct $TrimBegPct `
						-ImagesTrimEndPct $trim_end_pct
					
				}

				# Stop timer
				$timer_dat_jpg_extract.Stop()
				$elapsed_msec = [Math]::Round($timer_dat_jpg_extract.elapsed.totalmilliseconds,1)
				$dat_img_extract_time_ms = $elapsed_msec.ToString("F0")

				# Add timing to collector
				$makegif_timing_ht.Add('dat_img_extract_time_ms', $dat_img_extract_time_ms)
				
				# For debugging
				# Write-Host "DAT image extract time = $dat_img_extract_time_ms msec"
				# Write-Host ("$fn_name() extract_obj = " + ($extract_obj | ConvertTo-Json))
				# $msg = "$fn_name() extract_obj type = " + $extract_obj.GetType()
				# msg.exe $script:server_username /TIME:60 $msg
				# msg.exe $script:server_username /TIME:60 $extract_obj.Count
								
				# For testing - Simulate a bad DAT JPG extraction
				# $extract_obj = $null
				# $extract_obj = "ERR/ test"
				# $extract_obj = "FAIL: test"
				# $extract_obj = "ISSUE: test"

				if ($extract_obj) {
					
					# Check if $extract_obj is string like 'ERR/*s', 'FAIL:*', or 'ISSUE:*') {
					if ($extract_obj -is [String]) {
						
						if ($extract_obj -match 'FAIL: Unable to find DAT file') {
							
							# $gif_method = 'viaClip'
							$gif_method = 'viaClip/DatNotFound'
							$use_dat_file_flg = $false
							
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "Unable to find DAT file, trying BVR extraction instead" -ForegroundColor Green
							}
							
						} elseif ($extract_obj -match 'ISSUE: insufficient JPG images count in DAT file') {
							
							# $gif_method = 'viaClip'
							$gif_method = 'viaClip/LowDatImgCnt'
							$use_dat_file_flg = $false
							
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "DAT has an insufficient number of JPG images, trying BVR extraction instead..." -ForegroundColor Green
							}

						} else {
						
							return $extract_obj
						}
						
					} elseif ($extract_obj.Count -gt 0) {

						# SUCCESS

						if ($script:FromConsole -and !$script:Quiet) {
							Write-Host ($extract_obj | ConvertTo-Json)
						}

						$gif_method = "viaDat"
						$dat_file = $extract_obj['source']
						$dat_jpgs_cnt = $extract_obj['jpgstotal']
						$gif_images_cnt = ($extract_obj['jpgsextracted']).Count
						$gif_dur_ms = $DatImageIntervalMsec*($gif_images_cnt  - 1)
						
						# For debugging
						# Write-Host "$fn_name() dat_file = '$dat_file'"
						# Write-Host "$fn_name() dat_jpgs_cnt** = $dat_jpgs_cnt"
						# Write-Host "$fn_name() gif_images_cnt = $gif_images_cnt"
						# Write-Host "$fn_name() gif_dur_ms = $gif_dur_ms"

					} else {
						$err_msg = "ERR/no images extracted from DAT file"
						return $err_msg
					}
				} else {
					$err_msg = "ERR/DAT extraction failed; function ExtractDatImages() returned null"
					return $err_msg
				}
			}
			
			# For debugging
			#msg.exe $script:server_username /TIME:60 "gif_method = $gif_method"

			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			# 2. GET JPGs FROM CAMERA'S STREAM (-LIVE) -OR- ALERT'S VIDEO CLIP FILE
			# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
			#if ($gif_method -ne "viaDat") {
			if (-not ($use_dat_file_flg -eq $true)) {
				
				# FROM LIVE STREAM
				if ($LiveFlg) {

					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "`nGetting JPGs from the Camera's Live Video Stream..."
						Write-Host "---------------------------------------------------"
					}

					# For debugging
					#$bg_msg = "$fn_name()`nGetting JPGs from the Camera's Live Video Stream..."
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
					
					$gif_method = "viaLive"

					# Start performance timer - Get JPGS for -Live GIF
					$timer_live_clip_img_extract = [Diagnostics.Stopwatch]::StartNew()

					for ($j=0; $j -lt $gif_images_cnt; $j++) {

						# Build the URL
						$time = ($gif_images_cnt - $j - 1) * $step_ms
						#$uri = $script:bi_webserver + "/image/" + $Camera + "&cache=1&h=" + $Height
						#$uri = $script:bi_webserver + "/image/" + $Camera + "&h=" + $Height

						# Default is '&decode=0', sub-stream is '&decode=-1'
						$decode_arg = ""
						if ($script:jpg_images_from_substream_flg) {
							$decode_arg = "&decode=-1"
						}
						$height_arg = ""
						if ($script:jpg_images_match_gif_height_flg) {
							$height_arg = "&h=" + $Height
						}
						#$uri = $script:bi_webserver + "/image/" + $Camera + $decode_arg + "&cache=1&h=" + $Height
						$uri = $script:bi_webserver + "/image/" + $Camera + $decode_arg + "&cache=1" + $height_arg
						if ($script:post_w_creds_flg) {
							$uri += "&user=" + $script:bi_user_name + "&pw=" + $script:bi_user_pw
						}

						# Build the full filespec for the URL's output
						$jpg_filename_sfx = ([String]($j + 1)).PadLeft(2,'0')
						$filespec = $RootPath + "\" + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"
						
						# GRAB THE IMAGE & SAVE TO DISK FILE
						Invoke-RestMethod -UseBasicParsing -Uri $uri -OutFile $filespec | Out-Null

						# Give feedback on progress
						if ($script:FromConsole -and !$script:Quiet) {
							Write-Host ("Initiated request for $uri, after " + $j*$step_ms + " msec")
						}

						Start-Sleep -Milliseconds $step_ms

					}

					# Stop timer
					$timer_live_clip_img_extract.Stop()
					$elapsed_msec = [Math]::Round($timer_live_clip_img_extract.elapsed.totalmilliseconds,1)
					$live_clip_img_extract_ms = $elapsed_msec.ToString("F0")
					# Write-Host ("$fn_name() timer_live_clip_img_extract = $live_clip_img_extract_ms msec")

					# Add timing to collector
					$makegif_timing_ht.Add('live_clip_img_extract_ms', $live_clip_img_extract_ms)
					
				# FROM BVR FILE
				} else { 

					# For debugging
					# $bg_msg = "$fn_name()`nGetting JPGs from Blue Iris BVR file..."
					# Write-Host ("`n" + $bg_msg)
					PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
					
					if ($script:FromConsole -and !$script:Quiet) {
						#Write-Host "`nGetting JPGs from Blue Iris Clip..."
						Write-Host "`nBVR Clip Processing Summary:"
						Write-Host "----------------------------"
					}
					
					# DELAY FETCHING OF JPGS FROM THE BVR
					if ($LatestFlg) {
						
						if ($script:reg_cam_rec_is_continuous) {

							# Start performance timer - Delay before extracting JPGs from BVR
							$timer_bvr_extract_delay = [Diagnostics.Stopwatch]::StartNew()
								
							# The need for argument BvrDelayMsec (and var $bvr_delay_msec) was eliminated in Blue Iris 5.9.9.22.
							# ... the /time/ web server interface now supports MJPEG video for easy browser playback 
							#     and relative time specification (i.e., 1 minute ago). 
							# var $is_v59922 previously set
							if (!$is_v59922) {
								
								# DETERMINE HOW MUCH TIME TO PAUSE THE SCRIPT 
								# BEFORE USING THE /file/clips/ WEB SERVER INTERFACE
								
								if ($BvrDelayMsec) {
									# Use the custom value provided via cmdline argument -BvrDelayMsec
									$bvr_delay_msec = $BvrDelayMsec 
								} else {
									$bvr_delay_msec = 0  # Default
								}
								
								# If the latest Alert is 'fresh' (just happened), then delay the JPG fetch
								# ... Why? otherwise, the "/file/clips/" HTTP interface command will grab frames 
								#     from the start of the clip instead of where we want
								# ... The reason for this is that the BVR is still buffering and the HTTP command
								#     is (currently) unable to retrieve JPGs from the buffer.
								#     (Ken at blueirissupport told be once that this is doable, just not implemented yet)
						
								if ($alert_is_fresh -eq $true) {
									
									# THE FOLLOWING DELAY CHOICES ARE ARBITRARY
									
									# Delay for a period at least as long as the alert's duration
									#$bvr_delay_msec = [Int](1.10 * $al_dur_ms) + 1
									
									# Delay for a period at least as 2 x alert's pre-trigger record timed
									$bvr_delay_msec = 2 * $script:reg_cam_pretrig_rectime_msec + 1
									
									# Provide feedback if the delay will be noticeable in the console
									if ($script:FromConsole -and $bvr_delay_msec -gt 1500) {
										Write-Host ("Alert_is_fresh = $alert_is_fresh [alert's age (" + $alert_age_sec + " sec) < its original duration (" + ($al_dur_ms_orig/1000) + " sec)]")
										Write-Host ("HEADS-UP: Because Alert is 'fresh', delaying JPG fetch for $bvr_delay_msec msec (" + ($bvr_delay_msec/1000).ToString("F1") + " sec)...") -ForegroundColor green
									}
									
								}

								# For testing
								# $bvr_delay_msec = 0
								# $bvr_delay_msec = 5000

								# For debugging
								# Write-Host "$fn_name() bvr_delay_msec = $bvr_delay_msec msec"

								Start-Sleep -Milliseconds $bvr_delay_msec
							
							}

							# Stop timer
							$timer_bvr_extract_delay.Stop()
							$elapsed_msec = [Math]::Round($timer_bvr_extract_delay.elapsed.totalmilliseconds,1)
							if ($elapsed_msec -lt 100) {
								$bvr_extract_delay_time_ms = $elapsed_msec.ToString("F1")
							} else {
								$bvr_extract_delay_time_ms = $elapsed_msec.ToString("F0")
								
							}
							# Write-Host ("$fn_name() timer_bvr_extract_delay = $elapsed_msec_str msec")

							# Add timing to collector
							$makegif_timing_ht.Add('bvr_extract_delay_time_ms', $bvr_extract_delay_time_ms)

						}
					}

					# For debugging
					#$bg_msg = "$fn_name()`nGetting JPGs from Blue Iris BVR Clip..."
					#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on

					#$extracted_jpg_files = @()  # Extracted JPG filespecs, serialized like '[path\]img001.jpg' to '[path\]img010.jpg'

					# KEY POINT: 
					# In PowerShell, Measure-Command captures and runs a script block {} but it does so in its own scope, 
					# meaning that commands like return inside this block will exit only from the script block, not the 
					# containing function. When you use return inside a Measure-Command block, it essentially returns control 
					# to the Measure-Command scope rather than to the function that called it. 
					# So we use AND TRAP a 'throw' command to exit the function.
					
					try {

						$measure_obj = Measure-Command {

							# Build arrays FOR A RUNSPACE that will asynchronously execute many URLs 
							$urls = @()   # Blue Iris HTTP Interface commands - to extract single JPGs from an alert's BVR
							$files = @()  # target JPG filenames, serialized like 'img001.jpg' to 'img010.jpg' 
							$offset = $gif_offset

							for ($j=0; $j -lt $gif_images_cnt; $j++) {

								# Build the URL

								if ($alert_is_fresh -eq $true -and $is_v59922) {
									
									# USE THE /time/ WEB SERVER INTERFACE
									
									# KEY-POINT: the web server interface cannot seem to grab images from the video buffer that 
									# are fresher than about 10 seconds.
									# Currently we're using a slightly conservative value of 12 sec

									if ($BvrDelayMsec) {
										# Use the custom value provided via cmdline argument -BvrDelayMsec
										$buffer_offset_msec = $BvrDelayMsec 
									} else {
										$buffer_offset_msec = 10000  # Default
										$buffer_offset_msec = 0  # Default
									}									

									#$time = -1 * ($gif_images_cnt - $j) * $step_ms - $buffer_offset_msec
									$time = -$buffer_offset_msec - ($gif_images_cnt - $j) * $step_ms
									
									$height_arg = ""
									if ($script:jpg_images_match_gif_height_flg) {
										$height_arg = "&h=" + $Height
									}

									# URLS are like... http://192.168.1.3:8600/time/DW1?jpeg&h=520&pos=-120000
									#$uri = $script:bi_webserver + "/time/" + $al_cam + "?mode=jpeg" + $height_arg + "&cache=1" + "&pos=" + $time
									$uri = $script:bi_webserver + "/time/" + $al_cam + "?mode=jpeg" + $height_arg  + "&pos=" + $time

								} else {

									# USE THE /file/clips/ WEB SERVER INTERFACE

									$time = $offset + ($j * $step_ms)
									
									# Default is '&decode=0', sub-stream is '&decode=-1', main-stream is '&decode=1'
									# Per BI Help PDF... "Use x=0 for automatic sub/main stream selection based on resolution; 
									# the sub stream will also be upsampled until the main stream is available. Use x=1 to wait 
									# for the main stream or x=-1 to always select the sub-stream."
									$decode_arg = ""
									if ($script:jpg_images_from_substream_flg) {
										$decode_arg = "&decode=-1"
									}
									$height_arg = ""
									if ($script:jpg_images_match_gif_height_flg) {
										$height_arg = "&h=" + $Height
									}
									#$uri = $script:bi_webserver + "/file/clips/" + $al_path + "?time=" + $time + "&h=" + $Height
									#$uri = $script:bi_webserver + "/file/clips/" + $al_path + "?time=" + $time + $decode_arg + "&cache=1"
									$uri = $script:bi_webserver + "/file/clips/" + $al_path + "?time=" + $time + $decode_arg + "&cache=1" + $height_arg
									
								}

								# Build the full filespec for the URL's output
								$jpg_filename_sfx = ([String]($j + 1)).PadLeft(2,'0')
								$filespec = $RootPath + "\" + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"

								# Add user credentials to the URL
								if ($script:post_w_creds_flg) {
									$uri += "&user=" + $script:bi_user_name + "&pw=" + $script:bi_user_pw
								}

								$urls  += $uri
								$files += $filespec
							}

							# For debugging
							# Write-Host ("urls cnt = " + $urls.Count)
							# Write-Host ("urls = `n" + ($urls | Out-String))
							# Write-Host ("files = `n" + ($files | Out-String))

							# CALL LIST OF URLS ASYNCHRONOUSLY
							# REF: https://stackoverflow.com/a/31525044/13499984

							$maxConcurrentJobs = $gif_images_cnt
							$content = $urls

							# Create a RUNSPACE POOL where $maxConcurrentJobs is the 
							# maximum number of runspaces allowed to run concurrently    
							$RunspacePool = [runspacefactory]::CreateRunspacePool(1,$maxConcurrentJobs)

							# Open the runspace pool (very important)
							$RunspacePool.Open()

							# Provide feedback for a noticeably slow process
							if ($script:FromConsole -and !$script:Quiet) {
								Write-Host "Extracting JPGs from video clip file..."
							}

							# Initialize collectors
							$ps_instances = @()  
							$jobs = New-Object System.Collections.ArrayList
							
							$j = 0
							foreach ($url in $content) {
								
								$file = $files[$j]

								# For debugging
								# Write-Host "url = $url"
								# PopupBackgroundMsg -Seconds 30 -Message "$fn_name()`nextracting BVR $file" -UserName $script:server_username -InBackground $script:bg_msgs_on
								
								# Create a new PowerShell instance
								# REF: https://devblogs.microsoft.com/scripting/beginning-use-of-powershell-runspaces
								$ps_instance = [powershell]::Create()
								
								# Add the instance to a collector (to facilitate disposal later)
								$ps_instances += $ps_instance

								# Tell the instance to execute in our runspace pool
								$ps_instance.RunspacePool = $RunspacePool

								# Attach some code to the PS instance
								# ... "Invoke-RestMethod is perfect for quick APIs that have no special response information such as Headers or Status Codes, 
								#     whereas Invoke-WebRequest gives you full access to the Response object and all the details it provides."
								#$ps_instance.AddCommand("Invoke-WebRequest").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file) | Out-Null
								# NOTE: w/o Out-Null, the next line injects "Invoke-RestMethod" objects to the output stream
								$ps_instance.AddCommand("Invoke-RestMethod").AddParameter("UseBasicParsing",$true).AddParameter("Uri",$url).AddParameter("OutFile",$file) | Out-Null
								
								# BEGIN EXECUTION ASYNCHRONOUSLY
								#$ps_instance.BeginInvoke()
								#$ps_instance.BeginInvoke() | Out-Null
								$handle = $ps_instance.BeginInvoke()

								# For debugging
								#Write-Host ("handle = " + ($handle | ConvertTo-Json))

								# Fill collector
								$job_obj = [PSCustomObject]@{
									instance = $null
									data = $null
								}
								$job_obj.instance = $ps_instance
								$job_obj.data = $handle
								# NOTE: w/o Out-Null, the next line injects an integer to the output stream
								$jobs.Add($job_obj) | Out-Null  
				
								# Give feedback on progress
								if ($script:FromConsole -and !$script:Quiet) {
									#Write-Host ("Initiated request for {0}" -f $url)
									Write-Host "Initiated request for $url"
									#Write-Host ("Initiated request for $url`n" + ($handle | ConvertTo-Json))
									#Write-Host ("handle = " + ($handle | ConvertTo-Json -Depth 16))
									#Write-Host ("Available Runspaces in RunspacePool: {0}" -f $RunspacePool.GetAvailableRunspaces())
									<#Write-Host ("Remaining Jobs: {0}" -f @($jobs | Where {
										$_.data.iscompleted -ne 'Completed'
									}).Count) #>
								}
								
								# For debugging
								#$bg_msg = ("$fn_name()`nInitiated request for $url`n" + ($handle | ConvertTo-Json))
								#PopupBackgroundMsg -Seconds 30 -Message $bg_msg -UserName $script:server_username -InBackground $script:bg_msgs_on
								
								$j++
							}

							# For debugging
							#Write-Host ("Jobs: " + ($jobs | ConvertTo-Json))

							# WAIT FOR JPGS TO BE CREATED
							# ... NOTE: Tried 3 different methods seeking the most efficient 
							#     then later (2/3/25) added a 4th method which we're now using
							
							$method = 4  # <-- (See below)
							$expected_jpg_files_cnt = $files.Count
							$max_tries = 20  # do-loop break variable
							$sleep_ms = 25
							$iter = 0

							# Stat timer
							$bvr_extract_timer = [Diagnostics.Stopwatch]::StartNew()
							
							# METHOD #1
							# WAIT UNTIL NO JOBS REMAIN IN THE RUNSPACE POOL
							# ... 10 Tests (pwsh.exe (PS7)): 288, 250, 433, 218, 277, 244, 235, 273, 294, 261
							#     avg: 277 ; stdev = 57   (https://www.calculator.net/statistics-calculator.html)
							if ($method -eq 1) {
								do {

									$iter++

									# Get status of jobs in the runspace pool
									$remaining_jobs = @($jobs | Where { $_.data.iscompleted -ne 'Completed' }).Count
									
									# For debugging
									# Write-Host ("iter = $iter" + "; remaining_jobs = $remaining_jobs") 
										
									# Pause for a bit until the next check
									Start-Sleep -Milliseconds $sleep_ms
								
								} until ($remaining_jobs -eq 0 -or $iter -gt $max_tries)
								$success_flg = ($iter -le $max_tries)
								$extracted_jpg_files_cnt = 0
								$extracted_jpg_files = @(Get-Childitem -Path ($path + "\img*.jpg"))
								if ($null -ne $extracted_jpg_files) {
									$extracted_jpg_files_cnt = $extracted_jpg_files.Count
								}
							}

							# METHOD #2
							# WAIT UNTIL *ALL* JPGS CREATED OR NO JOBS REMAIN IN THE RUNSPACE POOL
							# ... 10 Tests (pwsh.exe (PS7)): 277, 363, 285, 310, 359, 248, 406, 273, 308, 311
							#     avg: 314 ; stdev = 46   (https://www.calculator.net/statistics-calculator.html)
							if ($method -eq 2) {
								do {

									$iter++

									$extracted_jpg_files_cnt = 0
									$extracted_jpg_files = @(Get-Childitem -Path ($path + "\img*.jpg"))
									if ($null -ne $extracted_jpg_files) {
										$extracted_jpg_files_cnt = $extracted_jpg_files.Count
									}

									# Get status of jobs in the runspace pool
									$remaining_jobs = @($jobs | Where { $_.data.iscompleted -ne 'Completed' }).Count
									
									# For debugging
									# Write-Host ("iter = $iter" + "; images created/expected = $extracted_jpg_files_cnt/$expected_jpg_files_cnt" + "; remaining_jobs = $remaining_jobs") 
										
									# Pause for a bit until the next check
									Start-Sleep -Milliseconds $sleep_ms
								
								} until ($extracted_jpg_files_cnt -eq $expected_jpg_files_cnt -or $iter -gt $max_tries) 
								$success_flg = ($iter -le $max_tries) 
							}

							if ($method -lt 3) {
								# Write-Host ("+++ iter = $iter" + "; images created/expected = $extracted_jpg_files_cnt/$expected_jpg_files_cnt" + "; remaining_jobs = $remaining_jobs") 
							}

							# METHOD #3
							# WAIT UNTIL *EACH* JPG FILE HAS BEEN CREATED, OR NO JOBS REMAIN IN THE RUNSPACE POOL
							# ... 10 Tests (pwsh.exe (PS7)): 199, 193, 322, 277, 272, 192, 233, 245, 235, 263
							#     avg: 241 ; sdev: 40  <-- lowest avg of the 3 methods
							# ... 10 Tests again (pwsh.exe (PS7)): 317, 134, 230, 263, 152, 248, 263, 277, 281, 240
							#     avg: 240 ; stdev = 54   (https://www.calculator.net/statistics-calculator.html)
							# ... tests in powershell.exe (PS5.1) & powershell_ise.exe (PS5.1) appear to be 50-100% slower
							if ($method -eq 3) {

								$all_done_flg = $false  # runspace pool is empty (no open jobs)
								$img_no = 1
								$iter_max = 0
								$waits_cntr = 0
								foreach ($jpg_filespec in $files) {  # Test each expected JPG file

									$iter = 0
									$success_flg = $true
									$sleep_ms = 25
									
									# Wait until expected JPG file exists
									# NOTE: this approach reasonably expects that img01.jpg will be created before img02.jpg, and so on. 
									do { 

										$iter++
										
										# Break out of do-loop?
										if ($iter -gt $max_tries) {
											$success_flg = $false
											$all_done_flg = $true  # used to break out of the foreach-loop
											break 
										}

										# Get status of jobs in the runspace pool
										# ... this is an alternate do-loop break variable
										$remaining_jobs = @($jobs | Where { $_.data.iscompleted -ne 'Completed' }).Count
									
										# For debugging
										# Write-Host ("image #$img_no" + ": iter = $iter" + ": waits_cntr = $waits_cntr" + "; remaining_jobs = $remaining_jobs") 
										
										# Test if 'all done' (runspace pool is empty)
										# NOTE: we will confirm the expected number of JPG files have been created files have bee created
										if ($remaining_jobs -eq 0) {
											$all_done_flg = $true  # used to break out of the foreach-loop
											break
										}

										# Keep track for the highest number of iterations required for any file (just for user feedback)
										if ($iter -gt $iter_max) {
											$iter_max = $iter
										}
										
										# Pause for a bit until the next JPG file existence check
										Start-Sleep -Milliseconds $sleep_ms
										
										$waits_cntr++

									} until (Test-Path $jpg_filespec)
									
									if ($all_done_flg) {
										# Stop testing files
										break
									}
									
									$img_no++
								}
							}

							# METHOD #4
							# WAIT UNTIL *EACH* JPG FILE IS FULLY WRITTEN BEFORE MOVING TO THE NEXT IMAGE
							# Suggested by ChatGPT (2025-02-03) ... to eliminate occasionally observed
							# ffmpeg [mjpeg] overread 8 Errors (non-fatal)
							# Note: we removed the runspace pool check because even if the job has "completed"
							# the file size might not yet be stable.
							if ($method -eq 4) {
								
								$all_done_flg = $false  
								$img_no = 1
								$iter_max = 0
								$waits_cntr = 0

								foreach ($jpg_filespec in $files) {  
									$iter = 0
									$success_flg = $true
									$sleep_ms = 25
									$prev_size = 0  

									# Wait for file existence & stability (size stops changing)
									do { 
										$iter++

										# Get file size if it exists
										if (Test-Path $jpg_filespec) {
											$current_size = (Get-Item $jpg_filespec).Length
										} else {
											$current_size = 0
										}

										# Break out if max retries exceeded
										if ($iter -gt $max_tries) {
											$success_flg = $false
											$all_done_flg = $true
											break 
										}

										# Debugging info
										# Write-Host ("image #$img_no: iter = $iter, waits_cntr = $waits_cntr, current_size = $current_size") 

										# Check if file size is stable (unchanged for 2 iterations)
										if ($current_size -gt 0 -and $current_size -eq $prev_size) {
											break  # File is stable, continue to next image
										}

										# Update previous size and sleep
										$prev_size = $current_size
										Start-Sleep -Milliseconds $sleep_ms
										$waits_cntr++

									} until ($false)  # Exits when stability condition is met

									if ($all_done_flg) {
										break
									}

									$img_no++
								}
								
							}

							# Stop timer
							$bvr_extract_timer.Stop()
							$elapsed_msec = [Math]::Round($bvr_extract_timer.elapsed.totalmilliseconds,1)
							$elapsed_msec_str = $elapsed_msec.ToString("F0")
							# Write-Host ("$fn_name() bvr_extract_timer = $elapsed_msec_str msec")
						
							# Confirmatory checks
							$remaining_jobs = @($jobs | Where { $_.data.iscompleted -ne 'Completed' }).Count
							# ... Get-Childitem Method returns $null if no files found matching -Path
							$extracted_jpg_files_cnt = 0
							$extracted_jpg_files = @(Get-Childitem -Path ($path + "\img*.jpg"))
							if ($null -ne $extracted_jpg_files) {
								$extracted_jpg_files_cnt = $extracted_jpg_files.Count
								if ($extracted_jpg_files_cnt -ne $expected_jpg_files_cnt) {
									$err_msg = "ERR/unexpected number of images extracted from BVR (found $extracted_jpg_files_cnt, expecting $expected_jpg_files_cnt)"
									throw $err_msg
								}
							} else {
								$err_msg = "ERR/no images extracted from BVR"
								throw $err_msg
							}

							# For debugging
							# Write-Host ("remaining_jobs = " + $remaining_jobs)
							# Write-Host "iter_max = $iter_max"
							# Write-Host "success_flg = $success_flg"
							# Write-Host ("count of JPG files extracted ($extracted_jpg_files_cnt) matches expected number ($expected_jpg_files_cnt)")
						
							# Feedback async status
							if ($success_flg) {
								$ascync_msg = "SUCCESS: $gif_images_cnt asynchronous web requests"
								$ascync_msg += "; $elapsed_msec_str msec total"
								if ($method -eq 3) {
									$ascync_msg += "; done after $img_no images"
									<# if ($iter_max -gt 1) {
										$ascync_msg += "; at least one required $iter_max $sleep_ms-msec wait(s)"
									} else {
										$ascync_msg += "; none required more than 1 $sleep_ms-msec wait"
									} #>
									$ascync_msg += "; required $waits_cntr $sleep_ms-msec wait(s)"
								} else {
									$ascync_msg += "; required $iter $sleep_ms-msec wait(s)"
								}
								if ($script:FromConsole -and !$script:Quiet) {
									Write-Host $ascync_msg
								}
							} else {
								$ascync_msg = "FAIL: $gif_images_cnt asynchronous web requests;"
								$ascync_msg += "; $elapsed_msec_str msec total"
								if ($method -eq 3) {
									#$ascync_msg += "; stopped at $img_no images"
									#$ascync_msg += "; JPG NOT ready after $max_tries $sleep_ms-msec waits ($jpg_filespec)"
									$ascync_msg += "; required $waits_cntr $sleep_ms-msec wait(s)"
								} else {
									$ascync_msg += "; required $iter $sleep_ms-msec wait(s)"
								}
								if ($script:FromConsole) {
									Write-Warning $ascync_msg
								}
							}	
							
							# For debugging
							#PopupBackgroundMsg -Seconds 30 -Message $ascync_msg -UserName $script:server_username -InBackground $bg_msgs_on
							
							#Write-Host (“Available Runspaces in RunspacePool: {0}” -f $RunspacePool.GetAvailableRunspaces())

							# Perform a proper cleanup to avoid possible resources issues / memory leaks
							<# $return = $job_obj | ForEach {
								$_.instance.EndInvoke($_.handle)
								$_.instance.Dispose()
							}
							$job_obj.clear() #>
							foreach ($ps_instance in $ps_instances) {
								$ps_instance.Dispose()  # Release resources
							}
							$ps_instances.Clear()
						}
					} 
					catch {
						
						# NOTE: "ScriptHalted" is the default value from using 'throw' command without an argument
						$loggedError = $_.Exception.Message
						
						if ($loggedError -ne "ScriptHalted") {  
							return $loggedError
						} else {
							$err_msg = "ERR/BVR extraction failed; unable to find JPGs like '" + $path + "\img*.jpg'"
							return $err_msg	
						}							
						
					}

					$bvr_img_extract_time_ms = $measure_obj.TotalMilliseconds.ToString("F0")
					$makegif_timing_ht.Add('bvr_img_extract_time_ms', $bvr_img_extract_time_ms)

				}

			}
				
			# ==============================================================
			# CREATE THE GIF FILE

			# BUT FIRST, TRY TO GET THE JPG FILE DIMENSIONS (FROM THE FIRST FILE)
			# ... try checking each file as needed, stopping after the first success
			$gif_jpg_WxH = "NA"  # Default
			$needle = $RootPath + "\" + $al_cam + "\img*.jpg"
			$files = @(Get-ChildItem -Path "$needle")

			# Start performance timer
			# ... typically takes < 10 msec
			$timer_get_dat_img_wxh = [Diagnostics.Stopwatch]::StartNew()

			foreach ($jpg_file in $files) {
				
				$img = [System.Drawing.Image]::FromFile($jpg_file); 
				$gif_jpg_WxH = "$($img.Width)" + "x" + "$($img.Height)"  # expecting result like '568x320'
				
				# For debugging
				# Write-Host "jpg_file = $jpg_file"
				# Write-Host ("img = " + ($img | ConvertTo-Json))
				# Write-Host "gif_jpg_WxH = $gif_jpg_WxH"

				$img.Dispose()  # Release resources
				
				# Quit once we have a good string
				# ... want to avoid 'x' string when $img object is null, or malformed
				$arr = $gif_jpg_WxH.Split('x')  # creates 2-element array, even if $gif_jpg_WxH = 'x'
				if ($arr[0].length -ne 0 -and $arr[1].length -ne 0) {
					break
				}
			}

			# Stop timer
			$timer_get_dat_img_wxh.Stop()
			$elapsed_msec = [Math]::Round($timer_get_dat_img_wxh.elapsed.totalmilliseconds,1)
			$dat_img_wxh_time_ms = $elapsed_msec.ToString("F1")

			# Add timing to collector
			$makegif_timing_ht.Add('dat_img_wxh_time_ms', $dat_img_wxh_time_ms)
			
			# For debugging
			# Write-Host "needle = $needle"
			# Write-Host ("files.FullName = " + ($files.FullName | ConvertTo-Json))
			# Write-Host "gif_jpg_WxH = $gif_jpg_WxH"
			# Write-Host "dat_img_wxh_time_ms = $dat_img_wxh_time_ms msec"
			
			# Prep loop vars
			$ffmpeg_ms_sum = 0
			$ffmpeg_gif_size_ms_sum = 0
			$input_fs  = $RootPath + "\" + $al_cam + "\img%02d.jpg"
			$output_fs = $RootPath + "\" + $al_cam + "\output.gif"
			$jpg_file_idx = $gif_images_cnt
			$i = 0

			do { 
			
				# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
				# This code section is executed only if the initial GIF file exceeds the
				# maximum size accepted by the Pushover API (5.0 MB)
				# https://pushover.net/api#attachments
				# --------------------------------------------------------------------------------
				# TRY REMOVING ONE JPG AT A TIME FROM THE CAMERA'S IMAGE FOLDER
				if ($i -gt 0) {
					$jpg_filename_sfx = ([String]$i).PadLeft(2,'0')	
					$jpg_remove_fs = $RootPath + $al_cam + "\img" + $jpg_filename_sfx + ".jpg"
					if (Test-Path -LiteralPath $jpg_remove_fs) {
						Remove-Item -LiteralPath $jpg_remove_fs
						$gif_images_cnt--
					} else {
						# Give up
						break
					}
				}
				# ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
				
				if ($script:FromConsole -and !$script:Quiet) {
					Write-Host "`nExecuting FFmpeg:"
					Write-Host "-----------------"
				}

				# For debugging
				#Write-Host "$fn_name() output_fs = '$output_fs'"
				#PopupBackgroundMsg -Seconds 30 -Message "$fn_name()`noutput_fs = $output_fs" -UserName $script:server_username -InBackground $script:bg_msgs_on
			
				# MAKE THE GIF
				# optional arguments: -nostats
				# -hide_banner ... suppress printing showing a copyright notice, build options and library versions
				# -loglevel error ... set logging level
				# -f image2 ... ensures that FFmpeg interprets the input as an image sequence
				# -framerate $FPS ... input-only, often used with image sequences, controls how quickly FFmpeg reads each image.=
				# -r $FPS ... primarily output-oriented, adjusts the final frame rate of the output file
				# -vf "scale=320:-1" ... resize the output video to a specific width, keeping the aspect ratio
				# -y ...overwrite output files

				$size_str = "scale=" + $gif_height + ":-1"
				$ffmpeg_time_ms = 'NA'

				try {
					
					$measure_obj = Measure-Command {
						ffmpeg `
						-hide_banner `
						-loglevel error `
						-f image2 `
						-framerate $FPS `
						-i $input_fs `
						-r $FPS `
						-vf $size_str `
						-y `
						$output_fs
					}
					
					$ffmpeg_time_ms = $measure_obj.TotalMilliseconds.ToString("F0")
					$ffmpeg_ms_sum += $ffmpeg_time_ms

					# Alternative approach to capture the ffmpeg command's response in a variable
					<# $ffmpeg_cmd = `
					"ffmpeg" `
					+ " -hide_banner" `
					+ " -loglevel error" `
					+ " -y" `
					+ " -f image2" `
					+ " -framerate $FPS" `
					+ " -i $input_fs" `
					+ " -r $FPS" `
					+ " -vf $size_str" `
					+ " $output_fs"

					$ffmpeg_response = Invoke-Expression "$ffmpeg_cmd -f null - 2>&1"
					
					# For debugging
					Write-Host "ffmpeg_response = $ffmpeg_response" #>
					
				}
				catch {

					# FFMPEG RESPONSES ENCOUNTERED (SO FAR)
					#   1. [mjpeg @ xxxxxxxx] overread 8
					#       "This is an error that is thrown when ffmpeg tries to read a JPEG image (or a frame from an MJPEG stream) 
					#      block-by-block, but there are still leftover bits."
					#      REF:  https://superuser.com/a/559996
					#   2. "could not open file: I/O Error"
					#      REF:  TBD
					#   3. "Error demuxing input file 0: I/O Error..."
					#      REF:  TBD

					$err_msg = "ERR/ffmpeg unable make GIF; $_"
					return $err_msg

				}					
				
				if (Test-Path -LiteralPath $output_fs) {
					if ($script:FromConsole -and !$script:Quiet) {
						Write-Host "success"
						# Write-Host "ffmpeg time = $ffmpeg_time_ms msec"
					}
				} else {
					$err_msg = "ERR/ffmpeg could not create the GIF file"
					return $err_msg
				}

				# GET SIZE OF GIF FILE
				# ... DEPRECATED: waiting for the file to be created appears to be unnecessary
				<# $gif_kb = 0
				$iter = 0
				$max_tries = 20 # do-loop break variable
				$success_flg = $true
				do { 
					if (-not (Test-Path $output_fs)) {
						Start-Sleep -Milliseconds 50
						$iter++
					}
					if ($iter -gt $max_tries) {
						$success_flg = $false
						break
					}
				} until (Test-Path $output_fs) #>
				
				if (Test-Path -LiteralPath $output_fs) {
					
					# Start performance timer
					$timer_get_gif_size = [Diagnostics.Stopwatch]::StartNew()

					$sizeKB = (Get-Item -Path $output_fs).Length/1kB

					$gif_getsize_time_ms = $measure_obj.TotalMilliseconds

					# Stop timer
					$timer_get_gif_size.Stop()
					$gif_getsize_time_ms = [Math]::Round($timer_get_gif_size.elapsed.totalmilliseconds,1)
					
					$ffmpeg_gif_size_ms_sum += $gif_getsize_time_ms

				} else {
					$sizeKB = 'NA'
				}
				
				# No need to limit GIF size if we're not using the Pushover API
				if ($NoPushFlg) {
					break
				}
				
				# Reduce the image count, but not below 1 image
				if ($i -gt 0) {
					$ImagesCnt--
					if ($ImagesCnt -lt 2) {
						break
					}
				}
				
				$i++
				
			} until ($sizeKB -lt 5000)
			
			if ($sizeKB -ne 'NA') {
				$gif_kb = [double]([Math]::Round($sizeKB,1).ToString("F1"))
			}

			# Add timings to collector
			$makegif_timing_ht.Add('ffmpeg_time_ms', $ffmpeg_ms_sum.ToString("F0"))
			$makegif_timing_ht.Add('gif_getsize_time_ms', $ffmpeg_gif_size_ms_sum.ToString("F1"))

			# ARCHIVE THE GIF FILE
			if (!$NoArchiveGifFlg) {
				
				# Start performance timer
				$timer_gif_archive = [Diagnostics.Stopwatch]::StartNew()

				if ($LiveFlg) {
					$gif_archive_filename = $archive_gif_path			
				} else {
					$gif_archive_filename = (Split-Path $al_file -leaf).Replace(".jpg",".gif")
				}
				$gif_archive_fs = $gif_archive_path + "\" + $gif_archive_filename			
				if (!(Test-Path -LiteralPath $gif_archive_fs)) {
					Copy-Item -path $output_fs -destination $gif_archive_fs
					$archive_gif_fs = $gif_archive_fs
				} else {
					# Save multiple copies, like 'filename(N).gif'
					$archive_gif_fs = `
						CopyFileAsSequence `
						-SourceFilespec $output_fs `
						-TargetFilespec $gif_archive_fs `
						-DestPath $gif_archive_path
				}

				# Stop timer
				$timer_gif_archive.Stop()
				$elapsed_msec = [Math]::Round($timer_gif_archive.elapsed.totalmilliseconds,1)
				$gif_archive_time_ms = $elapsed_msec.ToString("F1")

				# Add timing to collector
				$makegif_timing_ht.Add('gif_archive_time_ms', $gif_archive_time_ms)
				
				# For debugging
				# Write-Host "gif_archive_filename = $gif_archive_filename"
				# Write-Host "gif_archive_path = $gif_archive_path"
				# Write-Host "gif_archive_fs = $gif_archive_fs"
				# Write-Host "gif_archive_time_ms = $gif_archive_time_ms msec"
			}

		}
		
		# Construct object to be returned
		if ($gif_method -match 'viaDAT') {
			$source_fs = $dat_file
		} else {  # 'viaLive' or 'viaBVR'
			$source_fs = $bvr_filespec
		}
		
		# For debugging
		#Write-Host "$fn_name() gif_method = '$gif_method'"
		#Write-Host "$fn_name() source_fs = '$source_fs'"
		#Write-Host "$fn_name() dat_jpgs_cnt = $dat_jpgs_cnt"
		
		$rtn_obj = [ordered]@{}
		$rtn_obj.Add( 'method', $gif_method )
		$rtn_obj.Add( 'file', $output_fs )
		$rtn_obj.Add( 'delay', $BvrDelayMsec )
		$rtn_obj.Add( 'cnt', $gif_images_cnt )
		$rtn_obj.Add( 'datjpgs', $dat_jpgs_cnt )
		$rtn_obj.Add( 'height', $Height )
		$rtn_obj.Add( 'fps', $FPS )
		$rtn_obj.Add( 'kb', $gif_kb )
		$rtn_obj.Add( 'msec', $gif_dur_ms )
		$rtn_obj.Add( 'trimbeg', $TrimBegPct )
		# A custom TrimEnd takes precedence
		if ($TrimEndAutoPct -gt 0) {
			$rtn_obj.Add( 'trimend', 0 )
		} else {
			$rtn_obj.Add( 'trimend', $trim_end_pct )
		}
		$rtn_obj.Add( 'trimcust', $TrimEndAutoPct )
		$rtn_obj.Add( 'almsec', $al_dur_ms )
		$rtn_obj.Add( 'alfresh', $alert_is_fresh )
		$rtn_obj.Add( 'res', $gif_jpg_WxH )  # WxH
		$rtn_obj.Add( 'cam', $al_cam )
		$rtn_obj.Add( 'memo', $al_memo )
		$rtn_obj.Add( 'source', $source_fs )
		$rtn_obj.Add( 'dbrec', $clip_db_locator )
		$rtn_obj.Add( 'archive', $archive_gif_fs )
		$rtn_obj.Add( 'timing_ht', $makegif_timing_ht )
		
		# For debugging
		# Write-Host ("`n")
		# Write-Host ("$fn_name() makegif_timing_ht = " + ($makegif_timing_ht | ConvertTo-Json))
		# Show-HashtableAsTable -HashTable $makegif_timing_ht
		# Write-Host ("$fn_name() rtn_obj cnt= " + ($rtn_obj.Count))
		# Write-Host ("$fn_name() rtn_obj = " + ($rtn_obj | ConvertTo-Json))

		return $rtn_obj
	}

	function SendPushover() {
	# PURPOSE: Send custom notification via the Pushover API
	# ARGS:    $MsgStr          = (String) notification message; may contain HTML
	#          $MsgTitle        = (String) notification title
	#          $PriorityLevel   = (Integer) -2|-1|0|1|2; -2=lowest, -1=low(silent), 0=normal, 1=high-priority, 2=emergency
	#          $Sound           = (String) Pushover sound
	#          $Devices         = (String) Pushover device(s); comma-delimited
	#          $AttachPath      = (String) full filespec of resized alert image, or $null
	#          $Camera          = (String) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
	# RETURNS: (String) Pushover API response, or 'ERR/' message
	# ATTRIBUTIONS:
	#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V4 (7/12/2024)
	# V5 (9/25/2024) added -Sound & -Devices arguments
	# V6 (11/22/2024) removed argument -QuietFlg; misc tweaks
	# V7 (11/24/2024) added check for already loaded assembly 'System.Net.Http'
		
		param(
			[String]$MsgStr,      # [OPTIONAL] default is 'test'
			[String]$MsgTitle,    # [OPTIONAL] default is the App name associated with your Pushover App token
			[Int]$PriorityLevel,  # [OPTIONAL] default is 0
			[String]$Sound,       # [OPTIONAL] default is ""
			[String]$Devices,     # [OPTIONAL] default is ""
			[String]$AttachPath,  # [OPTIONAL] default is null
			[String]$Camera
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() MsgStr = '$MsgStr'"
		# Write-Host "$fn_name() MsgTitle = '$MsgTitle'"
		# Write-Host "$fn_name() PriorityLevel = $PriorityLevel"
		# Write-Host "$fn_name() Sound = $Sound"
		# Write-Host "$fn_name() Devices = $Devices"
		# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
		# Write-Host "$fn_name() Camera = '$Camera'"	
		# NOTE REGARDING MSG.EXE COMMANDS
		# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
		#msg.exe $script:server_username /TIME:20 "Pushover notification args...`npriority_level = $PriorityLevel`nAttachPath = $AttachPath`nMsgStr = $MsgStr"

		# HANDLE OPTIONAL COMMANDLINE ARGS
		if (!$MsgStr) {
			$MsgStr = "Alert w/ GIF attachment demo"
		}
		if (!$Camera) {
			$cam = ''
		} else {
			$cam = $Camera
		}
		if (!$Sound) {
			$Sound = $script:po_sound
		}
		if (!$Devices) {
			$Devices = $script:po_device
		}

		$rtn_str = $null
		
		#Prep the attachment
		if ($AttachPath) {
			#Get-Item -Path $AttachPath
			$attach_obj = (Get-Item -Path $AttachPath)
			$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
		}

		# For debugging
		#Write-Host ("$fn_name() attach_obj = " + $attach_obj)
		#Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
		#Write-Host ("$fn_name() po_tags = '" + $script:po_tags + "'")
		#Write-Host ("$fn_name() Camera = " + $Camera)
		
		# PREP CORE PARAMS
		# ... will be modified if sending an attachment
		$uri = 'https://api.pushover.net/1/messages.json'
		if ($script:po_tags) {
			$po_tags = ($script:po_tags).Replace("&CAM", $cam)
		} else {
			$po_tags = ''		
		}
		$parameters = @{
			token      = $script:po_app_token
			user       = $script:po_user_token
			priority   = $PriorityLevel
			retry      = $script:po_retry
			expire     = $script:po_expire
			tags       = $script:po_tags
			sound      = $Sound
			device     = $Devices
			html       = 1
			title      = $MsgTitle
			message    = $MsgStr
		}
		
		# For debugging
		#Write-Host ("parameters.Keys = " + $parameters.Keys)
		#Write-Host "`nPushover parameters..."
		<# [string[]]$arr = $parameters | Out-String -stream
		foreach ($item in $arr) {
			if ($item) {
				Write-Host $item
			}
		} #>
		
		# EXECUTE
		if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {
			
			# SEND BASIC (OTEXT-ONLY) NOTIFICATION

			# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
			#	  do not execute; it doesn't produce anything, no error, warning, or output.
			#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
			#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
			#     all web requests now use basic parsing only.
			# ... attachmnets
			#     https://psg-prod-centralus.cloudapp.net/packages/Poshover/0.1.450600-alpha/Content/Private%5CSend-MessageWithAttachment.ps1
			try {
				# Returns a PSCustomObject
				$api_response_obj = $parameters | Invoke-RestMethod -Uri $uri -Method Post
				# Alternative syntaxes
				#$parameters | Invoke-RestMethod -Uri $uri -Method Post | Out-Null
				#$api_response_obj = Invoke-RestMethod -Uri $uri -Method Post -Form $parameters
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "An error was encountered sending the Pushover text message:"
					Write-Error $_
				}
				$rtn_str = "ERR/an error was was encountered sending the Pushover text message"
				return $rtn_str
			}
				
			# Expecting string like "@{status=1; request=21d4f100-3127-4c62-aad9-27058400fd74}"
			[String]$response_str = ($api_response_obj | ConvertTo-Json -Depth 100) | ConvertFrom-Json
					
			# For debugging
			#Write-Host "$fn_name() w/o attachment api_response_obj = $response_str"

		} else {	
		
			# SEND NOTIFICATION WITH ATTACHMENT
			
			# Notes:
			# Modified code from:  https://tinyurl.com/mrxp7pn3
			# More refs: https://gist.github.com/markekraus/a0135d0a5f52b609eda513df4db71cab

			try {

				if (-not [type]::GetType("System.Net.Http.HttpClient")) {
					Add-Type -AssemblyName System.Net.Http
				}

				# NOTE: Originally, the $client= line below was not working when this script was 
				#        executed in the background via a Blue Iris 'run a program' action
				#        SOLUTION:  https://stackoverflow.com/a/70535083
				$client = [system.net.http.httpclient]::new()

				try {
					$content = [system.net.http.multipartformdatacontent]::new()
					foreach ($key in $parameters.Keys) {
						$textContent = [system.net.http.stringcontent]::new($parameters.$key)
						$content.Add($textContent, $key)
					}
					
					# For debugging
					#Write-Host ("content = " + $content | ConvertTo-Json)
					
					$jpegContent = [system.net.http.bytearraycontent]::new($attach_bytes)
					$jpegContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
					$jpegContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
					$jpegContent.Headers.ContentDisposition.Name = 'attachment'
					$jpegContent.Headers.ContentDisposition.FileName = $AttachPath
					$content.Add($jpegContent)

					#For debugging
					# Write-Host ("content = " + ($content | ConvertTo-Json))

					# Not needed.
					# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
					#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

					# Modify the message body
					Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.token, "********").Replace($parameters.user, "********"))"

					# For debugging
					#msg.exe $script:server_username /TIME:10 "executing pushover api cmd w/ attachment..."

					# DO IT!
					$api_response_obj = $client.PostAsync($uri, $content).Result
					
					# For debugging
					#Write-Host ("$fn_name() w/ attachment api_response_obj = " + ($api_response_obj | ConvertTo-Json -Depth 32))
					
					try {
						$response_str = $api_response_obj.Content.ReadAsStringAsync().Result | ConvertFrom-Json
						
						# For debugging
						#Write-Host "$fn_name() api_response_obj.Content = $response_str"
					}
					catch {
						$response_str = "WARNING: Unexpected Pushover API response; attachment may be too large"
					}
					
					# For debugging
					#Write-Host "$fn_name() w attachment api_response_obj = $response_str"
					
				}
				catch {
					$response_str = "ERR/unable to send Pushover notification"				
				}
				finally {
					$content.Dispose()  # Release resources
				}
			}
			finally {
				$client.Dispose()  # Release resources
			}
		}

		# For testing - Simulate a bad pushover response
		#$response_str = 'invalid'
		#$response_str = 'status=0'
		
		if ($response_str -match 'invalid') {
			$rtn_str = "ERR/Pushover API returned 'invalid'"
		} elseif (!($response_str -match 'status=1')) {
			$rtn_str = "ERR/failed to send Pushover notification"				
		} else {
			# SUCCESS!!
			$rtn_str = ("Pushover notification sent`nAPI response = " + $response_str)
		}
		
		return $rtn_str
	}

	function SendTelegram() {
	# PURPOSE: Send custom notification via the Telegram Bot API
	# ARGS:    $MsgStr          = (String) notification message; may contain HTML
	#          $PriorityLevel   = CURRENTLY UNSUPPORTED
	#          $AttachPath      = (String) full filespec of attachment, or $null
	#          $Camera          = (String) Blue Iris camera shortname (replaces '&CAM' in user-setting $po_tags)
	# RETURNS: (String) Telegram API response, or 'ERR/*' message
	#          API response is like 
	# GLOBALS: $script:FromConsole, $script:Quiet
	# ATTRIBUTIONS:
	#    handling attachments: https://tinyurl.com/mrxp7pn3 (PoshOver Powershell Package)
	#
	# NOTES: 
	# 1. Per the API, 'sendPhoto' max is 10 MB  https://core.telegram.org/bots/api#sendPhoto
	# 2. Per the API, 'sendVideo' max is 50 MB  https://core.telegram.org/bots/api#sendVideo
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1 (7/12/2024)
	# V2 (9/23/2024) fixed bug where $response_str could be type PSCutomObj
	# V3 (10/4/2024) fixed bug where $MsgStr cannot contain HTML '<br>' tags
	# V4 (11/12/2024) fixed bug where $response_str was actually a PSCustomObject; 
	#                 fixed bug where Telegram's sendVideo method stopped working for GIFs; changed to sendAnimation method
	# V5 (11/22/2024) removed arguments -FromConsoleFlg, -QuietFlg - now using globals; misc tweaks
	# V6 (11/24/2024) added check for already loaded assembly 'System.Net.Http'
	# V7 (11/26/2024) fixed bug where function returned a PSCustomObject instead of a string
		
		param(
			[String]$MsgStr,      # [OPTIONAL] default is 'test'
			[String]$AttachPath,  # [OPTIONAL] default is null
			[String]$Camera       # [OPTIONAL] default is ""
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() MsgStr = '$MsgStr'"
		# Write-Host "$fn_name() AttachPath = '$AttachPath'"	
		# Write-Host "$fn_name() Camera = '$Camera'"	
		# NOTE REGARDING MSG.EXE COMMANDS
		# ... You may need to replace 'msg.exe *' with 'msg.exe /server:computername username
		#msg.exe $script:server_username /TIME:20 "Pushover notification args...``nAttachPath = $AttachPath`nMsgStr = $MsgStr"

		# HANDLE OPTIONAL COMMANDLINE ARGS
		if (!$MsgStr) {
			$MsgStr = "test"
		} else {
			# The Telegram API does not support '<br>' tags
			$MsgStr = $MsgStr.Replace("<br>", "`n")
		}
		
		if (!$Camera) {
			$cam = ''
		} else {
			$cam = $Camera
		}

		$PsVersion = $PSVersionTable.PSVersion.Major

		$rtn_str = ""  # Default
		
		# PREP THE ATTACHMENT
		if ($AttachPath) {
			try {
				$attach_obj = Get-Item -Path $AttachPath -ErrorAction Stop
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "The specified attachment file could not be interpreted properly"
				}
				$err_msg = "ERR/the specified attachment file could not be interpreted properly"
				return $err_msg
			}
			$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
		}

		# For debugging
		# Write-Host ("$fn_name() cam = " + $cam)
		# Write-Host ("$fn_name() attach_obj = " + $attach_obj)
		# Write-Host ("$fn_name() attach_obj.Name = " + $attach_obj.Name)
		# Write-Host ("$fn_name() attach_bytes cnt = " + $attach_bytes.Count)
		# Write-Host ("$fn_name() attach_bytes = " + $attach_bytes)
		
		# PREP CORE PARAMS
		# ... will be modified if sending an attachment
		$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendMessage"

		$parameters = @{
			chat_id     = $script:tg_chat_id
			parse_mode  = 'html'
		}
		
		# EXECUTE
		if (!$AttachPath -or ($attach_bytes.Length -eq 0)) {

			# SEND BASIC NOTIFICATION (NO ATTACHMENT)
			
			$parameters.Add('text', $MsgStr)
		
			# For debugging
			#Write-Host ("parameters.Keys = " + $parameters.Keys)
			<# Write-Host "`nPushover parameters..."
			[string[]]$arr = $parameters | Out-String -stream
			foreach ($item in $arr) {
				if ($item) {
					Write-Host $item
				}
			} #>

			# ... NOTE: When running in the background, both Invoke-RestMethod & Invoke-WebRequest
			#	  do not execute; it doesn't produce anything, no error, warning, or output.
			#     REF: https://stackoverflow.com/a/50274056 looked like a possible solution
			#     however, since PS v6.0 switch -UseBasicParsing has been deprecated, and 
			#     all web requests now use basic parsing only.

			try {
				if ($PsVersion -gt 5) {
					# Returns a PSCustomObject
					$api_response_obj = `
						Invoke-RestMethod `
						-Uri $uri `
						-Method Post `
						-Form $parameters `
						-ErrorAction Stop
				} else {
					# Returns a PSCustomObject
					$api_response_obj = `
						Invoke-RestMethod `
						-Uri $uri `
						-Method Post `
						-Body $parameters `
						-ErrorAction Stop
				}
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "An error was encountered sending the Telegram text message:"
					Write-Error $_
				}
				$rtn_str = "ERR/an error was was encountered sending the Telegram text message"
				return $rtn_str
			}
			
			# Make string like "@{ok=True; result=}"
			[String]$response_str = ($api_response_obj | ConvertTo-Json -Depth 100) | ConvertFrom-Json
					
			# For debugging
			# Write-Host ("$fn_name() w/o attachment api_response_obj type = " + $api_response_obj.GetType())
			# Write-Host ("$fn_name() w/o attachment api_response_obj = " + ($api_response_obj | ConvertTo-Json))
			# Write-Host ("$fn_name() w/o attachment response_str type = " + $response_str.GetType())
			# Write-Host "$fn_name() w/o attachment api_response_str = $response_str"

		} 
		else {  # send notification w/ attachment
			
			# SEND NOTIFICATION WITH ATTACHMENT
			
			# Notes:
			# Modified code from:  https://www.powershellgallery.com/packages/PoshGram/0.8.4/Content/PoshGram.psm1

			try {
				$fileObject = Get-Item $AttachPath -ErrorAction Stop
			}
			catch {
				if ($script:FromConsole) {
					Write-Warning "The specified attachment file could not be interpreted properly"
				}
				$rtn_str = "ERR/the specified attachment file could not be interpreted properly"
				return $rtn_str
			}

			# For debugging
			# Write-Host ("$fn_name() fileObject = " + ($fileObject | ConvertTo-Json))

			if ($PsVersion -ge 7) {

				if ($AttachPath.EndsWith('.jpg')) {
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendPhoto"
					$parameters.Add('photo', $fileObject)
					$parameters.Add('caption', $MsgStr)
				} elseif ($AttachPath.EndsWith('.gif') -or $AttachPath.EndsWith('.mp4')) {
					#$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendVideo"
					#$parameters.Add('video', $fileObject)
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendAnimation"
					$parameters.Add('animation', $fileObject)
					$parameters.Add('caption', $MsgStr)			
				} else {
					if ($script:FromConsole) {
						Write-Warning "Unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					}
					$rtn_str = "ERR/unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					return $rtn_str			
				}
		
				# For debugging
				# Write-Host "$fn_name() uri = $uri"
				# Write-Host ("$fn_name() parameters.Keys = " + $parameters.Keys)
				# Write-Host "`nTelegram parameters..."
				# [string[]]$arr = $parameters | Out-String -stream
				<# foreach ($item in $arr) {
					if ($item) {
						Write-Host $item
					}
				} #>
					
				try {
					# Returns a PSCustomObject
					$api_response_obj = `
						Invoke-RestMethod `
						-Uri $uri `
						-Method Post `
						-Form $parameters `
						-ErrorAction Stop				
				}
				catch {
					if ($script:FromConsole) {
						Write-Warning "An error was encountered sending the Telegram video message:"
						Write-Error $_
					}
					$rtn_str = "ERR/an error was encountered sending the Telegram video message"
					return $rtn_str			
				}
					
				# Expecting string like "@{ok=True; result=}"
				[String]$response_str = ($api_response_obj | ConvertTo-Json -Depth 100) | ConvertFrom-Json

				# For debugging
				# Write-Host ("$fn_name() w/ attachment api_response_obj type = " + $api_response_obj.GetType())
				# Write-Host ("$fn_name() w/ attachment api_response_obj = " + ($api_response_obj | ConvertTo-Json-Depth 100))
				# Write-Host ("$fn_name() api_response_obj.ok = " + $api_response_obj.ok)
				# Write-Host ("$fn_name() w/ attachment response_str type = " + $response_str.GetType())
				# Write-Host "$fn_name() w/ attachmentapi_response_str = $response_str"
					
			} 
			else {  # $PsVersion -lt 7

				if ($AttachPath.EndsWith('.jpg')) {
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendPhoto"
					$parameters.Add('caption', $MsgStr)
				} elseif ($AttachPath.EndsWith('.gif') -or $AttachPath.EndsWith('.mp4')) {
					$uri = "https://api.telegram.org/bot" + $script:tg_api_token + "/sendVideo"
					$parameters.Add('caption', $MsgStr)			
				} else {
					if ($script:FromConsole) {
						Write-Warning "Unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					}
					$rtn_str = "ERR/unsupported attachment file (expecting *.jpg, *.gif, or *.mp4)"
					return $rtn_str			
				}
		
				# For debugging
				# Write-Host "$fn_name() uri = $uri"
				# Write-Host ("$fn_name() parameters.Keys = " + $parameters.Keys)
				# Write-Host "`nTelegram parameters..."
				<# [string[]]$arr = $parameters | Out-String -stream
				foreach ($item in $arr) {
					if ($item) {
						Write-Host $item
					}
				} #>
				
				try {
					
					# This command fails in PS5
					#$rtn_flg = Invoke-RestMethod -Uri $uri -Method Post -Body $parameters -ErrorAction Stop

					#Prep the attachment
					if ($AttachPath) {
						#Get-Item -Path $AttachPath
						$attach_obj = Get-Item -Path $AttachPath
						$attach_bytes = [System.IO.File]::ReadAllBytes($AttachPath)
					}
					if (-not [type]::GetType("System.Net.Http.HttpClient")) {
						Add-Type -AssemblyName System.Net.Http
					}

					# NOTE: Originally, the $client= line below was not working when this script was 
					#        executed in the background via a Blue Iris 'run a program' action
					#        SOLUTION:  https://stackoverflow.com/a/70535083
					$client = [system.net.http.httpclient]::new()

					try {
						$content = [system.net.http.multipartformdatacontent]::new()
						foreach ($key in $parameters.Keys) {
							$textContent = [system.net.http.stringcontent]::new($parameters.$key)
							$content.Add($textContent, $key)
						}
						
						# For debugging
						# Write-Host ("content = " + $content | ConvertTo-Json)
						
						$videoContent = [system.net.http.bytearraycontent]::new($attach_bytes)
						$videoContent.Headers.ContentType = [system.net.http.headers.mediatypeheadervalue]::new('image/jpeg')
						$videoContent.Headers.ContentDisposition = [system.net.http.headers.contentdispositionheadervalue]::new('form-data')
						$videoContent.Headers.ContentDisposition.Name = 'video'  # Telegram API 'video' parameter
						$videoContent.Headers.ContentDisposition.FileName = $AttachPath
						$content.Add($videoContent)

						#For debugging
						# Write-Host ("content = " + ($content | ConvertTo-Json))

						# Not needed.
						# REF: https://stackoverflow.com/questions/43014894/httpclient-and-stream-issues
						#$content.LoadIntoBufferAsync().GetAwaiter().GetResult()

						# Modify the message body
						Write-Verbose "Message body:`r`n$($content.ReadAsStringAsync().Result.Substring(0, 2000).Replace($parameters.chat_id, "********"))"

						# For debugging
						#msg.exe $script:server_username /TIME:10 "executing telegram api cmd w/ 'video' parameter..."

						# DO IT!
						$api_response_obj = $client.PostAsync($uri, $content).Result
						
						# For debugging
						# Write-Host ("api_response_obj w/ 'video' parameter = " + ($api_response_obj | ConvertTo-Json -Depth 100))
						
						try {
							[String]$response_str = $api_response_obj.Content.ReadAsStringAsync().Result | ConvertFrom-Json
							
							# For debugging
							# Write-Host ("$fn_name() response_str type = " + $response_str.GetType())
							# Write-Host "$fn_name() response_str = $response_str"
						}
						catch {
							$response_str = "WARNING: Unexpected Telegram API response; 'video' parameter may be too large"
						}
						
						# For debugging
						# Write-Host "api_response_obj w/ 'video' parameter = $response_str"
						
					}
					catch {
						$response_str = "ERR/unable to send Telegram notification"				
					}
					finally {
						$content.Dispose()  # Release resources
					}
				}
				finally {
					$client.Dispose()  # Release resources
				}
			}
			
		}

		# For testing - Simulate a bad Telegram response
		#$response_str = 'ok=False'
	
		if (!($response_str -match "ok=True")) {
			if ($script:FromConsole) {
				Write-Warning "failed to send Telegram notification"
			}
			$rtn_str = "ERR/failed to send Telegram notification"				
			
			# For debugging
			# Write-Error $_
			
		} else {
			# SUCCESS!!
			$rtn_str = $response_str
		}
		
		# For debugging
		# Write-Host ("$fn_name() rtn_str type = " + $rtn_str.GetType())
		# Write-Host ("$fn_name() rtn_str = " + $rtn_str)
		
		return $rtn_str
	}

	function LogScriptResultsToCsv() {
	# PURPOSE:  Log performance data to CSV file.
	# ARGS:     see param lines below
	# RETURNS:  (String) 'success' or 'ERR/unable to append row to CSV log'
	# NOTES:    1. CSV file is created automatically if it does not already exist. 
	#           2. CSV file's header row is generated from the object keys in variable $new_row_obj.
	# 
	# Created by jaydeel @ ipcamtalk.com
	# Copyright: (2024) jaydeel
	#
	# V1
	# V2 (11/24/2024) added column 'PSVersion'
		
		param(
			[String]$CsvFilespec,   # [REQUIRED]
			[String]$Timestamp,     # [REQUIRED]
			[String]$GifSourceFilespec,
			[String]$GifMethod,
			[String]$GifImagesCnt, 
			[String]$GifHeight, 
			[String]$GifFps, 
			[String]$GifTrimBeg, 
			[String]$GifTrimEnd, 
			[String]$GifTrimEndAuto, 
			[String]$GifDurSec, 
			[String]$GifSize, 
			[String]$GifJpgWxH,
			[Double]$MakeGifMsec,
			[Double]$PushoMsec,
			[Double]$TelegramMsec,
			[Double]$TotalMsec,
			[String]$AlertRecord,
			[String]$ScriptHostStr,
			[String]$PsVersionFullStr
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered function: $fn_name()"
		# Write-Host "$fn_name() CsvFilespec = $CsvFilespec"
		# Write-Host "$fn_name() Timestamp = $Timestamp"
		# Write-Host "$fn_name() GifSourceFilespec = $GifSourceFilespec"
		# Write-Host "$fn_name() AlertRecord = $AlertRecord"
		# Write-Host "$fn_name() GifTrimEndAuto = $GifTrimEndAuto"
		# Write-Host "$fn_name() ScriptHostStr = $ScriptHostStr"
		# Write-Host "$fn_name() PsVersionFullStr = $PsVersionFullStr"

		# HANDLE REQUIRED COMMANDLINE ARGS
		if (!$CsvFilespec) {
			$err_msg = "ERR/expecting 2 arguments (-CsvFilespec -Timestamp)"
			return $err_msg
		}
		if (!$Timestamp) {
			$err_msg = "ERR/expecting 2 arguments (-CsvFilespec -Timestamp)"
			return $err_msg
		}
		
		$rtn_str = 'success'
			
		# BUILD THE NEW ROW
		# ... Note: this object's keys are the header column labels
		$new_row_obj = [PSCustomObject]@{
			'Timestamp'      = $Timestamp
			'SourceFile'     = $GifSourceFilespec
			'GifMethod'      = $GifMethod
			'GifImagesCnt'   = $GifImagesCnt
			'GifHeight'      = $GifHeight
			'GifFPS'         = $GifFps
			'GifTrimBegPct'  = $GifTrimBeg
			'GifTrimEndPct'  = $GifTrimEnd
			'GifTrimCustPct' = $GifTrimEndAuto
			'GifDurSec'      = $GifDurSec
			'GifSizeKb'      = $GifSize
			'GifJpgWxH'      = $GifJpgWxH
			'GifMakeMs'      = $MakeGifMsec
			'PushoverMs'     = $PushoMsec
			'TelegramMs'     = $TelegramMsec
			'TotalMs'        = $TotalMsec
			'AlertRecId'     = $AlertRecord
			'ScriptHost'     = $ScriptHostStr
			'PSVersion'      = $PsVersionFullStr
		}
		
		# For debugging
		#msg.exe $script:server_username /TIME:10 "new_row_obj... `n$new_row_obj"

		# APPEND NEW ROW TO CSV FILE, creating the file if necessary
		if ($PSVersionTable.PSVersion.Major -ge 7) {
			# REFS: https://tinyurl.com/yvy4mjy6
			#       https://stackoverflow.com/a/70717331
			#       https://stackoverflow.com/a/60681762
			try {
				# NOTE: To simulate an Export-Csv error, remove the -Force argument and make file read-only
				$new_row_obj | Export-Csv -LiteralPath $CsvFilespec -Append -Force -UseQuotes AsNeeded
			} catch {
				$err_msg = "ERR/unable to append row to CSV file; $_"
				return $err_msg
			}
		} else {
			try {
				# NOTE: To simulate an Export-Csv error, remove the -Force argument and make file read-only
				$new_row_obj | Export-Csv -LiteralPath $CsvFilespec -Append -Force -NoTypeInformation
			} catch {
				$err_msg = "ERR/unable to append row to CSV file; $_"
				return $err_msg
			}
		}
		
		if (!(Test-Path -LiteralPath $CsvFilespec)) {
			$err_msg = "ERR/unable to write to CSV log ($CsvFilespec)"
			return $err_msg
		}
		
		return $rtn_str
	}

	function PostCustomBiLogEntry() {
	# PURPOSE:  Add custom message to the Blue Iris logfile
	# ARGS:     -MsgStr    = (string) NOTE: be sure spaces are replaced with '%20'
	#           -Level     = (integer) 0|1|2 (0:info, 1:warn, 2:error)
	#                         NOTE: level=2 can trigger a status alert if the 'Send only if there are new Errors' box 
	#                         is checked on the 'Status alerts' dialog (see button at bottom of the 'Blue Iris Status' (log) dialog)
	#           -SilentFlg = (boolean) $true prevents default output to the console or msg.exe (showing the message sent)
	# RETURNS:  (string) message as posted to the Blue Iris logfile. 
	#                   OR ERR message
	# REQUIRES:  none
	# GLOBALS:   Uses vars: $script:bi_webserver, $script:post_w_creds_flg, $script:bi_user_name, $script:bi_user_pw
	# KEY POINT: Message string maximum length is 128 characters (BI HTTP interface command '/admin?log=message&level=x')
	#            MESSAGES OVER 124 CHARACTERS IN LENGTH ARE TRUNCATED AND APPENDED WITH "..."
	#
	# V2 (9/30/2024) fixed illegal URL chars in $MsgStr; tweaked global var usage
	# V3 (10/12/2024) trimmed message to not exceed the Blue Iris' maximum length of 128 characters
	# V4 (11/14/2024) fixed failure to work when using Blue Iris credentials
	# V5 (12/8/2024] added explicit type check of $web_response_obj
	# V6 (2/13/2025] clarified message truncation

		param (
			[String]$MsgStr,
			[Int]$Level,
			[Boolean]$SilentFlg
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		if (!$script:bi_webserver) {
			$err_msg = "ERR/bad web server address"
			return $err_msg
		}
		
		$http = "http://" + $script:bi_webserver
		$msg_str = $MsgStr -Replace "`n", "%20" -Replace "\\r",  "%20" -Replace "\\n", "%20" -Replace " ", "%20"

		# Trim message if it exceeds Blue Iris maximum length (128 characters)
		# ... and append an ellipsis
		# ... test string (128 characters)
		#$msg_str = 'thisisalongmessagethatisonehundredandtwentyeightcharacterslongANDthisisalongmessagethatisonehundredandtwentyeightcharacterslong'
		<# if ($msg_str.Length -gt 127) {
			# Does not work... displays as 'â€¦'
			$msg_str = $msg_str.Substring(0, [math]::Min($msg_str.Length, 125)) + "…"
		} #> 
		if ($msg_str.Length -gt 124) {
			$msg_str = $msg_str.Substring(0, [math]::Min($msg_str.Length, 124)) + "..."
		} 

		$uri = "$http/admin?log=$msg_str&level=$Level"
		if ($script:post_w_creds_flg) {
			$uri += "&user=" + $script:bi_user_name + "&pw=" + $script:bi_user_pw
		}
		
		# Do it!
		try {
			
			# Returns a BasicHtmlWebResponseObject
			$web_response_obj = `
				Invoke-WebRequest `
				-Uri $uri
		} 
		catch {

			# For debugging
			# Write-Host "$fn_name() uri (FAILED) = $uri"
			# Write-Host ("$fn_name() web_response_obj (FAILED) =`n" + $web_response_obj)	
			# Write-Host "error = `n$error"
			# Write-Host $_.ScriptStackTrace
			# Write-Host ("$fn_name() " + $_.Exception.Message) -ForegroundColor Red
			# Write-Host $PSitem
			
			$err_msg = "ERR/" + $_.Exception.Message + "`nuri = " + $uri
			return $err_msg
		}
		
		# For debugging
		# Write-Host "$fn_name() uri = $uri"
		# Write-Host ("$fn_name() web_response_obj type = " + $web_response_obj.GetType().Name)
		# Write-Host ("$fn_name() web_response_obj =`n" + $web_response_obj)	
		
		# Expecting response_obj to a string like,  
		#	signal=green
		#	schedule=Home
		#	profile=1
		#	lock=0
		#	transport=0,0/1
		#	dio_inputs=0
		
		if ($web_response_obj -match 'signal=' -and $web_response_obj -match 'schedule=' -and $web_response_obj -match 'profile=') {
			# Normal response
		} else {
			# Unexpected response
			$err_msg = "ERR/failed to post message to Blue Iris logfile"
			return $err_msg
		}

		# HANDLE NOTIFICATIONS
		if (!$SilentFlg) {
			if (!$script:FromConsole) {
				msg.exe * /TIME:10 "$fn_name() Custom BI logfile message posted...`nlevel = $level`n$MsgStr"
			}
		}
		
		if ($MsgStr.StartsWith('ERR/')) {
			$rtn_str = "Unable to to post custom BI logfile message:`n'" + ($MsgStr.Replace("%20"," ")) + "'`n"
		} else {
			$rtn_str = "Custom BI logfile message posted:`n'" + ($MsgStr.Replace("%20"," ")) + "'`n"
		}

		return $rtn_str
	}

	function Convert-FromUnixDate() {
	# Returns string like '02/01/2024 21:34:00'

		param (
			[Int]$UnixDate,
			[String]$FormatStr  # like, "yyyy-MM-dd"
		)

		if ($FormatStr) {
			$rtn_str = [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate)) | Get-Date -f $FormatStr
		} else {
			$rtn_str = [timezone]::CurrentTimeZone.ToLocalTime(([datetime]'1/1/1970').AddSeconds($UnixDate))
		}
	   
	   return $rtn_str
	   
	}

	function PauseOutput() {
	# PURPOSE:   Pause lengthy output before exceeding console height
	# KEY-POINT: The Powershell ISE UI must be handled differently because it is a Forms window, not a console
	# ATTRIBUTES:  https://stackoverflow.com/a/28237896/13499984
	#
	# V2 (11/24/2024) added check for already loaded assembly
	
		param (
			[String]$Message
		)

		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host "$fn_name() -Message = '$Message'"

		# Check if running Powershell ISE
		$isPsISE = $Host.Name -match 'ISE Host'
		#Write-Host ("HostName = " + $Host.Name)
		# Write-Host "isISE = $isPsISE"
		
		# For testing
		#$isPsISE = $true

		# KEY-POINT: System.Windows.Forms is a namespace, not a type.
		# Therefore, we cannot test if the assembly is loaded by using 'if ([System.Windows.Forms]) {}'
		# In PowerShell, square brackets [] are used for type names, not namespaces.
		# However, [System.Windows.Forms.Form] can be used to check if the System.Windows.Forms assembly is loaded, 
		# as System.Windows.Forms.Form is a valid type within the System.Windows.Forms assembly
		if (-not [System.Windows.Forms.Form]) {
			Add-Type -AssemblyName System.Windows.Forms
		}

		if ($isPsISE) {
			[System.Windows.Forms.MessageBox]::Show("$Message")
		} else {
			Write-Host "$Message" -ForegroundColor Yellow
			$x = $host.ui.RawUI.ReadKey("NoEcho,IncludeKeyDown")
		}
	}

	function Format-Color() {
	# PURPOSE:  Colorize the command's output based on line-by-line pattern matching.
	#           Also supports pausing output based on console's line height
	# SOURCE:   https://www.bgreco.net/powershell/format-color/
	# REFS:     https://superuser.com/a/1709091  (colorize output)
	#           https://stackoverflow.com/a/4647985/13499984  (colorize output)
	#           https://stackoverflow.com/a/44872079/13499984  (measure window height)
	# REQUIRES: PauseOutput()
	#
	# V2 (11/24/2024) fixed logic for implementing an output pause (unsupported in powershell_ise.exe environment)

		param (
			[Hashtable]$InputObject = @{}
		)
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name
		
		$Colors      = $InputObject.Colors
		$SimpleMatch = $InputObject.SimpleMatch
		$PauseOutput = $InputObject.PauseOutput

		# For debugging
		# Write-Host "Entered $fn_name() ..."
		# Write-Host ("$fn_name() -InputObject =" + ($InputObject | ConvertTo-Json))
		# Write-Host ("$fn_name() Colors type = " + $Colors.GetType())
		# Write-Host ("$fn_name() Colors = " + ($Colors | ConvertTo-Json))
		# Write-Host "$fn_name() SimpleMatch = $SimpleMatch"
		# Write-Host "$fn_name() PauseOutput = $PauseOutput"
		# Write-Host ("$fn_name() console_lines = " + $Host.UI.RawUI.WindowSize.Height)
		
		# GET LINE HEIGHT OF CONSOLE WINDOW
		if ($Host.UI.SupportsVirtualTerminal) {   # Host is a true console
			# KEY-POINT: This feature supports only Consoles; 
			# The Powershell ISE UI does not have a height because it is a Forms window. Its viewport 
			# has a size, but it is not tied to the number of visible lines
			# For Powershell-ISE, $console_lines = $null
			$console_lines = $Host.UI.RawUI.WindowSize.Height
		}
		
		# For debugging
		# Write-Host "$fn_name() console_lines = $console_lines"
		# if ($null -eq $console_lines) {
			# Write-Host "$fn_name() console_lines = null"
		# }

		# WRITE OUT COLORIZED LINES
		$output_lines = ($input | Out-String) -replace "`r", "" -split "`n"
		$i = 0
		foreach($line in $output_lines) {
			$color = ''
			foreach($pattern in $Colors.Keys){
				if(!$SimpleMatch -and $line -match $pattern) { 
					$color = $Colors[$pattern] 
				} elseif ($SimpleMatch -and $line.IndexOf($pattern) -gt -1) { 
					$color = $Colors[$pattern]
				}
			}

			if($color) {
				Write-Host $line -ForegroundColor $color 			
			} else {
				Write-Host $line
			}

			# PAUSE OUTPUT IN CONSOLES
			if ($Host.UI.SupportsVirtualTerminal) {  # Host is a true console
				if ($PauseOutput) {
					if ($i -eq $console_lines - 5) {
						PauseOutput -Message 'Press any key to continue...'
					}
				}
			}

			$i++
		}
	}

	<# function SleepWithStatusBar() {
	# Source: https://gist.github.com/ctigeek/bd637eeaeeb71c5b17f4

		param (
			[double]$Seconds,
			[String]$Activity,
			[String]$Message
		)
		
		$doneDT = (Get-Date).AddSeconds($Seconds)
		while($doneDT -gt (Get-Date)) {
			$secondsLeft = $doneDT.Subtract((Get-Date)).TotalSeconds
			$percent = ($Seconds - $secondsLeft) / $Seconds * 100
			Write-Progress -Activity $Activity -Status $Message -SecondsRemaining $secondsLeft -PercentComplete $percent
			[System.Threading.Thread]::Sleep(500)
		}
		Write-Progress -Activity $Activity -Status $Message -SecondsRemaining 0 -Completed
	} #>

	<# function Convert-ByteArrayToHexString() {
	# SOURCE: https://www.sans.org/blog/powershell-byte-array-and-hex-functions/
	################################################################
	#.Synopsis
	# Returns a hex representation of a System.Byte[] array as
	# one or more strings. Hex format can be changed.
	#.Parameter ByteArray
	# System.Byte[] array of bytes to put into the file. If you
	# pipe this array in, you must pipe the [Ref] to the array.
	# Also accepts a single Byte object instead of Byte[].
	#.Parameter Width
	# Number of hex characters per line of output.
	#.Parameter Delimiter
	# How each pair of hex characters (each byte of input) will be
	# delimited from the next pair in the output. The default
	# looks like "0x41,0xFF,0xB9" but you could specify "\x" if
	# you want the output like "\x41\xFF\xB9" instead. You do
	# not have to worry about an extra comma, semicolon, colon
	# or tab appearing before each line of output. The default
	# value is ",0x".
	#.Parameter Prepend
	# An optional string you can prepend to each line of hex
	# output, perhaps like '$x += ' to paste into another
	# script, hence the single quotes.
	#.Parameter AddQuotes
	# A switch which will enclose each line in double-quotes.
	#.Example
	# [Byte[]] $x = 0x41,0x42,0x43,0x44
	# Convert-ByteArrayToHexString -ByteArray $x
	#
	# 0x41,0x42,0x43,0x44
	#.Example
	# [Byte[]] $x = 0x41,0x42,0x43,0x44
	# Convert-ByteArrayToHexString -ByteArray $x -Width 2 -Delimiter "\x" -AddQuotes
	#
	# "\x41\x42"
	# "\x43\x44"
	################################################################

		[CmdletBinding()] Param (
		[Parameter(Mandatory = $True, ValueFromPipeline = $True)] [System.Byte[]] $ByteArray,
		[Parameter()] [Int] $Width = 10,
		[Parameter()] [String] $Delimiter = ",0x",
		[Parameter()] [String] $Prepend = "",
		[Parameter()] [Switch] $AddQuotes )
		
		$fn_name = (Get-Variable MyInvocation).Value.MyCommand.Name

		# For debugging.
		#Write-Host "entered function $fn_name()..."
		 
		if ($Width -lt 1) { 
			$Width = 1
		}
		if ($ByteArray.Length -eq 0) { 
			Return 
		}
		$FirstDelimiter = $Delimiter -Replace "^[\,\:\t]",""
		$From = 0
		$To = $Width - 1
		
		# For debugging
		#Write-Host ("fn_name() ByteArray cnt = " + $ByteArray.Count)
		#Write-Host ("fn_name() ByteArray[0] = " + $ByteArray[0])
		#Write-Host ("fn_name() ByteArray[1] = " + $ByteArray[1])
		#Write-Host ("fn_name() ByteArray[" + ($ByteArray.Count - 1) + "] = " + $ByteArray[$ByteArray.Count - 1])
		#Write-Host "$fn_name() From = $From, To = $To"
		#Write-Host ("fn_name() ByteArray[" + $From + "] = " + $ByteArray[$From])
		#Write-Host ("fn_name() ByteArray[" + $To + "] = " + $ByteArray[$To])
		
		Do {
			$String = [System.BitConverter]::ToString($ByteArray[$From..$To])
			
			# For debugging
			#Write-Host ("fn_name() String = " + $String)
			
			$String = $FirstDelimiter + ($String -replace "\-",$Delimiter)
			if ($AddQuotes) { 
				$String = '"' + $String + '"' 
			}
			if ($Prepend -ne "") { 
				$String = $Prepend + $String 
			}
			$String
			$From += $Width
			$To += $Width
			
			# [ADDED BY JAYDEEL] added this code to avoid error under Set-StrictMode method 
			if ($To -gt ($ByteArray.Count - 1)) {
				$To = $ByteArray.Count - 1
			}
			
		} While ($From -lt $ByteArray.Length)
	} #>

}

# LOAD REQUIRED ASSEMBLIES
if (1 -eq 1) {  # Use to collapse this code section while editing

	# 1. Load 'System.Drawing'
	#    ... used by functions MakeGIF(), IsJpegFileValid()
	# WARNING: The following assembly IS ESSENTIAL when executing this script from Task Scheduler (or Blue Iris)
	#          (on the other hand, PS7 and Powershell ISE works just fine without it)
	#          REF: https://tinyurl.com/2p929nhz
	Add-Type -AssemblyName System.Drawing
	
	if (!$NoPush -and !$Demo) {

		# 2. Load 'System.Net.Http'
		#    ... used by functions SendPushover(), SendTelegram() 
		# KEY-POINT: System.Net.Http is a namespace, not a type.
		# Therefore, we cannot test if the assembly is loaded by using 'if ([System.Net.Http]) {}'
		# (In PowerShell, square brackets [] are used for type names, not namespaces).
		# However, (-not [type]::GetType("System.Net.Http.HttpClient")) can be used to check if the assembly is loaded, 
		# as 'HttpClient' is a valid type within the System.Net.Http assembly
		Add-Type -AssemblyName System.Net.Http
		
	}

	if ($Examples) {
		
		# 3. Load 'System.Windows.Forms'
		#    ... used by function PauseOutput()() (when using argument -Examples)
		# KEY-POINT: System.Windows.Forms is a namespace, not a type.
		# Therefore, we cannot test if the assembly is loaded by using 'if ([System.Windows.Forms]) {}'
		# (In PowerShell, square brackets [] are used for type names, not namespaces).
		# However, [System.Windows.Forms.Form] can be used to check if the assembly is loaded, 
		# as System.Windows.Forms.Form is a valid type within the System.Windows.Forms assembly
		Add-Type -AssemblyName System.Windows.Forms 
	
	}
	
}

# SET FLAG FOR DISABLING THE USE OF SAVED SESSION IDS FOR BLUE IRIS JSON CALLS 
if (1 -eq 1) {  # Use to collapse this code section while editing
	if ($require_new_json_session_keys_flg) {
		$JsonCallForceFlg = $true
		
		if ($FromConsole) {
			Write-Host ("HEADS-UP: BI JSON calls will always renew session keys; see user-setting `$require_new_json_session_keys_flg") -ForegroundColor Green
		}

	} else {
		$JsonCallForceFlg = $false

		# Deprecated...
		# KEY POINT: older versions of Blue Iris (<=5.8.0) appear not to tolerate using saved session keys (from the 1st JSON 'login' cmd)
		# The first stable software version that tolerates this appears to be 5.8.1.5
		# $is_v581 = IsBiVersion "5.8.1"
		# $JsonCallForceFlg = !$is_v581
		#if ($FromConsole -and !$is_v581) {
			# Write-Host ("HEADS-UP: BI JSON calls will always renew session keys (BI version is <=5.8.0)") -ForegroundColor Green
		#}
	}
	# Write-Host "JsonCallForceFlg = $JsonCallForceFlg"

	# For testing'camlist'
	# $JsonCallForceFlg = $true
	# $JsonCallForceFlg = $false
	# Write-Host "JsonCallForceFlg for testing = $JsonCallForceFlg"
}

# GET BLUE IRIS VERSION, WEBSERVER ADDRESS & FOLDER PATHS FROM THE REGISTRY
if (1 -eq 1) {  # Use to collapse this code section while editing

	# Start performance timer - BI registry setting get times
	# ... Note: this function call can be relatively time consuming (5-100 msec), so we save the value to avoid recalling it in a loop
	$timer_bi_regsettings_get = [Diagnostics.Stopwatch]::StartNew()

	# GET CURRENT VERSION OF BLUE IRIS
	# ... from Windows registry
	# ... Note: this function call can be relatively time consuming (5-100 msec), so we save the value to avoid recalling it in a loop
	$BiVersion = GetBiVersion
	# Abort if GET failed
	if ($null -eq $BiVersion) {
		$err_msg = "ERR/unable to get the Blue Iris software version from the Windows registry"
		return $err_msg
	}
	# Write-Host "BiVersion = $BiVersion"

	# GET BLUE IRIS WEB SERVER'S LAN IP AND PORT
	# ... from Windows registry
	# ... Note: this function call can be relatively time consuming (5-100 msec), so we save the value to avoid recalling it in a loop
	$bi_webserver = GetBiWebserverIpPort
	# Abort if GET failed
	if ($null -eq $bi_webserver) {
		$err_msg = "ERR/unable to get Blue Iris web server address:port from the registry"
		return $err_msg
	}
	# Write-Host "Bi webserver = $bi_webserver"

	# GET ALL BLUE IRIS STORAGE FOLDER PATHS
	# ... from Windows registry
	# ... needed only to validate populated user-setting $bi_new_folder_paths
	$variable_test = Get-Variable -Name bi_new_folder_paths -Scope Local -ErrorAction SilentlyContinue
	if ($null -ne $variable_test) {
		$reg_bi_folders_ht = GetBiStorageFolders  # returns ordered hashtable
		# Abort if GET failed
		if ($null -eq $reg_bi_folders_ht) {
			$err_msg = "ERR/unable to get Blue Iris 'Alerts' folder path from the Windows registry"
			return $err_msg
		}
		$bi_folders_paths = $reg_bi_folders_ht.Values
			
		# For debugging
		# Write-Host ("BI folder in registry = " + ($reg_bi_folders_ht | ConvertTo-Json))
		# Write-Host ("BI folder paths = " + ($bi_folders_paths | ConvertTo-Json))
	}

	# GET THE BLUE IRIS 'NEW' FOLDER PATH
	# ... from Windows registry
	$reg_bi_new_folder_path = GetBiNewFolder
	# Abort if GET failed
	if ($null -eq $reg_bi_new_folder_path) {
		$err_msg = "ERR/unable to get Blue Iris 'New' folder path from the Windows registry"
		return $err_msg
	}
	# Write-Host "reg_bi_new_folder_path = $reg_bi_new_folder_path"

	# GET THE BLUE IRIS 'ALERTS' FOLDER PATH
	# ... from Windows registry
	$reg_bi_alerts_folder_path = GetBiAlertsFolder
	# Abort if GET failed
	if ($null -eq $reg_bi_alerts_folder_path) {
		$err_msg = "ERR/unable to get Blue Iris 'Alerts' folder path from the Windows registry"
		return $err_msg
	}
	# Write-Host "reg_bi_alerts_folder_path = $reg_bi_alerts_folder_path"

	# GET THE BLUE IRIS 'TEMP' FOLDER PATH
	# ... from Windows registry
	$reg_bi_temp_folder_path = GetBiTempFolder
	# Abort if GET failed
	if ($null -eq $reg_bi_temp_folder_path) {
		$err_msg = "ERR/unable to get Blue Iris 'temp' folder path from the Windows registry"
		return $err_msg
	}
	# Write-Host "reg_bi_temp_folder_path = $reg_bi_temp_folder_path"

	# Stop timer
	$timer_bi_regsettings_get.Stop()
	$elapsed_msec = [Math]::Round($timer_bi_regsettings_get.elapsed.totalmilliseconds,1)
	$timer_bi_regsettings_get_time_ms = $elapsed_msec.ToString("F1")

	if ($FromConsole -and $DevMode) {
		Write-Host ("----------------------------------------") -ForegroundColor Cyan
		Write-Host ("(dev mode) Time to get various Blue Iris setting from the registry = $timer_bi_regsettings_get_time_ms msec") -ForegroundColor Cyan
		Write-Host ("(dev mode) --> version, webserver address, storage folder paths, temp folder path") -ForegroundColor Cyan
		#exit
	}
}

# GET SHORT- & LONG-NAMES, AND PROFILES for ALL BLUE IRIS CAMERAS
# ... via Blue Iris JSON interface
# ... required for RegGetValue() calls (or functions using RegGetValue())
# ... KEY POINT: MUST follow code getting Blue Iris webserver address & port
if (1 -eq 1) {  # Use to collapse this code section while editing

	# Start performance timer
	$timer_BIJsonCamListParse = [Diagnostics.Stopwatch]::StartNew()

	<# $camlist_ht is like...
	 {
		"DW1": {
			"optionDisplay": "DriveWay-1",
			"isEnabled": true,
		   "profile": 5
		},
		"DW2": {
			"optionDisplay": "DriveWay-2",
			"isEnabled": true
		   "profile": 5
		},
		---
	 }
	 -OR-
	 [string]error_msg
	#>
	$camlist_ht = `  # returns ordered hashtable
		BIJsonCamListParse `
		-CamProps 'optionDisplay,isEnabled,profile'
					
	# For debugging
	# Write-Host ("BIJsonCamListParse() camlist_ht type = " + $camlist_ht.GetType())
	# Write-Host ("BIJsonCamListParse() camlist_ht cnt = " + $camlist_ht.Count)
	# Write-Host ("BIJsonCamListParse() camlist_ht = " + ($camlist_ht | ConvertTo-Json))

	# Stop timer
	$timer_BIJsonCamListParse.Stop()
	$elapsed_msec = [Math]::Round($timer_BIJsonCamListParse.elapsed.totalmilliseconds,1)
	$BIJsonCamListParse_time_ms = $elapsed_msec.ToString("F1")

	# For debugging
	# Write-Host "BIJsonCamListParse() time ms = $BIJsonCamListParse_time_ms"
	
	if ($FromConsole -and $DevMode) {
		Write-Host ("(dev mode) time to execute BIJsonCamListParse() = $BIJsonCamListParse_time_ms msec") -ForegroundColor Cyan
	}

	# For testing - Simulate a bad JSON alertlist response
	#$camlist_ht = "ERR/test"
	
	if ($camlist_ht.Count -eq 0) {
		$err_msg = "ERR/unexpected response from BIJsonCamListParse(); returned object is empty"
		return $err_msg
	}					
	if ($camlist_ht -is [String]) {
		if ($camlist_ht.StartsWith('ERR/')) {
			return $camlist_ht	
		} else {
			$err_msg = "ERR/unexpected response from BIJsonCamListParse()"
			return $err_msg
		}
	}
			
	# For debugging
	# Write-Host ("hashtable keys = " + ([String[]]$camlist_ht.Keys | ConvertTo-Json))
	# Write-Host ("hashtable = " + ($camlist_ht | ConvertTo-Json))

	# Start performance timer
	$timer_make_bi_camera_lookup_objects = [Diagnostics.Stopwatch]::StartNew()

	$cam_shortnames_ary = [String[]]$camlist_ht.Keys
	$cam_lognnames_ht = @{}
	$cam_profiles_current_ht = @{}
	foreach ($item in $cam_shortnames_ary) {
		if ($camlist_ht[$item]['isEnabled']) {
			$cam_lognnames_ht.Add($item, $camlist_ht[$item]['optionDisplay'])
			$cam_profiles_current_ht.Add($item, $camlist_ht[$item]['profile'])
		}
	}

	# Stop timer
	$timer_make_bi_camera_lookup_objects.Stop()
	$elapsed_msec = [Math]::Round($timer_make_bi_camera_lookup_objects.elapsed.totalmilliseconds,1)
	$make_bi_camera_lookup_objects_time_ms = $elapsed_msec.ToString("F1")

	# For debugging
	# Write-Host "make_bi_camera_lookup_objects_time_ms = $make_bi_camera_lookup_objects_time_ms"
	
	if ($FromConsole -and $DevMode) {
		# Write-Host ("(dev mode) time to populate BI camera lookup objects = $make_bi_camera_lookup_objects_time_ms") -ForegroundColor Cyan
	}
			
	# For debugging
	# Write-Host ("cam_shortnames_ary = " + ($cam_shortnames_ary | ConvertTo-Json))
	# Write-Host ("cam_lognnames_ht = " + ($cam_lognnames_ht | ConvertTo-Json))
	# Write-Host ("cam_profiles_current_ht = " + ($cam_profiles_current_ht | ConvertTo-Json))
}

# GET VARIOUS BLUE IRIS CAMERA SETTINGS FROM THE REGISTRY
# ... KEY POINT: MUST follow code getting Blue Iris camera longmaes
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($FromConsole -and $DevMode) {
		Write-Host ""
	}

	# Start performance timer - BI version check times
	$timer_bicamregsettings_get = [Diagnostics.Stopwatch]::StartNew()

	# GET CAMERA's 'Video recording method', 'Pre-trigger record time' and 'Pre-trigger playback time'
	# and 'AI analyze one each msec'
	if ($PSBoundParameters.ContainsKey('Camera')) {

		$cam_longname = $cam_lognnames_ht.$Camera
		$cam_profile_current = $cam_profiles_current_ht.$Camera

		# For debugging
		# Write-Host "Camera = '$Camera'"
		# Write-Host "cam_longname = '$cam_longname'"
		# Write-Host "cam_profile_current = $cam_profile_current"

		# 1. VIDEO RECORDING METHOD
		# ... from Windows registry
		$reg_cam_rec_is_continuous = $false  # Default
		$response = `
			IsCamRecContinuous `
			-CamLongName $cam_longname `
			-CamProfileNo $cam_profile_current			

		if ($response -is [Boolean]) { 
			$reg_cam_rec_is_continuous = $response
		}

		# For debugging
		# Write-Host ("response type = " + $response.GetType())
		# Write-Host "response = $response"
		# Write-Host "reg_cam_rec_is_continuous = $reg_cam_rec_is_continuous"

		# 2. PRE-TRIGGER RECORD TIME
		# ... from Windows registry
		$reg_cam_pretrig_rectime_msec = 0  # Default
		$response = `
			GetPreTrigRecMsec `
			-CamLongName $cam_longname `
			-CamProfileNo $cam_profile_current			

		if ($response -is [Int]) { 
			$reg_cam_pretrig_rectime_msec = $response
		}

		# For debugging
		# Write-Host "Camera = '$Camera'"
		# Write-Host "cam_longname = '$cam_longname'"
		# Write-Host ("response type = " + $response.GetType())
		# Write-Host "response = $response"
		# Write-Host "reg_cam_pretrig_rectime_msec = $reg_cam_pretrig_rectime_msec"

		# 3. PRE-TRIGGER PLAYBACK TIME
		$reg_cam_pretrig_playback_msec = 0  # Default
		$response = `
			GetPreTrigPlaybackMsec `
			-CamLongName $cam_longname `
			-CamProfileNo $cam_profile_current			

		if ($response -is [Int]) { 
			$reg_cam_pretrig_playback_msec = $response
		}

		# For debugging
		# Write-Host "Camera = '$Camera'"
		# Write-Host "cam_longname = '$cam_longname'"
		# Write-Host ("response type = " + $response.GetType())
		# Write-Host "response = $response"
		# Write-Host "reg_cam_pretrig_playback_msec = $reg_cam_pretrig_playback_msec"

		# 4. AI ANALYZE ONE EACH MSEC
		# ... from Windows registry
		$reg_cam_ai_analyze_each_msec = 0  # Default
		$response = `
			GetAiAnalyzeOneEachMsec `
			-CamLongName $cam_longname `
			-CamProfileNo $cam_profile_current			

		if ($response -is [Int]) { 
			$reg_cam_ai_analyze_each_msec = $response
		}

		# For debugging
		# Write-Host "Camera = '$Camera'"
		# Write-Host "cam_longname = '$cam_longname'"
		# Write-Host ("response type = " + $response.GetType())
		# Write-Host "response = $response"
		# Write-Host "reg_cam_ai_analyze_each_msec = $reg_cam_ai_analyze_each_msec"

		if ($PSBoundParameters.ContainsKey('DatImageMsec')) {
			if ($dat_image_interval_msec -ne $reg_cam_ai_analyze_each_msec) {
				if ($FromConsole) {
					Write-Warning "argument -DatImageMsec value ($dat_image_interval_msec) does not match Alerts > AI configuration' value ($reg_cam_ai_analyze_each_msec)"
					Write-Host ("Using argument's value regardless; check if GIF's time span (last frame - first frame) makes sense") -ForegroundColor Yellow
				}
			}			
		} else {
			if ($reg_cam_ai_analyze_each_msec -ne 0) {
				$dat_image_interval_msec = $reg_cam_ai_analyze_each_msec
			} else {
				$dat_image_interval_msec = 500
			}
		}
	}

	# Stop timer
	$timer_bicamregsettings_get.Stop()
	$elapsed_msec = [Math]::Round($timer_bicamregsettings_get.elapsed.totalmilliseconds,1)
	$timer_bicamregsettings_get_time_ms = $elapsed_msec.ToString("F1")

	if ($FromConsole -and $DevMode) {
		Write-Host ("----------------------------------------") -ForegroundColor Cyan
		Write-Host ("(dev mode) Time to get various camera settings from the registry = $timer_bicamregsettings_get_time_ms msec") -ForegroundColor Cyan
		Write-Host ("(dev mode) --> " + $Camera + ": Video recording method, Pre-trigger record time, Pre-trigger playback time") -ForegroundColor Cyan
		# exit
	}
}

# VALIDATE USER SETTINGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# 1. Handle REQUIRED user settings
	
	$variable_test = Get-Variable -Name app_path -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/user-settings file is missing setting `$app_path"
		return $err_msg
	} else {
		if ($app_path.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$app_path is not set"
			return $err_msg
		}
	}
	$app_path = $app_path.TrimEnd('\')

	$bi_new_folder_paths_ary = @($reg_bi_new_folder_path)  # $reg_bi_new_folder_path already retrieved from the Registry
	$variable_test = Get-Variable -Name bi_new_folder_paths -Scope Local -ErrorAction SilentlyContinue
	if ($null -ne $variable_test) {
		$arr = $bi_new_folder_paths -Split ','								
		foreach ($item in $arr) {
			if ($item) {
				$path = $item.Trim().TrimEnd('\')
				if ($path) {
					if ($bi_folders_paths -contains $path) {
						if (!($bi_new_folder_paths_ary -contains $path)) {
							$bi_new_folder_paths_ary += $path
						}
					} else {
						$err_msg = "ERR/user-setting `$bi_new_folder_paths contains an unrecognized BI Storage Folder path: '$path'"
						return $err_msg
					}
				}
			}
		}
	}

	# For debugging
	# Write-Host ("bi_new_folder_paths_ary = " + ($bi_new_folder_paths_ary | ConvertTo-Json))

	if ($bi_new_folder_paths_ary.Count -gt 1) {
		if ($DevMode) {
			if ($FromConsole) {
				Write-Host "(dev mode) Blue Iris appears to be using MULTIPLE 'NEW' FOLDERS" -ForegroundColor Cyan
				Write-Host ($bi_new_folder_paths_ary | ConvertTo-Json) -ForegroundColor Cyan
			}
		}
	}

	$variable_test = Get-Variable -Name bi_user_name -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/User-settings file is missing setting `$bi_user_name"
		return $err_msg
	} else {
		if ($bi_user_name.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$bi_user_name is not set"
			return $err_msg
		}
	}

	$variable_test = Get-Variable -Name bi_user_pw -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$err_msg = "ERR/User-settings file is missing setting `$bi_user_pw"
		return $err_msg
	} else {
		if ($bi_user_pw.StartsWith('YOUR')) {
			$err_msg = "ERR/user-setting file setting `$bi_user_pw is not set"
			return $err_msg
		}
	}

	$push_service_ary = @()
	if ($PSBoundParameters.ContainsKey('PushService')) {
		$push_service_ary = $PushService
		foreach ($item in $push_service_ary) {
			if (!(@('pushover', 'telegram') -contains $item.ToLower())) {  # -contains operator is case-insensitive by default
				$err_msg = "ERR/argument -PushService must be 'Pushover' or 'Telegram' or 'Pushover','Telegram'"
				return $err_msg
			}
		}
	} else {
		$variable_test = Get-Variable -Name push_service -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$push_service_ary = @('pushover')
		} else {
			# Write-Host ("push_service type = " + $push_service.GetType())
			if ($push_service -is [String]) {  # user entered string like 'Pushover,Telegram'
				# Force to array
				$push_service_ary = $push_service -Split ','
			} else {  # user entered list like 'Pushover','Telegram'
				$push_service_ary = $push_service
			}
			foreach ($item in $push_service_ary) {
				if (!(@('pushover', 'telegram') -contains $item.ToLower())) {  # -contains operator is case-insensitive by default
					$err_msg = "ERR/user-setting file setting `$push_service must be 'Pushover' or 'Telegram' or 'Pushover,Telegram'"
					return $err_msg
				}
			}
		}
	}
	for ($i=0; $i -lt $push_service_ary.Count; $i++) {
		$push_service_ary[$i] = $push_service_ary[$i].ToLower()
	}
	# For debugging
	# Write-Host ("push_service_ary = " + ($push_service_ary | ConvertTo-Json))
	
	if ($push_service_ary -contains 'pushover') {

		$variable_test = Get-Variable -Name po_app_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$po_app_token"
			return $err_msg
		} else {
			if ($po_app_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$po_app_token is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name po_user_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$po_user_token"
			return $err_msg
		} else {
			if ($po_user_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$po_user_token' is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name po_priority -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_priority = 0
		}

		$variable_test = Get-Variable -Name po_sound -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_sound = ""
		}

		$variable_test = Get-Variable -Name po_device -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_device = ""
		}

		$variable_test = Get-Variable -Name po_retry -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_retry = 30
		}

		$variable_test = Get-Variable -Name po_expire -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_expire = 300
		}

		$variable_test = Get-Variable -Name po_tags -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$po_tags = ""
		}

	}
	
	if ($push_service_ary -contains 'telegram') {

		$variable_test = Get-Variable -Name tg_chat_id -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$tg_chat_id"
			return $err_msg
		} else {
			if ($tg_chat_id.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$tg_chat_id is not set"
				return $err_msg
			}
		}

		$variable_test = Get-Variable -Name tg_api_token -Scope Local -ErrorAction SilentlyContinue
		if ($null -eq $variable_test) {
			$err_msg = "ERR/user-settings file is missing setting `$tg_api_token"
			return $err_msg
		} else {
			if ($tg_api_token.StartsWith('YOUR')) {
				$err_msg = "ERR/user-setting file setting `$tg_api_token' is not set"
				return $err_msg
			}
		}
	}

	# 2. Handle OPTIONAL user settings
	
	$variable_test = Get-Variable -Name experienced_user_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$experienced_user_flg = $false
	}
	
	$variable_test = Get-Variable -Name require_new_json_session_keys_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$require_new_json_session_keys_flg = $false
	}

	$variable_test = Get-Variable -Name bg_msgs_on -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$bg_msgs_on = $true
	}

	$variable_test = Get-Variable -Name gif_images_cnt -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_images_cnt = 10
	}
	# Write-Host "gif_trim_beg = $gif_trim_beg"

	$variable_test = Get-Variable -Name jpg_images_from_substream_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$jpg_images_from_substream_flg = $false
	}
	# Write-Host "jpg_images_from_substream_flg = $jpg_images_from_substream_flg"

	$variable_test = Get-Variable -Name jpg_images_match_gif_height_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$jpg_images_match_gif_height_flg = $false
	}
	# Write-Host "jpg_images_match_gif_height_flg = $jpg_images_match_gif_height_flg"

	$variable_test = Get-Variable -Name gif_trim_beg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim_beg = 0
	}
	# Write-Host "gif_trim_end = $gif_trim_end"

	$variable_test = Get-Variable -Name gif_trim_end -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim_end = 0
	}
	# Write-Host "gif_trim_end = $gif_trim_end"

	$variable_test = Get-Variable -Name gif_trim_end_cust_obj -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim_end_cust_obj = @{}
	}
	# Write-Host ("gif_trim_end_cust_obj = " + ($gif_trim_end_cust_obj | ConvertTo-Json))

	$variable_test = Get-Variable -Name gif_height -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_height = 320
	}

	$variable_test = Get-Variable -Name gif_fps -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_fps = 2
	}

	$variable_test = Get-Variable -Name gif_trim -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_trim = 0
	}

	$variable_test = Get-Variable -Name gif_archive_lifetime_hours -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_archive_lifetime_hours = 360
	}

	$variable_test = Get-Variable -Name gif_DAT_reqd_images_cnt -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$gif_DAT_reqd_images_cnt = 4
	}

	$variable_test = Get-Variable -Name dat_is_fresh_mult -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$dat_is_fresh_mult = 2
	}

	$variable_test = Get-Variable -Name dat_image_interval_msec -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$dat_image_interval_msec = 500
	}

	$variable_test = Get-Variable -Name append_gif_method_tag_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$append_gif_method_tag_flg = $true
	}

	$variable_test = Get-Variable -Name append_performance_tag_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$append_performance_tag_flg = $true
	}

	$variable_test = Get-Variable -Name append_performance_tag_details_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$append_performance_tag_details_flg = $false
	}

	$variable_test = Get-Variable -Name post_to_BI_logfile_flg -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_to_BI_logfile_flg = $false
	}

	$variable_test = Get-Variable -Name post_all_events -Scope Local -ErrorAction SilentlyContinue
	if ($null -eq $variable_test) {
		$post_all_events = $false
	}
}

# ABORT IF USER SETTINGS PATHS NOT FOUND
if (!$app_path) {
	$err_msg = "ERR/setting '`$app_path' missing in user-settings file`n"
	if ($post_to_BI_logfile_flg) {
		$null = `
			PostCustomBiLogEntry `
			-MsgStr "$err_msg ($ScriptName)" `
			-Level 2 `
			-SilentFlg ($bg_msgs_on -eq $false)
	}
	if ($FromConsole) {
		return $err_msg
	} else {
		msg.exe $server_username /TIME:10 "$err_msg`n`n(via script: '$ScriptName')"
		exit
	}	
}

# MANAGE APP'S SUBFOLDERS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# Note: $app_path is user-defined (_user_settings.ps1 file)

	# 1. Home folder
	if (!(Test-Path -LiteralPath $app_path)) {
		New-Item -ItemType Directory -Force -Path $app_path | Out-Null
	}

	# 2. Create subfolder for archiving GIF files created by this script.
	$gif_archive_path = ($app_path + "\_gif_archive")  # like 'D:\pushover_gifs\_gif_archive'
	if (!(Test-Path -LiteralPath $gif_archive_path)) {
		New-Item -ItemType Directory -Force -Path $gif_archive_path
	}
	# ... And add/refresh a README to the folder
	$readme_path = "$gif_archive_path\_README.txt"
	if (!(Test-Path -LiteralPath $readme_path)) {
		New-Item -Path "$readme_path" | Out-Null
		$content = "GIF Files Archive Folder" +
			"`n------------------------" +
			"`n`nDESCRIPTION: This folder archives every GIF file created by script 'BI_gif_tool.ps1'." + 
			"`n`nTIP: This folder may be managed by the script to automatically prune old files." + 
			"`n     To implement this feature, see setting '`$gif_archive_lifetime_hours' in the user-settings file" +
			"`n`nNote: This folder and this README.txt file are programmatically created."

		# Ensure UTF-8 w/o BOM in PS5.1 & PS7 using version detection
		if ($PSVersionTable.PSVersion.Major -lt 7) {
			$utf8NoBom = New-Object System.Text.UTF8Encoding $false
			[System.IO.File]::WriteAllText($readme_path, $content, $utf8NoBom)
		} else {
			$content | Out-File -FilePath "$readme_path" -Encoding utf8NoBOM
		}
	}
		
	# 3. Create the Script's DevMode folder if necessary
	$DevmodePath = ($app_path + "\_devmode")  # like 'D:\pushover_gifs\_devmode'
	if (!(Test-Path -LiteralPath $DevmodePath)) {
		New-Item -ItemType Directory -Force -Path $DevmodePath | Out-Null
	}
	# ... and add/refresh a README to the folder
	if ($DevMode) {
		$readme_path = "$DevmodePath\_README.txt"
		if (!(Test-Path -LiteralPath $readme_path)) {
			New-Item -Path "$readme_path" | Out-Null
			$content = "Script 'DevMode' Folder" +
				"`n----------------------------" +
				"`n`nDESCRIPTION: This folder holds output created when using the -DevMode argument" + 
				"`n`nNote: This folder and this README.txt file are programmatically created."

			# Ensure UTF-8 w/o BOM in PS5.1 & PS7 using version detection
			if ($PSVersionTable.PSVersion.Major -lt 7) {
				$utf8NoBom = New-Object System.Text.UTF8Encoding $false
				[System.IO.File]::WriteAllText($readme_path, $content, $utf8NoBom)
			} else {
				$content | Out-File -FilePath "$readme_path" -Encoding utf8NoBOM
			}
		}
	}
}

# CHECK IF SCRIPT'S PATH IS IN ENVIRONMENT 'SYSTEM PATH' VARIABLE
if (!$experienced_user_flg) {
	if ($FromConsole) {

		$EnvPath = [Environment]::GetEnvironmentVariable('Path', 'Machine')

		# For debugging
		# Write-Host "ScriptPath = $ScriptPath"
		# Write-Host ("EnvPath = " + (($EnvPath -Split ';') | ConvertTo-Json))

		if ($null -ne $EnvPath) {
			if (!($EnvPath -split ';' -contains $ScriptPath)) {

				#return "ERR/this script's path '$ScriptPath' does not exist in the Environment System Path variable"
				Write-Host "HEADS-UP: this script's path '$ScriptPath' does not exist in the Environment System Path variable" -ForegroundColor Green
				
				# Attempt repair
				# ABORTED: when I tried this it added MANY other/duplicate entries
				#[Environment]::SetEnvironmentVariable("Path", $EnvPath + ";$ScriptPath", "Machine")
				#Write-Warning "Added this script's path '$ScriptPath' to the Environment Path"
			}
		} else {
			#Write-Warning "Unable to inspect Environment variable; be sure script's path is in the Environment Path"
			Write-Host "HEADS-UP: Unable to inspect Environment variable; be sure script's path is in the Environment Path" -ForegroundColor Green
		}

	}
}

# SHOW USAGE GUIDE
if ($Help -and $FromConsole) {
	Write-Host "================================================================================================================"
	Write-Host "BLUE IRIS GIF TOOL - MAKES GIF FILE & SENDS PUSHOVER/TELEGRAM NOTIFICATION WITH GIF ATTACHMENT"
	Write-Host "----------------------------------------------------------------------------------------------------------------"
	Write-Host "USAGE: $ScriptName -AlertRec [Args]          ... send GIF using BI database record, like '@123456' [1]"
	Write-Host "       $ScriptName -DbRecNo | -UI3Rec [Args] ... send GIF from ANY EXISTING ALERT with a duration [2,3]"
	Write-Host "       $ScriptName -Camera -Latest [Args]    ... send GIF from camera's most recent alert [4]"
	Write-Host "       $ScriptName -Camera -Live [Args]      ... send GIF from camera's active video stream"
	Write-Host "OPTIONAL ARGS:"
	Write-Host "       -IgnoreDat    ... make GIF from BVR clip instead of DAT file (DAT is the default)"
	Write-Host "       -DatImageMsec ... custom msec between AI images (applies to DAT-derived GIFs only) [5]"
	Write-Host "       -BvrDelayMsec ... custom msec to wait before extracting images (applies to BVR-derived GIFs only) [6]"
	Write-Host "       -Seconds      ... custom GIF duration (default is 10 seconds) (for -Live & BVR-derived GIFs only)"
	Write-Host "       -ImagesCnt -Height -FPS ... change GIF's default settings (per the user-settings file)"
	Write-Host "       -TrimBegPct -TrimEndPct ... make GIF from middle (active) portion of the Alert clip's duration [7]"
	Write-Host "       -AlertMemo ... auto-trim GIF for faster- or slower-moving objects, as labeled by AI service [8]"
	#Write-Host "       -NukeDatFile "  # REDACTED
	Write-Host "       -PushService ... override default push application [9]"
	Write-Host "       -Msg      ... notification message (may contain HTML)"
	Write-Host "       -Title    ... custom title for notification; default is 'via $ScriptName'"
	Write-Host "       -Priority ... notification priority: -2|-1|0|1|2 = lowest|silent|normal|high|emergency (Pushover only)"
	Write-Host "       -Sound    ... custom sound for notification (Pushover only)"
	Write-Host "       -Devices  ... custom device(s) for notification; comma-delimited (Pushover only)"
	Write-Host "       -NoPush -NoBiLog -NoArchive ... disable default actions (per the user-settings file) [10]"
	Write-Host "       -IgnoreDatAge ... ignores age of temporary 'DAT' files (forces use of the file)"
	Write-Host "       -Demo  ... skip notification/logfile/delete actions (useful when tuning GIF params)"
	Write-Host "       -Quiet ... suppress verbose output to the Console; show the JSON return object instead"
	Write-Host "       -Help     ... show this usage guide"
	Write-Host "       -Examples ... show usage examples"
	Write-Host "NOTES:"
	Write-Host "[1]  [-AlertRec X] value must be prefixed with '@'; for usage from Blue Iris 'On Alert' Actions, see [A]"
	Write-Host ("     TIP: ctrl-right-click any Alert in the clipslist to load its DB record to the clipboard (BI 5.7.8.4+)") -ForegroundColor Cyan
	Write-Host "[2]  [-DbRecNo X] is the numeric part of the Alert's database record (must NOT be prefixed with '@') "
	Write-Host "[3]  [-UI3Rec X] is the value after '&rec=' in the UI3 address bar (viewing an alert)"
	Write-Host "[4]  [-Latest] uses the most recent entry in the camera's alertlist; for usage from BI Actions, see [B]"
	Write-Host "[5]  [-DatImageMsec X] may be used to improve the estimation of the GIF's duration; by default, the script"
	Write-Host "     uses the value from BI > Camera settings > Alert > 'AI configuration' > 'Analyze one image each'"
	Write-Host "[6]  [-BvrDelayMsec X] may be needed if the BVR is still buffering (during which the BI HTTP interface command "
	Write-Host "     is unable to extract JPGs); by default, the delay is 2*(cam pre-trigger record time) via the registry"
	Write-Host "[7]  [-TrimBegPct X] may be negative to start GIF before the Alert's offset (BVR-derived GIFs only)"
	Write-Host "[8]  [-AlertMemo string] is intended for use via Blue Iris 'On Alert' Actions [C]; the string is used"
	Write-Host "     to right-trim the GIF if user-settings file variable '`$gif_trim_end_cust_obj' has been configured"
	#Write-Host "Note:  use [-NukeDatFile] to remove DAT file from Blue Iris storage (always archived first); "  # REDACTED
	#Write-Host "       ALWAYS executes a Database/Regenerate action (required per Blue Iris support)"
	Write-Host "[9]  [-PushService X] must be 'Pushover' or 'Telegram' or 'Pushover','Telegram' (use both)"
	Write-Host "[10] [-NoPush] disables sending a notification; ALSO disables management of GIF size to below 5.0 MB"
	Write-Host ""
	Write-Host "USING THIS SCRIPT W/ MACROS FROM A BLUE IRIS ACTION SET:"
	Write-Host ("[A] use [-AlertRec \&ALERT_DB]") -NoNewLine
	Write-Host (" <-- KEY POINT: a backslash (or backtick '``') prefix is ESSENTIAL") -ForegroundColor Cyan
	Write-Host ("[B] use [-Camera '&CAM' -Latest]") -NoNewLine
	Write-Host (" <-- alternative to [A]") -ForegroundColor Cyan
	Write-Host "[C] use [-AlertMemo '&MEMO']"
	Write-Host ""
	Write-Host "RETURNS: (json string) if successful, or (string) 'ERR/*' if failure"
	Write-Host "(json string) is like {'result: success, data: {gif_method: viaDat, gif_kb: 497.5, ...}}"
	#Write-Host "(String) is like 'ERR/description'"
	Write-Host "----------------------------------------------------------------------------------------------------------------"
	Write-Host ""
	exit
}

# SHOW EXAMPLES
if ($Examples -and $FromConsole) {
	$examples_str = `
	"==================================================================================================" `
	+ "`nBLUE IRIS GIF TOOL - EXAMPLES" `
	+ "`n--> to try these commands, copy/paste the colored text to the console prompt" `
	+ "`n--------------------------------------------------------------------------------------------------" `
	+ "`n  1. Send demo 10- or 15-second 'live' GIF notification for the specified camera" `
	+ "`n     $ScriptName -Camera DW1 -Live" `
	+ "`n     $ScriptName -Camera DW1 -Live -Seconds 15" `
	+ "`n  2. Send GIF notification for any alert record obtained by ctrl-right-clicking clips list entry" `
	+ "`n     (requires Blue Iris 5.7.8.4+)" `
	+ "`n     $ScriptName -AlertRec '@161248235'"  `
	+ "`n  3. Like #2, using only the numeric value for the alert's database record" `
	+ "`n     $ScriptName -DbRec 161248235"  `
	+ "`n  4. Like #2, using the value(s) copied from the UI3 address bar after the rec= argument" `
	+ "`n     $ScriptName -UI3Rec '161248235-2298624' .or. $ScriptName -UI3Rec 161248235"  `
	+ "`n  5. Send 10-sec GIF notification for specified camera's most recent alert, from the BVR file" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat"  `
	+ "`n  6. Like #5, but make a 15-sec GIF" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -Seconds 15"  `
	+ "`n  7. Like #5, but trim the end of alert duration by 40%" `
	+ "`n     (esp. useful w/ BVR-derived GIFs when the object is fast moving, e.g., a vehicle)" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -TrimEndPct 40"  `
	+ "`n  8. Like #5, but with larger, faster GIF" `
	+ "`n     (warning: larger GIFs may delay notification or exceed API's attachment size limit)" `
	+ "`n     (script will automatically try to right-trim frames to avoid exceeding 5.0 MB size limit)" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -ImagesCnt 15 -Height 500 -FPS 10"  `
	+ "`n  9. Like #5, using non-default push service" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -PushService 'telegram'"  `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -PushService 'pushover',telegram'"  `
	+ "`n 10. Like #5, but GIF spans only the center portion of alert clip (GIF duration now 6-seconds)" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -TrimEnd 20 -TrimBeg 20"  `
	+ "`n 11. Tune GIF without sending notification or posting to Blue Iris logfile (useful while testing)" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -TrimEnd 20 -TrimBeg 20 -NoPush -NoBiLog" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -TrimEnd 20 -TrimBeg 20 -Demo" `
	+ "`n 12. Like #5, but dynamically adjust the GIF's right trim based on the alert's AI results" `
	+ "`n     (this requires the customization of user-setting `$gif_trim_end_cust_obj)" `
	+ "`n     (this example is a simulation - the command would typically be used from a Blue Iris " `
	+ "`n     'Run a program' Action using a Blue Iris macro, i.e., -AlertMemo &MEMO" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -AlertMemo 'vehicle:93%' -Demo"  `
	+ "`n 13. Like #5, using custom notification sound (assumes the default service is Pushover)" `
	+ "`n     (learn more here https://tinyurl.com/mwchwa7n)" `
	+ "`n     $ScriptName -Camera DW1 -Latest -IgnoreDat -Sound 'this-is-it'"  `
	+ "`n FROM BLUE IRIS ACTION SET: "`
	+ "`n 14. Send GIF notification using active alert's database record macro $ALERT_DB" `
	+ "`n     `"[path\]$ScriptName`" -AlertRec \&ALERT_DB -Msg `"msgstr`"..."  `
	+ "`n     `"[path\]$ScriptName`" -AlertRec ``&ALERT_DB -Msg `"msgstr`"..."  `
	+ "`n 15. Send GIF notification using active alert's camera macro $ALERT_DB" `
	+ "`n     `"[path\]$ScriptName`" -Camera `"&CAM`" -Latest -Msg `"msgstr`"..."  `
	+ "`n 16. Like #13 & #14, and auto-trimming GIF based on AI results" `
	+ "`n     `"[path\]$ScriptName`" -AlertRec \&ALERT_DB -AlertMemo `"&MEMO`" -Msg `"msgstr`"..."  `
	+ "`n     `"[path\]$ScriptName`" -Camera `"&CAM`" -Latest -AlertMemo `"&MEMO`" -Msg `"msgstr`"..."  `
	+ "`n--------------------------------------------------------------------------------------------------" 

	# For inline hashtables, use unquoted keys and a semicolon separator
	# https://learn.microsoft.com/en-us/powershell/scripting/learn/deep-dives/everything-about-hashtable?view=powershell-7.4
	$InputObject = @{
		Colors = @{ "$ScriptName" = 'Cyan' };
		SimpleMatch = $true;
		PauseOutput = $true
	}
	# We use an object in the following line because you can only have one parameter accept 
	# input from the Pipeline by "Value" (for each value type, e.g string, int etc.)
	# REF: https://stackoverflow.com/a/53780647/13499984
	$examples_str | Format-Color $InputObject
	
	exit
}

# HANDLE COMMANDLINE ARGS
if (1 -eq 1) {  # Use to collapse this code section while editing

	# PERMANENTLY DISABLED
	if ($PSBoundParameters.ContainsKey('NukeDatFile')) {
		$NukeDatFile = $false
		if ($FromConsole) {
			#Write-Warning "Argument -NukeDatFile performs a DB/Regenerate action, which temporarily disables all camera recording;`nTip: it might be preferable to batch delete no-longer-needed DAT files via my 'BI_DAT_tool.ps1' script"
			Write-Warning "Argument -NukeDatFile has been deprecated (see changelog)"
		}
	}

	if ($PSBoundParameters.ContainsKey('NoArchiveDat')) {
		$NoArchiveDat = $false
		if ($FromConsole) {
			Write-Warning "Argument -NoArchiveDat has been deprecated (see changelog)"
		}
	}

	if ($PSBoundParameters.ContainsKey('Camera') -and !($PSBoundParameters.ContainsKey('Live') -or $PSBoundParameters.ContainsKey('Latest'))) {
		$err_msg = "ERR/expecting argument -Live or -Latest when using argument -Camera"
		return $err_msg
	}

	if ($PSBoundParameters.ContainsKey('Live') -and !$PSBoundParameters.ContainsKey('Camera')) {
		$err_msg = "ERR/expecting argument -Camera when using -Live"
		return $err_msg
	}

	if ($PSBoundParameters.ContainsKey('Latest') -and !$PSBoundParameters.ContainsKey('Camera')) {
		$err_msg = "ERR/expecting argument -Camera when using -Latest"
		return $err_msg
	}

	if (!$PSBoundParameters.ContainsKey('AlertRec') -and !$PSBoundParameters.ContainsKey('DbRecNo') -and !$PSBoundParameters.ContainsKey('UI3Rec') -and !$PSBoundParameters.ContainsKey('Camera')) {
		#$err_msg = "ERR/expecting an argument (-AlertRec, -AlertDB, -DbRecNo, -UI3Rec, or -Camera 'xxx' -Live)"
		$err_msg = "ERR/expecting an argument (-AlertRec, -DbRecNo, -UI3Rec, or -Camera)"
		return $err_msg
	}

	$deb_rec = $null  # Initialize
	$clip_db_locator = $null  # Initialize
	if ($PSBoundParameters.ContainsKey('AlertRec') -or $PSBoundParameters.ContainsKey('DbRecNo') -or $PSBoundParameters.ContainsKey('UI3Rec')) {

		if ($PSBoundParameters.ContainsKey('AlertRec') -and ($PSBoundParameters.ContainsKey('DbRecNo') -or $PSBoundParameters.ContainsKey('UI3Rec'))) {
			$err_msg = "ERR/do not simultaneously use -AlertRec and -DbRecNo or -UI3Rec"
			return $err_msg
		}
		if ($PSBoundParameters.ContainsKey('DbRecNo') -and ($PSBoundParameters.ContainsKey('AlertRec') -or $PSBoundParameters.ContainsKey('UI3Rec'))) {
			$err_msg = "ERR/do not simultaneously use -DbRecNo and -AlertRec or -UI3Rec"
			return $err_msg
		}
		if ($PSBoundParameters.ContainsKey('UI3Rec') -and ($PSBoundParameters.ContainsKey('AlertRec') -or $PSBoundParameters.ContainsKey('DbRecNo'))) {
			$err_msg = "ERR/do not simultaneously use -UI3Rec and -AlertRec or -DbRecNo"
			return $err_msg
		}
		if ($PSBoundParameters.ContainsKey('DbRecNo') -and $DbRecNo.Length -eq 0) {
			# This happens when using -DbRecNo @123456 (without quotes)
			$err_msg = "ERR/argument -DbRecNo must contain only digits"
			return $err_msg
		}	
		
		if ($PSBoundParameters.ContainsKey('AlertRec')) {
			$clip_db_locator = $AlertRec.Replace("\","")  # Why? prefix '\' may be used via Blue Iris Action set action
			if (!$clip_db_locator.StartsWith('@')) {
				$err_msg = "ERR/argument -AlertRec is expected to be prefixed with an '@' (like '@123456')"
				return $err_msg
			}
		}
			
		if ($PSBoundParameters.ContainsKey('DbRecNo')) {
			# Test if argument contains non-numeric characters
			$num_str = ($DbRecNo -replace "[^0-9]" , '').ToString()
			#Write-Host "num_str = $num_str"
			if ($num_str.Length -lt $DbRecNo.Length) {
				if ($DbRecNo.StartsWith('@')) {
					$err_msg = "ERR/argument -DbRecNo must be an integer; it should not be prefixed with '@'"
				} else {
					$err_msg = "ERR/argument -DbRecNo must contain only digits"
				}
				return $err_msg
			} else {
				$clip_db_locator = "@" + $DbRecNo.ToString()
			}
		}
		
		if ($PSBoundParameters.ContainsKey('UI3Rec')) {
			$ui3_rec = $UI3Rec.Trim()
			if ($ui3_rec.StartsWith('@')) {
				$err_msg = "ERR/argument -UI3Rec should not be prefixed with '@'"
				return $err_msg
			} else {
				$str = $ui3_rec.Replace('-', '')
				# Test if argument w/o hyphen contains any non-numeric characters
				$numbers_only_str = $str -replace "[^0-9]" , ''
				# Write-Host "numbers_only_str = $numbers_only_str"
				if ($numbers_only_str.Length -lt $str.Length) {
					$err_msg = "ERR/argument -UI3Rec must contain only digits, or digits separated by a hyphen"
					return $err_msg
				} else {
					# Use string like '103356469-2981206'
					$hypen_idx = $ui3_rec.IndexOf('-')
					if ($hypen_idx) {
						$arr = $ui3_rec.Split('-')
						$clip_db_locator = "@" + ($arr[0]).ToString()
					} else {
						$clip_db_locator = "@" + $DbRecNo.ToString()
					}
				}			
			}
		}
		# Write-Host "clip_db_locator = $clip_db_locator"
	}

	<# if ($PSBoundParameters.ContainsKey('BvrDelayMsec') -and !$PSBoundParameters.ContainsKey('IgnoreDat')) {
		# Not alwasy correct; script will try to use a BVR file if a DAT file cannot be found
		$err_msg = "ERR/argument -BvrDelayMsec should be used with argument -IgnoreDat"
		return $err_msg
	} #>

	if ($PSBoundParameters.ContainsKey('DatImageMsec') -and $PSBoundParameters.ContainsKey('IgnoreDat')) {
		$err_msg = "ERR/do not simultaneously use arguments -DatImageMsec and -IgnoreDat"
		return $err_msg
	}

	if ($PSBoundParameters.ContainsKey('DatImageMsec')) {
		if ($DatImageMsec -lt 0) {
			$err_msg = "ERR/argument '-DatImageMsec' must be greater than zero"
			return $err_msg
		}
		if ($DatImageMsec -gt 0) {
			$dat_image_interval_msec = $DatImageMsec
		}
	}

	$duration_sec = 10  # Default - for -Live & BVR-derived GIFs obly
	if ($PSBoundParameters.ContainsKey('Seconds')) {
		if ($Seconds -gt 0) {
			$duration_sec = $Seconds
		} else {
			$err_msg = "ERR/Argument -Seconds must be greater than zero"
			return $err_msg
		}
	}

	$bvr_delay_msec = 0  # Defaults
	if ($PSBoundParameters.ContainsKey('BvrDelayMsec')) {
		if ($BvrDelayMsec -lt 0) {
			$err_msg = "ERR/argument '-BvrDelayMsec' must be greater than zero"
			return $err_msg
		}
		if ($BvrDelayMsec -gt 0) {
			$bvr_delay_msec = $BvrDelayMsec
		}
	}
	# Write-Host "bvr_delay_msec = $bvr_delay_msec"
	
	if ($PSBoundParameters.ContainsKey('Height')) {
		if ($Height -lt 0) {
			$err_msg = "ERR/argument '-Height'  must be greater than zero"
			return $err_msg
		}
	}

	if ($PSBoundParameters.ContainsKey('FPS')) {
		if ($FPS -le 0) {
			$err_msg = "ERR/argument '-FPS' must be greater than zero"
			return $err_msg
		}
	}

	if ($PSBoundParameters.ContainsKey('TrimBegPct')) {
		if (($TrimBegPct -gt 99)) {
			$err_msg = "ERR/argument '-TrimBegPct' must less than 99"
			return $err_msg
		}
		$gif_trim_beg = $TrimBegPct
	}

	if ($PSBoundParameters.ContainsKey('TrimEndPct')) {
		if (($TrimEndPct -lt 0) -or ($TrimEndPct -ge 99)) {
			$err_msg = "ERR/argument '-TrimEndPct' must be between 0 and 99"
			return $err_msg
		}
		$gif_trim_end = $TrimEndPct
	}

	$trim_end_auto_flg = $false  # Default
	if ($PSBoundParameters.ContainsKey('AlertMemo')) {
		if ($gif_trim_end_cust_obj.Keys.count -gt 0) {
			$trim_end_auto_flg = $true
		}
	}
	# Write-Host "trim_end_auto_flg = $trim_end_auto_flg"

	# Change GIF's END TRIM if using AI label testing
	$clip_dur_adj = 0
	$cust_lbl_disp = ''
	$gif_trim_end_cust_pct = 0	
	
	if ($trim_end_auto_flg) {

		# NOTE: existence of argument -AlertMemo has already been verified
		$arr = $AlertMemo.Split(';')

		# For debugging
		# Write-Host ("AlertMemo array cnt = " + $arr.Count)
		# Write-Host ("AlertMemo array = " + ($arr | ConvertTo-Json))

		$al_memo_first_ai_str = ''
		foreach ($item in $arr) {
			if ($item.IndexOf('%') -gt 0) {
				$al_memo_first_ai_str = $item.Trim()
				break
			}
		}

		# For debugging
		# Write-Host "al_memo_first_ai_str = '$al_memo_first_ai_str'"

		if ($al_memo_first_ai_str.length -gt 0) {
			
			$al_memo_first_ai_str_pcs = $al_memo_first_ai_str.Split(',')
			$j = 0
			foreach ($al_memo_first_ai_str_pc in $al_memo_first_ai_str_pcs) {
				$arr = $al_memo_first_ai_str_pc.Split(':')
				$al_memo_first_ai_str_pcs[$j] = $arr[0]
				$j++
			}
			
			# We use just the first label. It is always the highest confidence.
			$first_lbl = $al_memo_first_ai_str_pcs[0]

			if ($gif_trim_end_cust_obj[$first_lbl]) {
				$cust_lbl_disp = $first_lbl  # Used for error messages
				$gif_trim_end_cust_pct = $gif_trim_end_cust_obj[$first_lbl]
			}

			# For debugging
			#Write-Host ("al_memo_first_ai_str_pcs = " + ($al_memo_first_ai_str_pcs | ConvertTo-Json))
			#Write-Host "cust_lbl_disp = $cust_lbl_disp"

			$clip_dur_adj = $TrimBegPct + $gif_trim_end_cust_pct
			
			if ($clip_dur_adj -gt 100) {
				$err_msg = "ERR/sum of clip duration adjustments cannot exceed <100% ('TrimBegPct/TrimEndPct/TrimEndAutoPct = $TrimBegPct/$TrimEndPct/$gif_trim_end_cust_pct)"
				return $err_msg
			} 
			<# if ($clip_dur_adj -lt 0) {
				$err_msg = "ERR/sum of clip duration adjustments cannot be <0% ('TrimBegPct/TrimEndPct/TrimEndAutoPct = $TrimBegPct/$TrimEndPct/$gif_trim_end_cust_pct)"
				return $err_msg
			} #> 

		} else {
			if ($FromConsole) {
				Write-Warning "expecting argument -AlertMemo to contain at least one substring like 'vehicle:93%'"
			}
		}
		
	} else {
		
		$clip_dur_adj = $TrimBegPct + $TrimEndPct
		
		if ($clip_dur_adj -gt 100) {
			$err_msg = "ERR/sum of clip duration adjustments cannot exceed <100% ('TrimBegPct/TrimEndPct = $TrimBegPct/$TrimEndPct)"
			return $err_msg
		} 
		<# if ($clip_dur_adj -lt 0) {
			$err_msg = "ERR/sum of clip duration adjustments cannot be <0% ('TrimBegPct/TrimEndPct = $TrimBegPct/$TrimEndPct)"
			return $err_msg
		} #> 
	}
	
	if ($gif_trim_end_cust_pct -gt 0) {
		$gif_trim_end = 0
	}

	# For debugging
	#Write-Host "clip_dur_adj = $clip_dur_adj"
	#Write-Host "gif_trim_beg = $gif_trim_beg"
	#Write-Host "gif_trim_end = $gif_trim_end"
	#Write-Host "gif_trim_end_cust_pct = $gif_trim_end_cust_pct"

	if ($PSBoundParameters.ContainsKey('Title')) {
		$notification_title = $Title
	} else {
		#$notification_title = "via $ScriptName - $Camera"
		$notification_title = "via $ScriptName ($ScriptVersion)"
	}

	if ($PSBoundParameters.ContainsKey('Priority')) {
		if (($Priority -ge -2) -and ($Priority -le 2)) {
			$po_priority = $Priority
		}
	}

	$gif_via_DAT_flg = $true  # Default
	if ($PSBoundParameters.ContainsKey('IgnoreDat')) {
		$gif_via_DAT_flg = $false
	}
	# Write-Host "gif_via_DAT_flg = $gif_via_DAT_flg"
	
	if ($PSBoundParameters.ContainsKey('IgnoreDat') -and $PSBoundParameters.ContainsKey('Live')) {
		$err_msg = "ERR/argument -IgnoreDat should not be used with argument -Live"
		return $err_msg
	}
	
	if ($PSBoundParameters.ContainsKey('ImagesCnt')) {
		$gif_images_cnt = $ImagesCnt
	}
	#Write-Host "gif_images_cnt = $gif_images_cnt"

	if ($PSBoundParameters.ContainsKey('Height')) {
		$gif_height = $Height
	}

	if ($PSBoundParameters.ContainsKey('FPS')) {
		$gif_fps = $FPS
	}
	
	if ($PSBoundParameters.ContainsKey('PushService')) {
		# Handled in VALIDATE USER SETTINGS code section
	}

	if ($PSBoundParameters.ContainsKey('Sound')) {
		$Sound = ""
	}

	if ($PSBoundParameters.ContainsKey('Devices')) {
		$Devices = ""
	}

	$nopush_flg = $false  # Default
	if ($PSBoundParameters.ContainsKey('NoPush')) {
		$nopush_flg = $true
	}

	$nobilog_flg = $false  # Default
	if ($PSBoundParameters.ContainsKey('NoBiLog')) {
		$nobilog_flg = $true
	}

	$noarchivegif_flg = $false  # Default
	if ($PSBoundParameters.ContainsKey('NoArchive')) {
		$noarchivegif_flg = $true
	} else {
		if ($gif_archive_lifetime_hours -eq 0) {
			$noarchivegif_flg = $true
		}
	}

	if ($Quiet) {
		$bg_msgs_on = $false
	}

	$demo_flg = $false  # Default
	if ($PSBoundParameters.ContainsKey('Demo')) {
		$demo_flg = $true
	}
	
	# Simulator Args
	#$Sim1_arg_exists = ($cmdline_args_passed -contains 'Sim1')
	#$Sim2_arg_exists = ($cmdline_args_passed -contains 'Sim2')
	#$Sim3_arg_exists = ($cmdline_args_passed -contains 'Sim3')
	#$Sim4_arg_exists = ($cmdline_args_passed -contains 'Sim4')
	#$Sim5_arg_exists = ($cmdline_args_passed -contains 'Sim5')
}

# Start overall performance timer
$timer_overall = [Diagnostics.Stopwatch]::StartNew()

# Initialize script's return object
$rtn_obj = @{}

# TRY TO MAKE GIF
if ($AlertRec -or $DbRecNo -or $UI3Rec -or $Live -or $Latest) {
	
	# Get System CPU% <== stopped getting value b/c it substantially increases the GIF make time
	<# if ($PSVersion -ge 7) {
		$cpu_pct_before = (Get-CimInstance -ClassName Win32_Processor | Select-Object -Property LoadPercentage).LoadPercentage
	} else {
		$cpu_pct_before = (Get-WmiObject win32_processor | Select-Object -Property LoadPercentage).LoadPercentage
	} #>
	
	# Handle Blue Iris installations using multiple 'New' folder paths
	if ($bi_new_folder_paths_obj.Count -gt 0) {
		$arr = $bi_new_folder_paths_obj.Keys
		if ($FromConsole -and !$Quiet) {
			Write-Host "Assuming Blue Iris is using multiple 'New' folders:"
			Write-Host ($arr | ConvertTo-Json)
		}
		# Fix REQUIRED param $reg_bi_new_folder_path if unset by user
		if ($reg_bi_new_folder_path.StartsWith('YOUR')) {
			$reg_bi_new_folder_path = $arr[0]
		}
	}

	# For debugging
	# Write-Host "AlertRec = $AlertRec"
	# Write-Host "RootPath = $app_path"
	# Write-Host "New = $app_path"
	# Write-Host "UseDatFile = $gif_via_DAT_flg"
	# msg.exe $server_username /TIME:10 "UseDatFile = $gif_via_DAT_flg"		
	# Write-Host "DatReqdImages = $gif_DAT_reqd_images_cnt"
	# Write-Host "ImagesCnt = $gif_images_cnt"
	# Write-Host "TrimEndAuto = $gif_trim_end_cust_obj"
	# Write-Host "Height = $gif_height"
	# Write-Host "FPS = $gif_fps"

	# Start performance timer - MakeGIF()
	$timer_makegif = [Diagnostics.Stopwatch]::StartNew()

	<# $makegif_obj is usually a hashtable like... 
	 {
		"method": "via Clip",
		"file": "D:\\pushover_gifs\\GT\\output.gif",
		"delay": 52612,
		"cnt": 10,
		"datjpgs": 10,
		"height": 320,
		"fps": 2,
		"kb": 226.2,
		"msec": 2625,
		"trimbeg": 0,
		"trimend": 0,
		"trimcust": 30,
		"almsec": 52612,
		"alfresh": True,
		"res": "1920x1080",
		"cam": "GT",
		"memo": "person:86%",
		"source": "GT.20230609_160000.bvr"
		"archive": "D:\pushover_gifs\_archive\GT.20230609_160000.971414.17-0.gif"
	   "dbrec": "@123456789"
		"timing_ht": @{}
	 }
	 or "ERR/message"
	#>
	$makegif_obj = `  # returns a hashtable
		MakeGIF `
		-RootPath $app_path `
		-BvrPaths $bi_new_folder_paths_ary `
		-DatPath $reg_bi_alerts_folder_path `
		-TempPath $reg_bi_temp_folder_path `
		-AlertRec $clip_db_locator `
		-LiveFlg $Live `
		-Camera $Camera `
		-Seconds $duration_sec `
		-UseDatFileFlg $gif_via_DAT_flg `
		-LatestFlg $Latest `
		-DatReqdImages $gif_DAT_reqd_images_cnt `
		-IgnoreDatAgeFlg $IgnoreDatAge `
		-DatIsFreshMult $dat_is_fresh_mult `
		-DatImageIntervalMsec $dat_image_interval_msec `
		-BvrDelayMsec $bvr_delay_msec `
		-NoPushFlg $nopush_flg `
		-NoArchiveGifFlg $noarchivegif_flg `
		-GifArchiveHours $gif_archive_lifetime_hours `
		-ImagesCnt $gif_images_cnt `
		-ImagesCntUsedFlg $PSBoundParameters.ContainsKey('ImagesCnt') `
		-Height $gif_height `
		-FPS $gif_fps `
		-TrimBegPct $gif_trim_beg `
		-TrimEndPct $gif_trim_end `
		-TrimEndAutoPct $gif_trim_end_cust_pct `
		-ClipDurAdjPct $clip_dur_adj
	
	# Stop performance timer
	$timer_makegif.Stop()
	$elapsed_msec = [Math]::Round($timer_makegif.elapsed.totalmilliseconds,1)
	$makegif_ms = $elapsed_msec.ToString("F1")
	$makegif_ms2 = $elapsed_msec.ToString("F0")
	# Write-Host ("$fn_name() json_alertlist_time_ms = $json_alertlist_time_ms msec")
	
	# For debugging
	# Write-Host ("MakeGIF() response type = " + $makegif_obj.GetType())
	# Write-Host ("MakeGIF() response cnt = " + $makegif_obj.Count)
	# Write-Host ("MakeGIF() response obj keys = " + $makegif_obj.Keys)
	# Write-Host ("MakeGIF() response = " + ($makegif_obj | ConvertTo-Json))

	# Trap various error conditions
	if ($null -eq $makegif_obj) {
		$err_msg = "ERR/unexpected MakeGIF() response; returned object is null"
		if ($FromConsole) {
			Write-Host ""
			Write-Host $err_msg -ForegroundColor Red
			Write-Host ""
		}
		return $err_msg
	}
	if ($makegif_obj -is [String]) {
		$err_msg = "ERR/unexpected response from MakeGIF(): '" + $makegif_obj + "'"
		if ($FromConsole) {
			Write-Host ""
			Write-Host $err_msg -ForegroundColor Red
			Write-Host ""
			return
		} else {
			return $err_msg
		}
	}
	if (!($makegif_obj.Keys -contains 'file')) {
		$err_msg = "ERR/unexpected MakeGIF() response; returned object is missing key: 'file'"
		if ($FromConsole) {
			Write-Host ""
			Write-Host $err_msg -ForegroundColor Red
			Write-Host ($makegif_obj | ConvertTo-Json).ToString() -ForegroundColor Red
			Write-Host ""
			return
		} else {
			return $err_msg
		}
	} else {
		if ($makegif_obj.file.Length -eq 0) {
			$err_msg = "ERR/unexpected MakeGIF() response; key 'file' is blank"
			if ($FromConsole) {
				Write-Host ""
				Write-Host $err_msg -ForegroundColor Red
				Write-Host ($makegif_obj | ConvertTo-Json).ToString() -ForegroundColor Red
				Write-Host ""
				return
			} else {
				return $err_msg
			}
		}
	}
	
	# Parse output for items to be displayed now, and logged later
	$gif_method        = $makegif_obj.method
	$gif_filespec      = $makegif_obj.file
	$delay_ms          = $makegif_obj.delay     # type is [Int]
	$gif_imgs_cnt      = $makegif_obj.cnt       # type is [Int]
	$dat_jpgs_cnt      = $makegif_obj.datjpgs   # type is [Int]
	$gif_height        = $makegif_obj.height    # type is [Int]
	$gif_fps           = $makegif_obj.fps       # type is [Int]
	$gif_ms            = $makegif_obj.msec      # type is [Int]
	$gif_trim_beg      = $makegif_obj.trimbeg   # type is [Int]
	$gif_trim_end      = $makegif_obj.trimend   # type is [Int]
	$gif_trim_end_cust = $makegif_obj.trimcust  # type is [Int]
	$gif_kb            = $makegif_obj.kb        # type is [double]
	$gif_jpg_wxh       = $makegif_obj.res
	$al_cam            = $makegif_obj.cam
	$al_msec           = $makegif_obj.almsec    # type is [Int]
	$al_fresh          = $makegif_obj.alfresh   # type is boolean or null
	$al_memo           = $makegif_obj.memo
	$gif_source_fs     = $makegif_obj.source
	$archive_fs        = $makegif_obj.archive
	$db_locator        = $makegif_obj.dbrec
	$timing_ht         = $makegif_obj.timing_ht
	
	# For debugging
	# Write-Host "gif_method = '$gif_method'"
	# msg.exe $server_username /TIME:10 "gif_method = $gif_method"
	# Write-Host "gif_filespec = '$gif_filespec'"
	# Write-Host "delay_ms = $delay_ms"
	# Write-Host "gif_imgs_cnt = $gif_imgs_cnt"
	# Write-Host "dat_jpgs_cnt = $dat_jpgs_cnt"
	# Write-Host "gif_height = $gif_height"
	# Write-Host "gif_fps = $gif_fps"
	# Write-Host "gif_ms = $gif_ms"
	# Write-Host "gif_trim_beg = $gif_trim_beg"
	# Write-Host "gif_trim_end = $gif_trim_end"
	# Write-Host "gif_trim_end_cust = $gif_trim_end_cust"
	# Write-Host "gif_kb = $gif_kb"
	# Write-Host "gif_jpg_wxh = $gif_jpg_wxh"
	# Write-Host "al_cam = $al_cam"
	# Write-Host "al_msec = $al_msec"
	# Write-Host "al_fresh = $al_fresh"
	# Write-Host "al_memo = $al_memo"
	# Write-Host "gif_source_fs = '$gif_source_fs'"
	# Write-Host "gif archive_fs = '$archive_fs'"
	# Write-Host "db_locator = '$db_locator'"
	# Write-Host ("timing_ht = " + ($timing_ht | ConvertTo-Json))

	$gif_kb2 = [Math]::Round($gif_kb,0).ToString("F0")
	$gif_sec = [Math]::Round($gif_ms/1000,1).ToString("F1")
	$gif_source_filename = (Split-Path $gif_source_fs -leaf).TrimStart('\')

	# Prep timings to be shown in notifications
	$json_alertlist_time_ms   = $timing_ht['json_alertlist_time_ms']
	#$get_cam_settings_time_ms = $timing_ht['get_cam_settings_time_ms']
	#$bi_version_chk_time_ms   = $timing_ht['bi_version_chk_time_ms']
	$json_clipstats_time_ms   = $timing_ht['json_clipstats_time_ms']
	$dat_img_ext_ms           = $timing_ht['dat_img_extract_time_ms']
	$bvr_ext_delay_ms         = $timing_ht['bvr_extract_delay_time_ms']
	$bvr_img_ext_ms           = $timing_ht['bvr_img_extract_time_ms']
	$get_bvr_time_ms          = $timing_ht['get_bvr_time_ms']
	$live_img_extract_ms      = $timing_ht['live_clip_img_extract_ms']
	$ffmpeg_ms                = $timing_ht['ffmpeg_time_ms']	
	#$makegif_prep_ms          = $timing_ht['makegif_prep_time_ms']
	$gif_archive_ms           = $timing_ht['gif_archive_time_ms']  # not included in $makegif_ms
	
	# Currently removed from 'other' calculation (usu too small to show)
	$get_cam_settings_time_ms = 0
	$bi_version_chk_time_ms1 = 0
	$bi_version_chk_time_ms2 = 0
	$makegif_prep_ms = 0

	if ($bvr_ext_delay_ms -and [Decimal]$bvr_ext_delay_ms -gt 10.0) {
		$other_ms = [Decimal]$makegif_ms - ( [Decimal]$json_alertlist_time_ms + [Decimal]$get_cam_settings_time_ms + [Decimal]$bi_version_chk_time_ms1 + [Decimal]$bi_version_chk_time_ms2 + [Decimal]$json_clipstats_time_ms + [Decimal]$dat_img_ext_ms + [Decimal]$bvr_ext_delay_ms + [Decimal]$bvr_img_ext_ms + [Decimal]$get_bvr_time_ms + [Decimal]$live_img_extract_ms + [Decimal]$ffmpeg_ms + [Decimal]$makegif_prep_ms )
	} else {
		$other_ms = [Decimal]$makegif_ms - ( [Decimal]$json_alertlist_time_ms + [Decimal]$get_cam_settings_time_ms + [Decimal]$bi_version_chk_time_ms1 + [Decimal]$bi_version_chk_time_ms2 + [Decimal]$json_clipstats_time_ms + [Decimal]$dat_img_ext_ms + [Decimal]$bvr_img_ext_ms + [Decimal]$get_bvr_time_ms + [Decimal]$live_img_extract_ms + [Decimal]$ffmpeg_ms + [Decimal]$makegif_prep_ms )
	}
		
	# For debugging
	# Write-Host "gif_source_filename = '$gif_source_filename'"
	# Write-Host "dat_img_ext_ms = $dat_img_ext_ms"
	# Write-Host "ffmpeg_ms = $ffmpeg_ms"

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nGIF File Summary:"
		Write-Host "-----------------"
		Write-Host "GIF created:  '$gif_filespec'"
		Write-Host "GIF method:   $gif_method"
		Write-Host "GIF specs:    ImageCnt, Height, Dur(msec), FPS, Size(KB) = $gif_imgs_cnt, $gif_height, $gif_ms, $gif_fps, $gif_kb"
		# Write-Host "GIF archived: '$archive_fs'"
		# Write-Host "GIF source:   '$gif_source_fs'"
		
	}
	if ($gif_kb -gt 5000) {
		if ($FromConsole) {
			Write-Warning "GIF file may exceed limit (5000 KB) for Pushover API"
		}
	}
	if (!$noarchivegif_flg) {
		if ($archive_fs -eq 'fail') {
			if ($FromConsole) {
				Write-Warning "unable to archive the GIF file (reason unknown)"
			}
		} elseif ($archive_fs.StartsWith("ERR/")) {
			if ($FromConsole) {
				Write-Warning "unable to archive the GIF file ($archive_fs)"
			}
		} else {
			if ($FromConsole -and !$Quiet) {
				Write-Host "GIF archived: '$archive_fs'"
			}
		}
	}

	if ($FromConsole -and !$Quiet) {
		Write-Host "GIF source:   '$gif_source_fs'"
	}

	# PERFORM MAINTENANCE ON GIF ARCHIVE FOLDER
	# KEY POINT: This timing value is NOT a component of GifMakeMsec
	if ($FromConsole -and !$Quiet) {
		Write-Host "`nGIF Archive Maintenance Summary:"
		Write-Host   "--------------------------------"
	}
	$gif_archive_prune_ms = "0"
	if (!$NoArchive) {

		if ($gif_archive_lifetime_hours -gt 0) {
			
			$gif_prune_filespec = $gif_archive_path  + "\*.gif"

			# Start performance timer - GIF archive maintenance
			$timer_gif_archive_prune = [Diagnostics.Stopwatch]::StartNew()

			# Returned array is like, @(found_files_cnt, pruned_files_cnt, oldest_file_hours, oldest_file_path)
			$arr = `
				PruneOldFiles `
				-Pattern $gif_prune_filespec `
				-Hours $gif_archive_lifetime_hours

			# Stop timer
			$timer_gif_archive_prune.Stop()
			$elapsed_msec = [Math]::Round($timer_gif_archive_prune.elapsed.totalmilliseconds,1)
			$gif_archive_prune_ms = $elapsed_msec.ToString("F0")
			
			# KEY POINT: This timing value is NOT a component of GifMakeMsec
			# $timing_ht.Add('gif_archive_prune_time_ms', $gif_archive_prune_ms)
			# $timing_ht.Add('makgif_time_adj_ms', "-" + $gif_archive_prune_ms)

			# For debugging
			# Write-Host "gif_prune_filespec = '$gif_prune_filespec'"
			# Write-Host ("GIF prune arr = " + ($arr | ConvertTo-Json))
			# Write-Host "gif_archive_prune_time_ms = $gif_archive_prune_ms msec"
			
			if ($FromConsole -and !$Quiet) {
				Write-Host ("Count of archived GIF files matching '$gif_prune_filespec' = " + $arr[0])
				if ($arr[1] -gt 0) {			
					Write-Host ("Count of archived GIF files pruned (age >= $gif_archive_lifetime_hours hrs) = " + $arr[1])
					Write-Host ("Count of remaining archived GIF files = " + ($arr[0] - $arr[1]))
				} else {
					Write-Host "No old GIF files found to prune (age >= $gif_archive_lifetime_hours hrs)"
					Write-Host ("Oldest GIF file age is " + $arr[2] + " hrs (" + $arr[3] + ")")
				}
				if ([Int]$gif_archive_prune_ms -gt 300) {
					$descriptor = "is relatively large"
					if ([Int]$gif_archive_prune_ms -ge 1000) {
						$descriptor = "is excessive"
					}								
					Write-Warning "pruning operation time $descriptor ($gif_archive_prune_ms msec);`nTo speed up script execution, consider reducing setting `$gif_archive_lifetime_hours (currently $gif_archive_lifetime_hours hrs)"
				}
			}
		} else {
			if ($FromConsole -and !$Quiet) {
				Write-Host "Monitoring of the archive folder has been DISABLED (via user-setting `$gif_archive_lifetime_hours)"
			}					
		}
		
	} else {
		if ($FromConsole -and !$Quiet) {
			Write-Host "Monitoring of the archive folder has been DISABLED via argument -NoArchive"
		}					
	}
	
	# For debugging
	# Write-Host ""
	# Write-Host "makegif() execution time elements:"
	# Write-Host ("timing_ht = " + ($timing_ht | ConvertTo-Json))
	# s$sortedTable = $timing_ht.GetEnumerator() | Sort-Object {[int]$_.Value} -Descending
	# $sortedTable | Format-Table -Property Key, Value

	# Get System CPU% <== stopped getting value b/c it substantially increases the GIF make time
	<# if ($PSVersion -ge 7) {
		$cpu_pct_after = (Get-CimInstance -ClassName Win32_Processor | Select-Object -Property LoadPercentage).LoadPercentage
	} else {
		$cpu_pct_after = (Get-WmiObject win32_processor | Select-Object -Property LoadPercentage).LoadPercentage
	} #>
	#$cpu_pct_avg = [Math]::Round(($cpu_pct_after + $cpu_pct_before)/2,0).ToString("F0")
	
	# Fill return object
	#$rtn_obj = @{}
	$rtn_data_obj = [ordered]@{}
	$rtn_obj['result'] = 'success'
	$rtn_data_obj['gif_method'] = $gif_method
	$rtn_data_obj['gif_make_ms'] = $makegif_ms
	$rtn_data_obj['gif_images'] = $gif_imgs_cnt
	$rtn_data_obj['dat_images'] = $dat_jpgs_cnt
	$rtn_data_obj['gif_height'] = $gif_height
	$rtn_data_obj['gif_fps'] = $gif_fps
	$rtn_data_obj['gif_ms'] = $gif_ms
	$rtn_data_obj['gif_trim_pct_beg'] = $gif_trim_beg
	$rtn_data_obj['gif_trim_pct_end'] = $gif_trim_end
	$rtn_data_obj['gif_trim_pct_cust'] = $gif_trim_end_cust
	$rtn_data_obj['gif_kb'] = $gif_kb
	$rtn_data_obj['gif_jpg_wxh'] = $gif_jpg_wxh
	$rtn_data_obj['gif_fs'] = $archive_fs
	#$rtn_data_obj['gif_source_fs'] = $reg_bi_new_folder_path + "\" + $gif_source_fs.Name
	$rtn_data_obj['gif_source_fs'] = $gif_source_fs
	#$rtn_data_obj['gif_cpu_pct'] = $cpu_pct_before.ToString() + "/" + $cpu_pct_after.ToString()
	$rtn_obj['data'] = $rtn_data_obj

}

# Stop overall performance timer
$timer_overall.Stop()
$total_ms = [Math]::Round($timer_overall.elapsed.totalmilliseconds,1) + $BvrDelayMsec
$total_ms_disp = $total_ms.ToString("F1")
# Write-Host "Overall performance time = $total_ms_disp msec"

# SEND NOTIFICATION
if (!$demo_flg -and !$nopush_flg) {

	if ($push_service_ary -contains 'pushover' -or $push_service_ary -contains 'telegram') {

		# BUILD ATTACHMENT PATH FOR ALL NOTIFICATIONS
		if ($gif_is_recent_flg) {
			$gif_image_path = $recent_gif_path
		} else {
			$gif_image_path = $app_path + "\" + $al_cam + "\output.gif"
		}
		
		# BUILD COMMON MESSAGE STRING FOR ALL NOTIFICATIONS
		if ($PSBoundParameters.ContainsKey('Msg')) {
			$notification_msg = $Msg		
		} else {
			# Default notification message
			$notification_msg = "Camera: $Camera"
			$notification_msg += "`nGIF: imgs=$gif_imgs_cnt, H=$gif_height, sec=$gif_sec, FPS=$gif_fps, KB=$gif_kb2"
			$notification_msg +="`nTrimBeg/End/EndCust: $gif_trim_beg/$gif_trim_end/$gif_trim_end_cust"
			$notification_msg +="`nSourceJpgSize: $gif_jpg_wxh"
			#$notification_msg +="`nAlertIsFresh: $al_fresh"
			if ($al_memo) {
				# If argument -AlertMemo was used, check if it different from the alertlist memo
				# If so, wrap the -AlertMemo value in slashes as a 'heads-up' in the notification message
				# (an example of a situation where this can occur is in testing)
				if ($AlertMemo -and $al_memo) {
					if ($al_memo -ne $AlertMemo) {
						$al_memo = "/" + $AlertMemo + "/"
					}
				}
				$notification_msg +="`nAlertMemo: " + $al_memo
			} else {
				$notification_msg +="`nAlertMemo: (none)"
			}
		}

		# ... Build optional 'DAT tag' for message string
		if ($append_gif_method_tag_flg) {
			if ($gif_method -eq 'viaDAT') {
				$method_str = "viaDAT:" + $dat_jpgs_cnt
			} else {
				$method_str = "$gif_method"
			}
			$notification_msg += "`nMethod: ($method_str)"
		}
		
		# ... Append optional script performance info to notification message string
		if ($append_performance_tag_flg) {
			$notification_msg += "`nAlertIsFresh: $al_fresh"
			#$notification_msg += "`nCpuPctAvg: " + $cpu_pct_avg + " (" + $cpu_pct_before + "," + $cpu_pct_after + ")"
			$notification_msg += "`nGifMakeMsec: " + $makegif_ms2
			if ($append_performance_tag_details_flg) {

				# KEY POINT: These timing values ARE components of GifMakeMsec
				#$notification_msg += "`n-PrepMsec: " + $makegif_prep_ms
				if ($Latest -and $json_alertlist_time_ms) {
					$notification_msg += "`n-JsonAlertlistMsec: " + $json_alertlist_time_ms
				}
				<# if ($get_cam_settings_time_ms) {
					$notification_msg += "`n-GetCamSettingsMsec: " + $get_cam_settings_time_ms
				} #>
				if ($bi_version_chk_time_ms1) {
					$notification_msg += "`n-BiVerCheckMsec: " + $bi_version_chk_time_ms1
				}
				if ($bi_version_chk_time_ms2) {
					$notification_msg += "`n-BiVerCheckMsec: " + $bi_version_chk_time_ms2
				}
				if ($json_clipstats_time_ms) {
					$notification_msg += "`n-JsonClipstatsMsec: " + $json_clipstats_time_ms
				}
				if ($get_bvr_time_ms) {
					$notification_msg += "`n-GetBvrFileMsec: " + $get_bvr_time_ms
				}
				if ($live_img_extract_ms) {
					$notification_msg += "`n-LiveImgExtractMsec: " + $live_img_extract_ms
				}
				if ($bvr_ext_delay_ms -and [Decimal]$bvr_ext_delay_ms -gt 10.0) {
					$notification_msg += "`n-BvrExtractDelayMsec: " + $bvr_ext_delay_ms
				}
				if ($bvr_img_ext_ms) {
					$notification_msg += "`n-BvrImgExtractMsec: " + $bvr_img_ext_ms
				}
				if ($dat_img_ext_ms) {
					$notification_msg += "`n-DatImgExtractMsec: " + $dat_img_ext_ms
				}
				$notification_msg += "`n-FFmpegMsec: " + $ffmpeg_ms
				if ([Int]$gif_archive_time_ms -gt 200) {
					$notification_msg += "`n-GifArchiveMsec: " + $gif_archive_time_ms
				}
				if ($other_ms) {
					$notification_msg += "`n-OtherMsec: " + $other_ms.ToString("F0")
				}
				
				# KEY POINT: This timing value is NOT a component of GifMakeMsec
				# (Note the lack of a '-' sub-bullet)
				if ([Int]$gif_archive_prune_ms -gt 200) {
					$notification_msg += "`nGifArchivePruneMsec: " + $gif_archive_prune_ms
				}
			}
		}
	}
	
	if ($push_service_ary -contains 'pushover') {

		# Start notification service performance timer
		$timer_notification_pushover = [Diagnostics.Stopwatch]::StartNew()
		
		if ($gif_kb -gt 0) {
		
			if ($PSBoundParameters.ContainsKey('Sound')) {
				$po_sound = $Sound
			}

			# Returns string - either API response, or "ERR/*"|"WARNING:*" message
			# ... returns 'ERR/*' or 
			#     API response is like '@{status=1; request=28f81c23-dc07-4a36-8523-f6428159284c}'
			$response_str = `
				SendPushover `
				-MsgStr $notification_msg `
				-MsgTitle $notification_title `
				-PriorityLevel $po_priority `
				-Sound $Sound `
				-Devices $Devices `
				-AttachPath $gif_image_path `
				-Camera $al_cam

			if ($FromConsole -and !$Quiet) {
				Write-Host "`nPushover Response:"
				Write-Host "------------------"
			}

			# For testing - Simulate a bad pushover response
			#$response_str = "ERR/test"
			#$response_str = "WARNING:test"
			
			# for debugging
			#Write-Host "SendPushover() response_str = $response_str"
			
			# Trap error conditions
			if ($response_str.StartsWith("ERR/") -or $response_str.StartsWith("WARNING")) {

				# POST MESSAGE TO BI LOGFILE
				if ($post_to_BI_logfile_flg) {
					$null = `
						PostCustomBiLogEntry `
						-MsgStr "$ScriptName - unexpected response trying to send Pushover notification: $response_str" `
						-Level 2 `
						-SilentFlg $Quiet
				}

				if ($FromConsole) {
					Write-Host ("ERR/unexpected response from SendPushover(): '" + $response_str + "'") -ForegroundColor Red
					Write-Warning "Notification not sent"
				}		
				
			} else {
					
				if ($FromConsole -and !$Quiet) {
					Write-Host $response_str
				}
				
				# For debugging
				# PopupBackgroundMsg -Seconds 30-Message "$response_str ($script:ScriptName)" -UserName $server_username -InBackground $bg_msgs_on 
			}
		}

		# Stop notification service performance timer
		$timer_notification_pushover.Stop()
	}
	
	if ($push_service_ary -contains 'telegram') {

		# Start notification service performance timer
		$timer_notification_telegram = [Diagnostics.Stopwatch]::StartNew()
		
		if ($gif_kb -gt 0) {
			
			$notification_msg = "<b>$notification_title</b>`n" + $notification_msg

			# Returns string - either API response, or "ERR/*"|"WARNING:*" message
			# API response is like '@{ok=True; result=}'
			$response_str = `
				SendTelegram `
				-MsgStr $notification_msg `
				-AttachPath $gif_image_path `
				-Camera $al_cam

			if ($FromConsole -and !$Quiet) {
				Write-Host "`nTelegram Response:"
				Write-Host "------------------"
			}

			# For testing - Simulate a bad API response
			#$response_str = "ERR/test"
			#$response_str = "WARNING:test"
			
			# For debugging
			# Write-Host ("SendTelegram response_str type = " + $response_str.GetType())
			# Write-Host "SendTelegram response_str = '$response_str'"
			
			# Trap error conditions
			if ($response_str.StartsWith("ERR/") -or $response_str.StartsWith("WARNING")) {

				# POST MESSAGE TO BI LOGFILE
				if ($post_to_BI_logfile_flg) {
					$null = `
						PostCustomBiLogEntry `
						-MsgStr "$ScriptName - unexpected response trying to send Telegram notification: $response_str" `
						-Level 2 `
						-SilentFlg $Quiet
				}

				if ($FromConsole) {
					Write-Host ("ERR/unexpected response from SendTelegram(): '" + $response_str + "'") -ForegroundColor Red
					Write-Warning "Notification not sent"
				}		
				
			} else {
					
				if ($FromConsole -and !$Quiet) {
					Write-Host $response_str
				}
				
				# For debugging
				# PopupBackgroundMsg -Seconds 30 -UserName $server_username -InBackground $bg_msgs_on -Message "$response_str `n($ScriptName)"
			}
		}

		# Stop notification service performance timer
		$timer_notification_telegram.Stop()
	}

} else {

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nNotification Summary:"
		Write-Host "---------------------"
		Write-Host "DISABLED by argument -NoPush or -Demo"
	}
	
}

$pusho_ms = "0.0"
$telegram_ms = "0.0"
if ($push_service_ary -contains 'pushover') {
	$pusho_ms = [Math]::Round($timer_notification_pushover.elapsed.totalmilliseconds,1).ToString("F1")
}
if ($push_service_ary -contains 'telegram') {
	$telegram_ms = [Math]::Round($timer_notification_telegram.elapsed.totalmilliseconds,1).ToString("F1")
}

# LOG SCRIPT PERFORMANCE INFO TO CSV FILE
# ... Prep values for logging
if (1 -eq 1) {  # Use to collapse this code section while editing

	$csvfilespec = $app_path + "\" + $ScriptBaseName + "_" + $ScriptVersion + "_log.csv"
	$timestamp   = Get-Date -Format "yyyy/MM/dd_HH:mm:ss"
	$total_ms    = [Decimal]$total_ms 
	$pusho_ms    = [Decimal]$pusho_ms 
	$telegram_ms = [Decimal]$telegram_ms 
	#$gif_ms      = [Decimal]$gif_ms 
	#$gif_sec     = [Decimal]$gif_sec

	if ($gif_kb -gt 0) {
		$images_cnt    = [Int]$gif_imgs_cnt
		$gif_height    = [Int]$gif_height
		$gif_fps       = [Int]$gif_fps
		$gif_trim_beg  = [Int]$gif_trim_beg
		$gif_trim_end  = [Int]$gif_trim_end
		$gif_trim_cust = [Int]$gif_trim_end_cust
		$gif_sec       = [Decimal]$gif_sec
		$gif_size      = [Int]$gif_kb
		$makegif_ms    = [Decimal]$makegif_ms
	} else {
		$images_cnt    = 0
		$gif_height    = 0
		$gif_fps       = 0
		$gif_trim_beg  = 0
		$gif_trim_end  = 0
		$gif_trim_cust = 0
		$gif_sec       = 0
		$gif_size      = 0
		$makegif_ms    = [Decimal]$makegif_ms 
	}

	$disp_total_ms    = [Math]::Round($total_ms,0)
	$disp_delay_ms    = [Math]::Round($delay_ms,0)
	$disp_makegif_ms  = [Math]::Round($makegif_ms,0)
	$disp_pusho_ms    = [Math]::Round($pusho_ms,0)
	$disp_telegram_ms = [Math]::Round($telegram_ms,0)

	# SHOW PERFORMANCE NOTICE
	# NOTE (2024-02-15): The 'JPG fetch delay' has been rendered irrelevant as a result of
	# changing the source of JPGs for -Live GIFs from the active BVR file to the live stream.
	# 'JPG fetch delay' is now ALWAYS ZERO. Despite this, we are continuing to report it in 
	# Blue Iris logfile entries. WHY? to not break other utilities that extract/use this script's
	# performance data from the BI logfile).
	if ($FromConsole -and !$Quiet) {
		$perf_msg = `
			"Performance Summary:" `
			+ "`n--------------------" `
			+ "`nJPG fetch delay:         $disp_delay_ms msec" `
			+ "`nmakegif execution time:  $disp_makegif_ms msec" 
		# $perf_msg += "`ntotal execution time:    $disp_total_ms msec"
		if ($push_service_ary -contains 'telegram') {
			$perf_msg += "`ntelegram execution time: $disp_telegram_ms msec"
		}
		if ($push_service_ary -contains 'pushover') {
			$perf_msg += "`npushover execution time: $disp_pusho_ms msec"
		}
		Write-Host "`n$perf_msg"

		# For debugging
		# Write-Host ""
		# Write-Host "makegif() execution time elements:"
		# Write-Host ("timing_ht = " + ($timing_ht | ConvertTo-Json))
		# $sortedTable = $timing_ht.GetEnumerator() | Sort-Object {[int]$_.Value} -Descending
		# $sortedTable | Format-Table -Property Key, Value

		# Output performance details
		if ($DevMode) {

			$sortedTable = $timing_ht.GetEnumerator() | Sort-Object {[int]$_.Value} -Descending			
			$sum_obj = ($timing_ht.Values | ForEach-Object {[int]$_}) | Measure-Object -Sum
			$sum_delta = ($makegif_ms - $sum_obj.Sum).ToString("F0")
			$delta_pct = ($sum_delta/$makegif_ms).ToString("P2")
			Write-Host ("-" * 55) -ForegroundColor Cyan
			Write-Host ("(dev mode) MAKEGIF EXECUTION TIME ANALYSIS:") -ForegroundColor Cyan
			$sortedTable | Format-Table -Property Key, Value
			Write-Host ("(dev mode) MakeGIF execution time    : " + $disp_makegif_ms + " msec") -ForegroundColor Cyan
			Write-Host ("(dev mode) sum of elements above     : " + $sum_obj.Sum + " msec") -ForegroundColor Cyan
			Write-Host ("(dev mode) unaccounted for ('other') : " + $sum_delta + " msec (" + $delta_pct + ")") -ForegroundColor Cyan
			Write-Host ("-" * 55) -ForegroundColor Cyan
		}

	}

	# ... Log the values
	#     RETURNS:  (array) @(images_cnt, gif_height, gif_fps, gif_ms, gif_size, gif_filespec)

	$csv_response = `
		LogScriptResultsToCsv `
		-CsvFilespec $csvfilespec `
		-Timestamp $timestamp `
		-GifSourceFilespec $gif_source_filename `
		-GifMethod $gif_method `
		-GifImagesCnt $images_cnt `
		-GifHeight $gif_height `
		-GifFps $gif_fps `
		-GifTrimBeg $gif_trim_beg `
		-GifTrimEnd $gif_trim_end `
		-GifTrimEndAuto $gif_trim_end_cust `
		-GifDurSec $gif_sec `
		-GifSize $gif_size `
		-GifJpgWxH $gif_jpg_wxh `
		-MakeGifMsec $makegif_ms `
		-PushoMsec $pusho_ms `
		-TelegramMsec $telegram_ms `
		-TotalMsec $total_ms `
		-AlertRecord $db_locator `
		-ScriptHostStr $script_host_str `
		-PsVersionFullStr $PSVersionFull

	# For debugging
	# Write-Host ("LogScriptResultsToCsv response = " + ($csv_response | ConvertTo-Json))
	
	# For testing
	# $csv_response = "ERR/simulated error"
	# $csv_response = $null

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nCSV Log Summary:"
		Write-Host "----------------"
	}

	if ($csv_response) {
		if ($csv_response.StartsWith("ERR/")) {
			if ($FromConsole) {
				Write-Warning "unable to update script's CSV log (Reason: $csv_response)"
			} else {
				$bg_msg = "Unable to update script's CSV log (Reason: $csv_response)`n`n$csvfilespec"
				PopupBackgroundMsg -Seconds 10 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
			}
		} else {
			if ($FromConsole -and !$Quiet) {
				Write-Host "Performance data logged to '$csvfilespec'"
			}
		}
	} else {
		if ($FromConsole) {
			Write-Warning "unable to update script's CSV log (Reason unknown)"
		} else {
			$bg_msg = "Unable to update script's CSV log (Reason unknown)`n`n$csvfilespec"
			PopupBackgroundMsg -Seconds 10 -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
		}
	}

}

# Build strings for optional output and BI logfile
$push_service_str = $push_service_ary -Join '+'
$push_service_str = $push_service_str.Replace('pushover','pusho').Replace('telegram','pusht')
$x = 0
if ($push_service_ary -eq 'pushover') {
	$x += [Int]$disp_pusho_ms
}
if ($push_service_ary -eq 'telegram') {
	$x += [Int]$disp_telegram_ms
}
$disp_notify_ms = $x

# SHOW OPTIONAL MESSAGE WHEN RUNNING FROM BACKGROUND
#if ($bg_msgs_on) {
if ($bg_msgs_on -and !$FromConsole) {
	if ($alert_rec) {
		$alert_rec_str = $alert_rec
	} elseif ($AlertRec) {
		$alert_rec_str = $AlertRec
	} elseif ($DbRecNo) {
		$alert_rec_str = $DbRecNo
	} elseif ($UI3Rec) {
		$alert_rec_str = $UI3Rec
	} elseif ($Camera -and $Latest) {
		$alert_rec_str = 'LATEST ALERT'
	} else {
		$alert_rec_str = '<unspecified>'
	}
	$msg_sec = 120
	$bg_msg = "Alert record: $alert_rec_str" + 
		"`n`nSource file: $gif_source_filename" +
		"`n`nGIF Height/FPS/Dur(s)/Size(KB): $gif_height / $gif_fps / $gif_sec / $gif_kb" + 
		"`n`nTimes (delay,makegif,$push_service_str,total) ... $disp_delay_ms, $disp_makegif_ms, $disp_notify_ms, $disp_total_ms msec" + 
		"`n`n(this msg auto-closes in $msg_sec seconds)"
		
	# KEY POINT: message length limit is 255 characters

	# For debugging.
	# Write-Host ("bg_msg len = " + $bg_msg.length)
	# Write-Host ("bg_msg:`n" + $bg_msg)
	
	PopupBackgroundMsg -Seconds $msg_sec -Message $bg_msg -UserName $server_username -InBackground $bg_msgs_on
}

# POST MESSAGE TO BLUE IRIS LOGFILE
# KEY POINT: the message string maximum length is 127 characters (BI HTTP interface command '/admin?log=message&level=x')
if (1 -eq 1) {  # Use to collapse this code section while editing

	if ($FromConsole -and !$Quiet) {
		Write-Host "`nBlue Iris Logfile Summary:"
		Write-Host "--------------------------"
	}
	#$bilogfile_msg = "No custom BI logfile message posted"
	if (!$demo_flg -and !$nobilog_flg) {
		if ($post_all_events) {
			if ($rtn_obj -is [String] -and $rtn_obj.StartsWith("ERR/")) {
				$log_level = 2
				$log_msg = "$scriptName - $rtn_obj"
			} elseif ($rtn_obj -is [String] -and $rtn_obj.StartsWith("WARNING:")) {
				$log_level = 1
				$log_msg = "$scriptName - $rtn_obj"
			} else {
				$log_level = 0
				# $log_msg should look like, 
				#   gif_tool [320 H, 2 FPS, 7.5 s, 155.1 KB] [0, 3014, 914, 3357 ms] viaDat pushover FD.20250211...
				$log_msg = "gif_tool [$gif_height H, $gif_fps FPS, $gif_sec s, $gif_kb KB] [$disp_delay_ms, $disp_makegif_ms, $disp_notify_ms, $disp_total_ms ms] $gif_method $push_service_str $gif_source_filename"
			}
			$response_str = `
				PostCustomBiLogEntry `
				-MsgStr "$log_msg" `
				-Level $log_level `
				-SilentFlg $true
				
			if ($FromConsole -and !$Quiet) {
				if (!$response_str.StartsWith('ERR/')) {
					Write-Host "Custom BI logfile message posted"
				} else {
					#Write-Host "Unable to post custom BI logfile message"
					Write-Host ($response_str) -ForegroundColor Red
				}
			}
		}
	} else {
		if ($FromConsole -and !$Quiet) {
			Write-Host "DISABLED by argument -NoBiLog or -Demo"
		}
	}
}

# HANDLE BACKGROUND EXECUTION RETURN OBJECT
if ($RtnHashtable) {
	if (!$FromConsole) {
		return $rtn_obj
	} else {
		if (!$Quiet) {
			Write-Host "Background return object:"
			Write-Host "-------------------------"
		}
		return ($rtn_obj | ConvertTo-Json)
	}
}

# Closing separator
if ($FromConsole -and !$Quiet) {
	Write-Host ""
}

#Stop-Transcript